/**
 *
 * Copyright (C) 2012-2014, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */
#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/uaccess.h>
#include <linux/types.h>
#include <linux/uio.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/time.h>
#if !defined(__i386__) && !defined(__x86_64__)
#include <linux/of.h>
#endif
#include <linux/gfp.h>
#include <linux/list.h>
#include <linux/rculist.h>
#include <linux/workqueue.h>
#include <linux/signal.h>
#include <linux/netlink.h>
#include <net/netlink.h>
#include <net/net_namespace.h>
#include "cpld_handler.h"

/************************************************
 *  Define
 ***********************************************/
#define DRV_NAME	"cpld_handler"
#define PFX		DRV_NAME ": "
#ifndef BUILD_VERSION
#define BUILD_VERSION 	""
#endif

#define CPLD_HANDLER_CHAR_DEVICE_NAME	"/dev/cpld/cpld0" // echo > /dev/cpld/cpld0 &
#define CPLD_HANDLER_DRV_VERSION	"1.00" BUILD_VERSION

/* Control IRQ / Polling option */
static bool cpld_no_work_queue;
module_param(cpld_no_work_queue, bool, 0444);
MODULE_PARM_DESC(cpld_no_work_queue, "Don't use work queue. Default = 0 (use work queue)");

/* When in IRQ mode read the device every THREAD_IRQ_SLEEP_SECS */
#define THREAD_IRQ_SLEEP_SECS	2
#define THREAD_IRQ_SLEEP_MSECS	(THREAD_IRQ_SLEEP_SECS * MSEC_PER_SEC)

struct nl_data_t {
    struct sock* sock; /* Netlink socket pointer */
    u32 pid;           /* User space PID waiting netlink msg */
    u32 seq;           /* Sequence number */
};

#if 0
/* CPLD_FAMILY the family itself for creation 
   a new Generic Netlink family named "CPLD_FAMILY"
*/
/* CPLD_FAMILY attributes */
enum {
	CPLD_FAMILY_A_UNSPEC,
	CPLD_FAMILY_A_MSG,
	__CPLD_FAMILY_A_MAX,
};
#define CPLD_FAMILY_A_MAX (__CPLD_FAMILY_A_MAX - 1)
/* CPLD_FAMILY attribute policy */
static struct nla_policy cpld_genl_policy[CPLD_FAMILY_A_MAX + 1] = {
	[CPLD_FAMILY_A_MSG] = { .type = NLA_NUL_STRING },
};
/* CPLD_FAMILY family definition */
static struct genl_family cpld_gnl_family = {
	.id = GENL_ID_GENERATE,
	.hdrsize = 0,
	.name = "CPLD_FAMILY",
	.version = 1,
	.maxattr = CPLD_FAMILY_A_MAX,
};

/* The operations for CPLD_FAMILY, which creates instances of
   the genl_ops structure. Up to 255 unique operations can be created
*/

/* handler */
static int cpld_echo(struct sk_buff *skb, struct genl_info *info)
{
       /* message handling code goes here; return 0 on success, negative
        * values on failure */
       return 0;
}
/* commands */
enum {
	CPLD_FAMILY_C_UNSPEC,
	CPLD_FAMILY_C_ECHO,
	__CPLD_FAMILY_C_MAX,
};
#define CPLD_FAMILY_C_MAX (__CPLD_FAMILY_C_MAX - 1)
/* operation definition */
static struct genl_ops cpld_gnl_ops_echo = {
	.cmd = CPLD_FAMILY_C_ECHO,
	.flags = 0,
	.policy = cpld_genl_policy,
	.doit = cpld_echo,
	.dumpit = NULL,
};
#endif
#define FALSE_INT_NOTIFY_THRESH 500
struct cpld_handler_dev {
	struct list_head 	action_list_1_level;
	struct list_head 	action_list_2_level;
	struct list_head 	action_list_3_level;
	struct list_head 	action_list_ha_level;
	struct list_head 	action_list_4_level;
	struct list_head 	action_list_5_level;
	struct cdev		*cdev;
	spinlock_t 		lock;
	wait_queue_head_t 	poll_wait;
	u8			int_occurred;
	int			int_disable_counter;
	int			irq;
	struct delayed_work	dwork;
	unsigned int		num;
	int			cpu;
	struct nl_data_t	nl;
	struct int_digest_blk_t *digest_blk;
	long                     false_int_cntr;
	u8			resched_on_exit;
};
static struct sock *sys_irq_sock;

enum CPLD_HANDLER_CHAR_DEVICE {
	CPLD_HANDLER_MAJOR	 = 232,
	CPLD_HANDLER_BASE_MINOR	 = 250,
	CPLD_HANDLER_MAX_DEVICES = 3
};
#define CPLD_HANDLER_BASE_DEV MKDEV(CPLD_HANDLER_MAJOR, CPLD_HANDLER_BASE_MINOR)

static struct cpld_handler_dev *p_dev_table[CPLD_HANDLER_MAX_DEVICES];
 
#ifndef NO_IRQ
#define NO_IRQ       ((unsigned int)(-1))
#endif

#if defined(__i386__) || defined(__x86_64__)
static int CPLD_IRQ1 = PIRQB;
module_param(CPLD_IRQ1, int, 0644);
static int CPLD_IRQ2 = PIRQC;
module_param(CPLD_IRQ2, int, 0644);
static int CPLD_IRQ3 = PIRQD;
module_param(CPLD_IRQ3, int, 0644);
static inline int _find_irq_num(int index)
{
        switch (index) {
                case 0: return CPLD_IRQ1;
                case 1: return CPLD_IRQ2;
                case 2: return CPLD_IRQ3;
                default: return -EINVAL;
        }
}
#else
static int CPLD_IRQ1 = 1;
module_param(CPLD_IRQ1, int, 0644);
static int CPLD_IRQ2 = 2;
module_param(CPLD_IRQ2, int, 0644);
static int CPLD_IRQ3 = 3;
module_param(CPLD_IRQ3, int, 0644);
const char *uih_name[] = {
	"/plb/opb/user_interrupt_handler",
	"/plb/opb/uih1",
	"/plb/opb/uih2",
};
static inline int _find_irq_num(int index)
{
	struct device_node *p_nd = of_find_node_by_path(uih_name[index]);
        if (p_nd) {
        	p_dev_table[index]->irq = of_irq_to_resource(p_nd, 0, NULL);
        	if (p_dev_table[index]->irq == NO_IRQ) {
        		printk (KERN_ALERT "%s IRQ not connected\n", __func__);
        		return -EINVAL; 
        	}
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
        	         "%s(%d) ind=%d %s irq=%d\n", __FUNCTION__, __LINE__,
        	         index, uih_name[index], p_dev_table[index]->irq);
        } else {
        	printk (KERN_ALERT "%s: node device pointer is NULL\n ", __func__);
        	return -EINVAL;
        }
        return p_dev_table[index]->irq;
}
#endif

const char *irq_names[] = {
	"sys_int_0",
	"sys_int_1",
	"sys_int_2",
};

int cpld_handler_action_subscribe(struct cpld_handler_action_entry *action_entry)
{
	unsigned long flags;
	struct cpld_handler_action_entry *p, *n;
	int rc = 0;
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
                "%s(%d) action_entry=%p index=%d usr_handle=0x%08x int_level=%d pre_action=%p post_action=%p\n",
                __FUNCTION__, __LINE__, action_entry, action_entry->index, 
                action_entry->usr_handle, action_entry->int_level,
                action_entry->pre_action, action_entry->post_action);
	if (action_entry->index >= CPLD_HANDLER_MAX_DEVICES)
		return -EFAULT;
	spin_lock_irqsave(&p_dev_table[action_entry->index]->lock, flags);
	switch (action_entry->int_level) {
		/* Verify that action_entry is not exist in action list.
		   it's enough to check action_list.
		   Here is tested action_list.
		*/
		case INT_LEVEL_1:
		case INT_LEVEL_1_2:
		case INT_LEVEL_1_2_3:
		case INT_LEVEL_HA:
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
                		"%s(%d) level=%d index=%d: entry=%p\n",
                		__FUNCTION__, __LINE__, action_entry->int_level, action_entry->index,
                		p_dev_table[action_entry->index]);
			list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_1_level,
						 action_list_1) {
				if (p == action_entry) {
					/* Already exist */
					printk(KERN_INFO "%s(%d) action=%p already exist\n",
						__FUNCTION__, __LINE__, p->pre_action);
					rc = -EEXIST;
					goto entry_exist;
				}
			}
			list_add_rcu(&action_entry->action_list_1,
					&p_dev_table[action_entry->index]->action_list_1_level);
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) p_dev=%p action_list=%d action_entry=%p\n",
				__FUNCTION__, __LINE__,
				p_dev_table[action_entry->index],
				!list_empty(&p_dev_table[action_entry->index]->action_list_1_level),
				action_entry);
				p = list_first_entry(&p_dev_table[action_entry->index]->action_list_1_level,
						    struct cpld_handler_action_entry, action_list_1);
				mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
					"%s(%d) p=%p pre_action=%p post_action=%p\n",
					__FUNCTION__, __LINE__, p, p->pre_action, p->post_action);
			if (action_entry->int_level == INT_LEVEL_1)
				break;
			if (action_entry->int_level == INT_LEVEL_HA) {
				/* perfrom HA action setting */
				list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_ha_level,
						         action_list_ha) {
					if (p == action_entry) {
						/* Already exist */
						printk(KERN_INFO "%s(%d) action=%p already exist\n",
							__FUNCTION__, __LINE__, p->pre_action);
						rc = -EEXIST;
						goto entry_exist;
					}
				}
				list_add_rcu(&action_entry->action_list_ha,
					     &p_dev_table[action_entry->index]->action_list_ha_level);
				mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
					"%s(%d) p_dev=%p action_list=%d action_entry=%p\n",
					__FUNCTION__, __LINE__,
					p_dev_table[action_entry->index],
					!list_empty(&p_dev_table[action_entry->index]->action_list_ha_level),
					action_entry); 
				break;
			}
			/* For INT_LEVEL_1_2 and INT_LEVEL_1_2_3 continue with next case */
		case INT_LEVEL_2:
		case INT_LEVEL_2_3:
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
                		"%s(%d) level=%d index=%d: entry=%p\n",
                		__FUNCTION__, __LINE__, action_entry->int_level, action_entry->index,
                		p_dev_table[action_entry->index]);
			list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_2_level,
						 action_list_2) {
				if (p == action_entry) {
					/* Already exist */
					printk(KERN_INFO "%s(%d) action=%p already exist\n",
						__FUNCTION__, __LINE__, p->pre_action);
					rc = -EEXIST;
					goto entry_exist;
				}
			}
			list_add_rcu(&action_entry->action_list_2,
					&p_dev_table[action_entry->index]->action_list_2_level);
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) p_dev=%p action_list=%d action_entry=%p\n",
				__FUNCTION__, __LINE__,
				p_dev_table[action_entry->index],
				!list_empty(&p_dev_table[action_entry->index]->action_list_2_level),
				action_entry);
			if ((action_entry->int_level == INT_LEVEL_2) ||
			    (action_entry->int_level == INT_LEVEL_1_2))
				break;
			/* For INT_LEVEL_2_3 continue with next case */
		case INT_LEVEL_3:
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
                		"%s(%d) level=%d index=%d: entry=%p\n",
                		__FUNCTION__, __LINE__, action_entry->int_level, action_entry->index,
                		p_dev_table[action_entry->index]);
			list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_3_level,
						 action_list_3) {
				if (p == action_entry) {
					/* Already exist */
					printk(KERN_INFO "%s(%d) action=%p already exist\n",
						__FUNCTION__, __LINE__, p->pre_action);
					rc = -EEXIST;
					goto entry_exist;
				}
			}
			list_add_rcu(&action_entry->action_list_3,
					&p_dev_table[action_entry->index]->action_list_3_level);
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) p_dev=%p action_list=%d action_entry=%p\n",
				__FUNCTION__, __LINE__,
				p_dev_table[action_entry->index],
				!list_empty(&p_dev_table[action_entry->index]->action_list_3_level),
				action_entry);
			break;
		case INT_LEVEL_4:
		case INT_LEVEL_4_5:
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
                		"%s(%d) level=%d index=%d: entry=%p\n",
                		__FUNCTION__, __LINE__, action_entry->int_level, action_entry->index,
                		p_dev_table[action_entry->index]);
			list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_4_level,
						 action_list_4) {
				if (p == action_entry) {
					/* Already exist */
					printk(KERN_INFO "%s(%d) action=%p already exist\n",
						__FUNCTION__, __LINE__, p->pre_action);
					rc = -EEXIST;
					goto entry_exist;
				}
			}
			list_add_rcu(&action_entry->action_list_4,
					&p_dev_table[action_entry->index]->action_list_4_level);
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) p_dev=%p action_list=%d action_entry=%p\n",
				__FUNCTION__, __LINE__,
				p_dev_table[action_entry->index],
				!list_empty(&p_dev_table[action_entry->index]->action_list_4_level),
				action_entry);
				p = list_first_entry(&p_dev_table[action_entry->index]->action_list_4_level,
						    struct cpld_handler_action_entry, action_list_4);
				mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
					"%s(%d) p=%p pre_action=%p post_action=%p\n",
					__FUNCTION__, __LINE__, p, p->pre_action, p->post_action);
			if (action_entry->int_level == INT_LEVEL_4)
				break;
			/* For INT_LEVEL_4_5 continue with next case */
			break;
		case INT_LEVEL_5:
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
                		"%s(%d) level=%d index=%d: entry=%p\n",
                		__FUNCTION__, __LINE__, action_entry->int_level, action_entry->index,
                		p_dev_table[action_entry->index]);
			list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_5_level,
						 action_list_5) {
				if (p == action_entry) {
					/* Already exist */
					printk(KERN_INFO "%s(%d) action=%p already exist\n",
						__FUNCTION__, __LINE__, p->pre_action);
					rc = -EEXIST;
					goto entry_exist;
				}
			}
			list_add_rcu(&action_entry->action_list_5,
					&p_dev_table[action_entry->index]->action_list_5_level);
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) p_dev=%p action_list=%d action_entry=%p\n",
				__FUNCTION__, __LINE__,
				p_dev_table[action_entry->index],
				!list_empty(&p_dev_table[action_entry->index]->action_list_5_level),
				action_entry);
				p = list_first_entry(&p_dev_table[action_entry->index]->action_list_5_level,
						    struct cpld_handler_action_entry, action_list_5);
				mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
					"%s(%d) p=%p pre_action=%p post_action=%p\n",
					__FUNCTION__, __LINE__, p, p->pre_action, p->post_action);
			break;
		default:
			printk(KERN_INFO "%s(%d) action=%p has unsupported int level=%d\n", 
				__FUNCTION__, __LINE__, action_entry, action_entry->int_level);
			rc = -EINVAL;
			goto entry_exist;
	}

entry_exist:
	spin_unlock_irqrestore(&p_dev_table[action_entry->index]->lock, flags);
	return rc;
}
EXPORT_SYMBOL(cpld_handler_action_subscribe);

int cpld_handler_action_unsubscribe(struct cpld_handler_action_entry *action_entry)
{
	unsigned long flags;
	struct cpld_handler_action_entry *p, *n;
	int rc = -EINVAL;
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) index=%d\n", __FUNCTION__, __LINE__, action_entry->index);
	if (action_entry->index >= CPLD_HANDLER_MAX_DEVICES)
		return -EFAULT;
	spin_lock_irqsave(&p_dev_table[action_entry->index]->lock, flags);
	switch (action_entry->int_level) {
		/* Verify that action_entry is not exist in action list.
		   it's enough to check action_list.
		   Here is tested action_list.
		*/
		case INT_LEVEL_1:
		case INT_LEVEL_1_2:
		case INT_LEVEL_1_2_3:
		case INT_LEVEL_HA:
			list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_1_level, action_list_1) {
				if (p == action_entry) {
					list_del_rcu(&action_entry->action_list_1);
					rc = 0;
				}
			}
			if (action_entry->int_level == INT_LEVEL_1)
				break;
			if (action_entry->int_level == INT_LEVEL_HA) {
				/* remove HA action setting */
				list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_ha_level, 
						         action_list_ha) {
					if (p == action_entry) {
						list_del_rcu(&action_entry->action_list_ha);
						rc = 0;
					}
				}
				break;
			}
			/* For INT_LEVEL_1_2 and INT_LEVEL_1_2_3 continue with next case */
		case INT_LEVEL_2:
		case INT_LEVEL_2_3:
			list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_2_level,
						 action_list_2) {
				if (p == action_entry) {
					list_del_rcu(&action_entry->action_list_2);
					rc = 0;
				}
			}
			if ((action_entry->int_level == INT_LEVEL_2) ||
			    (action_entry->int_level == INT_LEVEL_1_2))
				break;
			/* For INT_LEVEL_2_3 continue with next case */
		case INT_LEVEL_3:
			list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_3_level,
						 action_list_3) {
				if (p == action_entry) {
					list_del_rcu(&action_entry->action_list_3);
					rc = 0;
				}
			}
			break;
		case INT_LEVEL_4:
		case INT_LEVEL_4_5:
			list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_4_level,
						 action_list_4) {
				if (p == action_entry) {
					list_del_rcu(&action_entry->action_list_4);
					rc = 0;
				}
			}
			break;
		case INT_LEVEL_5:
			list_for_each_entry_safe(p, n, &p_dev_table[action_entry->index]->action_list_5_level,
						 action_list_5) {
				if (p == action_entry) {
					list_del_rcu(&action_entry->action_list_5);
					rc = 0;
				}
			}
			break;
		default:
			printk(KERN_INFO "%s(%d) action=%p has unsupported int level=%d\n", 
				__FUNCTION__, __LINE__, action_entry, action_entry->int_level);
			rc = -EINVAL;
			goto exit_unsubscribe;
	}
exit_unsubscribe:
	spin_unlock_irqrestore(&p_dev_table[action_entry->index]->lock, flags);
	return rc;
}
EXPORT_SYMBOL(cpld_handler_action_unsubscribe);

static int cpld_send_nl_handler_err_notification(struct cpld_handler_dev *p_dev, u8 msg_type, int err)
{
	struct sk_buff *skb;
	struct nlmsghdr *p_nlh;
	struct sys_irq_nl_msg *p_msg;
	size_t payload_len;
	int data_len;
	int rc = 0;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) p_dev=%p msg_type=%d rc=%d\n",
		__FUNCTION__, __LINE__, p_dev, msg_type, rc);

	data_len = sizeof(int); 
	payload_len = sizeof(*p_msg) + data_len;
	/* Allocate skb - it should be freed by netlink */
	skb = alloc_skb(NLMSG_SPACE(payload_len), GFP_KERNEL);
	if (!skb) {
		printk(KERN_INFO "Failed to allocate socket buffer\n");
		return -ENOMEM;
	}
	p_nlh = nlmsg_put(skb, p_dev->nl.pid, p_dev->nl.seq,
				msg_type, payload_len, 0);

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) p_dev=%p msg_type=%d data_len=%d payload_len=%zd\n",
		__FUNCTION__, __LINE__, p_dev, msg_type, data_len, payload_len);

	p_msg = (struct sys_irq_nl_msg*)NLMSG_DATA(p_nlh);
	p_msg->data_len = data_len;
	memset(p_msg->data, 0, data_len);
	memcpy(p_msg->data, &err, sizeof(int));

	rc = netlink_unicast(p_dev->nl.sock, skb, p_dev->nl.pid, MSG_DONTWAIT);
	if (rc < 0) {
		printk(KERN_INFO "Failed to send netlink message; rc %d\n", rc);
		goto nlmsg_failure;
	}
        rc = 0;
        p_dev->nl.seq++;

nlmsg_failure:
    return rc;
}

static int cpld_send_nl_msg(struct cpld_handler_dev *p_dev, u8 msg_type)
{
	struct int_blk_digest_send_t all_digest;
	struct int_module_digest_send_t module_digest;
	struct mod_digest_send digest_tab;
	struct sk_buff *skb;
	struct nlmsghdr *p_nlh;
	struct sys_irq_nl_msg *p_msg;
	int offset;
	u8 *data_ptr, *data_ptr1;
	size_t payload_len;
	int data_len;
	int rc = 0;
	int i;
	int j;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) p_dev=%p msg_type=%d max_size=%zd mod_num=%d all_change_cntr=%d\n",
		__FUNCTION__, __LINE__, p_dev, msg_type, sizeof(struct int_digest_blk_t),
		p_dev->digest_blk->mod_num, p_dev->digest_blk->all_change_cntr);
#if 0
	if (p_dev->digest_blk->mod_num <= 0) {
		p_dev->false_int_cntr++;
		if ((p_dev->false_int_cntr % FALSE_INT_NOTIFY_THRESH) == 0) {
			printk(KERN_INFO "Nothing to send netlink message mod_num=%d all_change_cntr=%d false int cntr=%lu\n",
				p_dev->digest_blk->mod_num,
				p_dev->digest_blk->all_change_cntr,
				p_dev->false_int_cntr);
			cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_FALSE_INT_ALARM, rc);
		}
		return 0;
	}
#endif

	data_len = sizeof(all_digest.mod_num) +
			p_dev->digest_blk->mod_num *
			(sizeof(module_digest.usr_handle) + sizeof(module_digest.num_entry)) +
			p_dev->digest_blk->all_change_cntr * sizeof(struct mod_digest_send);
	payload_len = sizeof(*p_msg) + data_len;
	/* Allocate skb - it should be freed by netlink */
	skb = alloc_skb(NLMSG_SPACE(payload_len), GFP_KERNEL);
	if (!skb) {
		printk(KERN_INFO "Failed to allocate socket buffer\n");
		return -ENOMEM;
	}
	p_nlh = nlmsg_put(skb, p_dev->nl.pid, p_dev->nl.seq,
				msg_type, payload_len, 0);

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) p_dev=%p msg_type=%d data_len=%d payload_len=%zd\n",
		__FUNCTION__, __LINE__, p_dev, msg_type, data_len, payload_len);
	p_msg = (struct sys_irq_nl_msg*)NLMSG_DATA(p_nlh);
	p_msg->data_len = data_len;
	memset(p_msg->data, 0, data_len);

	/* Notification message format:
        +---------------------------------------------------------+
        | Number of modules in message (1 octets)                 |
        +---------------------------------------------------------+
        | 1-st module user handle (4 octets)                      | maybe 1 octet is enough
        +---------------------------------------------------------+
        | 1-st module number of entries with interrupt indication |
        +---------------------------------------------------------+
        | 1-st module 1-st irq entry id (1 octet)                 |
        +---------------------------------------------------------+
        | 1-st module 1-st irq mask (1 octet)                     |
        +---------------------------------------------------------+
        | ...                                                     |
        +---------------------------------------------------------+
        | 1-st module M1 irq entry id (1 octet)                   |
        +---------------------------------------------------------+
        | 1-st module M1 irq mask (1 octet)                       |
        +---------------------------------------------------------+
        | ...                                                     |
        +---------------------------------------------------------+
        | N module user handle (4 octets)                         |
        +---------------------------------------------------------+
        | N module number of entries with interrupt indication    |
        +---------------------------------------------------------+
        | N module 1-st irq entry id (1 octet)                    |
        +---------------------------------------------------------+
        | N module 1-st irq mask (1 octet)                        |
        +---------------------------------------------------------+
        | ...                                                     |
        +---------------------------------------------------------+
        | N module MN irq entry id (1 octet)                      |
        +---------------------------------------------------------+
        | N module MN irq mask (1 octet)                          |
        +---------------------------------------------------------+
	*/

	data_ptr1 =  data_ptr = p_msg->data;
	all_digest.mod_num = p_dev->digest_blk->mod_num;
	offset = sizeof(all_digest.mod_num);
	memcpy(data_ptr, &all_digest.mod_num, offset);
	data_ptr += offset;
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX, "%s(%d) mod_num=%d\n",
		__FUNCTION__, __LINE__, p_dev->digest_blk->mod_num);

	for (i = 0; i < p_dev->digest_blk->mod_num; i++) {
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) i=%d p_dev=%p usr_handle=0x%08x change_cntr=%d offset=%d\n",
			__FUNCTION__, __LINE__, i, p_dev, p_dev->digest_blk->digest_table[i].usr_handle,
			p_dev->digest_blk->digest_table[i].change_cntr, offset);
		memset(&module_digest, 0, sizeof(struct int_module_digest_send_t));
		module_digest.usr_handle = p_dev->digest_blk->digest_table[i].usr_handle;
		module_digest.num_entry = p_dev->digest_blk->digest_table[i].change_cntr;
		offset = sizeof(module_digest.usr_handle) + sizeof(module_digest.num_entry);
		memcpy(data_ptr, &module_digest, offset);
		data_ptr += offset;
		for (j = 0; j < MAX_MGMT_IRQS; j++) {
			if (!p_dev->digest_blk->digest_table[i].int_digest[j].int_vector)
				continue;
			memset(&digest_tab, 0, sizeof(struct mod_digest_send));
			digest_tab.irq_entry_id = j;
			digest_tab.int_digest.int_vector = p_dev->digest_blk->digest_table[i].int_digest[j].int_vector;
			digest_tab.int_digest.int_status = p_dev->digest_blk->digest_table[i].int_digest[j].int_status;
			offset = sizeof(struct mod_digest_send);
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) j=%d int_digest vector=0x%08x status=0x%08x offset=%d %zd\n",
				__FUNCTION__, __LINE__, j,
				p_dev->digest_blk->digest_table[i].int_digest[j].int_vector,
				p_dev->digest_blk->digest_table[i].int_digest[j].int_status,
				offset, (data_ptr - p_msg->data));
			memcpy(data_ptr, &digest_tab, offset);
			data_ptr += offset;
			p_dev->digest_blk->digest_table[i].int_digest[j].int_vector = 0;
			p_dev->digest_blk->digest_table[i].int_digest[j].int_status = 0;
		}
		p_dev->digest_blk->digest_table[i].change_cntr = 0;
		p_dev->digest_blk->digest_table[i].usr_handle = 0;
	}

	rc = netlink_unicast(p_dev->nl.sock, skb, p_dev->nl.pid, MSG_DONTWAIT);
	if (rc < 0) {
		printk(KERN_INFO "Failed to send netlink message; rc %d\n", rc);
		goto nlmsg_failure;
	}
        rc = 0;
        p_dev->nl.seq++;

nlmsg_failure:
    return rc;
}

static void cpld_receive_nl_msg(struct sk_buff *skb)
{
	struct nlmsghdr *p_nlh;
	struct sys_irq_nl_msg *p_msg;
	sys_irq_reg_t irq_reg;
	struct cpld_handler_dev *p_dev = NULL;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX, "%s(%d)\n", __FUNCTION__, __LINE__);
	p_nlh = nlmsg_hdr(skb);
	if (!NLMSG_OK(p_nlh, skb->len)) {
		printk(KERN_INFO "Received corrupted netlink message data_len=%d\n",
			skb->len);
		return;
	}
	switch (p_nlh->nlmsg_type) {
		case SYS_IRQ_NL_REGISTER:
		case SYS_IRQ_NL_UNREGISTER:
			p_msg = NLMSG_DATA(p_nlh);
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) nlmsg_type=%d skb_len=%d hdr_len=%d, msg_len=%zd data_len=%d\n",
				__FUNCTION__, __LINE__, p_nlh->nlmsg_type, 
				skb->len, NLMSG_HDRLEN, sizeof(*p_msg), p_msg->data_len);
			if (skb->len - NLMSG_HDRLEN - sizeof(*p_msg) != p_msg->data_len) {
				printk(KERN_INFO "Received incorrect netlink message " \
					"%d - %d - %zu != %d\n",
					skb->len, NLMSG_HDRLEN, sizeof(*p_msg), p_msg->data_len);
				return;
			}
			memset(&irq_reg, 0, sizeof(sys_irq_reg_t));
			memcpy(&irq_reg, p_msg->data, p_msg->data_len);
			if (irq_reg.int_line >= CPLD_HANDLER_MAX_DEVICES) {
				printk(KERN_INFO "Received incorrect netlink message " \
					"int line is invalid = %d\n",
					irq_reg.int_line);
				return;
			}
			p_dev = p_dev_table[irq_reg.int_line];
			if (p_nlh->nlmsg_type == SYS_IRQ_NL_REGISTER) {
				p_dev->nl.pid = irq_reg.pid;
				mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
					"Netlink register pid=%d int line=%d\n",
					irq_reg.pid, irq_reg.int_line);
			}
			else {
				mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
					"Netlink unregister  pid=%d int line=%d\n",
					irq_reg.pid, irq_reg.int_line);
				p_dev->nl.pid = 0;
			}
			break;
		default:
			printk(KERN_INFO "Received unknown netlink message, type %d\n", p_nlh->nlmsg_type);
			break;
	}
}

static void cpld_reschedule_work(struct cpld_handler_dev *p_dev, unsigned long delay)
{
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) p_dev=%p\n", __FUNCTION__, __LINE__, p_dev);
	/* If work is already scheduled then subsequent schedules will not
	   change the scheduled time that's why it should be canceled first. */
	cancel_delayed_work(&p_dev->dwork);
	schedule_delayed_work(&p_dev->dwork, delay);
}

static void cpld_cleanup_entries(struct cpld_handler_dev *p_dev)
{
	struct cpld_handler_action_entry *p, *n;

	list_for_each_entry_safe(p, n, &p_dev->action_list_5_level, action_list_5) {
		memset(&p->digest_entry->change_cntr, 0, sizeof(p->digest_entry));
	}

	list_for_each_entry_safe(p, n, &p_dev->action_list_4_level, action_list_4) {
		memset(&p->digest_entry->change_cntr, 0, sizeof(p->digest_entry));
	}

	list_for_each_entry_safe(p, n, &p_dev->action_list_3_level, action_list_3) {
		memset(&p->digest_entry->change_cntr, 0, sizeof(p->digest_entry));
	}

	list_for_each_entry_safe(p, n, &p_dev->action_list_2_level, action_list_2) {
		memset(&p->digest_entry->change_cntr, 0, sizeof(p->digest_entry));
	}

	list_for_each_entry_safe(p, n, &p_dev->action_list_1_level, action_list_1) {
		memset(&p->digest_entry->change_cntr, 0, sizeof(p->digest_entry));
	}
}

/* CPLD Work Handler */
static void cpld_work_handler(struct work_struct *work)
{
	unsigned long flags;
	struct cpld_handler_dev *p_dev = container_of(work, struct cpld_handler_dev, dwork.work);
	struct cpld_handler_action_entry *p, *n;
	unsigned long delay;
	int entry_num = 0;
	u8 not_detected_l1 = 1;
	u8 not_detected_l2 = 1;
	u8 aggregation = 0;
	u8 recovered = 0;
	u8 resched_on_exit = 1;
	int rc = 0;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) p_dev=%p action_list=%d\n",
		__FUNCTION__, __LINE__,
		p_dev, !list_empty(&p_dev->action_list_1_level));
	memset(p_dev->digest_blk, 0, sizeof(struct int_digest_blk_t));
	/* Sync with digest block read critical section. 
	   Get auxiliary cache block, copy there data from
	   master cache block. Active work queue entry works at
	   up-down traverse with master cache block.
	*/
	/* Mask and read level 1 interrupts 1-st time, compare with cache block,
	   store result in auxiliary cache block.
	*/
	list_for_each_entry_safe(p, n, &p_dev->action_list_1_level, action_list_1) {
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX, 
			"%s(%d) p=%p n=%p\n", __FUNCTION__, __LINE__, p, n);
		p_dev->digest_blk->all_change_cntr = 0;
		if (p->pre_action) {
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) pre_act=%p post_act=%p subscriber_data=%p entry_num=%d ent=%zd ecs=%x scs=%x\n",
				__FUNCTION__, __LINE__, p->pre_action, p->post_action,
				p->subscriber_data, entry_num,
				sizeof(*p->digest_entry), p->event_checksum, p->status_checksum);
			rc = p->pre_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, entry_num, INT_LEVEL_1);
			if (rc < 0) {
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_HANDLER_ERR, rc);
				goto irq_handler_error_exit;
			}
			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) 2 pre_act=%p post_act=%p sdata=%p aggr=%d entry_num=%d ent=%zd ecs=%x scs=%x\n",
				__FUNCTION__, __LINE__, p->pre_action, p->post_action,
				p->subscriber_data, entry_num, p->aggregation,
				sizeof(*p->digest_entry), p->event_checksum, p->status_checksum);
			/* both event_checksum and status_checksum are tested for level 1, since for example,
			   for aggregation there is no event register */
			if (p->event_checksum || p->status_checksum) {
				not_detected_l1 = 0;
				p->event_checksum = 0;
				p->status_checksum = 0;
				if (!p->aggregation) {
					mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
						"%s(%d) entry_num=%d usr_handle=%x change_cntr=%d\n",
						__FUNCTION__, __LINE__, entry_num, p->digest_entry->usr_handle,
						p->digest_entry->change_cntr);
					/* Update digest and number of entries only for non aggregated entry */
					memcpy(&p_dev->digest_blk->digest_table[entry_num],
						p->digest_entry, sizeof(struct mod_digest));
					entry_num++;
				}
			}
			if ((p->digest_entry->change_cntr) && (p->int_level == INT_LEVEL_1))
				p_dev->digest_blk->all_change_cntr += p->digest_entry->change_cntr;

			if(p->recovered)
				recovered = p->recovered;

			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) cntr=%d all_change_cntr=%d\n",
				__FUNCTION__, __LINE__,
				p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr);
		}
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) cntr=%d all_change_cntr=%d\n",
			__FUNCTION__, __LINE__, p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr);
	}

	/* Continue traversing if any raised interrupt is detected.
	   If not - procceed to removing entry from work queue. 
	*/
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) not_detected_l1=%d p=%p n=%p all_change_cntr=%d recovered=%d\n",
		__FUNCTION__, __LINE__, not_detected_l1, p, n, p_dev->digest_blk->all_change_cntr, recovered);

	if (not_detected_l1)
		goto rem_entry_from_wq;

	/* Mask and read level 2 interrupts, compare with cache block.
	   Store result in master cache block.
	*/
	list_for_each_entry_safe(p, n, &p_dev->action_list_2_level, action_list_2) {
		if (p->pre_action) {
			rc = p->pre_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, entry_num, INT_LEVEL_2);
			if (rc < 0) {
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_HANDLER_ERR, rc);
				goto irq_handler_error_exit;
			}
			if (p->event_checksum) {
				not_detected_l2 = 0;
				p->event_checksum = 0;
				p->status_checksum = 0;
				/* Make OR between level 1 and level 2 interrups -
				   all of the are related to same main module
				   (same usr_handle).
				*/
	        		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		        		"%s(%d) int_level=%d all_change_cntr=%d entry_num=%d change_cntr=%d %d usr_handle=%x\n",
		        		__FUNCTION__, __LINE__, p->int_level,
		        		p_dev->digest_blk->all_change_cntr,
		        		entry_num, p_dev->digest_blk->digest_table[entry_num].change_cntr,
		        		p->digest_entry->change_cntr, p->digest_entry->usr_handle);

                		/* Inrcease number of entries only for separate level 2 or level 1_2 */
				if ((!p->aggregation) &&
				    ((p->int_level == INT_LEVEL_2) || (p->int_level == INT_LEVEL_1_2) ||
				     p->termination)) {
					mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
						"%s(%d) entry_num=%d usr_handle=%x change_cntr=%d entry_num=%d\n",
						__FUNCTION__, __LINE__, entry_num, p->digest_entry->usr_handle,
						p->digest_entry->change_cntr, entry_num);
					/* Update digest and number of entries only for non aggregated entry */
					memcpy(&p_dev->digest_blk->digest_table[entry_num],
						p->digest_entry, sizeof(struct mod_digest));
                			entry_num++;
				}
			}

			if ((p->digest_entry->change_cntr) && 
				((p->int_level == INT_LEVEL_1) || (p->int_level == INT_LEVEL_2) ||
				(p->int_level == INT_LEVEL_1_2) || p->termination))
				p_dev->digest_blk->all_change_cntr += p->digest_entry->change_cntr;

			if (p->aggregation)
				aggregation++;

			p->termination = 0;

			mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
				"%s(%d) p->digest_entry=%p usr_handle=0x%08x cntr=%d all_change_cntr=%d aggr=%d\n",
				__FUNCTION__, __LINE__, p->digest_entry, p->digest_entry->usr_handle,
				p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr,
				aggregation);
		}	
	}

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) level 2 not_detected_l2=%d aggr=%d\n",
		__FUNCTION__, __LINE__, not_detected_l2, aggregation);

	if (not_detected_l2) {
		if (!aggregation)
			goto ack_unmask_level_2_int;
#if !defined(__i386__) && !defined(__x86_64__)
		else
			goto ack_unmask_level_1_int;
#endif
	}

	/* Mask and read level 3 interrupts, compare with cache block,
	   store result in auxiliary cache block and acknowledge all
	   inserted interrupts and unmask those were masked.
	   Re-read interrupts and store in master cache block.
	   Level 3 interrupts is bottom level interrupt, so there
	   is no post_action method for them - all sequence:
	   mask and read -> acknowledge and unmask are perfromed
	   within pre_action method.
	*/
	aggregation = 0;
	list_for_each_entry_safe(p, n, &p_dev->action_list_3_level, action_list_3) {
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) int_level=%d cntr=%d all_change_cntr=%d\n",
			__FUNCTION__, __LINE__, p->int_level,
			p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr);
		if (p->pre_action) {
			p->event_checksum = 0;
			p->status_checksum = 0;
			rc = p->pre_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, entry_num, INT_LEVEL_3);
			if (rc < 0) {
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_HANDLER_ERR, rc);
				goto irq_handler_error_exit;
			}
			if (p->event_checksum) {
				mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
					"%s(%d) p->event_checksum=%d p->status_checksum=%d set entry %d\n",
					__FUNCTION__, __LINE__,
					p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr, entry_num);
				memcpy(&p_dev->digest_blk->digest_table[entry_num],
					p->digest_entry, sizeof(struct mod_digest));
                		entry_num++;
			}
			p->event_checksum = 0;
			p->status_checksum = 0;
		}
		if ((p->digest_entry->change_cntr) && 
			((p->int_level == INT_LEVEL_2) || (p->int_level == INT_LEVEL_2_3) ||
			(p->int_level == INT_LEVEL_1_2_3) || (p->int_level == INT_LEVEL_3)))
			p_dev->digest_blk->all_change_cntr += p->digest_entry->change_cntr;

		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) p_dev=%p digest_entry=%p digest_blk=%p cntr=%d all_change_cntr=%d usr_handle=0x%08x entry_num=%d\n",
			__FUNCTION__, __LINE__, p_dev, p->digest_entry, p_dev->digest_blk,
			p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr,
			p->digest_entry->usr_handle, entry_num);
	}

	list_for_each_entry_safe(p, n, &p_dev->action_list_3_level, action_list_3) {
		if (p->post_action) {
			rc = p->post_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, 0, INT_LEVEL_3);
			if (rc < 0) {
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_HANDLER_ERR, rc);
				goto irq_handler_error_exit;
			}
		}
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) p_dev=%p digest_entry=%p digest_blk=%p cntr=%d all_change_cntr=%d usr_handle=0x%08x entry_num=%d\n",
			__FUNCTION__, __LINE__, p_dev, p->digest_entry, p_dev->digest_blk,
			p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr,
			p->digest_entry->usr_handle, entry_num);
	}

#if 0
	/* Activate this block within Barracuda support - it should
	   change the above behaviour at level 3 the above p->post_action 
	   call should be removed.
	*/

	list_for_each_entry_safe(p, n, &p_dev->action_list_4_level, action_list_4) {
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) cntr=%d all_change_cntr=%d\n",
			__FUNCTION__, __LINE__,
			p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr);
		if (p->pre_action) {
			p->event_checksum = 0;
			p->status_checksum = 0;
			rc = p->pre_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, entry_num, INT_LEVEL_4);
			if (p->event_checksum) {
				memcpy(&p_dev->digest_blk->digest_table[entry_num],
					p->digest_entry, sizeof(struct mod_digest));
                		entry_num++;
			}
			p->event_checksum = 0;
			p->status_checksum = 0;
		}
	}

	list_for_each_entry_safe(p, n, &p_dev->action_list_5_level, action_list_5) {
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) cntr=%d all_change_cntr=%d\n",
			__FUNCTION__, __LINE__,
			p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr);
		if (p->pre_action) {
			p->event_checksum = 0;
			p->status_checksum = 0;
			rc = p->pre_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, entry_num, INT_LEVEL_5);
			if (p->event_checksum) {
				memcpy(&p_dev->digest_blk->digest_table[entry_num],
					p->digest_entry, sizeof(struct mod_digest));
                		entry_num++;
			}
			p->event_checksum = 0;
			p->status_checksum = 0;
		if ((p->digest_entry->change_cntr) && 
			((p->int_level == INT_LEVEL_4) || (p->int_level == INT_LEVEL_4_5)))
			p_dev->digest_blk->all_change_cntr += p->digest_entry->change_cntr;
		if (p->post_action) {
			rc = p->post_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, 0, INT_LEVEL_5);
			if (rc < 0) {
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_HANDLER_ERR, rc);
				goto irq_handler_error_exit;
			}
		}
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) cntr=%d all_change_cntr=%d\n",
			__FUNCTION__, __LINE__,
			p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr);
		}
	}

	/* Acknowledge and unmask level 4 interrupts, collected before
	   if any. Don't touch any new interrupt that possibly have been
           inserted during work queue processing.
           Re-read interrupts and store auxiliary cache block.
	*/
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) level 4 post_action not_detected_l1=%d not_detected_l4=%d\n",
		__FUNCTION__, __LINE__, not_detected_l1, not_detected_l4);
	list_for_each_entry_safe(p, n, &p_dev->action_list_4_level, action_list_4) {
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) pre_act=%p post_act=%p subscriber_data=%p entry_num=%d ent=%zd ecs=%x scs=%x\n",
			__FUNCTION__, __LINE__, p->pre_action, p->post_action,
			p->subscriber_data, entry_num,
			sizeof(*p->digest_entry), p->event_checksum, p->status_checksum);
		if (p->post_action) {
			rc = p->post_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, 0, INT_LEVEL_4);
			if (rc < 0) {
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_HANDLER_ERR, rc);
				goto irq_handler_error_exit;
			}
		}
	}

	/* Acknowledge and unmask level 3 interrupts, collected before
	   if any. Don't touch any new interrupt that possibly have been
           inserted during work queue processing.
           Re-read interrupts and store auxiliary cache block.
	*/
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) level 2 post_action not_detected_l1=%d not_detected_l2=%d\n",
		__FUNCTION__, __LINE__, not_detected_l1, not_detected_l3);
	list_for_each_entry_safe(p, n, &p_dev->action_list_3_level, action_list_3) {
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) pre_act=%p post_act=%p subscriber_data=%p entry_num=%d ent=%zd ecs=%x scs=%x\n",
			__FUNCTION__, __LINE__, p->pre_action, p->post_action,
			p->subscriber_data, entry_num,
			sizeof(*p->digest_entry), p->event_checksum, p->status_checksum);
		if (p->post_action) {
			rc = p->post_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, 0, INT_LEVEL_3);
			if (rc < 0) {
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_HANDLER_ERR, rc);
				goto irq_handler_error_exit;
			}
		}
	}
#endif

ack_unmask_level_2_int:
	/* Acknowledge and unmask level 2 interrupts, collected before
	   if any. Don't touch any new interrupt that possibly have been
           inserted during work queue processing.
           Re-read interrupts and store auxiliary cache block.
	*/
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) level 2 post_action not_detected_l1=%d not_detected_l2=%d\n",
		__FUNCTION__, __LINE__, not_detected_l1, not_detected_l2);
	list_for_each_entry_safe(p, n, &p_dev->action_list_2_level, action_list_2) {
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) pre_act=%p post_act=%p subscriber_data=%p entry_num=%d ent=%zd ecs=%x scs=%x\n",
			__FUNCTION__, __LINE__, p->pre_action, p->post_action,
			p->subscriber_data, entry_num,
			sizeof(*p->digest_entry), p->event_checksum, p->status_checksum);
		if (p->post_action) {
			rc = p->post_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, 0, INT_LEVEL_2);
			if (rc < 0) {
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_HANDLER_ERR, rc);
				goto irq_handler_error_exit;
			}
		}		

	}

#if !defined(__i386__) && !defined(__x86_64__)
ack_unmask_level_1_int:
#endif
	/* Read level 1 interrupts 2-nd time, compare with cache.
	   If new interrupts have been inserted during work queue
	   processing, then process high priority interrupts only.
	   Such interrupt doesn't have sub-hierarchical logic.
	   Add such interrupt to cache, then acknowledge them and
	   unmask. Re-read interrupts and store in auxiliary cache block.
	*/
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX, 
		"%s(%d) level 1 post_action not_detected_l1=%d\n",
		__FUNCTION__, __LINE__, not_detected_l1);
	list_for_each_entry_safe(p, n, &p_dev->action_list_1_level, action_list_1) {
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) pre_act=%p post_act=%p subscriber_data=%p entry_num=%d ent=%zd ecs=%x scs=%x\n",
			__FUNCTION__, __LINE__, p->pre_action, p->post_action,
			p->subscriber_data, entry_num,
			sizeof(*p->digest_entry), p->event_checksum, p->status_checksum);
		if (p->post_action) {
			rc = p->post_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, 0, INT_LEVEL_1);
			if (rc < 0) {
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_HANDLER_ERR, rc);
				goto irq_handler_error_exit;
			}
		}
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) cntr=%d all_change_cntr=%d\n",
			__FUNCTION__, __LINE__,
			p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr);
	}

	p_dev->digest_blk->mod_num = entry_num;

	/* In case working queue is not configured - activate poll_wait.
	   Then connecetd to file descriptor application will obtain interrupts'
	   digest throguh dedicated API.
	   In other case - use netlink to deliver interrupt digets directly to
	   subscriber.
	*/
	if (cpld_no_work_queue)
		wake_up(&p_dev->poll_wait);
	else {
		if (p_dev->digest_blk->mod_num > 0) {
			cpld_send_nl_msg(p_dev, SYS_IRQ_NL_INTERRUPT);
			p_dev->false_int_cntr = 0;
			p_dev->resched_on_exit = 1;
		}
		else {
			p_dev->false_int_cntr++;
			if ((p_dev->false_int_cntr % FALSE_INT_NOTIFY_THRESH) == 0) {
				printk(KERN_INFO "Nothing to send netlink message mod_num=%d all_change_cntr=%d false int cntr=%lu\n",
					p_dev->digest_blk->mod_num,
					p_dev->digest_blk->all_change_cntr,
					p_dev->false_int_cntr);
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_FALSE_INT_ALARM, rc);
			}
			/* Don't reschedule (only one time if previous signal was empty */
		        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		                "%s(%d) resched_on_exit=%d p_dev->resched_on_exit=%d\n",
		                __FUNCTION__, __LINE__, resched_on_exit, p_dev->resched_on_exit);
			resched_on_exit = p_dev->resched_on_exit;
			if (p_dev->resched_on_exit)
				p_dev->resched_on_exit = 0;
		}
	}

irq_handler_error_exit:
	cpld_cleanup_entries(p_dev);

	/* Control Work Handler sleep time.
	   While interrupt driven, there is no real need to poll the devices -
	   reschdeualing will be activated from interrupt handler.
	   However, there could be errors related to physical environment.
	   So, polling is once activated in THREAD_IRQ_SLEEP_SECS.
	*/

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX, 
		"%s(%d) int_disable_counter=%d\n",
		__FUNCTION__, __LINE__, p_dev->int_disable_counter);

	spin_lock_irqsave(&p_dev->lock, flags);
	if (p_dev->int_disable_counter > 0) {
		p_dev->int_disable_counter--;
		enable_irq(p_dev->irq);
	}

	if (resched_on_exit) {
		delay = msecs_to_jiffies(THREAD_IRQ_SLEEP_MSECS);
		delay = round_jiffies_relative(delay);
		cpld_reschedule_work(p_dev, 0/*delay*/);
	}

	spin_unlock_irqrestore(&p_dev->lock, flags);

	return;

rem_entry_from_wq:
	/* no interrupts have been detected - could be just check loop
	   initiated not by cpld_irq_handler, but by sequnce
	   cpld_reschedule_work -> cpld_work_handler*/
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX, 
		"%s(%d) rem_entry_from_wq: int_disable_counter=%d\n",
		__FUNCTION__, __LINE__, p_dev->int_disable_counter);

	list_for_each_entry_safe(p, n, &p_dev->action_list_1_level, action_list_1) {
		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) pre_act=%p post_act=%p subscriber_data=%p entry_num=%d ent=%zd ecs=%x scs=%x\n",
			__FUNCTION__, __LINE__, p->pre_action, p->post_action,
			p->subscriber_data, entry_num,
			sizeof(*p->digest_entry), p->event_checksum, p->status_checksum);
		if (p->post_action) {
			rc = p->post_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, 0, INT_LEVEL_1);
			if (rc < 0) {
				cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_HANDLER_ERR, rc);
				goto irq_handler_error_exit_on_rem_entry;
			}
		}
		if (recovered)
			cpld_send_nl_handler_err_notification(p_dev, SYS_IRQ_NL_RCVRD_ERR, rc);

		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s(%d) cntr=%d all_change_cntr=%d\n",
			__FUNCTION__, __LINE__,
			p->digest_entry->change_cntr, p_dev->digest_blk->all_change_cntr);
	}

irq_handler_error_exit_on_rem_entry:
	spin_lock_irqsave(&p_dev->lock, flags);
	if (p_dev->int_disable_counter > 0) {
		p_dev->int_disable_counter--;
		enable_irq(p_dev->irq);
	}
	spin_unlock_irqrestore(&p_dev->lock, flags);

	return;
}

static irqreturn_t cpld_irq_handler(int irq, void *dev_ptr)
{
	struct cpld_handler_dev *p_dev = (struct cpld_handler_dev *)dev_ptr;
	struct cpld_handler_action_entry *p, *n;
	unsigned long flags;
	int rc;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) p_dev=%p cpld_no_work_queue=%d ... disable irq\n",
		__FUNCTION__, __LINE__, p_dev, cpld_no_work_queue);

	disable_irq_nosync(p_dev->irq);
	spin_lock_irqsave(&p_dev->lock, flags);
	p_dev->int_disable_counter++;
	if (cpld_no_work_queue) {
		list_for_each_entry_safe(p, n, &p_dev->action_list_1_level, action_list_1) {
			rc = p->pre_action(p->subscriber_data, p_dev->digest_blk, p->digest_entry, 0, INT_LEVEL_1);
		}
	}
	else {
		cpld_reschedule_work(p_dev, 0);
	}

	spin_unlock_irqrestore(&p_dev->lock, flags);
	if (cpld_no_work_queue)
		wake_up(&p_dev->poll_wait);

	return IRQ_HANDLED;
}

static int register_to_interrupt(int index)
{
        char irq_name[16];
	//unsigned long flags;
	int rc = 0;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) index=%d\n", __FUNCTION__, __LINE__, index);
    	p_dev_table[index]->irq = _find_irq_num(index);
	p_dev_table[index]->int_disable_counter = 0;
	p_dev_table[index]->int_occurred = 0;
	p_dev_table[index]->resched_on_exit = 1;
	if (p_dev_table[index]->irq == NO_IRQ) {
		printk(KERN_INFO "Error NO IRQ\n");
		return -EINVAL;
	}

	p_dev_table[index]->nl.sock = sys_irq_sock;
	p_dev_table[index]->cpu = smp_processor_id();
	memset(irq_name, 0, sizeof(irq_name));
	sprintf(irq_name, "%s%d", "cpld_int_", index);
	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) index=%d irq=%d %s\n",
		__FUNCTION__, __LINE__, index, p_dev_table[index]->irq, irq_name);
	rc = request_irq(p_dev_table[index]->irq, cpld_irq_handler,
        		     IRQF_SHARED /*| IRQF_TIMER | IRQF_DISABLED | IRQF_NOBALANCING*/, 
        		     irq_names[index], p_dev_table[index]);
	if (rc) {
        	printk(KERN_INFO "Error on request_irq %d\n", p_dev_table[index]->irq);
        	return -EINVAL;
	}

#if defined(__i386__) || defined(__x86_64__)
        //disable_irq(p_dev_table[index]->irq);
        //irq_set_affinity_hint(p_dev_table[index]->irq, cpumask_of(p_dev_table[index]->cpu));
        //enable_irq(p_dev_table[index]->irq);
#endif

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d)\n", __FUNCTION__, __LINE__);
	return 0;
}

static unsigned int cpld_handler_poll(struct file *filp, poll_table *wait)
{
	unsigned int	mask = 0;
	unsigned long 	flags;
	struct cpld_handler_dev *p_dev = (struct cpld_handler_dev *)filp->private_data;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d)\n", __FUNCTION__, __LINE__);
	spin_lock_irqsave(&p_dev->lock, flags);
	if (p_dev->int_occurred) {
		mask |= POLLIN | POLLRDNORM;
		p_dev->int_occurred = 0;
	}
	spin_unlock_irqrestore(&p_dev->lock, flags);
	poll_wait(filp, &p_dev->poll_wait , wait);
	return mask;
}

static ssize_t cpld_handler_write(struct file *filp,
		const char __user *buf, size_t count, loff_t *pos)
{
	return 0;
}

static ssize_t cpld_handler_read(struct file *filp, char __user *buf,
		size_t count, loff_t *pos)
{
	int    err = 0;
	if ((count == 0) || (buf == NULL)) {
		err = -EINVAL;
	}
	return err;
}

static int cpld_handler_open_1(struct inode *inode, struct file *filp)
{
	int err = 0;
	int index = 0;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d)  private_data=%p\n", __FUNCTION__, __LINE__, p_dev_table[index]);
        filp->private_data = (void *)p_dev_table[index];
	spin_lock_init(&p_dev_table[index]->lock);
	INIT_DELAYED_WORK(&p_dev_table[index]->dwork, cpld_work_handler);
	init_waitqueue_head(&p_dev_table[index]->poll_wait);
	p_dev_table[index]->int_occurred = 0;
	p_dev_table[index]->resched_on_exit = 1;
	err = register_to_interrupt(index);
	if (err)
		printk(KERN_INFO "%s error registering to the interrupt\n",
			__func__);

	return err;
}

static int cpld_handler_open_2(struct inode *inode, struct file *filp)
{
	int err = 0;
	int index = 1;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) private_data=%p\n", __FUNCTION__, __LINE__, p_dev_table[index]);
        filp->private_data = (void *)p_dev_table[index];
	spin_lock_init(&p_dev_table[index]->lock);
	INIT_DELAYED_WORK(&p_dev_table[index]->dwork, cpld_work_handler);
	init_waitqueue_head(&p_dev_table[index]->poll_wait);
	p_dev_table[index]->int_occurred = 0;
	p_dev_table[index]->resched_on_exit = 1;
	err = register_to_interrupt(index);
	if (err)
		printk(KERN_INFO "%s error registering to the interrupt\n",
			__func__);

	return err;
}

static int cpld_handler_open_3(struct inode *inode, struct file *filp)
{
	int err = 0;
	int index = 2;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) private_data=%p\n", __FUNCTION__, __LINE__, p_dev_table[index]);
        filp->private_data = (void *)p_dev_table[index];
	spin_lock_init(&p_dev_table[index]->lock);
	INIT_DELAYED_WORK(&p_dev_table[index]->dwork, cpld_work_handler);
	init_waitqueue_head(&p_dev_table[index]->poll_wait);
	p_dev_table[index]->int_occurred = 0;
	p_dev_table[index]->resched_on_exit = 1;
	err = register_to_interrupt(index);
	if (err)
		printk(KERN_INFO "%s error registering to the interrupt\n",
			__func__);

	return err;
}

static int cpld_handler_close(struct inode *inode, struct file *filp)
{
	int err = 0;
	int flush;
	struct cpld_handler_dev *p_dev = (struct cpld_handler_dev *)filp->private_data;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d)\n", __FUNCTION__, __LINE__);
	if (p_dev->irq != NO_IRQ) {
#if defined(__i386__) || defined(__x86_64__)
        	//irq_set_affinity_hint(p_dev->irq, NULL);
#endif
		free_irq(p_dev->irq, p_dev);
		flush = cancel_delayed_work_sync(&p_dev->dwork);
		if (flush)
			flush_scheduled_work();
	}
	return err;
}

static long cpld_handler_ioctl(struct file *filp,
		unsigned int cmd, unsigned long data)
{
	int err = 0;
	unsigned long flags;
	struct cpld_handler_dev *p_dev = (struct cpld_handler_dev *)filp->private_data;

	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) cmd=%d cntr=%d private_data=%p\n",
		__FUNCTION__, __LINE__, cmd, p_dev->int_disable_counter, p_dev);

	if (cmd == CPLD_IRQ_HANDLER_DISABLE) { // disable interrupts
		spin_lock_irqsave(&p_dev->lock, flags);
		p_dev->int_disable_counter++;
		disable_irq_nosync(p_dev->irq);
		spin_unlock_irqrestore(&p_dev->lock, flags);
	}
	else if (cmd == CPLD_IRQ_HANDLER_ENABLE) { // enable interrupts
		spin_lock_irqsave(&p_dev->lock, flags);
		while (p_dev->int_disable_counter > 0) {
			p_dev->int_disable_counter--;
			enable_irq(p_dev->irq);
		}
		spin_unlock_irqrestore(&p_dev->lock, flags);
	}
	else {
		return -EINVAL;
	}
	return err;
}

static const struct file_operations cpld_handler_fops[] = {
	{
		.owner	 =	THIS_MODULE,
		.open	 =	cpld_handler_open_1,
		.read	 =	cpld_handler_read,
		.write	 =	cpld_handler_write,
		.poll	 =	cpld_handler_poll,
		.unlocked_ioctl = cpld_handler_ioctl,
		.release = 	cpld_handler_close
	},
	{
		.owner	 =	THIS_MODULE,
		.open	 =	cpld_handler_open_2,
		.read	 =	cpld_handler_read,
		.write	 =	cpld_handler_write,
		.poll	 =	cpld_handler_poll,
		.unlocked_ioctl = cpld_handler_ioctl,
		.release = 	cpld_handler_close
	},
	{
		.owner	 =	THIS_MODULE,
		.open	 =	cpld_handler_open_3,
		.read	 =	cpld_handler_read,
		.write	 =	cpld_handler_write,
		.poll	 =	cpld_handler_poll,
		.unlocked_ioctl = cpld_handler_ioctl,
		.release = 	cpld_handler_close
	},
};

int *mlnx_debug_level_symbol = NULL;
static int __init cpld_handler_init(void)
{
	int err = 0;
	int ind;
	int ind1;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
        struct netlink_kernel_cfg cfg = {
                .input = cpld_receive_nl_msg,
        };
#endif

	err = register_chrdev_region(CPLD_HANDLER_BASE_DEV, CPLD_HANDLER_MAX_DEVICES, CPLD_HANDLER_CHAR_DEVICE_NAME);
	if (err) {
		printk(KERN_INFO "%s failed to register region %x max=%d maj=%d min=%d name=%s err=%d\n",
			__func__, CPLD_HANDLER_BASE_DEV, CPLD_HANDLER_MAX_DEVICES,
			CPLD_HANDLER_MAJOR, CPLD_HANDLER_BASE_MINOR,
			CPLD_HANDLER_CHAR_DEVICE_NAME, err);
		goto out;
	}

	/* Allocate p_dev_table entries */
	for (ind = 0; ind < CPLD_HANDLER_MAX_DEVICES; ind++) {
		p_dev_table[ind] = kmalloc(sizeof(struct cpld_handler_dev), GFP_KERNEL);
		if (!p_dev_table[ind]) {
			err = -ENOMEM;
			goto out_mem_dev_tab;
		}
	}

	for (ind = 0; ind < CPLD_HANDLER_MAX_DEVICES; ind++) {
		p_dev_table[ind]->cdev = cdev_alloc();
		if (!p_dev_table[ind]->cdev) {
			printk(KERN_INFO "%s failed to allocate device for ind=%d\n",
				__func__, ind);
			err = -ENOMEM;
			goto out_chrdev;
		}
		p_dev_table[ind]->cdev->owner = THIS_MODULE;
		p_dev_table[ind]->cdev->ops   = &cpld_handler_fops[ind];
		err = cdev_add(p_dev_table[ind]->cdev, 
				MKDEV(CPLD_HANDLER_MAJOR, CPLD_HANDLER_BASE_MINOR + ind), 1);
		if (err) {
			printk(KERN_INFO "%s failed to add device for ind=%d err=%d\n",
				__func__, ind, err);
			goto out_cdev;
		}

		mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			"%s added device for ind=%d err=%d\n", __func__, ind, err);
 
		INIT_LIST_HEAD(&p_dev_table[ind]->action_list_1_level);
		INIT_LIST_HEAD(&p_dev_table[ind]->action_list_2_level);
		INIT_LIST_HEAD(&p_dev_table[ind]->action_list_3_level);
		INIT_LIST_HEAD(&p_dev_table[ind]->action_list_ha_level);
		INIT_LIST_HEAD(&p_dev_table[ind]->action_list_4_level);
		INIT_LIST_HEAD(&p_dev_table[ind]->action_list_5_level);
		p_dev_table[ind]->digest_blk = kmalloc(sizeof(struct int_digest_blk_t), GFP_KERNEL);
		if (!p_dev_table[ind]->digest_blk) {
			err = -ENOMEM;
			goto out_mem;
		}
		memset(p_dev_table[ind]->digest_blk, 0, sizeof(struct int_digest_blk_t));
	}
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
	sys_irq_sock = netlink_kernel_create(&init_net, NETLINK_SYS_IRQ_SOCK, &cfg);
#else
        sys_irq_sock = netlink_kernel_create(&init_net, NETLINK_SYS_IRQ_SOCK, 0,
                                                cpld_receive_nl_msg, NULL, THIS_MODULE);
#endif
	if (!sys_irq_sock) {
		printk(KERN_INFO "Failed to create netlink socket\n");
		err = -EIO;
		goto out_cdev;
	}
#if 0
	/* Register the CPLD_FAMILY family with the Generic Netlink operation.
	   This call registers the new family name with the Generic Netlink mechanism
	   and requests a new channel number which is stored in the genl_family struct,
	   replacing the GENL_ID_GENERATE value.  
	   Family is to be unregistered from Generic Netlink families on exit 
	   as the kernel does allocate resources for each registered family.
	*/
	err = genl_register_family(&cpld_gnl_family);
	if (err != 0) {
		printk(KERN_INFO "%s failed to register netlink family\n",
			__func__);
		goto netlink_release;
	}
	/* Register the operations for the family */
	err = genl_register_ops(&cpld_gnl_family, &cpld_gnl_ops_echo);
	if (err != 0) {
		printk(KERN_INFO "%s failed to register netlink operations\n",
			__func__);
		goto unreg_family;
	}
#endif

	mlnx_debug_level_symbol = __symbol_get("mlnx_debug_level");

	printk(KERN_INFO "%s - initialized\n", __func__);

	return 0;

#if 0
unreg_family:
	genl_unregister_family(&cpld_gnl_family);
netlink_release:
	netlink_kernel_release(sys_irq_sock);
#endif
out_mem:
	for (ind1 = ind - 1; ind1 >= 0; ind1--)
		kfree(p_dev_table[ind1]->digest_blk);
	ind = CPLD_HANDLER_MAX_DEVICES;
out_cdev:
	for (ind1 = ind - 1; ind1 >= 0; ind1--)
		cdev_del(p_dev_table[ind1]->cdev);
	ind = CPLD_HANDLER_MAX_DEVICES;
out_chrdev:
	unregister_chrdev_region(CPLD_HANDLER_BASE_DEV, CPLD_HANDLER_MAX_DEVICES);
	ind = CPLD_HANDLER_MAX_DEVICES;
out_mem_dev_tab:
	for (ind1 = ind - 1; ind1 >= 0; ind1--)
		kfree(p_dev_table[ind1]);
out:
	return err;
}

static void __exit cpld_handler_exit(void)
{
	int ind1;

	printk(KERN_INFO "%s - cleanup\n", __func__);
#if 0
	genl_unregister_ops(&cpld_gnl_family, &cpld_gnl_ops_echo);
	genl_unregister_family(&cpld_gnl_family);
#endif
        if (sys_irq_sock)
		netlink_kernel_release(sys_irq_sock);
	for (ind1 = CPLD_HANDLER_MAX_DEVICES - 1; ind1 >= 0; ind1--) {
		cdev_del(p_dev_table[ind1]->cdev);
		kfree(p_dev_table[ind1]->digest_blk);
	}
	unregister_chrdev_region(CPLD_HANDLER_BASE_DEV, CPLD_HANDLER_MAX_DEVICES);
	for (ind1 = CPLD_HANDLER_MAX_DEVICES- 1; ind1 >= 0; ind1--)
		kfree(p_dev_table[ind1]);

	if (mlnx_debug_level_symbol)
		__symbol_put("mlnx_debug_level");
}

/************************************************
 *  MODULE init/exit
 ***********************************************/
module_init(cpld_handler_init);
module_exit(cpld_handler_exit);

MODULE_AUTHOR("Vadim Pasternak <vadimp@mellanox.com>");
MODULE_DESCRIPTION("CPLD handler driver");
MODULE_LICENSE("GPL v2");






