/**
 *
 * Copyright (C) 2013-2014, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <asm/io.h>
#include <linux/fs.h>
#include <linux/sysfs.h>

#include "cpld_gpio.h"
#include "cpld.h"
#define out_be32(a, v) __raw_writel(__cpu_to_be32(v), (a));

#define in_be32(a) __be32_to_cpu(__raw_readl(a))

static unsigned long gpio0_base;
static unsigned long gpio1_base;

//extern int pin_type;

/* Map GPIO registers */
int cpld_gpio_init(void)
{
	/* Map GPIO0 registers */
	gpio0_base = (unsigned long)ioremap_nocache(GPIO0_BASE, GPIO_SIZE);
	if (!gpio0_base) {
		CPLD_LOG_ERROR("Failed to map GPIO0\n");
		return -ENOMEM;
	}

	/* Map GPIO1 registers */
	gpio1_base = (unsigned long)ioremap_nocache(GPIO1_BASE, GPIO_SIZE);
	if (!gpio1_base) {
		CPLD_LOG_ERROR("Failed to map GPIO1\n");
		iounmap((void *)gpio0_base);
		return -ENOMEM;
	}

    CPLD_LOG_DBG(3, "%s, %lu %lu\n", __FUNCTION__, gpio0_base, gpio1_base);
	return 0;
}

/* Unmap GPIO registers */
void cpld_gpio_release(void)
{
    CPLD_LOG_DBG(3, "%s\n", __FUNCTION__);

	if (gpio0_base)
		iounmap((void *)gpio0_base);

	if (gpio1_base)
		iounmap((void *)gpio1_base);
}

/* Configure I/O pin to be either GPIO or alternate function */
void gpio_config_pin(struct gpio_param_type* cpld_gpio)
{
	int j = 0;
	int offs = 0;
	unsigned long reg;
    unsigned long base = gpio0_base;
    int pin = cpld_gpio->pin;
    int in_out = cpld_gpio->in_out;
    int gpio_alt = cpld_gpio->alt_nb;
    int out_val = cpld_gpio->out_val;

    CPLD_LOG_DBG(3, "%s, pin %d, in_out %d, alt_nb %d, out_val %d\n", __FUNCTION__,\
                 cpld_gpio->pin, cpld_gpio->in_out, cpld_gpio->alt_nb, cpld_gpio->out_val);

	/* Select GPIO controller */
	if (pin >= GPIO_MAX) {
		base = gpio1_base;
		pin -= GPIO_MAX;
	}

	/* Select "High" GPIO registers (OSRH, TSRH, etc.) */
	if (pin >= GPIO_MAX/2) {
		offs = 4;
		j = pin - 16;
	}

	if (in_out == GPIO_IN) {

		switch (gpio_alt) {

		case GPIO_SEL:
			break;

		case GPIO_ALT1:
			reg = in_be32((void *)GPIO_IS1(base+offs))
				& ~(GPIO_MASK >> (j*2));
			reg = reg | (GPIO_IN_SEL >> (j*2));
			out_be32((void *)GPIO_IS1(base+offs), reg);
			break;

		case GPIO_ALT2:
			reg = in_be32((void *)GPIO_IS2(base+offs))
				& ~(GPIO_MASK >> (j*2));
			reg = reg | (GPIO_IN_SEL >> (j*2));
			out_be32((void *)GPIO_IS2(base+offs), reg);
			break;

		case GPIO_ALT3:
			reg = in_be32((void *)GPIO_IS3(base+offs))
				& ~(GPIO_MASK >> (j*2));
			reg = reg | (GPIO_IN_SEL >> (j*2));
			out_be32((void *)GPIO_IS3(base+offs), reg);
			break;
		}
	}

	if (in_out == GPIO_OUT) {
		u32 gpio_alt_sel = 0;

		switch (gpio_alt) {

		case GPIO_SEL:
			/*
			 * Setup output value
			 * 1 -> high level
			 * 0 -> low level
			 * else -> don't touch
			 */
			reg = in_be32((void *)GPIO_OR(base));
			if (out_val == GPIO_OUT_1)
				reg |= GPIO_VAL(pin);
			else if (out_val == GPIO_OUT_0)
				reg &= ~GPIO_VAL(pin);
			out_be32((void *)GPIO_OR(base), reg);

			reg = in_be32((void *)GPIO_TCR(base)) |
				GPIO_VAL(pin);
			out_be32((void *)GPIO_TCR(base), reg);

			reg = in_be32((void *)GPIO_OS(base+offs))
				& ~(GPIO_MASK >> (j*2));
			out_be32((void *)GPIO_OS(base+offs), reg);
			reg = in_be32((void *)GPIO_TS(base+offs))
				& ~(GPIO_MASK >> (j*2));
			out_be32((void *)GPIO_TS(base+offs), reg);
			break;

		case GPIO_ALT1:
			gpio_alt_sel = GPIO_ALT1_SEL;
			break;

		case GPIO_ALT2:
			gpio_alt_sel = GPIO_ALT2_SEL;
			break;

		case GPIO_ALT3:
			gpio_alt_sel = GPIO_ALT3_SEL;
			break;
		}

		if (gpio_alt_sel) {
			reg = in_be32((void *)GPIO_OS(base+offs))
				& ~(GPIO_MASK >> (j*2));
			reg = reg | (gpio_alt_sel >> (j*2));
			out_be32((void *)GPIO_OS(base+offs), reg);

			if (out_val == GPIO_OUT_1) {
				reg = in_be32((void *)GPIO_TCR(base))
					| GPIO_VAL(pin);
				out_be32((void *)GPIO_TCR(base), reg);
				reg = in_be32((void *)GPIO_TS(base+offs))
					& ~(GPIO_MASK >> (j*2));
				out_be32((void *)GPIO_TS(base+offs), reg);
			} else {
				reg = in_be32((void *)GPIO_TCR(base))
					& ~GPIO_VAL(pin);
				out_be32((void *)GPIO_TCR(base), reg);
				reg = in_be32((void *)GPIO_TS(base+offs))
					& ~(GPIO_MASK >> (j*2));
				reg = reg | (gpio_alt_sel >> (j*2));
				out_be32((void *)GPIO_TS(base+offs), reg);
			}
		}
	}
}

/* Set value of GPIO output pin */
void gpio_write_bit(u8 pin, u8 out_val)
{
	unsigned long reg;
	unsigned long base = gpio0_base;

	if (pin >= GPIO_MAX) {
		base = gpio1_base;
		pin -= GPIO_MAX;
	}

    CPLD_LOG_DBG(4, "%s, base 0x%lx pin %d, out_val %d\n", __FUNCTION__, base, pin , out_val);

	reg = in_be32((void *)GPIO_OR(base));

	if (out_val == GPIO_OUT_1)
		reg |= GPIO_VAL(pin);
	else if (out_val == GPIO_OUT_0)
		reg &= ~GPIO_VAL(pin);

	out_be32((void *)GPIO_OR(base), reg);
}

/* Get value of GPIO output pin */
int gpio_read_bit(u8 pin)
{
	unsigned long reg;
	unsigned long base = gpio0_base;

	if (pin >= GPIO_MAX) {
		base = gpio1_base;
		pin -= GPIO_MAX;
	}

	reg = in_be32((void *)GPIO_OR(base));

	return (reg & GPIO_VAL(pin)) ? 1 : 0;
}

/* Get value of GPIO input pin */
int gpio_read_in_bit(u8 pin)
{
	unsigned long reg;
	unsigned long base = gpio0_base;

	if (pin >= GPIO_MAX) {
		base = gpio1_base;
		pin -= GPIO_MAX;
	}

	reg = in_be32((void *)GPIO_IR(base));

	return (reg & GPIO_VAL(pin)) ? 1 : 0;
}

