/*
 *
 * Copyright (C) Mellanox Technologies Ltd. 2001-2014.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/err.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <linux/workqueue.h>
#include <linux/platform_device.h>

#include <linux/init.h>
#include <linux/fs.h>
#include <linux/i2c.h>
#include <linux/gpio.h>
#include <linux/spinlock.h>
#include <linux/netlink.h>
#include <asm/param.h>
#include <net/net_namespace.h>

#ifdef CONFIG_X86
#include "gpio_ifc.h"
#include "cpld_gpio_x86.h"
#else
#include "cpld_gpio.h"
#endif
#include "cpld.h"

#define CPLD_MOD_DESCRIPTION    "CPLD JTAG driver. Build:" " "__DATE__" "__TIME__
#define CPLD_MOD_VERSION       "1.1.0"

// Module params and default values

int dbg_lvl = DFLT_CPLD_DBG_LVL;
module_param(dbg_lvl, int, 0644);
int pin_type = 0;
#ifndef CONFIG_X86
module_param(pin_type, int, 0644);
int jtag_enb_mode = CPLD_JTAG_ENB_MODE1;
module_param(jtag_enb_mode, int, 0644);
#endif

struct cpld_device_t* cpld_device;

#ifndef CONFIG_X86
/* Generally IO expander set not required, done from cli.
   Can be requested from standalone jam application.      */
static bool jtag_enable = 0;
module_param(jtag_enable, bool, 0644);
static int cpld_io_port = DFLT_CPLD_IO_PORT;
module_param(cpld_io_port, int, 0644);

static int cpld_i2c_addr = (DFLT_CPLD_I2C_BUS << 8 | DFLT_CPLD_I2C_ADDR);
module_param(cpld_i2c_addr, int, 0644);
extern struct mutex __i2c_mlnx_lock;

struct gpio_param_type cpld_orig_gpio_tbl[] = CONFIG_CPLD_ORIG_GPIO_TABLE;
struct gpio_param_type cpld_jtag_gpio_tbl0[] = CONFIG_CPLD_JTAG_GPIO_TABLE0;
struct gpio_param_type cpld_jtag_gpio_tbl1[] = CONFIG_CPLD_JTAG_GPIO_TABLE1;

/* Common function for access CPLD i2c registers as well as IO expander */
static int cpld_i2c_read(struct i2c_adapter* adapt, u8 addr, u8 reg, u8 *data, u8 len)
{
	int rc;
    struct i2c_msg msg[2];

   	msg[0].addr = addr;
    msg[0].flags = 0;
   	msg[0].len = 1;
   	msg[0].buf = &reg;

   	msg[1].addr = addr;
    msg[1].flags = I2C_M_RD;
   	msg[1].len = len;
   	msg[1].buf = data;

    CPLD_LOG_DBG(4, "%s(%d) addr 0x%x, reg 0x%x, len %d\n", __FUNCTION__, __LINE__, addr, reg, len);

    mutex_lock(&__i2c_mlnx_lock);
    rc = i2c_transfer(adapt, msg, 2);
    mutex_unlock(&__i2c_mlnx_lock);
	if (rc != 2) {
        CPLD_LOG_ERROR("Failed i2c read %s reg 0x%x, error %d\n",
                       (cpld_device->adapt == adapt ? "cpld" : "io exp"), reg, rc);
        return rc;
    }
    else {
        CPLD_LOG_DBG(4, "%s(%d) Ok, data 0x%x\n", __FUNCTION__, __LINE__, *data);
        return 0;
    }
}

static int cpld_i2c_write(struct i2c_adapter* adapt, u8 addr, u8 reg, u8 data, u8 len)
{
	int rc;
	struct i2c_msg msg;
    u8 msgbuf[] = {reg, data};

	msg.addr = addr;
	msg.flags = 0;
	msg.len = 2;
	msg.buf = msgbuf;

    CPLD_LOG_DBG(4, "%s addr 0x%x, reg 0x%x, data 0x%x len %d\n", __FUNCTION__, addr, reg, data, len);

    mutex_lock(&__i2c_mlnx_lock);
	rc = i2c_transfer(adapt, &msg, 1);
    mutex_unlock(&__i2c_mlnx_lock);
	if (rc != 1) {
        CPLD_LOG_ERROR("Failed i2c write %s reg 0x%x, error %d\n",
                       (cpld_device->adapt == adapt ? "cpld" : "io exp"), reg, rc);
        return rc;
    }
    else
        return 0;
}

/* Doesn't called. IO expander set not required, now done from user space application */
static int cpld_ioexp_config(void)
{
    int rc;
    u8 val;

    CPLD_LOG_DBG(3, "%s, adapt %p, i2c_addr 0x%x, reg 0x%x\n", __FUNCTION__, cpld_device->io_adapt, \
                 cpld_device->io_i2c_addr, CPLD_IOEXP_CFG_REG);
    rc = cpld_i2c_read(cpld_device->io_adapt, cpld_device->io_i2c_addr, CPLD_IOEXP_CFG_REG, &val, 1);
    if (rc)
        return rc; /* Failure */

    val &= ~(1<<CPLD_IOEXP_JTAG_BIT);
    rc = cpld_i2c_write(cpld_device->io_adapt, cpld_device->io_i2c_addr, CPLD_IOEXP_CFG_REG, val, 1);
    return rc;
}

/* IO expander can be requested from standalone jam application.*/
static int cpld_set_jtag(u8 status)
{
    int rc;
    u8 val;

    CPLD_LOG_DBG(3, "%s(%d)\n", __FUNCTION__, __LINE__);

    if (!(cpld_device->io_adapt)) {
        CPLD_LOG_ERROR("IO expander i2c adapter isn't initialised.\n");
        return -EPERM;
    }

    rc = cpld_ioexp_config();
    if (rc) {
        CPLD_LOG_ERROR("Failed config io expansion output port.\n");
        return rc;
    }

    rc = cpld_i2c_read(cpld_device->io_adapt, cpld_device->io_i2c_addr, CPLD_IOEXP_OUT_REG, &val, 1);
    if (rc) {
        CPLD_LOG_ERROR("Failed read io expansion output port.\n");
        return rc;
    }

    if (status == CPLD_JTAG_ENB)
        val |= (1<<CPLD_IOEXP_JTAG_BIT);
    else
        val &= ~(1<<CPLD_IOEXP_JTAG_BIT);
    rc = cpld_i2c_write(cpld_device->io_adapt, cpld_device->io_i2c_addr, CPLD_IOEXP_OUT_REG, val, 1);
    if (rc)
        CPLD_LOG_ERROR("Failed write to io expansion output port.\n");

    return rc;
}

static int cpld_data_init(void)
{
    struct i2c_adapter *adapt;

	spin_lock_init(&cpld_device->cpld_lock);

    cpld_device->i2c_bus = (cpld_i2c_addr & CPLD_I2C_BUS_MASK) >> 8 ;
    cpld_device->i2c_addr = cpld_i2c_addr & CPLD_I2C_ADDR_MASK;

    adapt = i2c_get_adapter(cpld_device->i2c_bus);
	if (!adapt) {
        CPLD_LOG_ERROR("Failed to get i2c cpld adapter, bus %ul\n", cpld_device->i2c_bus);
		return -EINVAL;
    }
    else
        cpld_device->adapt = adapt;

    /* IO expander can be requested from standalone jam application. */
    if (jtag_enable) {
        cpld_device->io_i2c_bus = (cpld_io_port & CPLD_I2C_BUS_MASK) >> 8 ;
        cpld_device->io_i2c_addr = cpld_io_port & CPLD_I2C_ADDR_MASK;

        CPLD_LOG_DBG(3, "%s cpld_i2c_addr 0x%x/0x%x, io_i2c_expander 0x%x/0x%x\n", \
                     __FUNCTION__,  cpld_device->i2c_bus, cpld_device->i2c_addr, cpld_device->io_i2c_bus, cpld_device->io_i2c_addr);
        adapt = i2c_get_adapter(cpld_device->io_i2c_bus);
        if (!adapt) {
            CPLD_LOG_ERROR("Failed get i2c io expansion adapter, bus %u\n", cpld_device->io_i2c_bus);
            return -EINVAL;
        }
        else {
            cpld_device->io_adapt = adapt;
            CPLD_LOG_DBG(4, "%s(%d), cpld_i2c_adapt %p, io_i2c_adapt %p\n", __FUNCTION__, __LINE__,
                         cpld_device->adapt, cpld_device->io_adapt);
            return 0;
        }
    }
    else
        return 0;
}
#else
struct gpio_param_type cpld_orig_gpio_tbl[] = CONFIG_CPLD_ORIG_GPIO_TABLE;
struct gpio_param_type cpld_jtag_gpio_tbl0[] = CONFIG_CPLD_JTAG_GPIO_TABLE0;
struct gpio_param_type* cpld_jtag_gpio_tbl1 = NULL; /* There is no second config option on X86 */
#endif

static int cpld_select(u8 cpld, u8 status)
{
    int rc = 0;
#ifndef CONFIG_X86
    u8 val, jtag_reg, jtag_mode, mask, cpld_sel;
    char str[10];

    memset(str, 0, sizeof(str));
    if (status == CPLD_JTAG_ENB)
        strncpy(str, "enable", strlen("enable"));
    else
        strncpy(str, "disable", strlen("disable"));

// #ifdef CONFIG_X86
    /* Now CPLD upgrade is done in serial manner, thus
       no need in CPLD select */
    /*if (status == CPLD_JTAG_ENB) {
        switch (cpld) {
        case CPLD1:
            gpio_write_bit(GPIO35, GPIO_OUT_0);
            gpio_write_bit(GPIO16, GPIO_OUT_1);
            break;
        case CPLD2:
            gpio_write_bit(GPIO16, GPIO_OUT_0);
            gpio_write_bit(GPIO35, GPIO_OUT_1);
            break;
        default:
            CPLD_LOG_ERROR("Incorrect cpld number %d.\n", cpld);
            return -EINVAL;
            break;
        }
        gpio_write_bit(GPIO16, GPIO_OUT_1);
    }
    else {
        gpio_write_bit(GPIO16, GPIO_OUT_0);
       // gpio_write_bit(GPIO35, GPIO_OUT_0);
    }*/
    // val = 0;
// #else

    if (jtag_enb_mode == CPLD_JTAG_ENB_MODE1) {
        jtag_reg = CPLD_JTAG_REG1;
        jtag_mode = CPLD_JTAG_MODE1;
        mask = CPLD_JTAG_MASK1;
        cpld_sel = 1<<(cpld-1);
    }
    else {
        jtag_reg = CPLD_JTAG_REG2;
        jtag_mode = CPLD_JTAG_MODE2;
        mask = CPLD_JTAG_MASK2;
        cpld_sel = 1<<cpld;
    }
    CPLD_LOG_DBG(3, "Jtag_reg: 0x%x, enb: 0x%x, mask: 0x%x, cpld_sel: 0x%x\n", jtag_reg, jtag_mode, mask, cpld_sel);

    if (!(cpld_device->adapt)) {
        CPLD_LOG_ERROR("CPLD i2c adapter isn't initialised.\n");
        return -EPERM;
    }

    rc = cpld_i2c_read(cpld_device->adapt, cpld_device->i2c_addr, jtag_reg, &val, 1);
    if (rc) {
        CPLD_LOG_ERROR("Failed jtag %s cpld_num %d, read err: %d, bus %d addr 0x%x\n", \
                str, cpld, rc, cpld_device->i2c_bus, cpld_device->i2c_addr);
		return rc;
    }
    CPLD_LOG_DBG(2, "Jtag_reg: 0x%x, read_val: 0x%x\n", jtag_reg, val);

    val &= mask;
    if (status == CPLD_JTAG_ENB)
        val |= (jtag_mode | cpld_sel);
    else
        val &= ~(jtag_mode | cpld_sel);

    rc = cpld_i2c_write(cpld_device->adapt, cpld_device->i2c_addr, jtag_reg, val, 1);
    if (rc)
        CPLD_LOG_ERROR("Failed jtag %s cpld_num %d, write err: %d, bus %d addr 0x%x\n", \
                str, cpld, rc, cpld_device->i2c_bus, cpld_device->i2c_addr);
    else
        /*CPLD_LOG_DBG(2, "CPLD %s: cpld_num %d jtag_reg 0x%x val 0x%x\n", str, cpld, jtag_reg, val);*/
        CPLD_LOG_ERROR("CPLD %s: cpld_num %d jtag_reg 0x%x val 0x%x\n", str, cpld, jtag_reg, val);
#endif

    return rc;
}

static void cpld_gpio_jtag_config(u8 status)
{
    struct gpio_param_type* cpld_gpio_tbl;
    u8 i;

    CPLD_LOG_DBG(4, "%s, status %s\n", __FUNCTION__, ((status == CPLD_JTAG_ENB) ? "enable" : "disable"));

    if (status == CPLD_JTAG_DIS) {
        cpld_device->gpio_st = GPIO_ORIG_STATE;
        cpld_gpio_tbl = cpld_orig_gpio_tbl;
    } else {
        cpld_device->gpio_st = GPIO_JTAG_STATE;
        if (!pin_type)
            cpld_gpio_tbl = cpld_jtag_gpio_tbl0;
        else
            cpld_gpio_tbl = cpld_jtag_gpio_tbl1;
    }

    for (i = 0; i < CPLD_GPIO_MAX; i++)
        gpio_config_pin((cpld_gpio_tbl+i));
}

static int cpld_gpio_jtag_io(struct jtag_io_ioctl_param* io_param)
{
    int rc = CPLD_JTAG_IO_OK;
#ifdef CPLD_WARNING
	int tck1, tck2;
#endif

    CPLD_LOG_DBG(4, "%s\n", __FUNCTION__);

#ifdef CPLD_WARNING
	/* Read TCK */
	tck1 = gpio_read_bit(io_param->tck.pin);
#endif

	/* Write TMS */
    gpio_write_bit(io_param->tms.pin, io_param->tms.data);

#ifdef CPLD_WARNING
	/* Read TCK */
    tck2 = gpio_read_bit(io_param->tck.pin);

	/* Make sure TCK value did not change */
	if (tck1 != tck2) {
        rc = CPLD_JTAG_IO_WARN;
        cpld_device->stat.warn_num++;
        CPLD_LOG_ERROR("TCK VALUE CHANGED AFTER CHANGING TMS.\n");
    }
#endif

	/* Write TDI */
    gpio_write_bit(io_param->tdi.pin, io_param->tdi.data);

#ifdef CPLD_WARNING
	/* Read TCK */
	tck2 = gpio_read_bit(io_param->tck.pin);

	/* Make sure TCK value did not change */
	if (tck1 != tck2) {
        rc = CPLD_JTAG_IO_WARN;
        cpld_device->stat.warn_num++;
		CPLD_LOG_ERROR("TCK VALUE CHANGED AFTER CHANGING TDI!\n");
    }
#endif

	/* Write TCK */
    gpio_write_bit(io_param->tck.pin, io_param->tck.data);

	return rc;
}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
static int cpld_device_ioctl(struct inode *inode,
#else
static long cpld_device_ioctl(
#endif
                              struct file *file,
                              unsigned int ioctl_num,
                              unsigned long ioctl_param)
{
	int rc = 0;
    char ret;
	void *ptr;
    struct read_ioctl_param read_params;
	struct write_ioctl_param write_params;
    struct jtag_io_ioctl_param io_params;
    struct jtag_set_ioctl_param jtag_set_params;
    struct cpld_select_ioctl_param cpld_select_params;

	switch (ioctl_num) {
    case IOCTL_CPLD_JTAG_SET:
        cpld_device->stat.ioctl_jtag_set_num++;
        if (copy_from_user(&jtag_set_params,
			   	   (void __user *)ioctl_param,
				   sizeof(jtag_set_params))) {
			rc = -EFAULT;
			CPLD_LOG_ERROR("IOCTL_JTAG_SET: failed on copy_from_user\n");
			break;
		}
        CPLD_LOG_DBG(1, "IOCTL_JTAG_SET: %s\n", (jtag_set_params.status ? "enable" : "disable"));
        cpld_gpio_jtag_config(jtag_set_params.status);

#ifndef CONFIG_X86
        /* IO expander can be requested from standalone jam application.*/
        if (jtag_enable) {
            if (cpld_set_jtag(jtag_set_params.status)) {
                CPLD_LOG_ERROR("Failed %s cpld jtag\n", (jtag_set_params.status ? "enable" : "disable"));
                rc = -EFAULT;
            }
        }
#endif
        break;

    case IOCTL_CPLD_GPIO_READ:
        cpld_device->stat.ioctl_read_num++;
		if (copy_from_user(&read_params,
				(void __user *)ioctl_param,
				sizeof(read_params))) {
			rc = -EFAULT;
			CPLD_LOG_ERROR("IOCTL_GPIO_READ: failed on copy_from_user\n");
			break;
		}

		if (read_params.dir == GPIO_IN)
#ifdef CONFIG_X86
			read_params.data = gpio_read_bit(read_params.pin);
#else
            read_params.data = gpio_read_in_bit(read_params.pin);
#endif
		else
			read_params.data = gpio_read_bit(read_params.pin);

        CPLD_LOG_DBG(3, "IOCTL_GPIO_READ: pin=%d dir=%d data=%d\n", \
                     read_params.pin, read_params.dir, read_params.data);

		ptr = &(((struct read_ioctl_param*)ioctl_param)->data);
		if (copy_to_user((void __user *)ptr,
			  	&read_params.data, sizeof(read_params.data))) {
			rc = -EFAULT;
            CPLD_LOG_ERROR("IOCTL_GPIO_READ: failed on copy_to_user\n");
		}
		break;

    case IOCTL_CPLD_GPIO_WRITE:
		if (copy_from_user(&write_params,
				(void __user *)ioctl_param,
				sizeof(write_params))) {
			rc = -EFAULT;
			CPLD_LOG_ERROR("IOCTL_GPIO_WRITE: failed on copy_from_user\n");;
		}
        else {
            CPLD_LOG_DBG(3, "IOCTL_GPIO_WRITE: pin=%d data=%d\n",
                         write_params.pin, write_params.data);
            gpio_write_bit(write_params.pin, write_params.data);
        }
		break;

    case IOCTL_CPLD_JTAG_IO:
        cpld_device->stat.ioctl_jtag_io_num++;
        if (copy_from_user(&io_params,
				(void __user *)ioctl_param,
				sizeof(io_params))) {
			rc = -EFAULT;
			CPLD_LOG_ERROR("IOCTL_CPLD_JTAG_IO: failed on copy_from_user\n");
		}
        else {
            CPLD_LOG_DBG(3, "IOCTL_CPLD_JTAG_IO tms: pin=%d data=%d, "\
                         "tdi: pin=%d data=%d, tck: pin=%d data=%d\n", \
                         io_params.tms.pin, io_params.tms.data, io_params.tdi.pin, \
                         io_params.tdi.data, io_params.tck.pin, io_params.tck.data);
            ret = (char)cpld_gpio_jtag_io(&io_params);
            ptr = &(((struct jtag_io_ioctl_param*)ioctl_param)->rc);
            if (copy_to_user((void __user *)ptr, &ret, sizeof(io_params.rc))) {
    			rc = -EFAULT;
                CPLD_LOG_ERROR("IOCTL_CPLD_JTAG_IO: failed on copy_to_user\n");
    		}
            CPLD_LOG_DBG(3, "IOCTL_CPLD_JTAG_IO rc %d\n", ret);
        }
        break;

    case IOCTL_CPLD_SELECT:
        cpld_device->stat.ioctl_cpld_select_num++;
        if (copy_from_user(&cpld_select_params,
			   	   (void __user *)ioctl_param,
				   sizeof(cpld_select_params))) {
			rc = -EFAULT;
			CPLD_LOG_ERROR("IOCTL_CPLD_SELECT: failed on copy_from_user\n");
			break;
		}
        CPLD_LOG_DBG(1, "IOCTL_CPLD_SELECT: cpld %d %s\n", cpld_select_params.cpld, (cpld_select_params.status ? "enable" : "disable"));
        cpld_gpio_jtag_config(cpld_select_params.status);
        if (cpld_select(cpld_select_params.cpld, cpld_select_params.status)) {
            rc = -EFAULT;
            CPLD_LOG_ERROR("IOCTL_CPLD_SELECT: failed on cpld access\n");
        }
        break;

    default:
        CPLD_LOG_ERROR("Unknown ioctl num %d\n", ioctl_num);
        rc = -ENOTTY;
		break;
	}

	return rc;
}

ssize_t show_dbg_lvl(struct device *dev, struct device_attribute *attr, char *buf)
{
    ssize_t rc;

    rc = sprintf(buf, "CPLD dbg lvl: %d\n", dbg_lvl);
    return rc+1;
}

ssize_t store_dbg_lvl(struct device *dev, struct device_attribute *attr, const char *buf, size_t cnt)
{
    u32 lvl;
    s32 rc;

    lvl = (u32)simple_strtol(buf, NULL, 10);
    if (lvl > CPLD_DBG_MAX_LVL) {
        CPLD_LOG_ERROR("Incorrect input dbg lvl %d > %d\n", lvl, CPLD_DBG_MAX_LVL);
        rc = 0;
    }
    else {
        dbg_lvl = lvl;
        CPLD_LOG_DBG(1, "CPLD dbg level has been changed to %d\n", lvl);
    }

    return cnt;
}

ssize_t show_info(struct device *dev, struct device_attribute *attr, char *buf)
{
    ssize_t cnt;

#ifdef CPLD_WARNING
    cnt = sprintf(buf, "CPLD info:\nioctl_jtag_set_num:\t%u\n"\
                  "ioctl_cpld_select_num:\t%u\nioctl_jtag_io_num:\t%u\nwarning:\t\t%u\n"\
                  "ioctl_read_num:\t\t%u\ngpio state:\t\t%s\n",\
                  cpld_device->stat.ioctl_jtag_set_num, cpld_device->stat.ioctl_cpld_select_num, \
                  cpld_device->stat.ioctl_jtag_io_num, cpld_device->stat.warn_num,\
                  cpld_device->stat.ioctl_read_num, \
                  (cpld_device->gpio_st == GPIO_JTAG_STATE) ? "jtag" : "original");
#else
    cnt = sprintf(buf, "CPLD info:\nioctl_jtag_set_num:\t%u\n"\
                  "ioctl_cpld_select_num:\t%u\nioctl_jtag_io_num:\t%u\n"\
                  "ioctl_read_num:\t\t%u\ngpio state:\t\t%s\n",\
                  cpld_device->stat.ioctl_jtag_set_num, cpld_device->stat.ioctl_cpld_select_num, \
                  cpld_device->stat.ioctl_jtag_io_num, cpld_device->stat.ioctl_read_num, \
                  (cpld_device->gpio_st == GPIO_JTAG_STATE) ? "jtag" : "original");
#endif
    return cnt+1;
}

#ifndef CONFIG_X86
ssize_t show_id_vers(struct device *dev, struct device_attribute *attr, char *buf)
{
    u8 id, vers;
    ssize_t cnt;
    int rc;

    rc = cpld_i2c_read(cpld_device->adapt, cpld_device->i2c_addr, CPLD_ID_REG, &id, 1);
    if (rc) {
        CPLD_LOG_ERROR("Read fail, Reg: 0x%x, Error %d\n", CPLD_ID_REG, rc);
        cnt = sprintf(buf, "Fail read reg: 0x%x", CPLD_ID_REG);
        return cnt+1;
    }

    rc = cpld_i2c_read(cpld_device->adapt, cpld_device->i2c_addr, CPLD_VER_REG, &vers, 1);
    if (rc) {
        CPLD_LOG_ERROR("Read fail, Reg: 0x%x, Error %d\n", CPLD_VER_REG, rc);
        cnt = sprintf(buf, "Fail read reg: 0x%x", CPLD_VER_REG);
        return cnt+1;
    }

    cnt = sprintf(buf, "CPLD ID: 0x%x, VERS: 0x%x\n", id, vers);

    return cnt+1;
}
ssize_t show_params(struct device *dev, struct device_attribute *attr, char *buf)
{
    ssize_t cnt;
    char cpld_str[40];
    char io_exp_str[40];

    if (cpld_device->adapt)
        sprintf(cpld_str, "cpld i2c_addr 0x%x/0x%x", cpld_device->i2c_bus, cpld_device->i2c_addr);
    else
        sprintf(cpld_str, "cpld_i2c not initialised");

    if (cpld_device->io_adapt)
        sprintf(io_exp_str, "io expander i2c_addr 0x%x/0x%x", cpld_device->io_i2c_bus, cpld_device->io_i2c_addr);
    else
        sprintf(io_exp_str, "io expander not initialised");

    cnt = sprintf(buf, "CPLD driver params:\n%s\n%s\n", cpld_str, io_exp_str);

    return cnt+1;
}

ssize_t show_cpld_reg(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
    u8 reg, val;
    ssize_t cnt;
    int rc;

    reg = (u32)simple_strtol(&(kobj->name[strlen("reg")]), NULL, 10);

    rc = cpld_i2c_read(cpld_device->adapt, cpld_device->i2c_addr, reg, &val, 1);
    if (!rc) {
        CPLD_LOG_DBG(3, "Read: Reg: 0x%x Val: 0x%x\n", reg, val);
        cnt = sprintf(buf, "Reg: 0x%x = 0x%x\n", reg, val);
    }
    else {
        CPLD_LOG_ERROR("Read fail, Reg: 0x%x, Error %d\n", reg, rc);
        cnt = sprintf(buf, "Fail read reg: 0x%x", reg);
    }

    return cnt+1;
}

ssize_t store_cpld_reg(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t cnt)
{
    u8 reg, val;
    s32 rc;

    if (!buf || cnt == 1) {
        CPLD_LOG_ERROR("Incorrect empty str input\n");
        return cnt;
    }

    reg = (u32)simple_strtol(&(kobj->name[strlen("reg")]), NULL, 10);
    val = (u8)simple_strtol(buf, NULL, 16);

    rc = cpld_i2c_write(cpld_device->adapt, cpld_device->i2c_addr, reg, val, 1);
    if (!rc) {
        CPLD_LOG_DBG(3, "Write: Reg: 0x%x Val: 0x%x\n", reg, val);
    } else {
        CPLD_LOG_ERROR("Write fail, Reg: 0x%x, Error %d\n", reg, rc);
    }

    return cnt;
}
#endif

static DEVICE_ATTR(dbg_lvl, 0644, show_dbg_lvl, store_dbg_lvl);
static DEVICE_ATTR(info, 0444, show_info, NULL);

#ifndef CONFIG_X86
static DEVICE_ATTR(id_vers, 0444, show_id_vers, NULL);
static DEVICE_ATTR(params, 0444, show_params, NULL);

static struct attribute *cpld_attributes[] = {
	&dev_attr_dbg_lvl.attr,
    &dev_attr_id_vers.attr,
    &dev_attr_info.attr,
    &dev_attr_params.attr,
	NULL
};

static struct kobj_attribute reg_val_attr = __ATTR(val, 0644, show_cpld_reg, store_cpld_reg);
#else
static struct attribute *cpld_attributes[] = {
	&dev_attr_dbg_lvl.attr,
    &dev_attr_info.attr,
	NULL
};
#endif

static struct file_operations cpld_fops = {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
	.unlocked_ioctl = cpld_device_ioctl,
#else
	.ioctl = cpld_device_ioctl,
#endif
};

static struct platform_driver cpld_driver = {
	.driver = {
		.name = "cpld_jtag",
		.owner = THIS_MODULE,
	}
};

#ifndef CONFIG_X86
static int cpld_sysfs_reg_create(u8 reg)
{
    struct kobject *reg_kobj;
    char name[8];
    s32 rc;

    sprintf(name, "reg%02d", reg);
    reg_kobj = kobject_create_and_add(name, cpld_device->reg_dir);
    if (!reg_kobj) {
        CPLD_LOG_ERROR("Create sysfs %s entry failed\n", name);
        return CPLD_RC_FAILURE;
    }
    else {
        cpld_device->reg[reg] = reg_kobj;
        rc = sysfs_create_file(reg_kobj, &reg_val_attr.attr);
        if (rc) {
            CPLD_LOG_ERROR("Create %s sysfs file for %s failed, error %d\n", reg_val_attr.attr.name, name, rc);
            return CPLD_RC_FAILURE;
        }
    }

    return CPLD_RC_OK;
}

static void cpld_sysfs_reg_remove(u8 reg)
{
    sysfs_remove_file(cpld_device->reg[reg], &reg_val_attr.attr);
    kobject_del(cpld_device->reg[reg]);
    cpld_device->reg[reg] = NULL;
}
#endif
static int cpld_sysfs_create(void)
{
#ifndef CONFIG_X86
    int i, rc;
#endif

    cpld_device->attr_grp.attrs = cpld_attributes;
	if (sysfs_create_group(&(cpld_device->plat_dev->dev.kobj), &cpld_device->attr_grp)) {
        CPLD_LOG_ERROR("Failed to create sysfs cpld_jtag group\n");
		return CPLD_RC_FAILURE;
    }

#ifndef CONFIG_X86
    cpld_device->reg_dir = kobject_create_and_add("reg", &(cpld_device->plat_dev->dev.kobj));
    if (!cpld_device->reg_dir) {
        CPLD_LOG_ERROR("Failed to create sysfs register dir\n");
        return -EACCES;
    }

    for (i = 0; i < CPLD_MAX_REG; i++) {
        rc = cpld_sysfs_reg_create(i);
        if (rc)
            return rc;
    }
#endif

    return CPLD_RC_OK;
}

static void cpld_sysfs_remove(void)
{
#ifndef CONFIG_X86
    int i;

    for (i = 0; i < CPLD_MAX_REG; i++)
        cpld_sysfs_reg_remove(i);
    kobject_del(cpld_device->reg_dir);

#endif
    sysfs_remove_group(&(cpld_device->plat_dev->dev.kobj), &cpld_device->attr_grp);
}

static int __init cpld_init(void)
{
	int rc;

    cpld_device = kzalloc(sizeof(struct cpld_device_t), GFP_KERNEL);
    if (!cpld_device) {
        CPLD_LOG_ERROR("Failed to allocate cpld_device\n");
        return -ENOMEM;
    }

#ifndef CONFIG_X86
    /* Initialize GPIO handling */
	rc = cpld_gpio_init();
	if (rc < 0) {
		CPLD_LOG_ERROR("Failed to init GPIO\n");
        kfree(cpld_device);
		return rc;
	}
    /* Initialize CPLD handling */
    /* Initialize CPLD handling */
	rc = cpld_data_init();
	if (rc < 0) {
		CPLD_LOG_ERROR("Failed to init CPLD\n");
        kfree(cpld_device);
		return rc;
	}
#endif
    cpld_device->gpio_st = GPIO_ORIG_STATE;

    rc = platform_driver_register(&cpld_driver);
	if (rc) {
        CPLD_LOG_ERROR("Failed to register cpld platform driver\n");
        kfree(cpld_device);
		return rc;
    }

    cpld_device->plat_dev = platform_device_alloc("cpld_jtag", -1);
	if (!cpld_device->plat_dev) {
		rc = -ENOMEM;
        CPLD_LOG_ERROR("Failed to alloc cpld platform device\n");
		goto fail_platform_driver;
	}

	rc = platform_device_add(cpld_device->plat_dev);
	if (rc) {
        CPLD_LOG_ERROR("Failed to add cpld platform device\n");
		goto fail_platform_device1;
    }

    if (cpld_sysfs_create()) {
        CPLD_LOG_ERROR("Failed to create sysfs cpld group\n");
		goto fail_platform_device2;
    }


    /* Register character device */
	rc = register_chrdev(CPLD_JTAG_DEVICE_MAJOR, CPLD_DEVICE_NAME, &cpld_fops);
	if (rc < 0) {
		CPLD_LOG_ERROR("Failed to register character device\n");
		goto fail;
	}

    CPLD_LOG_DBG(3, "%s initialised\n", CPLD_DEVICE_NAME);
	return 0;

fail:
    unregister_chrdev(CPLD_JTAG_DEVICE_MAJOR, CPLD_DEVICE_NAME);

fail_platform_device2:
	platform_device_del(cpld_device->plat_dev);

fail_platform_device1:
	platform_device_put(cpld_device->plat_dev);

fail_platform_driver:
    platform_driver_unregister(&cpld_driver);

    kfree(cpld_device);

    return rc;
}

static void __exit cpld_exit(void)
{
    unregister_chrdev(CPLD_JTAG_DEVICE_MAJOR, CPLD_DEVICE_NAME);

    cpld_sysfs_remove();

	platform_device_unregister(cpld_device->plat_dev);
	platform_driver_unregister(&cpld_driver);

    if (cpld_device->gpio_st == GPIO_JTAG_STATE)
        cpld_gpio_jtag_config(CPLD_JTAG_DIS);

#ifndef CONFIG_X86
    cpld_gpio_release();
#endif

    kfree(cpld_device);
}

module_init(cpld_init);
module_exit(cpld_exit);

MODULE_AUTHOR("Mellanox Technologies (Michael Shych)");
MODULE_DESCRIPTION(CPLD_MOD_DESCRIPTION);
MODULE_VERSION(CPLD_MOD_VERSION);
MODULE_LICENSE("GPL v2");
MODULE_PARM_DESC(dbg_lvl, " debug level, 0-4, 0 - disabled");
#ifndef CONFIG_X86
MODULE_PARM_DESC(pin_type, " pin type, 0 - dflt:TD0=28,TDI=29; 1: vice versa");
MODULE_PARM_DESC(jtag_enb_mode, "jtag enable mode, 0 - dflt (Jackal), 1 - new (Coyote)");
MODULE_PARM_DESC(cpld_i2c_addr, " cpld i2c bus|address");
MODULE_PARM_DESC(cpld_io_port, " io expander i2c bus|address");
MODULE_PARM_DESC(jtag_enable, " set gpio cpld jtag enable");
#endif


