/*
 *
 * Copyright (C) Mellanox Technologies Ltd. 2001-2014.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/err.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <linux/workqueue.h>
#include <linux/device.h>
#include <linux/platform_device.h>

#include <linux/init.h>
#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/gpio.h>
#include <linux/spinlock.h>
#include <linux/netlink.h>
#include <asm/param.h>
#include <net/net_namespace.h>

#include "gpio_ifc.h"
#include "gpio_pch.h"

#define GPIO_MOD_DESCRIPTION    "x86 PCH GPIO driver. Build:" " "__DATE__" "__TIME__
#define GPIO_MOD_VERSION       "1.0.0"

int dbg_lvl = GPIO_DFLT_DBG_LVL;
module_param(dbg_lvl, int, 0644);

struct gpio_device_t* gpio_device;

static int gpio_sysfs_grp_create(u8 gpio);
static void gpio_sysfs_grp_remove(u8 gpio);

ssize_t show_dbg_lvl(struct device *dev, struct device_attribute *attr, char *buf)
{
    ssize_t rc;

    rc = sprintf(buf, "GPIO dbg lvl: %d\n", dbg_lvl);
    return rc+1;
}

ssize_t store_dbg_lvl(struct device *dev, struct device_attribute *attr, const char *buf, size_t cnt)
{
    u32 lvl;
    s32 rc;
   
    lvl = (u32)simple_strtol(buf, NULL, 10);
    if (lvl > GPIO_DBG_MAX_LVL) {
        GPIO_LOG_ERROR("Incorrect input dbg lvl %d > %d\n", lvl, GPIO_DBG_MAX_LVL);
        rc = 0;
    }
    else {
        dbg_lvl = lvl;
        GPIO_LOG_DBG(2, "GPIO dbg level has been changed to %d\n", lvl);
    }

    return cnt;
}

ssize_t store_export(struct device *dev, struct device_attribute *attr, const char *buf, size_t cnt)
{
    u32 gpio;

    gpio = (u32)simple_strtol(buf, NULL, 10);
    if (gpio >= GPIO_MAX_NUM)
        GPIO_LOG_ERROR("Incorrect gpio %d > %d\n", gpio, (GPIO_MAX_NUM-1));
    else {
        if (gpio_device->gpio_sysfs[gpio])
            GPIO_LOG_ERROR("Sysfs entry for gpio %d is already exist\n", gpio);
        else if (gpio_sysfs_grp_create(gpio))
                GPIO_LOG_ERROR("Failed to create sysfs entry of gpio %d\n", gpio);
            else
                GPIO_LOG_DBG(2, "Created sysfs entry of gpio %d\n", gpio);
    }

    return cnt;
}

ssize_t store_unexport(struct device *dev, struct device_attribute *attr, const char *buf, size_t cnt)
{
    u32 gpio;
    
    gpio = (u32)simple_strtol(buf, NULL, 10);
    if (gpio > GPIO_MAX_NUM)
        GPIO_LOG_ERROR("Incorrect gpio %d > %d\n", gpio, GPIO_MAX_NUM);
    else {
        if (!gpio_device->gpio_sysfs[gpio])
            GPIO_LOG_ERROR("Sysfs entry for gpio %d doesn't exist\n", gpio);
        else {
            gpio_sysfs_grp_remove(gpio);
            GPIO_LOG_DBG(2, "Removed sysfs entry of gpio %d\n", gpio);
        }
    }

    return cnt;
}

ssize_t show_use_sel(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
    u8 gpio;
    ssize_t cnt;

    gpio = (u32)simple_strtol(&(kobj->name[strlen("gpio")]), NULL, 10);
    if (gpio_is_gpio(gpio))
        cnt = sprintf(buf, "GPIO %d used as GPIO\n", gpio);
    else
        cnt = sprintf(buf, "GPIO %d used in native mode\n", gpio);

    return cnt+1;
}

ssize_t store_use_sel(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t cnt)
{
    u8 gpio, pin_num, val;
    u32 reg_val;
   
    if (!buf || cnt == 1) {
        GPIO_LOG_ERROR("Incorrect empty str input\n");
        return cnt;
    }

    gpio = (u32)simple_strtol(&(kobj->name[strlen("gpio")]), NULL, 10);
    pin_num = gpio % 32;
    val = (u8)simple_strtol(buf, NULL, 10);
    spin_lock(&gpiopch_lock);
    reg_val = gpio_read_reg(gpio, GPIO_USE_SEL);
    if (val)
        reg_val |= (1<<pin_num);
    else
        reg_val &= ~(1<<pin_num);
    gpio_write_reg(gpio, GPIO_USE_SEL, reg_val);
    spin_unlock(&gpiopch_lock);
    GPIO_LOG_DBG(3, "USE_SEL of gpio %d changed to %s mode\n", gpio,\
                 (val ? "GPIO" : "Native"));

    return cnt;
}

ssize_t show_io_sel(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
    u8 gpio;
    ssize_t cnt;

    gpio = (u32)simple_strtol(&(kobj->name[strlen("gpio")]), NULL, 10);
    if (gpio_is_gpio(gpio)) {
        if (gpio_is_dir_in(gpio))
            cnt = sprintf(buf, "GPIO %d signal direction input\n", gpio);
        else
            cnt = sprintf(buf, "GPIO %d signal direction output\n", gpio);
    } else
        cnt = sprintf(buf, "GPIO %d used in native mode\n", gpio);

    return cnt+1;
}

ssize_t store_io_sel(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t cnt)
{
    u8 gpio, pin_num, val;
    u32 reg_val;
   
    if (!buf || cnt == 1) {
        GPIO_LOG_ERROR("Incorrect empty str input\n");
        return cnt;
    }

    gpio = (u32)simple_strtol(&(kobj->name[strlen("gpio")]), NULL, 10);
    pin_num = gpio % 32;
    val = (u8)simple_strtol(buf, NULL, 10);
    spin_lock(&gpiopch_lock);
    reg_val = gpio_read_reg(gpio, GPIO_IO_SEL);
    if (val)
        reg_val |= (1<<pin_num);
    else
        reg_val &= ~(1<<pin_num);
    gpio_write_reg(gpio, GPIO_IO_SEL, reg_val);
    spin_unlock(&gpiopch_lock);
    GPIO_LOG_DBG(3, "IO_SEL of gpio %d changed to %s direction\n", gpio,\
                 (val ? "Input" : "Output"));

    return cnt;
}

ssize_t show_gpio_lvl(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
    u8 gpio, dir, lvl;
    ssize_t cnt;

    gpio = (u32)simple_strtol(&(kobj->name[strlen("gpio")]), NULL, 10);
    if (gpio_is_gpio(gpio)) {
        dir = gpio_is_dir_in(gpio);
        lvl = gpio_read_bit(gpio);
        cnt = sprintf(buf, "GPIO %d %s signal lvl is %d\n", gpio,\
                      (dir ? "input" : "output"), lvl);
    } else
        cnt = sprintf(buf, "GPIO %d used in native mode\n", gpio);

    return cnt+1;
}

ssize_t store_gpio_lvl(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t cnt)
{
    u8 gpio, pin_num, val;
   
    if (!buf || cnt == 1) {
        GPIO_LOG_ERROR("Incorrect empty str input\n");
        return cnt;
    }

    gpio = (u32)simple_strtol(&(kobj->name[strlen("gpio")]), NULL, 10);
    if (gpio_is_gpio(gpio)) {
        if (!gpio_is_dir_in(gpio)) {
            pin_num = gpio % 32;
            val = (u8)simple_strtol(buf, NULL, 10);
            gpio_write_bit(gpio,val);
            GPIO_LOG_DBG(3, "GPIO %d output level set to %d\n", gpio,\
                         (val ? 1 : 0));
        }
        else
            GPIO_LOG_DBG(3, "GPIO %d is configured in input direction\n", gpio);      
    }
    else
        GPIO_LOG_DBG(3, "GPIO %d used in native mode\n", gpio);
    
    return cnt;
}

ssize_t show_gpio_info(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
    u8 gpio, pin_num, use_sel, io_sel, lvl;
    ssize_t cnt; 

    gpio = (u8)simple_strtol(&(kobj->name[strlen("gpio")]), NULL, 10);
    pin_num = gpio % 32;
    GPIO_LOG_DBG(3, "Gpio %d, pin %d\n", gpio, pin_num);

    use_sel = gpio_is_gpio(gpio);
    if (use_sel == GPIO_SIG) {
        io_sel = gpio_is_dir_in(gpio);
        lvl = gpio_read_bit(gpio);
        cnt = sprintf(buf, "GPIO %d, GPIO mode, Dir: %s, Lvl: %d\n", \
                      gpio, (io_sel ? "In" : "Out"), (lvl ? 1 : 0));
    }
    else    
        cnt = sprintf(buf, "GPIO %d, Native mode\n", gpio);

    return cnt+1;
}

static DEVICE_ATTR(dbg_lvl, 0644, show_dbg_lvl, store_dbg_lvl);
static DEVICE_ATTR(export, 0222, NULL, store_export);
static DEVICE_ATTR(unexport, 0222, NULL, store_unexport);

static struct attribute *gpio_attributes[] = {
	&dev_attr_dbg_lvl.attr,
	&dev_attr_export.attr,
    &dev_attr_unexport.attr,
	NULL
};

static struct kobj_attribute use_sel_attr = __ATTR(use_sel, 0644, show_use_sel, store_use_sel);
static struct kobj_attribute io_sel_attr = __ATTR(io_sel, 0644, show_io_sel, store_io_sel);
static struct kobj_attribute lvl_attr = __ATTR(lvl, 0644, show_gpio_lvl, store_gpio_lvl);
static struct kobj_attribute info_attr = __ATTR(info, 0444, show_gpio_info, NULL);

static struct attribute *gpio_attrs[] = {
	&use_sel_attr.attr,
	&io_sel_attr.attr,
	&lvl_attr.attr,
    &info_attr.attr,
    NULL
};

static struct attribute_group gpio_attr_group = {
        .attrs = gpio_attrs,
};

static int gpio_sysfs_grp_create(u8 gpio)
{
    struct kobject *gpio_kobj;
    char name[8];
    int rc;

    sprintf(name, "gpio%d", gpio);
    gpio_kobj = kobject_create_and_add(name, gpio_device->gpio_dir);
    if (!gpio_kobj) {
        GPIO_LOG_ERROR("Create sysfs %s entry failed\n", name);
        return GPIO_RC_FAILURE;
    }
    else {        
        rc = sysfs_create_group(gpio_kobj, &gpio_attr_group);
        if (rc) {
            GPIO_LOG_ERROR("Create sysfs group for %s failed, error %d\n", name, rc);
            kobject_del(gpio_kobj);
            return GPIO_RC_FAILURE;
        }
        gpio_device->gpio_sysfs[gpio] = gpio_kobj;            
    }

    return GPIO_RC_OK;
}

static void gpio_sysfs_grp_remove(u8 gpio)
{
    sysfs_remove_group(gpio_device->gpio_sysfs[gpio], &gpio_attr_group);

    kobject_del(gpio_device->gpio_sysfs[gpio]);
    gpio_device->gpio_sysfs[gpio] = NULL;
}

static int gpio_sysfs_create(void)
{
    gpio_device->gpio_attr_grp.attrs = gpio_attributes;
	if (sysfs_create_group(&(gpio_device->plat_dev->dev.kobj), &gpio_device->gpio_attr_grp)) {
        GPIO_LOG_ERROR("Failed to create sysfs gpio group\n");
		return -EACCES;
    }

    gpio_device->gpio_dir = kobject_create_and_add("gpio_pins", &(gpio_device->plat_dev->dev.kobj));
    if (!gpio_device->gpio_dir) {
        GPIO_LOG_ERROR("Failed to create sysfs gpio dir\n");
        return -EACCES;
    }

    return GPIO_RC_OK;
}

static void gpio_sysfs_remove(void)
{
    int i;

    sysfs_remove_group(&(gpio_device->plat_dev->dev.kobj), &gpio_device->gpio_attr_grp);
    for (i = 0; i < GPIO_MAX_NUM; i++) {
        if (gpio_device->gpio_sysfs[i])
            gpio_sysfs_grp_remove(i);
    }

    kobject_del(gpio_device->gpio_dir);
}

static int __init gpio_init(void)
{
	int rc;
    struct pci_dev* pdev = NULL;

    /* Initialize GPIO handling */
    pdev = pci_get_bus_and_slot(LPC_IFC_BUS_ID, PCI_DEVFN(LPC_IFC_SLOT_ID, LPC_IFC_FUNC_ID));
    if (!pdev) {
        GPIO_LOG_ERROR("LPC controler bus:%d slot:%d func:%d doesn't found\n",\
                       LPC_IFC_BUS_ID, LPC_IFC_SLOT_ID, LPC_IFC_FUNC_ID);
        return -EFAULT;
    }
    GPIO_LOG_DBG(3, "PCI_dev found: 0x%x:0x%x 0x%x:0x%x\n", pdev->vendor, pdev->device,\
                 pdev->subsystem_vendor, pdev->subsystem_device);

    gpio_device = kzalloc(sizeof(struct gpio_device_t), GFP_KERNEL);
    if (!gpio_device) {
        GPIO_LOG_ERROR("Failed to allocate gpio_device\n");
        return -ENOMEM;
    }

	gpio_data_init(gpio_device, pdev);
    pci_dev_put(pdev);

    gpio_device->plat_dev = platform_device_alloc(GPIO_DEVICE_NAME, -1);
	if (!gpio_device->plat_dev) {
		rc = -ENOMEM;
        GPIO_LOG_ERROR("Failed to alloc gpio platform device\n");
		goto fail_platform_device1;
	}

	rc = platform_device_add(gpio_device->plat_dev);
	if (rc) {
        GPIO_LOG_ERROR("Failed to add gpio platform device\n");
		goto fail_platform_device1;
    }

    if (gpio_sysfs_create()) {
        GPIO_LOG_ERROR("Failed to create sysfs gpio group\n");
		goto fail_platform_device2;
    }

    gpio_device->resource = request_region(gpio_device->gpio_base, GPIO_IOLEN, GPIO_DEVICE_NAME);
    if (!gpio_device->resource) {
        rc = -EIO;
        GPIO_LOG_ERROR("Fail requesting ioregion 0x%x len 0x%x for %s\n", gpio_device->gpio_base, GPIO_IOLEN, "gpio");
        goto fail_platform_device2;
    }
    else
        GPIO_LOG_DBG(3, "Access to ioport 0x%x is granted\n", gpio_device->gpio_base);

    GPIO_LOG_DBG(3, "%s initialised\n", GPIO_DEVICE_NAME);
	return 0;

fail_platform_device2:
    if (gpio_device->plat_dev->resource)
        release_region(gpio_device->gpio_base, GPIO_IOLEN);
	platform_device_del(gpio_device->plat_dev);

fail_platform_device1:
    if (gpio_device->plat_dev)
        platform_device_put(gpio_device->plat_dev);

    kfree(gpio_device);

    return rc;
}

static void __exit gpio_exit(void)
{
    release_region(gpio_device->gpio_base, GPIO_IOLEN);

    gpio_sysfs_remove();

	platform_device_unregister(gpio_device->plat_dev);

    kfree(gpio_device);
}

module_init(gpio_init);
module_exit(gpio_exit);

MODULE_AUTHOR("Mellanox Technologies (Michael Shych)");
MODULE_DESCRIPTION(GPIO_MOD_DESCRIPTION);
MODULE_VERSION(GPIO_MOD_VERSION);
MODULE_LICENSE("GPL v2");

MODULE_PARM_DESC(dbg_lvl, " debug level, 0-4, 0 - disabled");
