/*
 *
 * Copyright (C) Mellanox Technologies Ltd. 2001-2014.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <asm/io.h>
#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/gpio.h>
#include <linux/platform_device.h>

#include "gpio_ifc.h"
#include "gpio_pch.h"

spinlock_t gpiopch_lock;
static DEFINE_SPINLOCK(access_gpiopch_lock);
static DEFINE_SPINLOCK(access_gpiodev_lock);

static u32 gpio_reg_offs[3][4] ={
    {
        GPIO_USE_SEL1_OFFS,
        GPIO_IO_SEL1_OFFS,
        GPIO_LVL1_OFFS,
        GPIO_RST1_OFFS
    },
    {
        GPIO_USE_SEL2_OFFS,
        GPIO_IO_SEL2_OFFS,
        GPIO_LVL2_OFFS,
        GPIO_RST2_OFFS
    },
    {
        GPIO_USE_SEL3_OFFS,
        GPIO_IO_SEL3_OFFS,
        GPIO_LVL3_OFFS,
        GPIO_RST3_OFFS
    }
};

inline u32 gpio_read_reg(u8 gpio, enum gpio_reg_type type)
{
    u8 grp_num = gpio / 32;
    u32 reg;

    reg = gpio_device->gpio_base + gpio_reg_offs[grp_num][type];
    GPIO_LOG_DBG(4, "gpio %d read reg 0x%x\n", gpio, reg);
    return inl(reg);
}

inline void gpio_write_reg(u8 gpio, enum gpio_reg_type type, u32 val)
{
    u8 grp_num = gpio / 32;
    u32 reg;

    reg = gpio_device->gpio_base + gpio_reg_offs[grp_num][type];
    GPIO_LOG_DBG(4, "gpio %d write reg 0x%x val 0x%x\n", gpio, reg, val);
    outl(val, reg);
}

int gpio_read_bit(u8 gpio)
{
    u8 pin_num = gpio % 32;
    u32 reg_val; 
    
    spin_lock(&gpiopch_lock);
	reg_val = gpio_read_reg(gpio, GPIO_LVL);
    spin_unlock(&gpiopch_lock);
    return (GPIO_VAL(pin_num, reg_val) ? 1 : 0);
}

void gpio_write_bit(u8 gpio, u8 val)
{
    u8 pin_num = gpio % 32;
    u32 reg_val; 

    spin_lock(&gpiopch_lock);
    reg_val = gpio_read_reg(gpio, GPIO_LVL);
    if (val)
        reg_val |= (1<<pin_num); 
    else
        reg_val &= ~(1<<pin_num);
	gpio_write_reg(gpio, GPIO_LVL, reg_val);
    spin_unlock(&gpiopch_lock);
}

/* Vector write of several (num) gpio's.
   All gpio should be in the same register (bitmap),
   i.e. 0-31, 32-63, 64-95 */
void gpio_write_vec(struct gpio_pin** gpios, u8 num)
{
    u8 pin_num, gpio, i;
    u32 reg_val;
    struct gpio_pin* curr;

    spin_lock(&gpiopch_lock);
    gpio = gpios[0]->gpio;
    reg_val = gpio_read_reg(gpio, GPIO_LVL);
    for (i=0; i<num; i++) {
        curr = gpios[i];
        pin_num = curr->gpio % 32;
        if (curr->val)
            reg_val |= (1<<pin_num); 
        else
            reg_val &= ~(1<<pin_num);
    }
    gpio_write_reg(gpio, GPIO_LVL, reg_val);
    spin_unlock(&gpiopch_lock);
}

int gpio_is_gpio(u8 gpio)
{
    u8 pin_num = gpio % 32;
    u32 reg_val;

    spin_lock(&gpiopch_lock);
    reg_val = gpio_read_reg(gpio, GPIO_USE_SEL);
    spin_unlock(&gpiopch_lock);
    return (GPIO_VAL(pin_num, reg_val) ? 1 : 0);
}

int gpio_is_dir_in(u8 gpio)
{
    u8 pin_num = gpio % 32;
    u32 reg_val;

    spin_lock(&gpiopch_lock);
    reg_val = gpio_read_reg(gpio, GPIO_IO_SEL);
    spin_unlock(&gpiopch_lock);
    return (GPIO_VAL(pin_num, reg_val) ? 1 : 0);
}

void gpio_config_pin(struct gpio_param_type* gpio_param)
{
    u8 pin_num;
    u32 reg_val;

    if (!gpio_param) {
        GPIO_LOG_ERROR("%s NULL param\n", __FUNCTION__);
        return;
    }
    pin_num = gpio_param->gpio %32;

    GPIO_LOG_DBG(1, "Config gpio: %d, mode %s, dir %s, out_val %d\n", gpio_param->gpio, \
                 gpio_param->mode == NATIVE_SIG ? "NATIVE_SIG" : "GPIO_SIG",\
                 gpio_param->dir == GPIO_OUT ? "GPIO_OUT" : "GPIO_IN",\
                 gpio_param->out_val);

    spin_lock(&gpiopch_lock);
    reg_val = gpio_read_reg(gpio_param->gpio, GPIO_USE_SEL);

    if (gpio_param->mode == NATIVE_SIG) {
        if (GPIO_VAL(pin_num, reg_val)) {
            reg_val &= ~(1<<pin_num);
            gpio_write_reg(gpio_param->gpio, GPIO_USE_SEL, reg_val);
        }
    }
    else {
        if (!GPIO_VAL(pin_num, reg_val)) {
            reg_val |= (1<<pin_num);
            gpio_write_reg(gpio_param->gpio, GPIO_USE_SEL, reg_val);
        }
        reg_val = gpio_read_reg(gpio_param->gpio, GPIO_IO_SEL);
        if (gpio_param->dir == GPIO_OUT){
            if (GPIO_VAL(pin_num, reg_val)) {
                reg_val &= ~(1<<pin_num);
                gpio_write_reg(gpio_param->gpio, GPIO_IO_SEL, reg_val);
            }
        }
        else {
            if (!GPIO_VAL(pin_num, reg_val)) {
                reg_val |= (1<<pin_num);
                gpio_write_reg(gpio_param->gpio, GPIO_IO_SEL, reg_val);
            }
        }
    }
    spin_unlock(&gpiopch_lock);
    if ((gpio_param->mode == GPIO_SIG) && (gpio_param->dir == GPIO_OUT) && (gpio_param->out_val != GPIO_OUT_NO_CHG))
        gpio_write_bit(gpio_param->gpio, gpio_param->out_val);
}

void gpio_data_init(struct gpio_device_t* gpio_dev, struct pci_dev* pdev)
{
    u32 val;

    pci_read_config_dword(pdev, GPIOBASE_OFFS, &val);
    gpio_dev->gpio_base = val & GPIOBASE_MASK;
    GPIO_LOG_DBG(3, "GPIO Base 0x%x, read val 0x%x\n", gpio_dev->gpio_base, val);

    gpio_dev->lock = access_gpiodev_lock;
    gpiopch_lock = access_gpiopch_lock;
}

EXPORT_SYMBOL(gpio_read_bit);
EXPORT_SYMBOL(gpio_write_bit);
EXPORT_SYMBOL(gpio_write_vec);
EXPORT_SYMBOL(gpio_is_gpio);
EXPORT_SYMBOL(gpio_is_dir_in);
EXPORT_SYMBOL(gpio_config_pin);
EXPORT_SYMBOL(gpiopch_lock);

