/**
 *
 * Copyright (C) 2011-2014, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <asm/uaccess.h>
#include <linux/delay.h>
#include <linux/mod_devicetable.h>
#include <linux/i2c.h>
#include <linux/gpio.h>
#include "mlnx_log.h"
#include "mlnx_if.h"
#include "mlnx_if_int.h"

static DEFINE_MUTEX(mlnx_ioctl_mutex);

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
static int mlnx_device_ioctl(struct inode *inode,
#else
static long mlnx_device_ioctl(
#endif
			     struct file *file,
			     unsigned int ioctl_num, 
			     unsigned long ioctl_param)
{
	int rc = 0;
	void *ptr = NULL;
	u8 ioctl_buffer[MLNX_IOCTL_MAX_SIZE];
	struct mlnx_cpld_ioctl_params *p_cpld_params =
		(struct mlnx_cpld_ioctl_params *)ioctl_buffer;
	struct mlnx_gpio_ioctl_params *p_gpio_params =
		(struct mlnx_gpio_ioctl_params *)ioctl_buffer;
	struct mlnx_i2c_sgmii_ioctl_params *p_i2c_sgmii_params =
		(struct mlnx_i2c_sgmii_ioctl_params *)ioctl_buffer;
	struct mlnx_i2c_switchx_ioctl_params *p_i2c_switchx_params =
		(struct mlnx_i2c_switchx_ioctl_params *)ioctl_buffer;
	struct mlnx_i2c_virt_bus_ioctl_params *p_i2c_virt_bus_params =
		(struct mlnx_i2c_virt_bus_ioctl_params *)ioctl_buffer;
	struct mlnx_i2c_devices_ioctl_params *p_i2c_devices_params =
		(struct mlnx_i2c_devices_ioctl_params *)ioctl_buffer;
	struct mlnx_i2c_switch_ioctl_params *p_i2c_switch_params =
		(struct mlnx_i2c_switch_ioctl_params *)ioctl_buffer;
	struct mlnx_i2c_virt_bus_scan_ioctl_params *p_i2c_scan_params =
		(struct mlnx_i2c_virt_bus_scan_ioctl_params *)ioctl_buffer;
	struct mlnx_gpio_i2c_ioctl_params *p_i2c_gpio_params =
		(struct mlnx_gpio_i2c_ioctl_params *)ioctl_buffer;
	struct mlnx_gpio_i2c_del_ioctl_params *p_i2c_gpio_del_params =
		(struct mlnx_gpio_i2c_del_ioctl_params *)ioctl_buffer;
	struct mlnx_i2c_one_device_ioctl_params *p_i2c_one_device_params =
		(struct mlnx_i2c_one_device_ioctl_params *)ioctl_buffer;
	struct mlnx_rdwr_smbus_ioctl_params_2 *p_smbus_params =
		(struct mlnx_rdwr_smbus_ioctl_params_2 *)ioctl_buffer; 
	struct mlnx_cpld_subscribe_ioctl_params *p_cpld_subscribe = 
		(struct mlnx_cpld_subscribe_ioctl_params *)ioctl_buffer;
	struct mlnx_cpld_set_ioctl_params *p_cpld_set =
		(struct mlnx_cpld_set_ioctl_params *)ioctl_buffer;
	struct mlnx_cpld_get_data_ioctl_params *p_cpld_get_data = 
		(struct mlnx_cpld_get_data_ioctl_params *)ioctl_buffer;
	struct mlnx_watchdog_config_ioctl_params *p_watchdog_config = 
		(struct mlnx_watchdog_config_ioctl_params *)ioctl_buffer;
	struct mlnx_watchdog_enable_disable_ioctl_params *p_watchdog_ena_dis =
		(struct mlnx_watchdog_enable_disable_ioctl_params *)ioctl_buffer;

	memset(&ioctl_buffer, 0, MLNX_IOCTL_MAX_SIZE);

	switch (ioctl_num) {
	case IOCTL_MLNX_I2C_READ:
 	case IOCTL_MLNX_I2C_WRITE:
		rc = mlnx_rw_i2c_device((struct mlnx_rdwr_i2c_ioctl_params *)ioctl_param);		 
		break;
	case IOCTL_MLNX_I2C_SGMII_READ:
		if (copy_from_user(p_i2c_sgmii_params, (void __user *)ioctl_param, sizeof(*p_i2c_sgmii_params))) {
			rc = -EFAULT;
			break;
		}
		/* Copy data read from I2C device to userspace */
		ptr = ((struct mlnx_i2c_sgmii_ioctl_params *)ioctl_param)->data;
		if (copy_to_user((void __user *)ptr, p_i2c_sgmii_params->data, sizeof(*p_i2c_sgmii_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_I2C_SGMII_WRITE:
		if (copy_from_user(p_i2c_sgmii_params, (void __user *)ioctl_param, sizeof(*p_i2c_sgmii_params))) {
			rc = -EFAULT;
			break;
		}
		//
		/* Copy data read from I2C device to userspace */
		//ptr = ((struct mlnx_rdwr_i2c_ioctl_params *)ioctl_param)->data;
		if (copy_to_user((void __user *)ptr, p_i2c_sgmii_params->data, p_i2c_sgmii_params->datalen)) {
			rc = -EFAULT;
			break;
		}
		break;
	case IOCTL_MLNX_I2C_SWITCHX_READ:
		if (copy_from_user(p_i2c_switchx_params, (void __user *)ioctl_param, sizeof(*p_i2c_switchx_params))) {
			rc = -EFAULT;
			break;
		}
		//
		/* Copy data read from I2C device to userspace */
		//ptr = ((struct mlnx_rdwr_i2c_ioctl_params *)ioctl_param)->data;
		if (copy_to_user((void __user *)ptr, p_i2c_switchx_params->data, p_i2c_switchx_params->datalen)) {
			rc = -EFAULT;
			break;
		}
		break;
	case IOCTL_MLNX_I2C_SWITCHX_WRITE:
		break;
	case IOCTL_MLNX_CPLD_RW:
		if (copy_from_user(p_cpld_params, (void __user *)ioctl_param, sizeof(*p_cpld_params))) {
			rc = -EFAULT;
			break;
		}
		//rc = mlnx_cpld_read_reg(p_cpld_params->reg, p_cpld_params->data);
		//if (rc < 0)
		//	return rc;
		//ptr = &(((struct mlnx_cpld_ioctl_params *)ioctl_param)->data);
		if (copy_to_user((void __user *)ptr, &p_cpld_params->data, sizeof(p_cpld_params->data))) {
			rc = -EFAULT;
			break;
		}	
		break;
	case IOCTL_MLNX_LPC_RW:
		rc = mlnx_rw_lpc_device((struct mlnx_lpc_ioctl_params *)ioctl_param);
		break;
	case IOCTL_MLNX_GPIO_SET_DIR:
		if (copy_from_user(p_gpio_params, (void __user *)ioctl_param, sizeof(*p_gpio_params))) {
			rc = -EFAULT;
			break;
		}
		/* Export gpio - this will result in the pseudo-files for pin n
	   	   showing up under /sys/class/gpio/gpio<n> as:
	   	   /sys/class/gpio/gpio<n>/direction
	   	   /sys/class/gpio/gpio<n>/value
	   	   For this pin should be set direction to either 'in' or 'out'.
	   	   And then pin could be set and get through GPIO set/get interface.
		*/
        	mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
	        	"%s(%d) pin=%x dir=%d data=%d flags=%x\n",
                	__FUNCTION__, __LINE__, p_gpio_params->pin, p_gpio_params->dir,
                	p_gpio_params->data, p_gpio_params->flags);


		rc = gpio_request(p_gpio_params->pin, "sysfs");
		if (rc < 0) {
	     		if (rc == -EBUSY)
	     			rc = 0;
	     		else
	     			rc = -EFAULT;
			break;
		}
		if (p_gpio_params->flags & MLNX_GPIO_EXPORT) {
			rc = gpio_export(p_gpio_params->pin, true);

        		mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
	        		"%s(%d) rc=%d\n", __FUNCTION__, __LINE__, rc);

			if (rc < 0) {
	     			rc = -EFAULT;
				break;
			}
		}
		else if (p_gpio_params->flags & MLNX_GPIO_UNEXPORT) {
			gpio_unexport(p_gpio_params->dir);
			break;
		}

		if (p_gpio_params->flags & MLNX_GPIO_DIR) {
                	if (p_gpio_params->dir == GPIO_HIGH)
				rc = gpio_direction_output(p_gpio_params->pin, 1);
                	else if ( (p_gpio_params->dir == GPIO_OUT) || (p_gpio_params->dir == GPIO_LOW) )
				rc = gpio_direction_output(p_gpio_params->pin, 0);
                	else
				rc = gpio_direction_input(p_gpio_params->pin);

        		mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
	        		"%s(%d) rc=%d\n", __FUNCTION__, __LINE__, rc);

			if (rc < 0) {
				gpio_unexport(p_gpio_params->dir);
	     			rc = -EFAULT;
			}
		}
		break;
	case IOCTL_MLNX_GPIO_READ:
		if (copy_from_user(p_gpio_params, (void __user *)ioctl_param, sizeof(*p_gpio_params))) {
			rc = -EFAULT; 
			break;
		}
		p_gpio_params->data = __gpio_get_value(p_gpio_params->pin);
		ptr = &(((struct mlnx_gpio_ioctl_params *)ioctl_param)->data);
		if (copy_to_user((void __user *)ptr, &p_gpio_params->data, sizeof(p_gpio_params->data))) {
			rc = -EFAULT;
			break;
		}	
		break; 
	case IOCTL_MLNX_GPIO_WRITE:
		if (copy_from_user(p_gpio_params, (void __user *)ioctl_param, sizeof(*p_gpio_params))) {
			rc = -EFAULT;
			break;
		}
		switch (p_gpio_params->data) {
			case MLNX_WRITE_0:
				__gpio_set_value(p_gpio_params->pin, p_gpio_params->data);
				break;
			case MLNX_WRITE_1:
				__gpio_set_value(p_gpio_params->pin, p_gpio_params->data);
				break;
			case MLNX_WRITE_0_1:
				__gpio_set_value(p_gpio_params->pin, 0);
				if (p_gpio_params->delay > 0)
					udelay(1000*p_gpio_params->delay);
				__gpio_set_value(p_gpio_params->pin, 1);
				break;
			case MLNX_WRITE_1_0:
				__gpio_set_value(p_gpio_params->pin, 1);
				if (p_gpio_params->delay > 0)
					udelay(1000*p_gpio_params->delay);
				__gpio_set_value(p_gpio_params->pin, 0);
				break;
			default:
				rc = -EFAULT;
				break;
		}
		break;
	case IOCTL_MLNX_ADD_I2C_VIRTUAL_BUSES:
		if (copy_from_user(p_i2c_virt_bus_params, (void __user *)ioctl_param, sizeof(*p_i2c_virt_bus_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_add_i2c_virtual_buses(p_i2c_virt_bus_params);
		if (copy_to_user((void __user *)ioctl_param, p_i2c_virt_bus_params, sizeof(*p_i2c_virt_bus_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_DELETE_I2C_VIRTUAL_BUSES:
		if (copy_from_user(p_i2c_switch_params, (void __user *)ioctl_param, sizeof(*p_i2c_switch_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_delete_i2c_virtual_buses(p_i2c_switch_params);
		break;
	case IOCTL_MLNX_ADD_I2C_ONE_DEVICE_TO_VBUS:
		if (copy_from_user(p_i2c_one_device_params, (void __user *)ioctl_param, sizeof(*p_i2c_one_device_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_add_one_device_to_vbus(p_i2c_one_device_params);
		break; 
	case IOCTL_MLNX_DELETE_I2C_ONE_DEVICE_FROM_VBUS:
		if (copy_from_user(p_i2c_one_device_params, (void __user *)ioctl_param, sizeof(*p_i2c_one_device_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_delete_one_device_from_vbus(p_i2c_one_device_params);
		break;  
	case IOCTL_MLNX_BLOCK_I2C_VIRTUAL_BUSES:
		if (copy_from_user(p_i2c_switch_params, (void __user *)ioctl_param, sizeof(*p_i2c_switch_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_block_i2c_virtual_buses(p_i2c_switch_params);
		break;
	case IOCTL_MLNX_UNBLOCK_I2C_VIRTUAL_BUSES:
		if (copy_from_user(p_i2c_switch_params, (void __user *)ioctl_param, sizeof(*p_i2c_switch_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_unblock_i2c_virtual_buses(p_i2c_switch_params);
		break;
	case IOCTL_MLNX_RESET_I2C_VIRTUAL_BUSES_AND_DEVICES:
		if (copy_from_user(p_i2c_switch_params, (void __user *)ioctl_param, sizeof(*p_i2c_switch_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_reset_i2c_virtual_buses_and_devices(p_i2c_switch_params);
		break;
	case IOCTL_MLNX_ADD_I2C_DEVICES_TO_VBUS:
		if (copy_from_user(p_i2c_devices_params, (void __user *)ioctl_param, sizeof(*p_i2c_devices_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_add_devices_to_vbus(p_i2c_devices_params);
		break;
	case IOCTL_MLNX_DELETE_I2C_DEVICES_FROM_VBUS:
		if (copy_from_user(p_i2c_switch_params, (void __user *)ioctl_param, sizeof(*p_i2c_switch_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_delete_devices_from_vbus(p_i2c_switch_params);
		break;
	case IOCTL_MLNX_TEST_I2C_VBUS:
		if (copy_from_user(p_i2c_scan_params, (void __user *)ioctl_param, sizeof(*p_i2c_scan_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_test_vbus_exist(p_i2c_scan_params);
		if (copy_to_user((void __user *)ioctl_param, p_i2c_scan_params, sizeof(*p_i2c_scan_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_TEST_I2C_DEV:
		if (copy_from_user(p_i2c_one_device_params, (void __user *)ioctl_param, sizeof(*p_i2c_one_device_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_test_device_exist(p_i2c_one_device_params);
		if (copy_to_user((void __user *)ioctl_param, p_i2c_one_device_params, sizeof(*p_i2c_one_device_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_SCAN_I2C_SWITCH_DEVICES:
		if (copy_from_user(p_i2c_scan_params, (void __user *)ioctl_param, sizeof(*p_i2c_scan_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_scan_switch_devices(p_i2c_scan_params);
		if (copy_to_user((void __user *)ioctl_param, p_i2c_scan_params, sizeof(*p_i2c_scan_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_ADD_I2C_GPIO_VIRTUAL_BUSES:
		if (copy_from_user(p_i2c_gpio_params, (void __user *)ioctl_param, sizeof(*p_i2c_gpio_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_add_gpio_i2c_buses(p_i2c_gpio_params);
		break;
	case IOCTL_MLNX_DELETE_I2C_GPIO_VIRTUAL_BUSES:
		if (copy_from_user(p_i2c_gpio_del_params, (void __user *)ioctl_param, sizeof(*p_i2c_gpio_del_params))) {
			rc = -EFAULT;
			break;
		}
		rc = mlnx_del_gpio_i2c_buses(p_i2c_gpio_del_params);
		break;
	case IOCTL_MLNX_SMBUS_READ:
 	case IOCTL_MLNX_SMBUS_WRITE:
		rc = mlnx_rw_smbus_device((struct mlnx_rdwr_smbus_ioctl_params *)ioctl_param);
		break;
 	case IOCTL_MLNX_SMBUS_RW:
		if (copy_from_user(p_smbus_params, (struct mlnx_rdwr_smbus_ioctl_params_2 __user *)ioctl_param,
 	 	 	sizeof(struct mlnx_rdwr_smbus_ioctl_params_2)))
			return -EFAULT;
		rc = mlnx_rw_smbus(p_smbus_params);
		p_smbus_params->rc = rc;
		if (copy_to_user((void __user *)ioctl_param, p_smbus_params,
					sizeof(struct mlnx_rdwr_smbus_ioctl_params_2))) {
			rc = -EFAULT;
		}
		break;
    case IOCTL_MLNX_CPLD_IRQ_SUBSCRIBE:
		if (copy_from_user(p_cpld_subscribe, (struct mlnx_cpld_subscribe_ioctl_params __user *)ioctl_param,
 	 	 	sizeof(struct mlnx_cpld_subscribe_ioctl_params)))
			return -EFAULT;
		rc = mlnx_cpld_irq_subscribe(p_cpld_subscribe);
		p_cpld_subscribe->rc = rc;
		if (copy_to_user((void __user *)ioctl_param, p_cpld_subscribe,
					sizeof(struct mlnx_cpld_subscribe_ioctl_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_CPLD_IRQ_UNSUBSCRIBE:
		if (copy_from_user(p_cpld_subscribe, (struct mlnx_cpld_subscribe_ioctl_params __user *)ioctl_param,
 	 	 	sizeof(struct mlnx_cpld_subscribe_ioctl_params)))
			return -EFAULT;
		rc = mlnx_cpld_irq_unsubscribe(p_cpld_subscribe);
		p_cpld_subscribe->rc = rc;
		if (copy_to_user((void __user *)ioctl_param, p_cpld_subscribe,
					sizeof(struct mlnx_cpld_subscribe_ioctl_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_CPLD_IRQ_SET:
		if (copy_from_user(p_cpld_set, (struct mlnx_cpld_set_ioctl_params __user *)ioctl_param,
 	 	 	sizeof(struct mlnx_cpld_set_ioctl_params)))
			return -EFAULT;
		rc = mlnx_cpld_irq_set(p_cpld_set);
		p_cpld_set->rc = rc;
		if (copy_to_user((void __user *)ioctl_param, p_cpld_set,
					sizeof(struct mlnx_cpld_set_ioctl_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_CPLD_IRQ_GET_DATA:
		if (copy_from_user(p_cpld_get_data, (struct mlnx_cpld_get_data_ioctl_params __user *)ioctl_param,
 	 	 	sizeof(struct mlnx_cpld_get_data_ioctl_params)))
			return -EFAULT;
		rc = mlnx_cpld_irq_get_data(p_cpld_get_data);
		p_cpld_get_data->rc = rc;
		if (copy_to_user((void __user *)ioctl_param, p_cpld_get_data,
					sizeof(struct mlnx_cpld_get_data_ioctl_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_WATCHDOG_CONFIG:
		if (!mlnx_watchdog_config_symbol)
			return -EFAULT;
		if (copy_from_user(p_watchdog_config,
 	 	 	(struct mlnx_watchdog_config_ioctl_params __user *)ioctl_param,
 	 	 	sizeof(struct mlnx_watchdog_config_ioctl_params)))
			return -EFAULT;
		if (p_watchdog_config->device_type == I2C_BUS_TYPE)
			rc = mlnx_watchdog_config_symbol(p_watchdog_config, i2c_bus_access_func);
		else if (p_watchdog_config->device_type == LPC_BUS_TYPE)
			rc = mlnx_watchdog_config_symbol(p_watchdog_config, lpc_bus_access_func);
		else
			return -EFAULT;
		p_watchdog_config->rc = rc;
		if (copy_to_user((void __user *)ioctl_param, p_watchdog_config,
					sizeof(struct mlnx_watchdog_config_ioctl_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_WATCHDOG_ENA_DIS:
		if (!mlnx_watchdog_enable_disable_symbol)
			return -EFAULT;
		if (copy_from_user(p_watchdog_ena_dis, 
 	 	 	(struct mlnx_watchdog_enable_disable_ioctl_params __user *)ioctl_param,
 	 	 	sizeof(struct mlnx_watchdog_enable_disable_ioctl_params)))
			return -EFAULT;
		rc = mlnx_watchdog_enable_disable_symbol(p_watchdog_ena_dis);
		p_watchdog_ena_dis->rc = rc;
		if (copy_to_user((void __user *)ioctl_param, p_watchdog_ena_dis,
					sizeof(struct mlnx_watchdog_enable_disable_ioctl_params))) {
			rc = -EFAULT;
		}
		break;
	case IOCTL_MLNX_VPD_READ:
	case IOCTL_MLNX_VPD_WRITE:
		rc = mlnx_vpd_rw_limited((struct mlnx_rdwr_vpd_i2c_ioctl_params *)ioctl_param);

		break;
	default:
		break;
	}

	return rc;
}

static long _mlnx_device_ioctl_unblocked(struct file *file, unsigned int ioctl, unsigned long arg)
{
	int rc;
	int __user *argp = (int __user *)arg;
	u8 ioctl_buffer[MLNX_IOCTL_MAX_SIZE];
	struct mlnx_cpld_leds_config_ioctl_params *p_cpld_led_config;
	struct mlnx_asic_fan_config_ioctl_params *p_asic_fan_config;
	struct mlnx_asic_temp_sensor_config_ioctl_params *p_asic_temp_sensor;

	if (!argp)
		return -EINVAL;

	switch (ioctl) {
	case IOCTL_MLNX_CPLD_CONFIG_LED:
		if (!mlnx_cpld_led_config_symbol_hook)
			return -EFAULT;
		p_cpld_led_config = (struct mlnx_cpld_leds_config_ioctl_params *)ioctl_buffer;
		if (copy_from_user(p_cpld_led_config,
 	 	 	(struct mlnx_cpld_leds_config_ioctl_params __user *)argp,
 	 	 	sizeof(struct mlnx_cpld_leds_config_ioctl_params)))
			return -EFAULT;

		rc = mlnx_cpld_led_config_symbol_hook(p_cpld_led_config);
		break;
	case IOCTL_MLNX_ASIC_CONFIG_FAN:
		if (!mlnx_asic_fan_config_symbol_hook)
			return -EFAULT;
		p_asic_fan_config = (struct mlnx_asic_fan_config_ioctl_params *)ioctl_buffer;
		if (copy_from_user(p_asic_fan_config,
 	 	 	(struct mlnx_asic_fan_config_ioctl_params __user *)argp,
 	 	 	sizeof(struct mlnx_asic_fan_config_ioctl_params)))
			return -EFAULT;
		rc = mlnx_asic_fan_config_symbol_hook(p_asic_fan_config);
		break;
	case IOCTL_MLNX_ASIC_CONFIG_TEMP:
		if (!mlnx_asic_temp_sensor_config_symbol_hook)
			return -EFAULT;
		p_asic_temp_sensor = (struct mlnx_asic_temp_sensor_config_ioctl_params *)ioctl_buffer;
		if (copy_from_user(p_asic_temp_sensor,
 	 	 	(struct mlnx_asic_temp_sensor_config_ioctl_params __user *)argp,
 	 	 	sizeof(struct mlnx_asic_temp_sensor_config_ioctl_params)))
			return -EFAULT;
		rc = mlnx_asic_temp_sensor_config_symbol_hook(p_asic_temp_sensor);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
static long _mlnx_device_ioctl(struct file *file,
		 	       unsigned int ioctl_num,
		 	       unsigned long ioctl_param)
{
	long rc;

	mutex_lock(&mlnx_ioctl_mutex);
	rc = mlnx_device_ioctl(file, ioctl_num, ioctl_param);
	mutex_unlock(&mlnx_ioctl_mutex);

	return rc;
}
#endif

static unsigned int mlnx_device_poll(struct file *filp, struct poll_table_struct *wait)
{
        return 0;
}

static const struct file_operations fops = {
	.poll =  mlnx_device_poll,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
	.unlocked_ioctl = _mlnx_device_ioctl,
#else
	.ioctl = mlnx_device_ioctl,
#endif
};

static const struct file_operations fops_unblocked = {
	.poll =  mlnx_device_poll,
	.unlocked_ioctl = _mlnx_device_ioctl_unblocked,
};

/* Register character device */
static int __init mlnx_if_init(void)
{
	int ret;

	if ((ret = register_chrdev(MLNX_IF_DEVICE_MAJOR, MLNX_IF_DEVICE_NAME, &fops)) < 0) {
		printk(KERN_WARNING "Failed to register character device ret=%d\n", ret);
		return -EIO;
	}
	if ((ret = register_chrdev(MLNX_IF_DEVICE_MAJOR_UNBLOCKED, MLNX_IF_DEVICE_NAME_UNBLOCKED, &fops_unblocked)) < 0) {
		unregister_chrdev(MLNX_IF_DEVICE_MAJOR, MLNX_IF_DEVICE_NAME);
		printk(KERN_WARNING "Failed to register character device ret=%d\n", ret);
		return -EIO;
	}

	mlnx_i2c_init();

	return 0;
}

static void __exit mlnx_if_exit(void)
{
	mlnx_i2c_exit();
	unregister_chrdev(MLNX_IF_DEVICE_MAJOR_UNBLOCKED, MLNX_IF_DEVICE_NAME_UNBLOCKED);
	unregister_chrdev(MLNX_IF_DEVICE_MAJOR, MLNX_IF_DEVICE_NAME);
}

module_init(mlnx_if_init);
module_exit(mlnx_if_exit);

MODULE_LICENSE("GPL v2");
