/**
 *
 * Copyright (C) 2012-2014, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */
#ifndef __SX_GLUE_IF_H_
#define __SX_GLUE_IF_H_

#include <linux/skbuff.h>

typedef int (*sgmii_rx)(struct sk_buff *skb);

/* Hook APIs are used by low level drivers to provide APIs for SDK */
extern void* (*sx_sgmii_register_rx_hook)(sgmii_rx rx_fn);
extern int (*sx_sgmii_xmit_hook)(void* hndl, struct sk_buff *skb);

extern int (*i2c_write_hook)(int i2c_dev_id, int offset, int len, u8 *in_out_buf);
extern int (*i2c_read_hook)(int i2c_dev_id, int offset, int len, u8 *in_out_buf);
extern int (*i2c_write_dword_hook)(int i2c_dev_id, int offset, u32 val);
extern int (*i2c_read_dword_hook)(int i2c_dev_id, int offset, u32 *val);
extern int (*i2c_enforce_hook)(int i2c_dev_id);
extern int (*i2c_release_hook)(int i2c_dev_id);
extern int (*i2c_get_local_mbox_hook)(int i2c_dev_id, u32 *mb_size_in, u32 *mb_offset_in, 
					u32 *mb_size_out, u32 *mb_offset_out);
extern int (*i2c_get_fw_rev_hook)(int i2c_dev_id, u64 *fw_rev);
extern int (*i2c_set_go_bit_stuck_hook)(int i2c_dev_id);

/* SDK APIs */

/* 
  Description: register SGMII receive calback routine
  Input:
               rx_fn - callback routine
  Output:
  Return:
               pointer to handle, which is to be used by
               caller at transmit operations.
*/
void* sx_sgmii_register_rx(sgmii_rx rx_fn);

/* 
  Description: SGMII transmit
  Input:
               hndl - handle as it returned by
                      sx_sgmii_register_rx
               skb - pointer to user buffer
  Output:
  Return:
               0 - on success
               negative values (standard Linux error code) on failure.
*/
int sx_sgmii_xmit(void* hndl, struct sk_buff *skb);

/* 
  Description: writing data blok to i2c device
  Input:
               i2c_dev_id - device id of particular switchx
               offset     - offset from which device to be written
               len        - number of bytes to be written
               in_out_buf - point to input write buffer
  Return code:
               0 - on success
               negative values (standard Linux error code) on failure.
*/
int i2c_write(int i2c_dev_id, int offset, int len, u8 *in_out_buf);

/* 
  Description: reading data blok from i2c device
  Input:
               i2c_dev_id - device id of particular switchx
               offset     - offset from which device to be read
               len        - number of bytes to be read
               in_out_buf - pointer to input read buffer
  Return code:
               0 - on success
               negative values (standard Linux error code) on failure.
*/
int i2c_read(int i2c_dev_id, int offset, int len, u8 *in_out_buf);

/* 
  Description: writing double word to i2c device
  Input:
               i2c_dev_id - device id of particular switchx
               offset     - offset from which device to be written
               val        - pointer to input double word to where data to be put
  Return code:
               0 - on success
               negative values (standard Linux error code) on failure.
*/
int i2c_write_dword(int i2c_dev_id, int offset, u32 val);

/* 
  Description: reading double word from i2c device
  Input:
               i2c_dev_id - device id of particular switchx
               offset     - offset from which device to be read
               val        - pointer to input double word from where data to be get
  Return code:
               0 - on success
               negative values (standard Linux error code) on failure.
*/
int i2c_read_dword(int i2c_dev_id, int offset, u32 *val);

/* 
  Description: lock i2c bus, till i2c_release is called.
               Used to support multi-transaction.
  Input:
               i2c_dev_id - device id of particular switchx
  Return code:
               0 - on success
               negative values (standard Linux error code) on failure.
*/
int i2c_enforce(int i2c_dev_id);

/* 
  Description: unlock i2c bus, always is called after i2c_enforce.
               Used to support multi-transaction.
  Input:
               i2c_dev_id - device id of particular switchx
  Return code:
               0 - on success
               negative values (standard Linux error code) on failure.
*/
int i2c_release(int i2c_dev_id);

/* 
  Description: obtain local mailbox parameters through i2c request,
               using ASIC command interface.
  Input:
               i2c_dev_id - device id of particular switchx
  Output:
               mb_size_in - pointer to input mailbox size
               mb_offset_in - pointer to input mailbox offset
               mb_size_out - pointer to output mailbox size
               mb_offset_out  - pointer to output mailbox offset
  Return code:
               0 - on success
               negative values (standard Linux error code) on failure.
*/
int i2c_get_local_mbox(int i2c_dev_id, u32 *mb_size_in, u32 *mb_offset_in, 
			           u32 *mb_size_out, u32 *mb_offset_out);

/* 
  Description: obtain ASIC firmare version through i2c request,
               using ASIC command interface (query firmware command).
  Input:
               i2c_dev_id - device id of particular switchx
  Output:
               fw_ver - pointer to firmware version of ASIC
  Return code:
               0 - on success
               negative values (standard Linux error code) on failure.
*/
int i2c_get_fw_rev(int i2c_dev_id, u64 *fw_rev);

/* 
  Description: callback indicating that firmware expereinces
               problem with GO bit release. In such case
               software and firmware will not be able to
               communication over i2c, since GO bit is the
               semaphore between them.
  Input:
               i2c_dev_id - device id of particular switchx
  Output:
  Return code:
               0 - on success
               negative values (standard Linux error code) on failure.
*/
int i2c_set_go_bit_stuck(int i2c_dev_id);
 

#endif /* __SX_GLUE_IF_H_ */
 
