/**
 *
 * Copyright (C) 2013-2014, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/uaccess.h>
#include <linux/types.h>

#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/slab.h>

#include <linux/uio.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/time.h>
#if !defined(__i386__) && !defined(__x86_64__)
#include <linux/of.h>
#endif
#include <linux/gfp.h>
#include <linux/list.h>
#include <linux/rculist.h>
#include <linux/signal.h>
#include <linux/spinlock.h>
#include "mlnx_log.h"
#include "watchdog.h"
 

/************************************************
 *  Define
 ***********************************************/
#define DRV_NAME	"watchdog"
#define PFX		DRV_NAME ": "
#ifndef BUILD_VERSION
#define BUILD_VERSION 	""
#endif

/* After watchdog device is closed, check force_boot. If:
 * force_boot == 0, then force boot on next watchdog interrupt after close,
 * force_boot == 1, then force boot immediately when device is closed.
 */
static int force_boot;
module_param(force_boot, int, 0);
MODULE_PARM_DESC(force_boot,
		"A value of 1 means that the driver will reboot\n"
		"the system immediately if /dev/mlnx_watchdog1/2/3/4, devices are closed\n"
		"A value of 0 means that when /dev/mlnx_watchdog1/2/3/4 device are closed\n"
		"the watchdog timer will be refreshed for one more interval\n"
		"of length: timer_set. At the end of this interval, the\n"
		"watchdog timer will reset the system."
		);

struct mlnx_watchdog_notify_entry {
	struct list_head subscriber_list;  
	mlnx_watchdog_notify_func notifunc;
	void *subscriber_data;
};

struct mlnx_watchdog_cfg {
	int dev_id;                     /* Device id */
	union {
		struct mlnx_i2c_dev_params i2c_params;
		struct mlnx_lpc_dev_params lpc_params;
	} hwdev;
	__u16 flags;	                /* user specific flags */
	__u16 device_type;              /* device bus type, i2c, lpc, etc */
	rw_bus_access_func rw_func;
	struct watchdog_config_entry wd_entry;
	struct notifier_block mlnx_notifier;
	struct timer_list wd_soft_timer;
	unsigned long heartbeat_time;
	u32 sw_timer_avail;
	spinlock_t heartbeat_lock;
	spinlock_t notifier_lock;
	struct list_head notifier_list; 
};

struct mlnx_watchdog_dev {
	char name[24];
	struct watchdog_sub_entry se_wd;
	ulong driver_open;
	ulong driver_closed;
	u8 timer_started;
	u8 timer_set;
	u32 threshold;
	u32 soft_threshold;
	u8 wd_def_mask;
	u8 wd_mask;
	u8 act_def_mask;
	u8 set_mask;
	u8 action_mask;
	struct watchdog_config_entry wd_entry;
	struct mlnx_watchdog_ctrl ctrl;
	struct timer_list enable_delay_timer;
};

static struct mlnx_watchdog_dev *p_wd_dev[MAX_MGMT_WD];
static struct mlnx_watchdog_cfg wd_cfg;
static struct mlnx_watchdog_cfg *p_wd_cfg = &wd_cfg;

/* Stop all watchdogs and software timer */
static int mlnx_wd_stop(void)
{
	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d)\n", __func__, __LINE__);

	if(p_wd_cfg->heartbeat_time)
		del_timer_sync(&p_wd_cfg->wd_soft_timer);
	return 0;
}

/* Notifier for system down */
static int mlnx_wd_notify_sys(struct notifier_block *this,
			       unsigned long code,
			       void *another_unused)
{
	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) code=%lu\n",
					__func__, __LINE__, code);

	if (code == SYS_DOWN || code == SYS_HALT)
		/* Turn off the watchdog timer. */
		mlnx_wd_stop();
	return NOTIFY_DONE;
}

/* Forces restart, if force_reboot is set */
static void mlnx_wd_fire(struct mlnx_watchdog_dev *p_dev)
{
	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) p_dev=%p\n",
					__func__, __LINE__, p_dev);

	if (p_dev->ctrl.force_sw_reboot) {
		pr_crit("Initiating system reboot\n");
		emergency_restart();
		pr_crit("Reboot was not activated\n");
	}
	else {
		pr_crit("Immediate reboot is not enable\n");
		pr_crit("System will reset when watchdog timer is expired\n");
	}
}

/* Software timer handler */
static void mlnx_wd_soft_heartbeat(unsigned long arg)
{
	struct mlnx_watchdog_cfg *w = (struct mlnx_watchdog_cfg *)arg;
	struct mlnx_watchdog_notify_entry *p, *n;
	unsigned long flags;
	u8 curr_counter_val;
	int ind;

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) p_wd_cfg=%p w=%p\n",
					__func__, __LINE__, p_wd_cfg, w);

	/* Check if for any watchdogs threshold is exceeded */
	for (ind = 0; MAX_MGMT_WD; ind++) {
		mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) %d inc_thresh=%x timer_value=%x\n",
						__func__, __LINE__, ind,
						p_wd_dev[ind]->ctrl.inc_thresh, p_wd_dev[ind]->ctrl.timer_value);
		if (p_wd_dev[ind]->ctrl.inc_thresh && p_wd_dev[ind]->ctrl.timer_value) {
			p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
						p_wd_dev[ind]->se_wd.wd_counter,
						p_wd_dev[ind]->se_wd.wd_counter,
						1, &curr_counter_val, _B_LOCK);
		}

		mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) curr_counter_val[%d]=%x inc_thresh=%x timer_value=%x\n",
						__func__, __LINE__, ind, curr_counter_val,
						p_wd_dev[ind]->ctrl.inc_thresh, p_wd_dev[ind]->ctrl.timer_value);

		if (p_wd_dev[ind]->ctrl.inc_thresh && (curr_counter_val >= p_wd_dev[ind]->ctrl.inc_thresh)) {
			/* Counter is expired - activate sw_action if set */
			if (p_wd_dev[ind]->ctrl.sw_action & _wd_sw_notify) {
				mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) sw_action[%d]=%x\n",
							__func__, __LINE__, ind, p_wd_dev[ind]->ctrl.sw_action);
				if(!list_empty(&p_wd_cfg->notifier_list)) {
					spin_lock_irqsave(&p_wd_cfg->notifier_lock, flags);
					list_for_each_entry_safe(p, n, &p_wd_cfg->notifier_list, subscriber_list) {
						p->notifunc(p->subscriber_data);
					}
					spin_unlock_irqrestore(&p_wd_cfg->notifier_lock, flags);
				}
			}

			if (p_wd_dev[ind]->ctrl.sw_action & _wd_sw_reboot) {
				mlnx_wd_fire(p_wd_dev[ind]);
				return;
			}
		}
	}

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) heartbeat_time=0x%0lx\n",
					__func__, __LINE__, w->heartbeat_time);

	spin_lock(&w->heartbeat_lock);

	mod_timer(&w->wd_soft_timer, round_jiffies(jiffies + w->heartbeat_time));

	spin_unlock(&w->heartbeat_lock);
}

/* Software timer handler */
static void mlnx_wd_enable_kick(unsigned long arg)
{
	struct mlnx_watchdog_dev *p_dev = (struct mlnx_watchdog_dev *)arg;
	u8 curr_action = 0;
	u8 curr_val = 0;
	u8 curr_val_prot = 0;

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) p_dev=%p\n",
					__func__, __LINE__, p_dev);

	del_timer(&p_dev->enable_delay_timer);

	/* Set action and timer */
	if (p_dev->se_wd.wd_action != CPLD_MASK_NOP8) {                                   /* Watchdog action address */
		curr_action = p_dev->action_mask & 0xff;

		mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) write 0x%02x at 0x%02x\n",
			__func__, __LINE__, curr_action, p_dev->se_wd.wd_action);

		p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
					p_dev->se_wd.wd_action,
					p_dev->se_wd.wd_action,
					0, &curr_action, _B_LOCK);                         /* Set watchdog action */
	}
	if (p_dev->se_wd.wd_timer_lsb != CPLD_MASK_NOP8) {

		mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) write 0x%02x at 0x%02x\n",
					__func__, __LINE__, p_dev->ctrl.timer_value,
					p_dev->se_wd.wd_timer_lsb);

		p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
					p_dev->se_wd.wd_timer_lsb,
					p_dev->se_wd.wd_timer_lsb,
					0, &(p_dev->ctrl.timer_value), _B_LOCK);           /* Set watchdog timer */
	}
	p_dev->timer_started = 1;
 
	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) timer_started= %d %d %d %d\n",
				__func__, __LINE__,
				p_wd_dev[0]->timer_started, p_wd_dev[1]->timer_started,
				p_wd_dev[2]->timer_started, p_wd_dev[3]->timer_started);


	/* Write each clear protection bit 0 to remove protection,
	 	 write each clear bit 1 to clear */
	curr_val |= p_dev->set_mask;
	curr_val_prot &= ~p_dev->set_mask;

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) set_mask=%x write to: addr %x val %x addr %x val %x\n",
				__func__, __LINE__, p_dev->set_mask, p_wd_cfg->wd_entry.wd_clear_wp,
				 curr_val_prot, p_wd_cfg->wd_entry.wd_clear, curr_val_prot);
	if (p_wd_cfg->wd_entry.wd_clear_wp != CPLD_MASK_NOP8)             /* Watchdog clear write protection address */
		p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
					p_wd_cfg->wd_entry.wd_clear_wp,
					p_wd_cfg->wd_entry.wd_clear_wp,
					0, &curr_val_prot, _B_LOCK);      /* Remove protection for configured watchdogs */
	if (p_wd_cfg->wd_entry.wd_clear != CPLD_MASK_NOP8)                /* Watchdog clear address */
		p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
					p_wd_cfg->wd_entry.wd_clear,
					p_wd_cfg->wd_entry.wd_clear,
					0, &curr_val_prot, _B_LOCK);      /* Clear bits for configured watchdogs */
 
	if(p_wd_cfg->sw_timer_avail)
		setup_timer(&p_wd_cfg->wd_soft_timer, mlnx_wd_soft_heartbeat, (unsigned long)p_wd_cfg);
 

}

static int mlnx_wd_keepalive(struct mlnx_watchdog_dev *p_dev)
{
	u8 curr_val = 0;

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d)\n", __func__, __LINE__);

	/* temporarily disable the timer */
	p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
				p_dev->se_wd.wd_timer_lsb,
				p_dev->se_wd.wd_timer_lsb,
				0, &curr_val, _B_LOCK);

	spin_lock(&p_wd_cfg->heartbeat_lock);

	mod_timer(&p_wd_cfg->wd_soft_timer,
		  round_jiffies(jiffies + p_wd_cfg->heartbeat_time));

	spin_unlock(&p_wd_cfg->heartbeat_lock);

	return 0;
}

static int mlnx_wd_set_heartbeat(struct mlnx_watchdog_dev *p_dev, u8 margin)
{
	u8 curr_val;
	int err = 0;

	curr_val = p_dev->wd_mask & 0xff;

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) new margin=0x%02x\n",
			__func__, __LINE__, margin);

	if ((margin == 0) || (margin > 0x0f))
		return -EINVAL;

	if (p_wd_cfg->wd_entry.wd_clear != CPLD_MASK_NOP8) {
		err = p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
						p_wd_cfg->wd_entry.wd_clear,
						p_wd_cfg->wd_entry.wd_clear,
						0, &curr_val, _B_LOCK);
	}

	if (p_dev->se_wd.wd_timer_lsb != CPLD_MASK_NOP8) {
		curr_val = margin;
		err = (err >= 0) ? p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
							p_dev->se_wd.wd_timer_lsb,
							p_dev->se_wd.wd_timer_lsb,
							0, &curr_val, _B_LOCK) : err;
	}

	return err;
}

/**
 *	mlnx_wd_write:
 *	@filp: file handle to the watchdog device
 *	@buf: buffer to write (unused as data does not matter here
 *	@count: count of bytes
 *	@ppos: pointer to the position to write. No seeks allowed
 *
 *	A write to a watchdog device is defined as a keepalive signal.
 *	Any write of data is OK, since content is meaningless.
 */
static ssize_t mlnx_wd_write(struct file *filp,
				const char __user *buf,
				size_t count,
				loff_t *pos)
{
	int err = 0;
	u8 curr_val = 0;
	struct mlnx_watchdog_dev *p_dev = (struct mlnx_watchdog_dev *)filp->private_data;

	if (p_dev->timer_started) {
		if (p_wd_cfg->wd_entry.wd_clear != CPLD_MASK_NOP8) {
			curr_val = p_dev->wd_mask & 0xff;
			err = p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
						p_wd_cfg->wd_entry.wd_clear,
						p_wd_cfg->wd_entry.wd_clear,
						0, &curr_val, _B_LOCK);
		}
		if (err < 0)
			return -EFAULT;

		mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) write=0x%02x at 0x%02x\n",
				__func__, __LINE__, curr_val, p_wd_cfg->wd_entry.wd_clear);
 
		/* Watchdog already started, keep it alive */
		if (p_wd_cfg->sw_timer_avail)
			mlnx_wd_keepalive(p_dev);
	}
	else {
		/* Start watchdog with timer value set by init */
		mlnx_wd_set_heartbeat(p_dev, p_dev->timer_set);
	}

	return count;
}

/**
 *	mlnx_wd_read:
 *	@filp: file handle to the watchdog device
 *	@buf: buffer to write 1 byte into
 *	@count: length of buffer
 *	@ptr: offset (no seek allowed)
 *
 *	Function reports the current value of watchdog counter.
 */
static ssize_t mlnx_wd_read(struct file *filp,
				char __user *buf,
				size_t count,
				loff_t *pos)
{
	int wd_cntr = 0;
	u8 curr_val = 0;
	int err;
	struct mlnx_watchdog_dev *p_dev = (struct mlnx_watchdog_dev *)filp->private_data;

	if ((count == 0) || (buf == NULL)) {
		err = -EINVAL;
	}

	if (p_dev->timer_started) {
		if (p_dev->se_wd.wd_counter != CPLD_MASK_NOP8) {
			err = p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
						p_dev->se_wd.wd_counter,
						p_dev->se_wd.wd_counter,
						1, &curr_val, _B_LOCK);
			if (err < 0)
				return -EFAULT;

			wd_cntr = (int)curr_val;
			if (copy_to_user(buf, &wd_cntr, 1))
				return -EFAULT;
		}
	}

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) wd_cntr=%d\n",
					__func__, __LINE__, wd_cntr);

	return 1;
}

/**
 *	mlnx_wd1_open - mlnx_wd4_open:
 *	@inode: inode of device
 *	@filp: file handle to the watchdog device
 *
 *	Watchdog device is to be opened.
 */
static int mlnx_wd1_open(struct inode *inode, struct file *filp)
{
	struct mlnx_watchdog_dev *p_dev = (struct mlnx_watchdog_dev *)filp->private_data;
	int err;

	filp->private_data = p_dev = p_wd_dev[0];

	/* Set flag to indicate that watchdog device is open */
	if (test_and_set_bit(0, &p_dev->driver_open))
		return -EBUSY;

	err = nonseekable_open(inode, filp);

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) private_data=%p err=%d (%s)\n",
		 __FUNCTION__, __LINE__, filp->private_data, err, p_dev->name);

	return err; 
}

static int mlnx_wd2_open(struct inode *inode, struct file *filp)
{
	struct mlnx_watchdog_dev *p_dev = (struct mlnx_watchdog_dev *)filp->private_data;
	int err;

	filp->private_data = p_dev = p_wd_dev[1];

	/* Set flag to indicate that watchdog device is open */
	if (test_and_set_bit(0, &p_dev->driver_open))
		return -EBUSY;

	err = nonseekable_open(inode, filp);

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) private_data=%p err=%d (%s)\n",
		 __FUNCTION__, __LINE__, filp->private_data, err,
		 p_dev->name);

	return err; 
}

static int mlnx_wd3_open(struct inode *inode, struct file *filp)
{
	struct mlnx_watchdog_dev *p_dev = (struct mlnx_watchdog_dev *)filp->private_data;
	int err;

	filp->private_data = p_dev = p_wd_dev[2];

	/* Set flag to indicate that watchdog device is open */
	if (test_and_set_bit(0, &p_dev->driver_open))
		return -EBUSY;

	err = nonseekable_open(inode, filp);

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) private_data=%p err=%d (%s)\n",
		 __FUNCTION__, __LINE__, filp->private_data, err,
		 p_dev->name);

	return err; 
}

static int mlnx_wd4_open(struct inode *inode, struct file *filp)
{
	struct mlnx_watchdog_dev *p_dev = (struct mlnx_watchdog_dev *)filp->private_data;
	int err;

	filp->private_data = p_dev = p_wd_dev[3];

	/* Set flag to indicate that watchdog device is open */
	if (test_and_set_bit(0, &p_dev->driver_open))
		return -EBUSY;

	err = nonseekable_open(inode, filp);

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) private_data=%p err=%d (%s)\n",
		 __FUNCTION__, __LINE__, filp->private_data, err,
		 p_dev->name);

	return err; 
}

/**
 *	mlnx_wd_release:
 *	@inode: inode to board
 *	@filp: file handle to watchdog device
 *
 *	Watchdog device is to be closed.
 */
static int mlnx_wd_release(struct inode *inode, struct file *filp)
{
	struct mlnx_watchdog_dev *p_dev = (struct mlnx_watchdog_dev *)filp->private_data;

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) private_data=%p\n",
		 __FUNCTION__, __LINE__, p_dev);

	if (!test_and_clear_bit(0, &p_dev->driver_open)) {
		pr_debug("%s without open\n", __func__);
		return -ENOTTY;
	}

	if (!p_dev->timer_started) {
		/* Just close, since timer has not been started */
		pr_debug("closed, without starting timer\n");
		return 0;
	}

	/* Mark device as closed */
	p_dev->driver_closed = 1;
 
#if 0
	/*
	 * This watchdog should not be closed, after the timer
	 * is started with the WDIPC_SETTIMEOUT ioctl
	 * If force_boot is set mlnx_wd_fire(p_dev) will cause an
	 * immediate reset. If force_boot is not set, the watchdog
	 * timer is refreshed for one more interval. At the end
	 * of that interval, the watchdog timer will reset the system.
	 */
	pr_crit("Unexpected close of %s\n", p_dev->name);
	/* Refresh the timer for one more interval */
	if (p_wd_cfg->sw_timer_avail)
		mlnx_wd_keepalive(p_dev);
	/* Reboot system (if force_boot is set) */
	mlnx_wd_fire(p_dev);
	/* We should only reach this point if force_boot is not set */
#endif

	return 0;
}

static long mlnx_wd_ioctl(struct file *filp,
		unsigned int cmd, unsigned long data)
{
	void __user *argp = (void __user *)data;
	u32 __user *p = argp;
	u32 new_margin;
	int err = 0;
	struct mlnx_watchdog_dev *p_dev = (struct mlnx_watchdog_dev *)filp->private_data;
 
	switch (cmd) {
	case MLNX_WDIOC_GETSTATUS:
	case MLNX_WDIOC_GETBOOTSTATUS:
		return put_user(0, p);
	case MLNX_WDIOC_KEEPALIVE:
		return 0;
	case MLNX_WDIOC_SETTIMEOUT:
		if (get_user(new_margin, p))
			return -EFAULT;

		if (mlnx_wd_set_heartbeat(p_dev, (u8)new_margin))
			return -EINVAL;
		return 0;
	case MLNX_WDIOC_GETTIMEOUT:
		return put_user(p_dev->soft_threshold, p);
	case MLNX_WDIOC_GETSUPPORT:
	default:
		return -ENOTTY;
	}

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) cmd=%d private_data=%p\n",
		 __FUNCTION__, __LINE__, cmd, p_dev);
	return err;
}

/* Kernel Interfaces */
static const struct file_operations mlnx_wd_fops[] = {
	{
		.owner	 =	THIS_MODULE,
		.llseek  =      no_llseek,
		.open	 =	mlnx_wd1_open,
		.read	 =	mlnx_wd_read,
		.write	 =	mlnx_wd_write,
		.unlocked_ioctl = mlnx_wd_ioctl,
		.release = 	mlnx_wd_release,
	},
	{
		.owner	 =	THIS_MODULE,
		.llseek  =      no_llseek,
		.open	 =	mlnx_wd2_open,
		.read	 =	mlnx_wd_read,
		.write	 =	mlnx_wd_write,
		.unlocked_ioctl = mlnx_wd_ioctl,
		.release = 	mlnx_wd_release,
	},
	{
		.owner	 =	THIS_MODULE,
		.llseek  =      no_llseek,
		.open	 =	mlnx_wd3_open,
		.read	 =	mlnx_wd_read,
		.write	 =	mlnx_wd_write,
		.unlocked_ioctl = mlnx_wd_ioctl,
		.release = 	mlnx_wd_release,
	},
	{
		.owner	 =	THIS_MODULE,
		.llseek  =      no_llseek,
		.open	 =	mlnx_wd4_open,
		.read	 =	mlnx_wd_read,
		.write	 =	mlnx_wd_write,
		.unlocked_ioctl = mlnx_wd_ioctl,
		.release = 	mlnx_wd_release,
	},
};

static struct miscdevice mlnx_wd_miscdevs[] = {
        {
                .minor  = CPLD_WATCHDOG1_MINOR,
                .name   = "mlnx_watchdog1",
                .fops   = &mlnx_wd_fops[0],
        },
        {
                .minor  = CPLD_WATCHDOG2_MINOR,
                .name   = "mlnx_watchdog2",
                .fops   = &mlnx_wd_fops[1],
        },
        {
                .minor  = CPLD_WATCHDOG3_MINOR,
                .name   = "mlnx_watchdog3",
                .fops   = &mlnx_wd_fops[2],
        },
        {
                .minor  = CPLD_WATCHDOG4_MINOR,
                .name   = "mlnx_watchdog4",
                .fops   = &mlnx_wd_fops[3],
        },
};

int mlnx_watchdog_config(struct mlnx_watchdog_config_ioctl_params *wd_config_params,
			 rw_bus_access_func rw_func)
{
	int ind;
	struct i2c_adapter *adap = NULL;

	mlnx_log(1, MLNX_LOG_WD_MAX, "%s rw_func=%p\n", __func__, rw_func);

	p_wd_cfg->dev_id = wd_config_params->dev_id;
	p_wd_cfg->flags = wd_config_params->flags;
	p_wd_cfg->device_type = wd_config_params->device_type;
	p_wd_cfg->rw_func = rw_func;

	memcpy(&p_wd_cfg->wd_entry, &wd_config_params->wd_entry, 
		sizeof(struct watchdog_config_entry));
	memcpy(&p_wd_cfg->hwdev, &wd_config_params->hwdev, 
		sizeof(wd_config_params->hwdev));

	if (p_wd_cfg->device_type == I2C_BUS_TYPE) {
		adap = i2c_get_adapter(p_wd_cfg->hwdev.i2c_params.bus_id);
		if (!adap) {
			printk(KERN_INFO "%s: bad adaptor for device 0x%08x\n",
				__FUNCTION__, wd_config_params->dev_id);
			return -EFAULT;
		}
		p_wd_cfg->hwdev.i2c_params.adap = adap;
	}

	for (ind = 0; ind < MAX_MGMT_WD; ind++) {
	        mlnx_log(3, MLNX_LOG_WD_MAX, "%s ind=%d\n", __func__, ind);
                memcpy(&p_wd_dev[ind]->se_wd, &p_wd_cfg->wd_entry.se_wd[ind],
                	sizeof(struct watchdog_sub_entry));
                memcpy(&p_wd_dev[ind]->ctrl, &wd_config_params->ctrl[ind],
                	sizeof(struct mlnx_watchdog_ctrl));
                sprintf(p_wd_dev[ind]->name, "%s%d", "/dev/mlnx_watchdog", ind + 1);

	        mlnx_log(3, MLNX_LOG_WD_MAX, "%s %d action=%x counter=%x\n",
		                        __func__, ind,
		                        p_wd_cfg->wd_entry.se_wd[ind].wd_action,
		                        p_wd_cfg->wd_entry.se_wd[ind].wd_counter);
	}
	/* Check if any of whatchdogs is configured for software reset contol */
	if ((p_wd_dev[0]->ctrl.inc_thresh && p_wd_dev[0]->ctrl.timer_value) ||
	    (p_wd_dev[1]->ctrl.inc_thresh && p_wd_dev[1]->ctrl.timer_value) ||
	    (p_wd_dev[2]->ctrl.inc_thresh && p_wd_dev[2]->ctrl.timer_value) ||
	    (p_wd_dev[3]->ctrl.inc_thresh && p_wd_dev[3]->ctrl.timer_value)) {
		p_wd_cfg->heartbeat_time = min4(p_wd_dev[0]->ctrl.timer_value,
						p_wd_dev[1]->ctrl.timer_value,
						p_wd_dev[2]->ctrl.timer_value,
						p_wd_dev[3]->ctrl.timer_value);
		/* Software clock maintained by the kernel measures time in jiffies.
		   The size of a jiffy is determined by the value of the kernel constant HZ.
		   The value of HZ varies across kernel versions and hardware platforms.
		   On i386 the HZ value is a kernel configuration parameter can be 100,
		   250 (the default), 300 or 1000, yielding a jiffies value of, respectively, 
		   0.01, 0.004, 0.00(3) or 0.001 seconds.   
		   Configuration timer value is provided as power of 2 - so it set as a power
		   of 2 of minimal from for timer values and than it's multiplied by HZ/250
		   to convert timer accuracy from msec to jiffies resolution */
	        mlnx_log(3, MLNX_LOG_WD_MAX, "%s p_wd_cfg->heartbeat_time=0x%0lx -> 0x%0lx\n", __func__, 
	        	 p_wd_cfg->heartbeat_time, (p_wd_cfg->heartbeat_time << 1));
		p_wd_cfg->heartbeat_time = (p_wd_cfg->heartbeat_time << 1);
		p_wd_cfg->sw_timer_avail = 1;
	}
	else {
		p_wd_cfg->heartbeat_time = 0;
		p_wd_cfg->sw_timer_avail = 0;
	}
	/* Get default action mask for each watchdog */
	for (ind = 0; ind < MAX_MGMT_WD; ind++) {
		if (p_wd_dev[ind]->se_wd.wd_action != CPLD_MASK_NOP8) {             /* Watchdog action address */
			p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
						p_wd_dev[ind]->se_wd.wd_action,
						p_wd_dev[ind]->se_wd.wd_action,
						1, &p_wd_dev[ind]->act_def_mask, _B_LOCK);   /* Get watchdog current action */
		}
	}

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s act_def_mask 0x%02x 0x%02x 0x%02x 0x%02x\n" \
				"wd_def_mask 0x%02x 0x%02x 0x%02x 0x%02x\n",
				__func__, 
				p_wd_dev[0]->act_def_mask, p_wd_dev[1]->act_def_mask,
				p_wd_dev[2]->act_def_mask, p_wd_dev[3]->act_def_mask,
				p_wd_dev[0]->wd_def_mask, p_wd_dev[1]->wd_def_mask,
				p_wd_dev[2]->wd_def_mask, p_wd_dev[3]->wd_def_mask);
  
	return 0;
}
EXPORT_SYMBOL(mlnx_watchdog_config);

int mlnx_watchdog_enable_disable(struct mlnx_watchdog_enable_disable_ioctl_params *wd_ed_params)
{
	int ind;
	int ind1;
	u8 ena = 0;
	u8 set_mask = 0;
	u8 curr_val = 0;
	u8 curr_val_prot = 0;
	u8 curr_action = 0;
	u8 clear_timer = 0;
	u8 set_wd_action_mask[MAX_MGMT_WD];

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d)\n", __func__, __LINE__);

	if (p_wd_cfg->wd_entry.wd_clear_wp != CPLD_MASK_NOP8)           /* Watchdog clear write protection address */
		p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
					p_wd_cfg->wd_entry.wd_clear_wp,
					p_wd_cfg->wd_entry.wd_clear_wp,
					1, &curr_val, _B_LOCK);          /* Get watchdog protection setting */
	if (p_wd_cfg->wd_entry.wd_clear != CPLD_MASK_NOP8)               /* Watchdog clear address */
		p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
					p_wd_cfg->wd_entry.wd_clear_wp,
					p_wd_cfg->wd_entry.wd_clear_wp,
					1, &curr_val_prot, _B_LOCK);     /* Get watchdog setting */

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) wd_ed_params->wd=%x\n",
				__func__, __LINE__, wd_ed_params->wd);

	if (wd_ed_params->ena)
		ena = 1;
	for (ind = 0; ind < 8; ind++) {
		mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) signal(%d)=0x%02x\n",
					__func__, __LINE__, ind,
					p_wd_cfg->wd_entry.sig[ind].signal_id);
		switch(p_wd_cfg->wd_entry.sig[ind].signal_id) {
			case _wd1_timer_clear:
        			if (!(wd_ed_params->wd & _wd1_))
        				break;
				set_mask |= (ena << ind);
				break;
			case _wd2_timer_clear:
        			if (!(wd_ed_params->wd & _wd2_))
        				break;
				set_mask |= (ena << ind);
				break;
			case _wd3_timer_clear:
        			if (!(wd_ed_params->wd & _wd3_))
        				break;
				set_mask |= (ena << ind);
				break;
			case _wd4_timer_clear:
        			if (!(wd_ed_params->wd & _wd4_))
        				break;
				set_mask |= (ena << ind);
				break;
			case _wd1_counter_clear:
        			if (!(wd_ed_params->wd & _wd1_))
        				break;
				set_mask |= (ena << ind);
				break;
			case _wd2_counter_clear:
        			if (!(wd_ed_params->wd & _wd2_))
        				break;
				set_mask |= (ena << ind);
				break;
			case _wd3_counter_clear:
        			if (!(wd_ed_params->wd & _wd3_))
        				break;
				set_mask |= (ena << ind);
				break;
			case _wd4_counter_clear:
        			if (!(wd_ed_params->wd & _wd4_))
        				break;
				set_mask |= (ena << ind);
				break;
			case _none_wd_sig:
				break;;
			default:
				return -EINVAL;
		}
	}

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) set_mask=0x%02x\n",
				__func__, __LINE__, set_mask);

	for (ind1 = 0; ind1 < MAX_MGMT_WD; ind1++) {
		mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) hw_action=0x%02x\n",
					__func__, __LINE__, p_wd_dev[ind1]->ctrl.hw_action);

		if (!(wd_ed_params->wd & (1 << ind1)))
			continue; /* Skip watchdog */

		for (ind = 0; ind < 8; ind++) {
			mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) signal(%d,%d)=0x%02x\n",
						__func__, __LINE__, ind1, ind,
						p_wd_dev[ind1]->se_wd.sig[ind].signal_id);

			switch(p_wd_dev[ind1]->se_wd.sig[ind].signal_id) {
        			case _wd_reset_local:
        			case _wd_int_to_local:
        			case _wd_int_to_far:
        			case _wd_power_off_local:
        			case _wd_increment_cntr:
        				set_wd_action_mask[ind1] |= (ena << ind);
        				break;
        			case _none_wd_sig:
        			default:
					break;
			}
			set_wd_action_mask[ind1] &= p_wd_dev[ind1]->ctrl.hw_action;
		}
		mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) action_mask[%d]=0x%02x timer_value=0x%02x wd_def_mask=0x%02x\n" \
					"se_wd.wd_timer_lsb=0x%02x wd_action=0x%02x wd_counter=0x%02x\n",
					__func__, __LINE__, ind1, set_wd_action_mask[ind1],
					p_wd_dev[ind1]->ctrl.timer_value,
					p_wd_dev[ind1]->wd_def_mask, p_wd_dev[ind1]->se_wd.wd_timer_lsb,
					p_wd_dev[ind1]->se_wd.wd_action, p_wd_dev[ind1]->se_wd.wd_counter);
		p_wd_dev[ind1]->wd_mask = p_wd_dev[ind1]->wd_def_mask;
		p_wd_dev[ind1]->action_mask = set_wd_action_mask[ind1];
		p_wd_dev[ind1]->set_mask = set_mask;

		if (ena) {
			if (wd_ed_params->enable_delay) {
				/* If enable delay is requested - complete the bottom half
				   code within mlnx_wd_enable_kick upon timer expiration.
				   Set timer as (wd_ed_params->enable_delay << 1) msec:
				   cuurent jffies plus msec value mult msec resolution (HZ/1000)  */
                		p_wd_dev[ind1]->enable_delay_timer.expires = jiffies + (wd_ed_params->enable_delay << 1) *
                								HZ/1000;
               			p_wd_dev[ind1]->enable_delay_timer.function = mlnx_wd_enable_kick;
               			p_wd_dev[ind1]->enable_delay_timer.data = (unsigned long)p_wd_dev[ind1];
                		add_timer(&p_wd_dev[ind1]->enable_delay_timer);

				continue;
			}
			/* Set action and timer */
			if (p_wd_dev[ind1]->se_wd.wd_action != CPLD_MASK_NOP8) {                  /* Watchdog action address */
				curr_action = set_wd_action_mask[ind1] & 0xff;

				mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) write 0x%02x at 0x%02x\n",
					__func__, __LINE__, curr_action, p_wd_dev[ind1]->se_wd.wd_action);

				p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
							p_wd_dev[ind1]->se_wd.wd_action,
							p_wd_dev[ind1]->se_wd.wd_action,
							0, &curr_action, _B_LOCK);                         /* Set watchdog action */
			}
			if (p_wd_dev[ind1]->se_wd.wd_timer_lsb != CPLD_MASK_NOP8) {

				mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) write 0x%02x at 0x%02x\n",
							__func__, __LINE__, p_wd_dev[ind1]->ctrl.timer_value,
							p_wd_dev[ind1]->se_wd.wd_timer_lsb);

				p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
							p_wd_dev[ind1]->se_wd.wd_timer_lsb,
							p_wd_dev[ind1]->se_wd.wd_timer_lsb,
							0, &(p_wd_dev[ind1]->ctrl.timer_value), _B_LOCK);   /* Set watchdog timer */
			}
			p_wd_dev[ind1]->timer_started = 1;
		}
		else {
			/* Set action to default and clear timer */
			if (p_wd_dev[ind1]->se_wd.wd_action != CPLD_MASK_NOP8) {                            /* Watchdog action address */
				p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
							p_wd_dev[ind1]->se_wd.wd_action,
							p_wd_dev[ind1]->se_wd.wd_action,
							0, &p_wd_dev[ind1]->act_def_mask, _B_LOCK);          /* Clear watchdog action */
			}
			if (p_wd_dev[ind1]->se_wd.wd_counter != CPLD_MASK_NOP8) {
				p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
							p_wd_dev[ind1]->se_wd.wd_timer_lsb,
							p_wd_dev[ind1]->se_wd.wd_timer_lsb,
							0, &clear_timer, _B_LOCK);                            /* Clear watchdog timer */
			}
			p_wd_dev[ind1]->timer_started = 0;
		}
	}

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) timer_started= %d %d %d %d\n",
				__func__, __LINE__,
				p_wd_dev[0]->timer_started, p_wd_dev[1]->timer_started,
				p_wd_dev[2]->timer_started, p_wd_dev[3]->timer_started);

	if (ena) {
		/* Write each clear protection bit 0 to remove protection,
	 	  write each clear bit 1 to clear */
		curr_val |= set_mask;
		curr_val_prot &= ~set_mask;

		mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) set_mask=%x write to: addr %x val %x addr %x val %x\n",
					__func__, __LINE__, set_mask, p_wd_cfg->wd_entry.wd_clear_wp, curr_val_prot,
					p_wd_cfg->wd_entry.wd_clear, curr_val_prot);
		if (p_wd_cfg->wd_entry.wd_clear_wp != CPLD_MASK_NOP8)            /* Watchdog clear write protection address */
			p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
						p_wd_cfg->wd_entry.wd_clear_wp,
						p_wd_cfg->wd_entry.wd_clear_wp,
						0, &curr_val_prot, _B_LOCK);      /* Remove protection for configured watchdogs */
		if (p_wd_cfg->wd_entry.wd_clear != CPLD_MASK_NOP8)                /* Watchdog clear address */
			p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
						p_wd_cfg->wd_entry.wd_clear,
						p_wd_cfg->wd_entry.wd_clear,
						0, &curr_val_prot, _B_LOCK);      /* Clear bits for configured watchdogs */
	}
	else {
		/* Write each clear protection bit 0 to remove protection,
	 	  write each clear bit 1 to clear */
		curr_val &= set_mask;
		curr_val_prot |= set_mask;

		mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) write to: addr %x val %x addr %x val %x\n",
					__func__, __LINE__, p_wd_cfg->wd_entry.wd_clear_wp, curr_val,
					p_wd_cfg->wd_entry.wd_clear, curr_val_prot);
		if (p_wd_cfg->wd_entry.wd_clear != CPLD_MASK_NOP8)                /* Watchdog clear address */
			p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
						p_wd_cfg->wd_entry.wd_clear,
						p_wd_cfg->wd_entry.wd_clear,
						0, &curr_val, _B_LOCK);           /* Clear bits for configured watchdogs */
		if (p_wd_cfg->wd_entry.wd_clear_wp != CPLD_MASK_NOP8)             /* Watchdog clear write protection address */
			p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
						p_wd_cfg->wd_entry.wd_clear_wp,
						p_wd_cfg->wd_entry.wd_clear_wp,
						0, &curr_val_prot, _B_LOCK);      /* Set protection for configured watchdogs */
	}

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d)\n", __func__, __LINE__);

	if (ena) {
		if(p_wd_cfg->sw_timer_avail)
			setup_timer(&p_wd_cfg->wd_soft_timer, mlnx_wd_soft_heartbeat, (unsigned long)p_wd_cfg);
	}
	else {
		if(p_wd_cfg->sw_timer_avail)
			del_timer_sync(&p_wd_cfg->wd_soft_timer);
	}

	return 0;
}
EXPORT_SYMBOL(mlnx_watchdog_enable_disable);

int mlnx_watchdog_clear(wd_obj_e wd_id, wd_clear_e wd_clr)
{
	int err = 0;
	u8 curr_val = 0;
	u8 mask = 0x00;
	int ind;

	if ((wd_id > _wd_all_) || (wd_clr > _clear_all))
		return -EINVAL;
	if (wd_clr == _none_clear)
		return 0;

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) wd_id=%x wd_clr=%x\n",
			__func__, __LINE__, wd_id, wd_clr);

	if (wd_clr & _clear_timer)
		mask |= 0x0f;
	if (wd_clr & _clear_counter)
		mask |= 0xf0;

	for (ind = 0; ind < MAX_MGMT_WD; ind++) {
		if (!(wd_id & (1 << ind)))
			continue;

		if (p_wd_dev[ind]->timer_started) {
			if (p_wd_cfg->wd_entry.wd_clear != CPLD_MASK_NOP8) {
				curr_val = p_wd_dev[ind]->wd_mask & mask;
				err = (err >= 0) ? p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
									p_wd_cfg->wd_entry.wd_clear,
									p_wd_cfg->wd_entry.wd_clear,
									0, &curr_val, _B_LOCK) : err;
			}
			if (err < 0)
				return -EFAULT;
	
			mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) curr_val=0x%02x\n",
						__func__, __LINE__, curr_val);

			/* Watchdog already started, keep it alive */
			if (p_wd_cfg->sw_timer_avail)
				mlnx_wd_keepalive(p_wd_dev[ind]);
		}
		else {
			/* Start watchdog with timer value set by init */
			mlnx_wd_set_heartbeat(p_wd_dev[ind], p_wd_dev[ind]->timer_set);
		}
	}

	return 0;
}
EXPORT_SYMBOL(mlnx_watchdog_clear);

int mlnx_watchdog_get_counter(wd_obj_e wd_id, u8 *wd_cntr)
{
	int ind;
	int err;

	if ((wd_id != _wd1_) && (wd_id != _wd2_) &&
		(wd_id != _wd2_) && (wd_id != _wd3_)) {
		printk(KERN_INFO "%s Only one watchdog can be specified, while wd_id=0x%02x\n",
			__func__, wd_id);
		return -EINVAL;
	}

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) wd_id=%x\n",
			__func__, __LINE__, wd_id);

	for (ind = 0; ind < MAX_MGMT_WD; ind++) {
		if (!(wd_id & (1 << ind)))
			continue;
		if (p_wd_dev[ind]->timer_started) {
			if (p_wd_dev[ind]->se_wd.wd_counter != CPLD_MASK_NOP8) {
				err = p_wd_cfg->rw_func((union hw_dev*)&p_wd_cfg->hwdev,
							p_wd_dev[ind]->se_wd.wd_counter,
							p_wd_dev[ind]->se_wd.wd_counter,
							1, wd_cntr, _B_LOCK);
				if (err < 0)
					return -EFAULT;

				mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) wd_cntr=%x\n",
						__func__, __LINE__, *wd_cntr);
						
				return 0;
			}
		}
	}

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) wd_cntr=%x\n",
			__func__, __LINE__, *wd_cntr);

	return 0;
}
EXPORT_SYMBOL(mlnx_watchdog_get_counter);

int mlnx_watchdog_modify_timer(wd_obj_e wd_id, u8 new_margin)
{
	int ind;
	int err = 0;
	if (new_margin > 0x0f)
		return -EINVAL;
	if ((wd_id != _wd1_) && (wd_id != _wd2_) &&
		(wd_id != _wd2_) && (wd_id != _wd3_)) {
		printk(KERN_INFO "%s Only one watchdog can be specified, while wd_id=0x%02x\n",
			__func__, wd_id);
		return -EINVAL;
	}

	mlnx_log(3, MLNX_LOG_WD_MAX, "%s(%d) wd_id=%x\n",
			__func__, __LINE__, wd_id);

	for (ind = 0; ind < MAX_MGMT_WD; ind++) {
		if (!(wd_id & (1 << ind)))
			continue;
		if (p_wd_dev[ind]->timer_started)
			err = (err >= 0) ? mlnx_wd_set_heartbeat(p_wd_dev[ind], new_margin) : err;
	}
	return err;
}
EXPORT_SYMBOL(mlnx_watchdog_modify_timer);


int mlnx_watchdog_notify_subscribe(mlnx_watchdog_notify_func notifunc,
					void *subscriber_data)
{
	int err = 0;
	unsigned long flags;
	struct mlnx_watchdog_notify_entry *p, *n, *e;

	spin_lock_irqsave(&p_wd_cfg->notifier_lock, flags);
 
	list_for_each_entry_safe(p, n, &p_wd_cfg->notifier_list, subscriber_list) {
		if (p->notifunc == notifunc) {
			/* Already exist */
			printk(KERN_INFO "%s(%d) action=%p already exist\n",
				__FUNCTION__, __LINE__, p->notifunc);
			err = -EEXIST;
			goto entry_exist;
		}
	}
	e = kzalloc(sizeof(*e), GFP_ATOMIC);
        if (!e)
                return -ENOMEM;

	e->subscriber_data = subscriber_data;
	e->notifunc = notifunc;
 
	list_add_rcu(&e->subscriber_list, &p_wd_cfg->notifier_list);
 
entry_exist:
	spin_unlock_irqrestore(&p_wd_cfg->notifier_lock, flags);

	return err;
}
EXPORT_SYMBOL(mlnx_watchdog_notify_subscribe);

int mlnx_watchdog_notify_unsubscribe(mlnx_watchdog_notify_func notifunc,
					void *subscriber_data)
{
	unsigned long flags;
	struct mlnx_watchdog_notify_entry *p, *n;
 
	spin_lock_irqsave(&p_wd_cfg->notifier_lock, flags);
 
	list_for_each_entry_safe(p, n, &p_wd_cfg->notifier_list, subscriber_list) {
		if (p->notifunc == notifunc) {
			p->subscriber_data = subscriber_data;
			p->notifunc = notifunc;
			list_del_rcu(&p->subscriber_list);
			kfree(p);
			return 0;
		}
	}
 
	spin_unlock_irqrestore(&p_wd_cfg->notifier_lock, flags);

	return EINVAL;
}
EXPORT_SYMBOL(mlnx_watchdog_notify_unsubscribe);
 
int *mlnx_debug_level_symbol = NULL;
static int __init mlnx_wd_init(void)
{
	int err = 0;
	int ind;
	int ind1;

	memset(p_wd_cfg, 0, sizeof(struct mlnx_watchdog_cfg));

	p_wd_cfg->mlnx_notifier.notifier_call = mlnx_wd_notify_sys;
	err = register_reboot_notifier(&p_wd_cfg->mlnx_notifier);
	if (err) {
		pr_err("cannot register notifier %d)\n", err);
		goto register_reboot_error;
	}

	/* Allocate p_wd_dev entries */
	for (ind = 0; ind < MAX_MGMT_WD; ind++) {
		p_wd_dev[ind] = kzalloc(sizeof(struct mlnx_watchdog_dev), GFP_KERNEL);
		if (!p_wd_dev[ind]) {
			err = -ENOMEM;
			goto out_mem_dev_tab;
		}
	}

	for (ind = 0; ind < MAX_MGMT_WD; ind++) {
		err = misc_register(&mlnx_wd_miscdevs[ind]);

		if (err) {
			pr_err("cannot register miscdev %d err =%d\n",
		       		mlnx_wd_miscdevs[ind].minor, err);
			goto misc_register_error;
		}
	}
 
	p_wd_dev[0]->wd_def_mask = _wd1_timer_clear_m|_wd1_counter_clear_m;
	p_wd_dev[1]->wd_def_mask = _wd2_timer_clear_m|_wd2_counter_clear_m;
	p_wd_dev[2]->wd_def_mask = _wd3_timer_clear_m|_wd3_counter_clear_m;
	p_wd_dev[3]->wd_def_mask = _wd4_timer_clear_m|_wd4_counter_clear_m;

	INIT_LIST_HEAD(&p_wd_cfg->notifier_list);
	spin_lock_init(&p_wd_cfg->notifier_lock);
	spin_lock_init(&p_wd_cfg->heartbeat_lock);

	mlnx_debug_level_symbol = __symbol_get("mlnx_debug_level");

	printk(KERN_INFO "%s four wd entries allocated %p %p %p %p def mask %x %x %x %x\n",
		__func__, p_wd_dev[0], p_wd_dev[1], p_wd_dev[2], p_wd_dev[3],
		p_wd_dev[0]->wd_def_mask, p_wd_dev[1]->wd_def_mask,
		p_wd_dev[2]->wd_def_mask, p_wd_dev[3]->wd_def_mask);

	return 0;

misc_register_error:
	for (ind1 = ind - 1; ind1 >= 0; ind1--)
		misc_deregister(&mlnx_wd_miscdevs[ind1]);
	ind1 = MAX_MGMT_WD;

out_mem_dev_tab:
	for (ind1 = ind - 1; ind1 >= 0; ind1--)
		kfree(p_wd_dev[ind1]);

	unregister_reboot_notifier(&p_wd_cfg->mlnx_notifier);

register_reboot_error:
	mlnx_wd_stop();

	return err;
}

static void __exit mlnx_wd_exit(void)
{
	int ind1;
	unsigned long flags;
	struct mlnx_watchdog_notify_entry *p, *n;

	if(!list_empty(&p_wd_cfg->notifier_list)) {
		spin_lock_irqsave(&p_wd_cfg->notifier_lock, flags);
		list_for_each_entry_safe(p, n, &p_wd_cfg->notifier_list, subscriber_list) {
			list_del_rcu(&p->subscriber_list);
			kfree(p);
		}
		spin_unlock_irqrestore(&p_wd_cfg->notifier_lock, flags);
	}

	for (ind1 = MAX_MGMT_WD - 1; ind1 >= 0; ind1--) {
		misc_deregister(&mlnx_wd_miscdevs[ind1]);
		kfree(p_wd_dev[ind1]);
	}

	unregister_reboot_notifier(&p_wd_cfg->mlnx_notifier);

	mlnx_wd_stop();

	memset(p_wd_cfg, 0, sizeof(struct mlnx_watchdog_cfg));

	if (mlnx_debug_level_symbol)
		__symbol_put("mlnx_debug_level");
}

/************************************************
 *  MODULE init/exit
 ***********************************************/
module_init(mlnx_wd_init);
module_exit(mlnx_wd_exit);

MODULE_AUTHOR("Vadim Pasternak <vadimp@mellanox.com>");
MODULE_DESCRIPTION("Mellanox watchdog driver");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS_MISCDEV(CPLD_WATCHDOG1_MINOR);
MODULE_ALIAS_MISCDEV(CPLD_WATCHDOG2_MINOR);
MODULE_ALIAS_MISCDEV(CPLD_WATCHDOG3_MINOR);
MODULE_ALIAS_MISCDEV(CPLD_WATCHDOG4_MINOR);
MODULE_VERSION(MLNX_WDT_VER);






