/*
 *
 * Copyright (C) Mellanox Technologies Ltd. 2001-2014.  ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of Mellanox Technologies Ltd.
 * (the "Company") and all right, title, and interest in and to the software product,
 * including all associated intellectual property rights, are and shall
 * remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/wait.h>

#include "cpld.h"

#include <sx/sxd/sxd_dpt.h>
#include <sx/sxd/kernel_user.h>
#include <sx/sxd/sxd_access_register.h>

static void help(void)
{
    printf("cpld_ver_upd -v/u -s <system model> -c <cpld_name> [-f <file name>] [-h]\n");
    printf("v - get CPLD version\n");
    printf("u - update CPLD\n");
    printf("s - system model name. Supported systems:\n");
    printf("\t %s, %s, %s, %s, %s, %s\n", SYST_JACKAL, SYST_DINGO, SYST_COYOTE,
           SYST_COYOTEX, SYST_PIRANHAX, SYST_NEPTUNE);
    printf("c - particular CPLD name or ALL for get version of all CPLDs\n");
    printf("f - full path and file name for CPLD update\n");
    printf("h - this help\n");
}

static int check_system_name(char* syst_name, enum system_model* syst)
{
    if (!strncmp(syst_name, SYST_JACKAL, strlen(SYST_JACKAL)))
        *syst = Jackal;
    else if (!strncmp(syst_name, SYST_DINGO, strlen(SYST_DINGO)))
        *syst = Dingo;
    else if (!strncmp(syst_name, SYST_COYOTE, strlen(SYST_COYOTE)))
        *syst = Coyote;
    else if (!strncmp(syst_name, SYST_COYOTE, strlen(SYST_COYOTE)))
        *syst = Coyote;
    else if (!strncmp(syst_name, SYST_COYOTEX, strlen(SYST_COYOTEX)))
        *syst = Coyotex;
    else if (!strncmp(syst_name, SYST_PIRANHAX, strlen(SYST_PIRANHAX)))
        *syst = Piranhax;
    else if (!strncmp(syst_name, SYST_NEPTUNE, strlen(SYST_NEPTUNE)))
        *syst = Neptune;
    else
        return -1;
    return 0;
}

static int check_cpld_name(char* cpld_name, enum cpld_type* cpld)
{
    if (!strncmp(cpld_name, ALL_CPLD, strlen(ALL_CPLD)))
        *cpld = All;
    else if (!strncmp(cpld_name, CPLD_TOR, strlen(CPLD_TOR)))
        *cpld = Cpld_tor;
    else if (!strncmp(cpld_name, CPLD_PORT1, strlen(CPLD_PORT1)))
        *cpld = Cpld_port1;
    else if (!strncmp(cpld_name, CPLD_PORT2, strlen(CPLD_PORT2)))
        *cpld = Cpld_port2;
    else if (!strncmp(cpld_name, CPLD_MEZZ, strlen(CPLD_MEZZ)))
        *cpld = Cpld_mezz;
    else if (!strncmp(cpld_name, CPLD_MNGB, strlen(CPLD_MNGB)))
        *cpld = Cpld_mngb;
    else if (!strncmp(cpld_name, CPLD_SWB, strlen(CPLD_SWB)))
        *cpld = Cpld_swb;
    else
        return -1;

    return 0;
}

static char* system_name_str(enum system_model syst, char* str)
{
    switch (syst)
    {
    case Jackal:
        snprintf(str, strlen(SYST_JACKAL)+1, SYST_JACKAL);
        break;
    case Dingo:
        snprintf(str, strlen(SYST_DINGO)+1, SYST_DINGO);
        break;
    case Coyote:
        snprintf(str, strlen(SYST_COYOTE)+1, SYST_COYOTE);
        break;
    case Coyotex:
        snprintf(str, strlen(SYST_COYOTEX)+1, SYST_COYOTEX);
        break;
    case Piranhax:
        snprintf(str, strlen(SYST_PIRANHAX)+1, SYST_PIRANHAX);
        break;
    case Neptune:
        snprintf(str, strlen(SYST_NEPTUNE)+1, SYST_NEPTUNE);
        break;
    default:
        snprintf(str, 1, " ");
        break;
    }

    return str;
}

static char* cpld_name_str(enum cpld_type cpld, char* str)
{
    switch (cpld)
    {
    case Cpld_tor:
        snprintf(str, strlen(CPLD_TOR)+1, CPLD_TOR);
        break;
    case Cpld_port1:
        snprintf(str, strlen(CPLD_PORT1)+1, CPLD_PORT1);
        break;
    case Cpld_port2:
        snprintf(str, strlen(CPLD_PORT2)+1, CPLD_PORT2);
        break;
    case Cpld_mezz:
        snprintf(str, strlen(CPLD_MEZZ)+1, CPLD_MEZZ);
        break;
    case Cpld_mngb:
        snprintf(str, strlen(CPLD_MNGB)+1, CPLD_MNGB);
        break;
    case Cpld_swb:
        snprintf(str, strlen(CPLD_SWB)+1, CPLD_SWB);
        break;
    case All:
        snprintf(str, strlen(ALL_CPLD)+1, ALL_CPLD);
        break;
    default:
        snprintf(str, 1, " ");
        break;
    }
    return str;
}

static int helper_get_shell_str(const char* syscmd, char* str, int str_num,
                                int max_len)
{
    FILE *fp;
    char *tmp;
    int i, rc = 0;

    fp = popen(syscmd, "r");
    if (fp == NULL)
    {
        fprintf(stderr, "Failed to run command: %s\n", syscmd);
        return -1;
    }

    for (i = 0; i < str_num; i++)
    {
        tmp = str + max_len * i;
        if (!fgets(tmp, max_len, fp))
        {
            fprintf(stderr, "Failed to get info in command: %s\n", syscmd);
            rc = -1;
        }
    }

    pclose(fp);
    return rc;
}

static int helper_execute_shell_cmd(const char* syscmd, int quiet)
{
    int exit_stat, rc = 0;

    rc = system(syscmd);
    if (rc == -1)
    {
        fprintf(stderr, "Fail to call system command %s\n", syscmd);
        rc = -1;
    }
    else
    {
        exit_stat = WEXITSTATUS(rc);
        if (exit_stat != EXIT_SUCCESS)
        {
            if (!quiet)
                fprintf(stderr, "Fail to make %s, Error %s\n", syscmd,
                        strerror(exit_stat));
            rc = -1;
        }
    }
    return rc;
}

static struct system_info* get_system_info(enum system_model syst)
{
    struct system_info* syst_info = NULL;

    switch (syst)
    {
    case Jackal:
        syst_info = &jackal_info;
        break;
    case Dingo:
        syst_info = &dingo_info;
        break;
    case Coyote:
        syst_info = &coyote_info;
        break;
    case Coyotex:
        syst_info = &coyotex_info;
        break;
    case Piranhax:
        syst_info = &piranhax_info;
        break;
    case Neptune:
        syst_info = &neptune_info;
        break;
    default:
        break;
    }

    return syst_info;
}

static struct cpld_info* get_cpld_info(struct system_info* syst_info,
                                       enum cpld_type cpld)
{
    struct cpld_info* found_cpld = NULL;
    int i;

    for (i = 0; i < CPLD_MAX_NUM; i++)
    {
        if (cpld == syst_info->cplds[i].type)
        {
            found_cpld = &(syst_info->cplds[i]);
            break;
        }
    }
    return found_cpld;
}

static int cpld_access(uint16_t bus, uint16_t addr, \
                       enum access_type acc, enum rw_operation rw, \
                       int offs, uint8_t* data)
{
    int rc;
    char syscmd[255];
    char str_output[16];
    uint16_t lpc_addr;

    if (acc != i2c_access && acc != lpc_access)
    {
        fprintf(stderr, "System has incorrect access %d.\n", acc);
        return -1;
    }

    if (acc == i2c_access)
    {
        if (rw)
            snprintf(syscmd, 254, "%s/mellaggra _read %d %x %x %x %x",
                     EXECUTABLE_PATH, bus, addr, offs, DFLT_ADDR_WDTH,
                     DFLT_ACCESS_LEN);
        else
            snprintf(syscmd, 254, "%s/mellaggra _write %d %x %x %x %x %x",
                     EXECUTABLE_PATH, bus, addr, offs, DFLT_ADDR_WDTH,
                     DFLT_ACCESS_LEN, *data);
    }
    else
    {
        rc = helper_execute_shell_cmd("modprobe lpci2c", 0);
        if (rc != 0) {
            fprintf(stderr, "lpci2c load has failed. Please review dmesg\n");
            return rc;
        }
        lpc_addr = CPLD_BASE_ADDR + offs;
        if (rw)
            snprintf(syscmd, 254,
                     "%s/iorw -r -b 0x%x -l 0x%x | awk '{print $5}' | cut -c 3-",
                     EXECUTABLE_PATH, lpc_addr, DFLT_ACCESS_LEN);
        else
            snprintf(syscmd, 254, "%s/iorw -w -b 0x%x -l 0x%x -v 0x%x",
                     EXECUTABLE_PATH, lpc_addr, DFLT_ACCESS_LEN, *data);
    }

    if (rw)
    {
        rc = helper_get_shell_str(syscmd, str_output, 1, 16);
        if (!rc) *data = (uint8_t)strtol(str_output, NULL, 16);
    }
    else
    {
        rc = helper_execute_shell_cmd(syscmd, 0);
    }

    return rc;
}

static int find_sx_dev_id(enum system_model syst, int* sx_dev_id)
{
    char syst_str[24];

    switch (syst)
    {
    case Dingo:
    case Piranhax:
    case Neptune:
        *sx_dev_id = 1;
        break;
    default:
        *sx_dev_id = 0;
        fprintf(stderr, "There is no need to find sx dev for system %s\n",
                system_name_str(syst, syst_str));
        return -1;
    }

    return 0;
}

static int install_cpld_module(enum system_model syst)
{
    int rc;
    char param_list[20];
    char syscmd[128];
    char syst_str[24];

    switch (syst)
    {
    case Jackal:
        snprintf(param_list, 20, "pin_type=1");
        break;
    case Coyotex:
        snprintf(param_list, 20, " ");
        break;
    default:
        fprintf(stderr, "No need in CPLD FU module for system %s\n",
                system_name_str(syst, syst_str));
        return -1;
    }

    // Check if module is aleady installed
    strncpy(syscmd, "lsmod | grep -q cpld_jtag", 28);
    rc = helper_execute_shell_cmd(syscmd, 1);
    if (!rc)
        return 0;

    snprintf(syscmd, 128, "modprobe cpld_jtag %s", param_list);
    rc = helper_execute_shell_cmd(syscmd, 0);
    if (rc)
    fprintf(stderr, "%s failed to install cpld_jtag module\n", \
            system_name_str(syst, syst_str));

    return rc;
}

static int find_cpld_num(enum cpld_type cpld, int* cpld_num)
{
    *cpld_num = 0;

    switch (cpld)
    {
    case Cpld_tor:
        *cpld_num = 1;
        break;
    case Cpld_port1:
        *cpld_num = 2;
        break;
    case Cpld_port2:
        *cpld_num = 3;
        break;
    default:
        fprintf(stderr, "Incorrect cpld type %d\n", cpld);
        return -1;
    }

    return 0;
}

static int cpld_select(enum system_model syst, enum cpld_type cpld,
                       int cpld_num)
{
    int rc;
    uint8_t val;
    char syst_str[24], cpld_str[24];
    enum access_type acc;
    struct system_info* syst_info;
    struct cpld_info* cpld_in;
    uint16_t bus, addr;

    val = 1 << cpld_num;
    val |= 1;

    syst_info = get_system_info(syst);
    if (!syst_info) {
        fprintf(stderr, "Failed get system info for %s\n", \
                system_name_str(syst, syst_str));
        return -1;
    }
    cpld_in = get_cpld_info(syst_info, Cpld_tor);
    if (!cpld_in) {
        fprintf(stderr, "System %s doesn't have cpld %s\n", \
                system_name_str(syst, syst_str), cpld_name_str(Cpld_tor, cpld_str));
        return -1;
    }

    switch (syst) {
    case Jackal:
    case Coyote:
        acc = i2c_access;
        bus = cpld_in->bus;
        addr = cpld_in->addr_idx;
        break;
    case Coyotex:
        bus = 0xff;
        addr = CPLD_BASE_ADDR;
        acc = lpc_access;
        break;
    default:
        fprintf(stderr, "Select doesn't required for system %s\n", \
                system_name_str(syst, syst_str));
        return -1;

    }

    rc = cpld_access(bus, addr, acc, write_op, CPLD_FU_REG, &val);
    if (rc)
        fprintf(stderr, "%s failed to select CPLD %s for update\n",
                system_name_str(syst, syst_str), cpld_name_str(cpld, cpld_str));

    return rc;
}

static int perform_update_cmd(enum system_model syst, enum cpld_type cpld,
                              const char* cpld_update, int cpld_num,
                              int sx_dev_id)
{
    int rc;
    char upd_cmd[256];

    switch (syst)
    {
    case Jackal:
        snprintf(upd_cmd, 256, "%s/jam -n%d -t1 -e -aPROGRAM %s",
                 EXECUTABLE_PATH, cpld_num, cpld_update);
        break;
    case Coyote:
        snprintf(upd_cmd, 256, "%s/jamdpt -aPROGRAM %s",
                 EXECUTABLE_PATH, cpld_update);
        break;
    case Coyotex:
        if (cpld == Cpld_mezz)
            snprintf(upd_cmd, 256, "%s/jam -e -aPROGRAM %s",
                     EXECUTABLE_PATH, cpld_update);
        else
            snprintf(upd_cmd, 256, "%s/jamdpt -aPROGRAM %s",
                     EXECUTABLE_PATH, cpld_update);
        break;
    case Dingo:
    case Piranhax:
    case Neptune:
        snprintf(upd_cmd, 256, "cpldupdate --dev mlnxsw-%d %s",
                 sx_dev_id, cpld_update);
        break;
    default:
        return -1;
    }

    printf("CPLD update cmd: %s\n", upd_cmd);
    rc = helper_execute_shell_cmd(upd_cmd, 0);
    if (rc)
        fprintf(stderr, "Failed to perform %s in CPLD update \n", upd_cmd);

    return rc;
}

static int get_cpld_vers_reg(enum cpld_type cpld, u_int8_t* ver_reg)
{
    switch (cpld)
    {
    case Cpld_tor:
    case Cpld_mngb:
        *ver_reg = CPLD1_VERSION_REG;
        break;
    case Cpld_swb:
        *ver_reg = CPLD2_VERSION_REG;
        break;
    case Cpld_mezz:
        *ver_reg = CPLD_MEZZ_VERS_REG;
        break;
    case Cpld_port1:    // Accessed by FW
    case Cpld_port2:
    default:
        return -1;
    }

    return 0;
}

static int sdk_access_init(enum system_model syst)
{
    int rc = 0;

    switch (syst) {
    case Jackal:
    case Coyote:
    case Coyotex:
    case Piranhax:
    case Neptune:
        rc = sxd_access_reg_init(0, NULL, SX_LOG_ERROR);
        if (rc)
            fprintf(stderr, "sxd_access_reg_init failed with status 0x%x\n", rc);
        break;
    default:
        return 0;
    }
    return rc;
}

static int get_cpld_fw_version(u_int8_t cpld_fw_idx, u_int8_t* cpld_ver)
{
    struct ku_msci_reg msci_reg;
    sxd_reg_meta_t reg_meta;
    int rc;

    *cpld_ver = 0;
    memset(&msci_reg, 0, sizeof(struct ku_msci_reg));
    memset(&reg_meta, 0, sizeof(sxd_reg_meta_t));
    msci_reg.index = cpld_fw_idx;
    reg_meta.dev_id = 1;
    reg_meta.access_cmd = SXD_ACCESS_CMD_GET;
    rc = sxd_access_reg_msci(&msci_reg, &reg_meta, 1, NULL, NULL);
    if (rc)
    {
        fprintf(stderr, "Failed to read CPLD (idx %d) vers from SX\n", cpld_fw_idx);
        return -1;
    }
    *cpld_ver = (uint8_t) msci_reg.version;

    return 0;
}

static int cpld_version(enum system_model syst, enum cpld_type cpld)
{
    int i, ret, rc = 0;
    struct system_info* syst_info;
    struct cpld_info* curr_cpld;
    enum access_type acc;
    char syst_str[24], cpld_str[24];
    uint8_t cpld_fw_idx;
    uint8_t ver_reg;
    uint8_t cpld_ver;
    uint16_t bus, addr;

    syst_info = get_system_info(syst);
    if (!syst)
    {
        fprintf(stderr, "Failed get info for system %s\n",
                system_name_str(syst, syst_str));
        return -1;
    }

    if (cpld == All)
    {
        for (i = 0; i < syst_info->cpld_num; i++)
        {
            ret = cpld_version(syst, syst_info->cplds[i].type);
            rc += ret;
        }
        return rc;
    }

    curr_cpld = get_cpld_info(syst_info, cpld);
    if (!curr_cpld)
    {
        fprintf(stderr, "System %s doesn't have cpld %s\n",
                system_name_str(syst, syst_str), cpld_name_str(cpld, cpld_str));
        return -1;
    }
    acc = curr_cpld->acc_type;

    switch (acc)
    {
    case i2c_access:
    case lpc_access:
        bus = curr_cpld->bus;
        addr = curr_cpld->addr_idx;
        rc = get_cpld_vers_reg(cpld, &ver_reg);
        if (rc)
        {
            fprintf(stderr, "System %s cpld %s can't find vers_reg\n",
                    system_name_str(syst, syst_str), cpld_name_str(cpld, cpld_str));
            return -1;
        }
        rc = cpld_access(bus, addr, acc, read_op, ver_reg, &cpld_ver);
        if (rc)
        {
            fprintf(stderr, "System %s cpld %s can't access to vers_reg\n",
                    system_name_str(syst, syst_str), cpld_name_str(cpld, cpld_str));
            return -1;
        }
        break;
    case fw_access:
        cpld_fw_idx = curr_cpld->addr_idx;
        rc = get_cpld_fw_version(cpld_fw_idx, &cpld_ver);
        if (rc)
        {
            fprintf(stderr,
                    "System %s cpld %s failed to get vers through FW\n",
                    system_name_str(syst, syst_str), cpld_name_str(cpld, cpld_str));
            return -1;
        }
        break;
    default:
        fprintf(stderr, "System %s has incorrect access.\n", system_name_str(syst, syst_str));
        return -1;
    }

    printf("CPLD:\t%s\t0x%02x\n", cpld_name_str(cpld, cpld_str), cpld_ver);
    return rc;
}

static int cpld_update(enum system_model syst, enum cpld_type cpld,
                       char* fname)
{
    int cpld_num, sx_dev_id, rc = 0;

    switch (syst)
    {
    case Jackal:
        rc = install_cpld_module(syst);
        if (!rc)
        {
            rc = find_cpld_num(cpld, &cpld_num);
            if (!rc)
                rc = perform_update_cmd(syst, cpld, fname, cpld_num, 0);
        }
        break;
    case Coyote:
        rc = find_cpld_num(cpld, &cpld_num);
        if (!rc)
        {
            rc = cpld_select(syst, cpld, cpld_num);
            if (!rc)
                rc = perform_update_cmd(syst, cpld, fname, cpld_num, 0);
        }
        break;
    case Coyotex:
        if (cpld == Cpld_mezz)
        {
            rc = install_cpld_module(syst);
            if (!rc)
                rc = perform_update_cmd(syst, cpld, fname, 1, 0);
        }
        else
        {
            rc = find_cpld_num(cpld, &cpld_num);
            if (!rc)
            {
                rc = cpld_select(syst, cpld, cpld_num);
                if (!rc)
                    rc = perform_update_cmd(syst, cpld, fname, cpld_num, 0);
            }
        }
        break;
    case Dingo:
    case Piranhax:
    case Neptune:
        rc = find_sx_dev_id(syst, &sx_dev_id);
        if (!rc)
            rc = perform_update_cmd(syst, cpld, fname, 1, sx_dev_id);
        break;
    default:
        return -1;
    }

    return rc;
}

int main(int argc, char *argv[])
{
    int opt, rc = 0;
    char fname[255], syst_name[32], cpld_name[32];
    int ver_check = -1;
    enum system_model syst = Incorrect_syst;
    enum cpld_type cpld = Incorrect_cpld;
    char syst_str[24], cpld_str[24];

    if (argc < 2)
    {
        fprintf(stderr, "Incorrect number of parameters, should be at least 2\n");
        help();
        exit(-1);
    }

    memset(syst_name, 0, sizeof(syst_name));
    memset(fname, 0, sizeof(fname));
    memset(cpld_name, 0, sizeof(cpld_name));

    while ((opt = getopt(argc, argv, "s:c:f:vuh")) != -1)
    {
        switch (opt)
        {
        case 's':
            strncpy(syst_name, optarg, 31);
            break;
        case 'v':
            ver_check = 1;
            break;
        case 'u':
            ver_check = 0;
            break;
        case 'c':
            strncpy(cpld_name, optarg, 31);
            break;
        case 'f':
            strncpy(fname, optarg, 254);
            break;
        case 'h':
            help();
            break;
        default:
            fprintf(stderr, "Incorrect option input %c\n", opt);
            help();
            exit(-1);
            break;
        }
    }

    if (ver_check < 0)
    {
        fprintf(stderr, "Version or update option is not specified\n");
        help();
        exit(-1);
    }
    rc = check_system_name(syst_name, &syst);
    if (rc)
    {
        fprintf(stderr, "Incorrect system name %s\n", syst_name);
        help();
        exit(-1);
    }
    rc = check_cpld_name(cpld_name, &cpld);
    if (rc)
    {
        fprintf(stderr, "Incorrect cpld name %s\n", cpld_name);
        help();
        exit(-1);
    }

    if (ver_check) {
        rc = sdk_access_init(syst);
        if (!rc) {
            rc = cpld_version(syst, cpld);
            if (rc)
            {
                fprintf(stderr, "System %s failed to get cpld %s version\n",
                        system_name_str(syst, syst_str), cpld_name_str(cpld, cpld_str));
            }
        }
    }
    else
    {
        if (!fname[0])
        {
            fprintf(stderr, "Full path and CPLD update file name should be provided\n");
            help();
            exit(-1);
        }
        rc = cpld_update(syst, cpld, fname);
        if (rc)
            fprintf(stderr, "System %s failed to update cpld %s\n",
                    system_name_str(syst, syst_str), cpld_name_str(cpld, cpld_str));
        else
            printf("System %s success to update cpld %s\n",
                   system_name_str(syst, syst_str), cpld_name_str(cpld, cpld_str));
    }

    exit(rc);
}

