/**
 *
 * Copyright (C) 2012-2014, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/i2c.h>
#include <linux/hwmon.h>
#include <linux/hwmon-sysfs.h>
#include <linux/err.h>
#include <linux/mutex.h>
#include <linux/sysfs.h>
#include <linux/platform_device.h>
#include <linux/mod_devicetable.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/delay.h>
#include "sx_glue_if.h"
#include "switchx.h"
#include "mlnx_i2c_if.h"
#include "mlnx_log.h"

static unsigned short cir_reg_num = 1;
module_param(cir_reg_num, ushort, 1);
MODULE_PARM_DESC(cir_reg_num, "Command Interface Register to be used, default for i2c is 1");

#define SX_I2C_DEV_SIGNATURE 0xf0e1c2b3
#define SXDBGMODE 1

#define SWITCHX_ADDR32 4
#define MAX_I2C_BUFF_SIZE 256

#define BUS_ID_FROM_ID(i2c_dev_id)   ( (i2c_dev_id >> 8) & 0x0000ffff)
#define ADDR_FROM_DEV_ID(i2c_dev_id) ( i2c_dev_id & 0x7f )
#define GET_WIDTH(off)               ( (off > 0xffff) ? 4 : (off > 0xff) ? 2 : 1 )
	 
#define _FL_ struct file *fl,

#if defined(__i386__) || defined(__x86_64__)
#define CPLD_BAG_WA 0
#endif

#define LPCI2C_BLOCK_MAX 32


/* Addresses to scan */
static const unsigned short normal_i2c[] = { 0x48, I2C_CLIENT_END };

static struct i2c_mux_priv_cache cache_template[SX_ROUTE_DEPTH - 1];
extern int __last_accessed_bus_id;
extern struct i2c_mux_priv_cache *__p_last_bus_cache_path[SX_ROUTE_DEPTH - 1];
extern struct mutex __i2c_mlnx_lock;
extern int (*sx_i2c_get_probe_status_hook)(struct i2c_client *client, u8 *probe_status, 
					   int *probe_cntr, u8 *go_bit_status,
					   u8 *hcr_acess_status);

/* Each client has this additional data */
struct switchx_data {
	struct list_head list;
	int signature;
	struct device *i2c_dev;
	struct i2c_client *client;
	struct reg_cir_layout reg_cir;
	u32 mb_size_in;
	u32 mb_offset_in;
	u32 mb_size_out;
	u32 mb_offset_out;
	int retries;
	int timeout;
	int retry_cntr;
	u64 transact_id;
	int bus_id;
	int i2c_dev_addr;
	u8 cache[SX_MAX_REG_SIZE];
	struct mutex cmd_lock;
#define SX_ROUTE_DEPTH 4
	struct i2c_adapter *parent[SX_ROUTE_DEPTH];
	struct i2c_mux_priv *mux_priv[SX_ROUTE_DEPTH];
	int (*deselect[SX_ROUTE_DEPTH - 1])(struct i2c_adapter *, void *mux_dev, u32 chan_id);
	u8 route_enforced;
#define GO_BIT_STAT_OK    0
#define GO_BIT_STAT_STUCK 1
	u8 go_bit_status;
	u8 hcr_access_status;
};

/* Container structure */
struct switchx_probe_failed {
	struct list_head list;
	int bus_id;
	int probe_cntr;
	u8 probe_status;
};

/* Container structure */
struct switchx_config {
	struct list_head switchx_list;
	struct list_head probe_failed_list;
	u32 reg_cir_offset;
	u32 reg_sw_reset_offset;
	u32 last_access_dev_id;
	int last_accessed_bus_id;
};
static struct switchx_config switchx_cfg;

static struct kmem_cache *switchx_probe_failed_cache __read_mostly;

static inline int translate_offset_addr(u32 addr, u32 offset, u32 len, u8 width, char *msgbuf,
				  	struct i2c_msg *msg, u8 *in_out_buf, u8 flag);

#define ICR_REG_WRITE_GO_BIT(event, op_modifier, op, icr_reg)                          \
        cpu_to_be32(((1 << GO_BIT) | (event ? (1 << EVENT_BIT) : 0) |                  \
                        (op_modifier << OPMOD_SHIFT) | op), icr_reg);

#define SMBUS_XFER(client, data, rw_smbus_buf, read_write, command, size)              \
	orig_jiffies = jiffies;                                                        \
	for (try = 0; try <= data->retries; try++) {                                   \
		rc = i2c_smbus_xfer(client->adapter, client->addr, client->flags,      \
		   			read_write, command, size, &rw_smbus_buf);     \
		if (rc >= 0) {		                                               \
			break;                                                         \
		}                                                                      \
		if (time_after(jiffies, orig_jiffies + data->timeout)) {               \
			rc = -EIO;                                                     \
			mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX, "%s try=%d rc=%d\n", __FUNCTION__, try, rc);            \
			return rc;                                                     \
		}                                                                      \
	}                                                                              \
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX, "%s(%d) try=%d rc=%d\n", __FUNCTION__, __LINE__, try, rc);

#define I2C_XFER(client, data, msgs, try, orig_jiffies, read_write)            \
	orig_jiffies = jiffies;                                                \
	for (try = 0; try <= data->retries; try++) {                           \
		rc = i2c_transfer(client->adapter, msgs, read_write);          \
		if (rc > 0) {		                                       \
			break;                                                 \
		}                                                              \
		if (time_after(jiffies, orig_jiffies + data->timeout)) {       \
			rc = -EIO;                                             \
			mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX, "%s try=%d rc=%d\n", __FUNCTION__, try, rc);    \
			return rc;                                             \
		}                                                              \
	}                                                                      \
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX, "%s try=%d rc=%d\n", __FUNCTION__, try, rc);

#define I2C_SET_BUF_XFER(client, data, try, orig_jiffies, tbuf, msgs, msgbuf, offset,         \
				size, width, rw_flag)                                         \
	translate_offset_addr(client->addr, offset, size,                                     \
				width, msgbuf, msgs, tbuf, rw_flag);                          \
	I2C_XFER(client, data, msgs, try, orig_jiffies, rw_flag);                             \
	if (rc < 0) {                                                                         \
		printk(KERN_INFO "%s(%d) rc=%d\n", __FUNCTION__, __LINE__, rc);               \
		return rc;                                                                    \
	}
 
#ifdef SMBUS_USED
#define BUS_XFER(c,d,rw,p,q,s) SMBUS_XFER(client, data, rw_smbus_buf, read_write, command, size)
#else
#define BUS_XFER(c,d,rw,p,q,s) I2C_XFER(client, data, msgs, try, orig_jiffies, read_write)
#endif


#ifdef SMBUS_USED

static inline int smbbus_xfer_chunk(struct i2c_client *client, u32 offset, u32 datalen,
				    u8 rw_oper, u8 *in_out_buf)
{
	struct switchx_data *data = i2c_get_clientdata(client);
	int chunk_size = I2C_SMBUS_BLOCK_MAX;
	union i2c_smbus_data rw_smbus_buf;
	char read_write;
	u8 command = 0;
	int size;
	u8 *p_buf = in_out_buf;
	u32 len = datalen;
	unsigned long orig_jiffies;
	int try = 0;
        int rc = 0;
	memset(&rw_smbus_buf, 0, sizeof(union i2c_smbus_data));

	if (rw_oper == I2C_SMBUS_WRITE) {
		/* Write first buffer with addres and data */
		size = I2C_SMBUS_BLOCK_DATA;//I2C_SMBUS_I2C_BLOCK_DATA;
		command = (offset >> 24) & 0x00ff;
		/* Stole three data bytes for address pointing */
		rw_smbus_buf.block[1] = (offset >> 16) & 0x00ff;
		rw_smbus_buf.block[2] = (offset >> 8) & 0x00ff;
		rw_smbus_buf.block[3] = offset & 0x00ff;
		read_write = I2C_SMBUS_WRITE;
		chunk_size = (len > I2C_SMBUS_BLOCK_MAX - 4) ?
				(I2C_SMBUS_BLOCK_MAX - 4) : len;
		rw_smbus_buf.block[0] = chunk_size + 3;
		memcpy(&rw_smbus_buf.block[4], p_buf, chunk_size);
		mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
			"%s(%d) bus=%d addr=0x%x offset=0x%08x datalen=%d len=%d chunk_size=%d rc=%d command=%x block 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n", 
			__FUNCTION__, __LINE__, client->adapter->nr, client->addr,  
			offset, datalen, len, chunk_size, rc, command, rw_smbus_buf.block[0],
			rw_smbus_buf.block[1], rw_smbus_buf.block[2], rw_smbus_buf.block[3], rw_smbus_buf.block[4],
			rw_smbus_buf.block[5], rw_smbus_buf.block[6], rw_smbus_buf.block[7], rw_smbus_buf.block[8],
			rw_smbus_buf.block[9], rw_smbus_buf.block[10], rw_smbus_buf.block[11], rw_smbus_buf.block[12],
			rw_smbus_buf.block[13], rw_smbus_buf.block[14], rw_smbus_buf.block[15], rw_smbus_buf.block[16],
			rw_smbus_buf.block[17], rw_smbus_buf.block[18], rw_smbus_buf.block[19], rw_smbus_buf.block[20],
			rw_smbus_buf.block[21], rw_smbus_buf.block[22], rw_smbus_buf.block[23], rw_smbus_buf.block[24],
			rw_smbus_buf.block[25], rw_smbus_buf.block[26], rw_smbus_buf.block[27], rw_smbus_buf.block[28],
			rw_smbus_buf.block[29], rw_smbus_buf.block[30], rw_smbus_buf.block[31]);
		//SMBUS_XFER(client, NULL, rw_smbus_buf, read_write, command, size);
		rc = i2c_smbus_write_i2c_block_data(client, command, rw_smbus_buf.block[0],
							&rw_smbus_buf.block[1]);
		if (rc < 0)
			return rc;	
		offset += chunk_size;
		p_buf += chunk_size;
		datalen -= chunk_size;
		//SMBUS_XFER(client, NULL, rw_smbus_buf, read_write, command, size);
		memset(&rw_smbus_buf, 0, sizeof(union i2c_smbus_data));
	}
	switch (datalen) {
		case 1:
		case 2:
			if (datalen == 1)
				size = I2C_SMBUS_BYTE_DATA;
			else
				size = I2C_SMBUS_WORD_DATA;
			read_write = rw_oper;
			if (rw_oper == I2C_SMBUS_WRITE) {
				command = offset; /// ???? 
				memcpy(&rw_smbus_buf.block[0], p_buf, datalen);
			}
			mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		    		"%s(%d) bus=%d addr=0x%x rw_oper=%d size=%d cmd=%d offset=0x%08x datalen=%d len=%d\n", 
		    		__FUNCTION__, __LINE__, client->adapter->nr, client->addr, 
		    		rw_oper, size, command, offset, datalen, len);
			SMBUS_XFER(client, rw_smbus_buf, read_write, command, size);
			if (rw_oper == I2C_SMBUS_READ) {		 
				memcpy(p_buf, &rw_smbus_buf.block[0], datalen);
			}
			mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		    		"%s(%d) bus=%d addr=0x%x size=%d cmd=%d offset=0x%08x datalen=%d len=%d\n", 
		    		__FUNCTION__, __LINE__, client->adapter->nr, client->addr, 
		    		size, command, offset, datalen, len);
			break;
		default:
			while (len > 0) {
			command = 0;
			len = datalen;
			//chunk_size = I2C_SMBUS_BLOCK_MAX;
			size = I2C_SMBUS_I2C_BLOCK_DATA;//I2C_SMBUS_BLOCK_DATA;
			command = (offset >> 24) & 0x00ff;
			/* Stole three data bytes for address pointing */
			rw_smbus_buf.block[1] = (offset >> 16) & 0x00ff;
			rw_smbus_buf.block[2] = (offset >> 8) & 0x00ff;
			rw_smbus_buf.block[3] = offset & 0x00ff;
			read_write = I2C_SMBUS_WRITE;
			chunk_size = (len > I2C_SMBUS_BLOCK_MAX - 4) ?
					(I2C_SMBUS_BLOCK_MAX - 4) : len;
			rw_smbus_buf.block[0] = chunk_size + 3;
			if (rw_oper == I2C_SMBUS_WRITE)
				memcpy(&rw_smbus_buf.block[4], p_buf, chunk_size);
				if (len > chunk_size)
					datalen = chunk_size;
				else
					datalen = len;
				read_write = rw_oper;
				//if (datalen == 2) {
				//	size = I2C_SMBUS_WORD_DATA;
				//} else if (datalen == 1) {
				//	size = I2C_SMBUS_BYTE;
				//} else {
				//	rw_smbus_buf.block[0] = datalen;
				//	size = I2C_SMBUS_I2C_BLOCK_DATA;
				//}
				mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		    			"%s(%d) bus=%d addr=0x%x rw_oper=%d size=%d cmd=%d offset=0x%x datalen=%d len=%d chunk_size=%d tran=%d\n", 
		    			__FUNCTION__, __LINE__, client->adapter->nr, client->addr, 
		    			rw_oper, size, command, offset, datalen, len, chunk_size,
		    			rw_smbus_buf.block[0]);
				if (datalen == 1) {
				//	if (rw_oper == I2C_SMBUS_READ)
				//	 	rc = *p_buf = i2c_smbus_read_byte(client);
				//	else
				// 	 	rc = i2c_smbus_write_byte(client, *p_buf);
				//	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX, "%s(%d) rc=%d p_buf = 0x%x\n", 
		   	    	//	 	__FUNCTION__, __LINE__, rc, *p_buf);
				//	if (rc < 0)
				//		return rc;
				}
				else {
					if (rw_oper == I2C_SMBUS_WRITE) {
						memcpy(&rw_smbus_buf.block[4], p_buf, chunk_size);
					}
					SMBUS_XFER(client, NULL, rw_smbus_buf, read_write, command, size);
					if (rw_oper == I2C_SMBUS_READ) {
						if (len <= 2) {
							memcpy(p_buf, &rw_smbus_buf.block[0], datalen);
						} else {
							memcpy(p_buf, &rw_smbus_buf.block[1], datalen);
						}
					}
					mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		   	    			"%s(%d) rc=%d block %x %x %x %x %x %x %x in_out_buf %x %x %x %x %x %x %x %x\n", 
		   	    			__FUNCTION__, __LINE__, rc, rw_smbus_buf.block[0],
		   	    			rw_smbus_buf.block[1], rw_smbus_buf.block[2], 
		   	    			rw_smbus_buf.block[3], rw_smbus_buf.block[4],
                                		rw_smbus_buf.block[5], rw_smbus_buf.block[6],
                                		*in_out_buf, *(in_out_buf + 1), *(in_out_buf + 2), *(in_out_buf + 3),
                                		*(in_out_buf + 4), *(in_out_buf + 5), *(in_out_buf + 6),
                                		*(in_out_buf + 7));
				}				
				len -= datalen;
				offset += datalen;
				p_buf += datalen;
				datalen = len;
				memset(&rw_smbus_buf, 0, sizeof(union i2c_smbus_data));
				mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		   	    		"%s(%d) offset=0x%x datalen=%d len=%d\n", 
		   	    		__FUNCTION__, __LINE__, offset, datalen, len);
			} /*while*/		 
		break; /*default*/
	} /* switch(datalen)*/
	return rc;
}

static inline int translate_offset_addr(struct i2c_client *client, u32 offset, u32 len, u8 width, char *msgbuf,
				  	struct i2c_msg *msg, u8 *in_out_buf, u8 flag)
{
	//struct switchx_data *data = i2c_get_clientdata(client);
	union i2c_smbus_data rw_smbus_buf;
	u8 command;
	int size;
	int rc = 0;

	if (flag == I2C_SMBUS_READ) {
		memset(&rw_smbus_buf, 0, sizeof(union i2c_smbus_data));
		size = I2C_SMBUS_BLOCK_DATA;
		command = (offset >> 24) & 0x00ff;
		rw_smbus_buf.block[0] = 3;
		rw_smbus_buf.block[1] = (offset >> 16) & 0x00ff;
		rw_smbus_buf.block[2] = (offset >> 8) & 0x00ff;
		rw_smbus_buf.block[3] = offset & 0x00ff;
		mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
			"%s(%d) point addr bus=%d addr=0x%x offset=0x%08x len=%d rc=%d size=%d command=0x%x block 0x%x 0x%x 0x%x 0x%x  0x%x\n", 
			__FUNCTION__, __LINE__, client->adapter->nr, client->addr, offset, len, rc,
			size, command, rw_smbus_buf.block[0], rw_smbus_buf.block[1], 
			rw_smbus_buf.block[2], rw_smbus_buf.block[3], rw_smbus_buf.block[4]);
		/* Write before read to set device address for reading */
		rc = i2c_smbus_write_i2c_block_data(client, command, 3, &rw_smbus_buf.block[1]);
		rc = (rc >= 0) ? smbbus_xfer_chunk(client, offset, len, I2C_SMBUS_READ, in_out_buf) : rc;
		mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		  	 "%s(%d) rc=%d offset=0x%08x len=%d in_out_buf %x %x %x %x %x %x %x %x\n", 
		  	 __FUNCTION__, __LINE__, rc, offset, len,
		  	 *in_out_buf, *(in_out_buf + 1), *(in_out_buf + 2), *(in_out_buf + 3),
		  	 *(in_out_buf + 4), *(in_out_buf + 5), *(in_out_buf + 6), *(in_out_buf + 7));
	}
	else {
		rc = smbbus_xfer_chunk(client, offset, len, I2C_SMBUS_WRITE, in_out_buf);
	}
	return (rc >= 0) ? 1 : rc;
}

static inline int icr_reg_write(struct i2c_client *client, struct switchx_data *data,
				u16 token, u32 in_modifier,
				u8 op_modifier, u16 op)
{
	u32 icr_reg_buf[7];
	int rc = 0;
	memset(icr_reg_buf, 0, 28);
	icr_reg_buf[2] = cpu_to_be32(in_modifier);
	icr_reg_buf[5] = cpu_to_be32(token << 16);
	icr_reg_buf[6] = cpu_to_be32((op_modifier << OPMOD_SHIFT) | op);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		  "%s(%d) icr_reg_buf[2]=0x%08x icr_reg_buf[5]=0x%08x icr_reg_buf[6]=0x%08x\n", 
		  __FUNCTION__, __LINE__, icr_reg_buf[2], icr_reg_buf[5], icr_reg_buf[6]); 
	rc = translate_offset_addr(client, SX_BAR0_OFFSET_SECOND, 28, 4, NULL,
				   NULL, (u8 *)icr_reg_buf, 0);
	if (rc < 0) {
		printk(KERN_INFO "%s(%d) rc=%d\n", __FUNCTION__, __LINE__, rc);
		return rc;
	}
	/* Write go bit as ceparate */
	icr_reg_buf[6] |= cpu_to_be32(1 << GO_BIT);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		  "%s(%d) icr_reg_buf[6]=0x%08x\n", 
		  __FUNCTION__, __LINE__, icr_reg_buf[6]); 
	rc = translate_offset_addr(client, SX_BAR0_OFFSET_SECOND + STATUS_OFFSET, 4, 4, NULL,
				   NULL, (u8 *)&icr_reg_buf[6], 0);
	if (rc < 0) {
		printk(KERN_INFO "%s(%d) rc=%d\n", __FUNCTION__, __LINE__, rc);
		return rc;
	}
	return rc;
}

#else /*SMBUS_USED*/

static inline int icr_reg_write(struct i2c_client *client, struct switchx_data *data,
				u16 token, u32 in_modifier,
				u8 op_modifier, u16 op)
{
	unsigned long orig_jiffies;
	int try;
	struct i2c_msg msgs[1];
	u8 msgbuf[SWITCHX_ADDR32];
	u32 icr_reg_buf[7];
	int rc = -EIO;
	memset(icr_reg_buf, 0, 28);
	icr_reg_buf[2] = cpu_to_be32(in_modifier);
	icr_reg_buf[5] = cpu_to_be32(token << 16);
	icr_reg_buf[6] = cpu_to_be32((op_modifier << OPMOD_SHIFT) | op);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		 "%s(%d) icr_reg_buf[2]=0x%08x icr_reg_buf[5]=0x%08x icr_reg_buf[6]=0x%08x\n", 
		  __FUNCTION__, __LINE__, icr_reg_buf[2], icr_reg_buf[5], icr_reg_buf[6]); 

	I2C_SET_BUF_XFER(client, data, try, orig_jiffies,
				(u8 *)icr_reg_buf, msgs, msgbuf,
				SX_BAR0_OFFSET_SECOND, 28, SWITCHX_ADDR32, 0);

	/* Write go bit as ceparate */
	icr_reg_buf[6] |= cpu_to_be32(1 << GO_BIT);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		  "%s(%d) icr_reg_buf[6]=0x%08x\n", 
		  __FUNCTION__, __LINE__, icr_reg_buf[6]); 
	I2C_SET_BUF_XFER(client, data, try, orig_jiffies,
				(u8 *)&icr_reg_buf[6], msgs, msgbuf,
				SX_BAR0_OFFSET_SECOND + STATUS_OFFSET, 4, SWITCHX_ADDR32, 0);

	return rc;
}

static inline int icr_reg_read(struct i2c_client *client, u32 icr_reg_buf, u8 *status, u8 *go)
{
	int rc = 0;
	*status = icr_reg_buf & 0x000000ff; //>> 24;
	*go = icr_reg_buf & (1 << GO_BIT);
	return rc;
}

static inline void convert_mbox(u32 *p_mbox)
{
	*p_mbox = cpu_to_be32(*p_mbox);
	*(p_mbox + 1) = cpu_to_be32(*(p_mbox + 1));
}

static inline void convert_query_fw_mbox(u32 *p_query_fw_mbox)
{
	int i;
	u32 *p_mbox = p_query_fw_mbox;
	int dword_size = sizeof(struct query_fw_output_mbox_layout) / 4;
	if (sizeof(struct query_fw_output_mbox_layout) % 4)
		dword_size += 1;
	for (i = 0; i < dword_size; i++, p_mbox++)
		*p_mbox = cpu_to_be32(*p_mbox);
}

static inline void convert_oper_tlv(u32 *p_oper_tlv)
{
	int i;
	u32 *p_tlv = p_oper_tlv;
	int dword_size = sizeof(struct operation_tlv) / 4;
	if (sizeof(struct operation_tlv) % 4)
		dword_size += 1;
	for (i = 0; i < dword_size; i++, p_tlv++)
		*p_tlv = cpu_to_be32(*p_tlv);
}

static inline void convert_reg_accs_tlv(u32 *p_reg_tlv)
{
	int i;
	u32 *p_tlv = p_reg_tlv;
	int dword_size = sizeof(struct reg_access_tlv) / 4;
	if (sizeof(struct reg_access_tlv) % 4)
		dword_size += 1;
	for (i = 0; i < dword_size; i++, p_tlv++)
		*p_tlv = cpu_to_be32(*p_tlv);
}

static inline void convert_end_tlv(u32 *p_end_tlv)
{
	int i;
	u32 *p_tlv = p_end_tlv;
	int dword_size = sizeof(struct end_tlv) / 4;
	if (sizeof(struct end_tlv) % 4)
		dword_size += 1;
	for (i = 0; i < dword_size; i++, p_tlv++)
		*p_tlv = cpu_to_be32(*p_tlv);
}

static inline int translate_offset_addr(u32 addr, u32 offset, u32 len, u8 width, char *msgbuf,
				  struct i2c_msg *msg, u8 *in_out_buf, u8 flag)
{
	char *msg0;
	struct i2c_msg *msg_i2c = msg;

	if (flag == I2C_M_RD) {
		msg0 = msgbuf;
		switch (width) {
			case 4:
				*msg0++ = offset >> 24;
				*msg0++ = offset >> 16;
				*msg0++ = offset >> 8;
				*msg0 = offset;
				break;
			case 2:
				offset &= 0xffff;
				*msg0++ = offset >> 8;
				*msg0 = offset;
				break;
			case 1:
				offset &= 0xff;
				*msg0 = offset;
				break;
		}
		msg_i2c->addr = addr;
		msg_i2c->buf = msgbuf;
		msg_i2c->len = width;
		msg_i2c++;
		msg_i2c->addr = addr;
		msg_i2c->flags = I2C_M_RD;
		msg_i2c->buf = in_out_buf;
		msg_i2c->len = len;
	}
	else {
		msg0 = in_out_buf;
		switch (width) {
			case 4:
				*msg0++ = offset >> 24;
				*msg0++ = offset >> 16;
				*msg0++ = offset >> 8;
				*msg0 = offset;
				break;
			case 2:
				offset &= 0xffff;
				*msg0++ = offset >> 8;
				*msg0 = offset;
				break;
			case 1:
				offset &= 0xff;
				*msg0 = offset;
				break;
		}
		msg_i2c->addr = addr;
		msg_i2c->flags = 0;
		msg_i2c->len = width + len;
		msg_i2c->buf = in_out_buf;
	}
	return 1;
}
#endif /*#else SMBUS_USED*/

#define ASIC_HEALTH_ADDR     0x21
#define ASIC_HEALTH_BUS_OFF  0x06
#define ASIC_HEALTH_ADDR_OFF 0x12
#define ASIC_HEALTH_MASK     0x60
#define ASIC_HEALTH_GOOD     0x40
static int asic_bad_health(int bus_id)
{
	struct i2c_adapter *adap = NULL;
	struct i2c_msg msgs[2];
	unsigned long orig_jiffies;
	int try;
	u8 msgbuf[1];
	u8 asic_health;
	int rc;

	if (bus_id == 2)
		return 0;

	adap = i2c_get_adapter(bus_id + ASIC_HEALTH_BUS_OFF);
	if (!adap) {
		mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		        "%s: bad adaptor for bus_id=%d\n",
			__FUNCTION__, bus_id + ASIC_HEALTH_BUS_OFF);
		return -EFAULT;
	}

	memset(msgs, 0, 2*sizeof(struct i2c_msg));
	rc = translate_offset_addr(ASIC_HEALTH_ADDR, ASIC_HEALTH_ADDR_OFF, 1, 1,
				   msgbuf, msgs, &asic_health, I2C_M_RD);


	mutex_lock(&__i2c_mlnx_lock);
	orig_jiffies = jiffies;
	for (try = 0; try <= 3; try++) {
		rc = i2c_transfer(adap, msgs, 2);
		if (rc > 0) {		
			break;
		}
		if (time_after(jiffies, orig_jiffies + 500)) {
			mutex_unlock(&__i2c_mlnx_lock);
			printk(KERN_INFO "%s(%d) bus_id=%d can not read asic_healt\n",
				__FUNCTION__, __LINE__, bus_id);
			return -EIO;
		}
	}
	mutex_unlock(&__i2c_mlnx_lock);

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s bus=%d addr=0x%02x off=0x%02x health byte 0x%02x 0x%02x good 0x%02x asic_health 0x%02x rc=%d\n",
		__FUNCTION__, bus_id + ASIC_HEALTH_BUS_OFF, ASIC_HEALTH_ADDR, ASIC_HEALTH_ADDR_OFF,
		asic_health, (asic_health & ASIC_HEALTH_MASK), ASIC_HEALTH_GOOD, 
		((asic_health & ASIC_HEALTH_MASK) >> 5), rc);

	if ((asic_health & ASIC_HEALTH_MASK) != ASIC_HEALTH_GOOD) {
		printk(KERN_INFO "%s(%d) bus_id=%d asic_health=%x\n",
			__FUNCTION__, __LINE__, bus_id, ((asic_health & ASIC_HEALTH_MASK) >> 5));
		return -EFAULT;
	}

	return 0;
}

static int sx_i2c_software_reset(struct device *dev)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct switchx_data *data = i2c_get_clientdata(client);
	unsigned long orig_jiffies;
	int try;
	u8 tbuf[sizeof(struct reg_sw_reset_layout) + SWITCHX_ADDR32];
	struct reg_sw_reset_layout sw_res;
	int rc = -EIO;
	struct i2c_msg msgs[1];
	u8 msgbuf[SWITCHX_ADDR32];
        
        memset(&sw_res, 0, sizeof(struct reg_sw_reset_layout));
        sw_res.sw_reset = SW_RESET;
	memset(tbuf, 0, sizeof(struct reg_sw_reset_layout) + SWITCHX_ADDR32);
	memcpy(tbuf + SWITCHX_ADDR32, (u8 *)&sw_res, sizeof(struct reg_sw_reset_layout));

	translate_offset_addr(client->addr, switchx_cfg.reg_sw_reset_offset, sizeof(struct reg_sw_reset_layout),
				SWITCHX_ADDR32, msgbuf, msgs, tbuf, 0);

	orig_jiffies = jiffies;
	for (try = 0; try <= data->retries; try++) {
		rc = i2c_transfer(client->adapter, msgs, 1);
		if (rc > 0) {		
			break;
		}
		if (time_after(jiffies, orig_jiffies + data->timeout)) {
			rc = -EIO;
			break;
		}
	}

	return rc;
}
#if !defined(__i386__) && !defined(__x86_64__)
#define LEN_ACQ_GO ( sizeof(struct reg_cir_layout) - offsetof(struct reg_cir_layout, status) )
#define OFF_ACQ_GO ( switchx_cfg.reg_cir_offset + offsetof(struct reg_cir_layout, status) )
#else
#define LEN_ACQ_GO 4
#define OFF_ACQ_GO ( switchx_cfg.reg_cir_offset + 24 )
#endif
static int acquire_go_bit_complete(struct i2c_client *client,
			           struct switchx_data *data,
			           struct reg_cir_layout *reg_cir,
			           int len,
			           int offset,
			           u8 *in_out_buf,
			           int (*f_finish)(struct i2c_client *client, 
					           struct switchx_data *data, 
					           struct reg_cir_layout *reg_cir,
					           int len,
					           int offset,
					           u8 *in_out_buf))
{
	struct reg_cir_layout _reg_cir;
	int rc = -EIO;
	int try = 0;
	unsigned long orig_jiffies;
	struct i2c_msg msgs[2];
	u8 msgbuf[SWITCHX_ADDR32];
#if defined(__i386__) || defined(__x86_64__)
	u8 status = 1;
	u8 go = 1;
#else
	_reg_cir.go = 1; // Just for cleanup
#endif

	memset(&_reg_cir, 0, sizeof(struct reg_cir_layout));
	memset(msgs, 0, 2*sizeof(struct i2c_msg));
 
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s(%d) before getting GO len=%d offset=%d LEN_ACQ_GO=%d OFF_ACQ_GO=0x%08x\n",
		__FUNCTION__, __LINE__, len, offset, LEN_ACQ_GO, OFF_ACQ_GO);
#if defined(__i386__) || defined(__x86_64__)
	translate_offset_addr(client->addr, OFF_ACQ_GO, LEN_ACQ_GO/* + 4*/, SWITCHX_ADDR32, msgbuf, msgs,
				((u8 *)(&_reg_cir) + CIR_STATUS_OFF), I2C_M_RD);
#else
	translate_offset_addr(client->addr, OFF_ACQ_GO, LEN_ACQ_GO, SWITCHX_ADDR32, msgbuf, msgs,
				(u8 *)&_reg_cir.status, I2C_M_RD);
#endif
 
	orig_jiffies = jiffies;
	for (try = 0; try <= data->retries; try++) {
		rc = i2c_transfer(client->adapter, msgs, 2);
#if defined(__i386__) || defined(__x86_64__)
		rc = (rc >= 0) ? icr_reg_read(client, *((u32 *)((u8 *)(&_reg_cir) + CIR_STATUS_OFF)), &status, &go) : rc;
		if ((rc >= 0) && (go == CIR_GO_SW)) {
			mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
				"%s(%d) go=%x status=%x rc=%d\n",
				__FUNCTION__, __LINE__, go, status, rc);
			_reg_cir.status = status;
			_reg_cir.go = go;
#else
		if ((rc > 0) && (_reg_cir.go == CIR_GO_SW)) {
#endif
			if(f_finish)
				rc = f_finish(client, data, &_reg_cir, len, offset, in_out_buf);	
			break;
		}
		if (time_after(jiffies, orig_jiffies + data->timeout)) {
			rc = -EIO;
			break;
		}
		cond_resched();
	}
	data->retry_cntr += try;

	return (try <= data->retries) ? rc : -EIO;
}

static int acquire_go_bit(struct i2c_client *client,
		     	  struct switchx_data *data,
		     	  struct reg_cir_layout *reg_cir,
			  int len,
			  int offset,
		     	  u8 *in_out_buf,
		     	  int (*f_start)(struct i2c_client *client, 
				         struct switchx_data *data, 
				         struct reg_cir_layout *reg_cir,
				         int len,
					 int offset,
				         u8 *in_out_buf),
		     	  int (*f_finish)(struct i2c_client *client, 
				          struct switchx_data *data, 
				          struct reg_cir_layout *reg_cir,
				          int len,
					  int offset,
				          u8 *in_out_buf))
{
	struct reg_cir_layout _reg_cir;
	int try = 0;
	int rc = -EIO;
	struct i2c_msg msgs[2];
	u8 msgbuf[SWITCHX_ADDR32];
	unsigned long orig_jiffies;
#if defined(__i386__) || defined(__x86_64__)
	u8 status = 0;
	u8 go = 1;
#endif
	memset(msgs, 0, 2*sizeof(struct i2c_msg));
	memset(&_reg_cir, 0, sizeof(struct reg_cir_layout));
	_reg_cir.go = 1; // Just for cleanup
	mutex_lock(&data->cmd_lock);

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s(%d) before getting GO len=%d offset=%d LEN_ACQ_GO=%d OFF_ACQ_GO=0x%08x 0x%08x 0x%08x\n",
		__FUNCTION__, __LINE__, len, offset, LEN_ACQ_GO, OFF_ACQ_GO,
		(int)offsetof(struct reg_cir_layout, status), CIR_STATUS_OFF);

#if defined(__i386__) || defined(__x86_64__)
	translate_offset_addr(client->addr, OFF_ACQ_GO, LEN_ACQ_GO, SWITCHX_ADDR32, msgbuf, msgs,
				((u8 *)(&_reg_cir) + CIR_STATUS_OFF), I2C_M_RD);
#else
	translate_offset_addr(client->addr, OFF_ACQ_GO, LEN_ACQ_GO, SWITCHX_ADDR32, msgbuf, msgs,
				(u8 *)&_reg_cir.status, I2C_M_RD);
#endif

	orig_jiffies = jiffies;
	for (try = 0; try <= data->retries; try++) {
		rc = i2c_transfer(client->adapter, msgs, 2);

#if defined(__i386__) || defined(__x86_64__)
		rc = (rc >= 0) ? icr_reg_read(client, *((u32 *)((u8 *)(&_reg_cir) + CIR_STATUS_OFF)), &status, &go) : rc;
		if ((rc >= 0) && (go == CIR_GO_SW) /*&& (!status)*/) {
			mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
				"%s(%d) got GO go=%x status=%x go=%x opcode=%x opcode_mod=%x rsrvd=%x event=%x status=%x buf %x %x %x %x\n",
				__FUNCTION__, __LINE__, go, status, _reg_cir.go,
        			_reg_cir.opcode, _reg_cir.opcode_modifier, _reg_cir.reserved2,
        			_reg_cir.event, _reg_cir.status,
        			*((u8 *)(&_reg_cir) + CIR_STATUS_OFF), *((u8 *)(&_reg_cir) + CIR_STATUS_OFF + 1),
        			*((u8 *)(&_reg_cir) + CIR_STATUS_OFF + 2), *((u8 *)(&_reg_cir) + CIR_STATUS_OFF + 3));
#else
		if ((rc >= 0) && (_reg_cir.go == CIR_GO_SW)) {
			mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
				"%s got GO after try=%d status=%x param 0x%08x 0x%08x 0x%08x 0x%08x\n",
				__FUNCTION__, try, _reg_cir.status, _reg_cir.in_param_h, 
				_reg_cir.in_param_l, _reg_cir.out_param_h, _reg_cir.out_param_l);
#endif
			if (f_start) {
				rc = f_start(client, data, reg_cir, len, offset, in_out_buf);
				if (!rc)
					break;
				if(f_finish)
					rc = acquire_go_bit_complete(client, data, reg_cir, len,
									offset, in_out_buf, f_finish);
			}
			else {
				rc = 0;
			}
			break;
		}
		if (time_after(jiffies, orig_jiffies + data->timeout)) {
			rc = -EIO;
			break;
		}
		cond_resched();
	}

	mutex_unlock(&data->cmd_lock);
	data->retry_cntr += try;

	return  (try <= data->retries) ? rc : -EIO;
}

static int _cmd_start(struct i2c_client *client, 
		      struct switchx_data *data,
		      struct reg_cir_layout *reg_cir,
		      int len, int offset, u8 *in_out_buf)
{
	struct i2c_msg msgs[1];
	u8 msgbuf[SWITCHX_ADDR32];
	u8 tbuf[MAX_I2C_BUFF_SIZE];
	int size = sizeof(struct reg_cir_layout);
#if defined(__i386__) || defined(__x86_64__)
	u32 icr_reg = 0;
	u32 icr_reg_buf[7];
#else
	int off = offsetof(struct reg_cir_layout, status);
	int size1 = size - off;
#endif
	int rc;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s(%d) cir 0x%08x go=%x stat=%x tok=%d opcode=%d opmod=%d event=%d inmod %x par %x %x %x %x\n", 
		__FUNCTION__, __LINE__, switchx_cfg.reg_cir_offset, reg_cir->go, reg_cir->status,  
		reg_cir->token, reg_cir->opcode, reg_cir->opcode_modifier, reg_cir->event, reg_cir->input_modifier,
		reg_cir->in_param_h, reg_cir->in_param_l, reg_cir->out_param_h, reg_cir->out_param_l);
	if ((reg_cir->opcode == SX_CONFIG_PROFILE_OP) && (reg_cir->opcode_modifier == SX_CONF_PROFILE_SET)) {
		/* Write out input mailbox */
		memset(msgs, 0, sizeof(struct i2c_msg));
		memset(tbuf, 0, len + SWITCHX_ADDR32);
		memcpy(tbuf + SWITCHX_ADDR32, in_out_buf, len);
		translate_offset_addr(client->addr, data->mb_offset_in + offset, len, SWITCHX_ADDR32,
					msgbuf, msgs, tbuf, 0);
		rc = i2c_transfer(client->adapter, msgs, 1);
        	/* !!! The CONFIG_PROFILE set command can be executed on the device only once at startup.
       		    !!! Failure of the CONFIG_PROFILE command leaves the Hardware in an indeterminate state.
        	   !!! It is required to perform software reset to the device following an unsuccessful 
        	   !!! command completion.
        	   Reset switch in case access failure */
        	//if (rc < 0)
                //	sx_i2c_software_reset(&client->dev);
		if (rc < 0)
			return -EIO;
	} 
	else if (reg_cir->opcode == SX_ACCESS_REG_OP) {
		mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
			"%s write input at 0x%08x (len=%d)\n", 
			__FUNCTION__, data->mb_offset_in, len);
		memset(msgs, 0, sizeof(struct i2c_msg));
		memset(tbuf, 0, len + SWITCHX_ADDR32);
		memcpy(tbuf + SWITCHX_ADDR32, in_out_buf, len);
		rc = translate_offset_addr(client->addr, data->mb_offset_in, len, SWITCHX_ADDR32,
					msgbuf, msgs, tbuf, 0);
		rc = i2c_transfer(client->adapter, msgs, 1);
		if (rc < 0)
			return -EIO;
	}
	 
	/* Write out Command Interface Register */
        reg_cir->token = (u16)client->adapter->nr; /* Put bus id as token */
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s(%d) write cir at 0x%08x go=%x stat=%x tok=%d opcode=%d opmod=%d event=%d inmod %x par %x %x %x %x\n", 
		__FUNCTION__, __LINE__, switchx_cfg.reg_cir_offset,  
		reg_cir->go, reg_cir->status, reg_cir->token,
		reg_cir->opcode, reg_cir->opcode_modifier, reg_cir->event, reg_cir->input_modifier,
		reg_cir->in_param_h, reg_cir->in_param_l, reg_cir->out_param_h, reg_cir->out_param_l);
	memset(msgs, 0, sizeof(struct i2c_msg));
	memset(tbuf, 0, size + SWITCHX_ADDR32);
#if !defined(__i386__) && !defined(__x86_64__)
	memcpy(tbuf + SWITCHX_ADDR32, (u8 *)reg_cir, size);
	rc = translate_offset_addr(client->addr, switchx_cfg.reg_cir_offset, size, SWITCHX_ADDR32,
				msgbuf, msgs, tbuf, 0);

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s(%d) tbuf=0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n" \
		"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n" \
		"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n" \
		"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
		  __FUNCTION__, __LINE__,
		  *(tbuf), *(tbuf+1), *(tbuf+2), *(tbuf+3), *(tbuf+4), *(tbuf+5), *(tbuf+6), *(tbuf+7),
		  *(tbuf+8), *(tbuf+9), *(tbuf+10), *(tbuf+11), *(tbuf+12), *(tbuf+13), *(tbuf+14), *(tbuf+15),
		  *(tbuf+16), *(tbuf+17), *(tbuf+18), *(tbuf+19), *(tbuf+20), *(tbuf+21), *(tbuf+22), *(tbuf+23),
		  *(tbuf+24), *(tbuf+25), *(tbuf+26), *(tbuf+27), *(tbuf+28), *(tbuf+29), *(tbuf+30), *(tbuf+31)); 

	rc = i2c_transfer(client->adapter, msgs, 1);
#else
	memset(icr_reg_buf, 0, 28);
	icr_reg_buf[2] = cpu_to_be32(reg_cir->input_modifier);
	icr_reg_buf[5] = cpu_to_be32(reg_cir->token << 16);
	icr_reg_buf[6] = cpu_to_be32((reg_cir->opcode_modifier << OPMOD_SHIFT) | reg_cir->opcode |
					(reg_cir->event << EVENT_BIT));

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		  "%s(%d) icr_reg_buf[2]=0x%08x icr_reg_buf[5]=0x%08x icr_reg_buf[6]=0x%08x\n", 
		  __FUNCTION__, __LINE__, icr_reg_buf[2], icr_reg_buf[5], icr_reg_buf[6]); 

	if (CPLD_BAG_WA) 
		memcpy(tbuf + SWITCHX_ADDR32, (u8 *)&icr_reg_buf, 28);
	else
		memcpy(tbuf + SWITCHX_ADDR32, (u8 *)&icr_reg_buf[1], 28 - CPLD_BAG_WA);

	translate_offset_addr(client->addr, SX_BAR0_OFFSET_SECOND + CPLD_BAG_WA, 28 - CPLD_BAG_WA, SWITCHX_ADDR32,
				msgbuf, msgs, tbuf, 0);

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s(%d) tbuf=0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n" \
		"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n" \
		"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n" \
		"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
		  __FUNCTION__, __LINE__,
		  *(tbuf), *(tbuf+1), *(tbuf+2), *(tbuf+3), *(tbuf+4), *(tbuf+5), *(tbuf+6), *(tbuf+7),
		  *(tbuf+8), *(tbuf+9), *(tbuf+10), *(tbuf+11), *(tbuf+12), *(tbuf+13), *(tbuf+14), *(tbuf+15),
		  *(tbuf+16), *(tbuf+17), *(tbuf+18), *(tbuf+19), *(tbuf+20), *(tbuf+21), *(tbuf+22), *(tbuf+23),
		  *(tbuf+24), *(tbuf+25), *(tbuf+26), *(tbuf+27), *(tbuf+28), *(tbuf+29), *(tbuf+30), *(tbuf+31)); 

	rc = i2c_transfer(client->adapter, msgs, 1);
#endif
	if (rc < 0)
		return -EIO;

	/* Write out Command Interface Register GO bit - do it by separate transaction */
        reg_cir->go = CIR_GO_HW;                   /* Go (Software ownership for the CIR) */
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s(%d) write cir at 0x%08x go=%x stat=%x tok=%d opcode=%d opmod=%d event=%d param %x %x %x %x\n", 
		__FUNCTION__, __LINE__, switchx_cfg.reg_cir_offset + STATUS_OFFSET,
		reg_cir->go, reg_cir->status, reg_cir->token, reg_cir->opcode, 
		reg_cir->opcode_modifier, reg_cir->event, reg_cir->in_param_h, reg_cir->in_param_l,
		reg_cir->out_param_h, reg_cir->out_param_l);

	memset(msgs, 0, sizeof(struct i2c_msg));
	memset(tbuf, 0, size + SWITCHX_ADDR32);
#if !defined(__i386__) && !defined(__x86_64__)
	memcpy(tbuf + SWITCHX_ADDR32, (u8 *)&reg_cir->status, size1);
	translate_offset_addr(client->addr, switchx_cfg.reg_cir_offset + off, size1, SWITCHX_ADDR32,
				msgbuf, msgs, tbuf, 0);

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		  "%s(%d) at offset=0x%08x size=%d buf=0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
		  __FUNCTION__, __LINE__, switchx_cfg.reg_cir_offset + off, size1,
		  *(tbuf), *(tbuf+1), *(tbuf+2), *(tbuf+3), *(tbuf+4), *(tbuf+5), *(tbuf+6), *(tbuf+7)); 

#else
        icr_reg = cpu_to_be32((1 << GO_BIT) | (reg_cir->event?(1 << EVENT_BIT):0) | 
                                (reg_cir->opcode_modifier << OPMOD_SHIFT) | reg_cir->opcode);
	memcpy(tbuf + SWITCHX_ADDR32, (u8 *)&icr_reg, 4);
	translate_offset_addr(client->addr, switchx_cfg.reg_cir_offset + STATUS_OFFSET, 4, SWITCHX_ADDR32,
				msgbuf, msgs, tbuf, 0);

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		  "%s(%d) at offset=0x%08x size=4 (0x%08x) buf=0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
		  __FUNCTION__, __LINE__, switchx_cfg.reg_cir_offset + STATUS_OFFSET, icr_reg,
		  *(tbuf), *(tbuf+1), *(tbuf+2), *(tbuf+3), *(tbuf+4), *(tbuf+5), *(tbuf+6), *(tbuf+7));
#endif

	rc = i2c_transfer(client->adapter, msgs, 1);
	if (rc < 0)
		return -EIO;

	return rc;
}

static int query_fw_cmd_finish_imm(struct i2c_client *client, 
			           struct switchx_data *data,
			           struct reg_cir_layout *reg_cir,
			           int len, int offset, u8 *in_out_buf)
{
	struct i2c_msg msgs[2];
	u8 msgbuf[SWITCHX_ADDR32];
	struct reg_cir_layout _reg_cir;
	int rc;

	switch (reg_cir->status) {
		case SX_STAT_OK:
			break;
		case SX_STAT_INTERNAL_ERR:
			printk(KERN_INFO "%s: Internal bus error occurred while processing command, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case SX_STAT_BAD_OP:
			printk(KERN_INFO "%s: Operation not supported, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case SX_STAT_BAD_SYS_STATE:
			printk(KERN_INFO "%s: Switch is not initialized, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case SX_STAT_EXCEED_LIM:
			printk(KERN_INFO "%s: Switch is not initialized, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	default:
			printk(KERN_INFO "%s: Unexpected status %x is reported, client %s addr 0x%02x\n",
				__FUNCTION__, reg_cir->status, client->name, client->addr);
			// return -EIO;
	}

        /* Read all register, in spite only out_param_h and out_param_l are necessary anyway it happens once */
	memset(&_reg_cir, 0, sizeof(struct reg_cir_layout));
	memset(msgs, 0, 2*sizeof(struct i2c_msg));

	rc = translate_offset_addr(client->addr, switchx_cfg.reg_cir_offset,
				sizeof(struct reg_cir_layout),
				SWITCHX_ADDR32, msgbuf, msgs, (u8 *)&_reg_cir, I2C_M_RD);

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s len=%d(%d) 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x " \
		"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, len, (int)sizeof(struct reg_cir_layout),
		*in_out_buf, *(in_out_buf+1), *(in_out_buf+2), *(in_out_buf+3),
		*(in_out_buf+4), *(in_out_buf+5), *(in_out_buf+6), *(in_out_buf+7),
		*(in_out_buf+8), *(in_out_buf+9), *(in_out_buf+10), *(in_out_buf+11), 
		*(in_out_buf+12), *(in_out_buf+13), *(in_out_buf+14), *(in_out_buf+15));

	rc = i2c_transfer(client->adapter, msgs, 2);

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s  cir  inmod=%x o=%x stat=%x tok=%d opcode=%d opmod=%d event=%d param %x %x %x %x\n", 
		__FUNCTION__, _reg_cir.input_modifier, _reg_cir.go, _reg_cir.status, _reg_cir.token,
		_reg_cir.opcode, _reg_cir.opcode_modifier, _reg_cir.event, _reg_cir.in_param_h, 
		_reg_cir.in_param_l, _reg_cir.out_param_h, _reg_cir.out_param_l);
 
	memcpy(in_out_buf, &_reg_cir.out_param_h, len);

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s len=%d(%d) 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x " \
		"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, len, (int)sizeof(struct reg_cir_layout),
		*in_out_buf, *(in_out_buf+1), *(in_out_buf+2), *(in_out_buf+3),
		*(in_out_buf+4), *(in_out_buf+5), *(in_out_buf+6), *(in_out_buf+7),
		*(in_out_buf+8), *(in_out_buf+9), *(in_out_buf+10), *(in_out_buf+11), 
		*(in_out_buf+12), *(in_out_buf+13), *(in_out_buf+14), *(in_out_buf+15));

	return rc;
}

static int _cmd_finish_mbox(struct i2c_client *client, 
			    struct switchx_data *data,
			    struct reg_cir_layout *reg_cir,
			    int len, int offset, u8 *in_out_buf)
{
	int rc = -EIO;
	struct i2c_msg msgs[2];
	u8 msgbuf[SWITCHX_ADDR32];

	switch (reg_cir->status) {
		case SX_STAT_OK:
			break;
		case SX_STAT_INTERNAL_ERR:
			printk(KERN_INFO "%s: Internal bus error occurred while processing command, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
		case SX_STAT_BAD_PARAM:
			printk(KERN_INFO "%s: Bad parameters, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case SX_STAT_BAD_OP:
			printk(KERN_INFO "%s: Operation not supported, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case SX_STAT_BAD_SYS_STATE:
			printk(KERN_INFO "%s: Bad system state, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case SX_STAT_EXCEED_LIM:
			printk(KERN_INFO "%s: Exceeds device limits, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	default:
			printk(KERN_INFO "%s: Unexpected status is reported, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
	}  

	memset(msgs, 0, 2*sizeof(struct i2c_msg));
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s  cir read data at offset = 0x%08x width = %d len =%d param %x %x %x %x\n", 
		__FUNCTION__, data->mb_offset_out + offset, SWITCHX_ADDR32, len, reg_cir->in_param_h, 
		reg_cir->in_param_l, reg_cir->out_param_h, reg_cir->out_param_l); 
#if !defined(__i386__) && !defined(__x86_64__)
	rc = translate_offset_addr(client->addr, data->mb_offset_out + offset, len, SWITCHX_ADDR32,
				msgbuf, msgs, in_out_buf, I2C_M_RD);      	
	rc = i2c_transfer(client->adapter, msgs, 2);
#else
	if (len <= LPCI2C_BLOCK_MAX) {   
		rc = translate_offset_addr(client->addr, data->mb_offset_out + offset, len, SWITCHX_ADDR32,
					msgbuf, msgs, in_out_buf, I2C_M_RD);      	
		rc = i2c_transfer(client->adapter, msgs, 2);
	}
	else {
        	int num = len/LPCI2C_BLOCK_MAX;
        	u8 *tbuf = in_out_buf;
        	int off = data->mb_offset_out + offset;
        	int chunk_size = (len > LPCI2C_BLOCK_MAX) ? LPCI2C_BLOCK_MAX : len;
        	int i;
 
        	if (len%LPCI2C_BLOCK_MAX)
                	num++;
        	for (i = 0; i < num; i++) {
			chunk_size = (len > LPCI2C_BLOCK_MAX) ? LPCI2C_BLOCK_MAX : len;

			mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
				"%s i=%d (num=%d) chunk_size=%d off=0x%08x\n" \
				"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x " \
				"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n " \
				"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x " \
				"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
				__FUNCTION__, i, num, chunk_size, off,
				*(in_out_buf+16*i), *(in_out_buf+1+16*i), *(in_out_buf+2+16*i), *(in_out_buf+3+16*i),
				*(in_out_buf+4+16*i), *(in_out_buf+5+16*i), *(in_out_buf+6+16*i), *(in_out_buf+7+16*i),
				*(in_out_buf+8+16*i), *(in_out_buf+9+16*i), *(in_out_buf+10+16*i), *(in_out_buf+11+16*i), 
				*(in_out_buf+12+16*i), *(in_out_buf+13+16*i), *(in_out_buf+14+16*i), *(in_out_buf+15+16*i),
				*(in_out_buf+16+16*i), *(in_out_buf+17+16*i), *(in_out_buf+18+16*i), *(in_out_buf+19+16*i), 
				*(in_out_buf+20+16*i), *(in_out_buf+21+16*i), *(in_out_buf+22+16*i), *(in_out_buf+23+16*i),
				*(in_out_buf+24+16*i), *(in_out_buf+25+16*i), *(in_out_buf+26+16*i), *(in_out_buf+27+16*i), 
				*(in_out_buf+28+16*i), *(in_out_buf+29+16*i), *(in_out_buf+30+16*i), *(in_out_buf+31+16*i));

			rc = translate_offset_addr(client->addr, off, chunk_size, SWITCHX_ADDR32,
							msgbuf, msgs, tbuf, I2C_M_RD);
			rc = i2c_transfer(client->adapter, msgs, 2);

			mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
				"%s i=%d (num=%d) chunk_size=%d off=0x%08x rc=%d msgbuf 0x%02x 0x%02x 0x%02x 0x%02x\n" \
				"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x " \
				"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n " \
				"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x " \
				"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
				__FUNCTION__, i, num, chunk_size, off, rc,
				*(msgbuf), *(msgbuf+1), *(msgbuf+2), *(msgbuf+3),
				*(in_out_buf+16*i), *(in_out_buf+1+16*i), *(in_out_buf+2+16*i), *(in_out_buf+3+16*i),
				*(in_out_buf+4+16*i), *(in_out_buf+5+16*i), *(in_out_buf+6+16*i), *(in_out_buf+7+16*i),
				*(in_out_buf+8+16*i), *(in_out_buf+9+16*i), *(in_out_buf+10+16*i), *(in_out_buf+11+16*i), 
				*(in_out_buf+12+16*i), *(in_out_buf+13+16*i), *(in_out_buf+14+16*i), *(in_out_buf+15+16*i),
				*(in_out_buf+16+16*i), *(in_out_buf+17+16*i), *(in_out_buf+18+16*i), *(in_out_buf+19+16*i), 
				*(in_out_buf+20+16*i), *(in_out_buf+21+16*i), *(in_out_buf+22+16*i), *(in_out_buf+23+16*i),
				*(in_out_buf+24+16*i), *(in_out_buf+25+16*i), *(in_out_buf+26+16*i), *(in_out_buf+27+16*i), 
				*(in_out_buf+28+16*i), *(in_out_buf+29+16*i), *(in_out_buf+30+16*i), *(in_out_buf+31+16*i));

			if (rc < 0)
				return -EIO;
			tbuf += chunk_size;
			off += chunk_size;
			len -= chunk_size;
			memset(msgs, 0, 2*sizeof(struct i2c_msg));
        	}
	}
#endif

	if (rc < 0)
		return -EIO;	
	return 0;
}

static int query_fw_cmd(struct device *dev, enum query_fw_opcode_mod opmod, 
			enum query_fw_input_mod inmod, u8 event, 
                        int len, int offset, u8 *in_out_buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct switchx_data *data = i2c_get_clientdata(client);
	struct reg_cir_layout *reg_cir = &data->reg_cir;
	int rc = 0;
 
        memset(reg_cir, 0, sizeof(struct reg_cir_layout));
        reg_cir->opcode = SX_QUERY_FW_OP;
        reg_cir->opcode_modifier = opmod;
        reg_cir->event = event;
        reg_cir->input_modifier = inmod;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
         	"%s(%d) offset=0x%08x op=%x event=%d opmod=%x inmod=%x stat=%x tok=%x data=%p reg_cir=%p len=%d\n",
         	__FUNCTION__, __LINE__, offset, reg_cir->opcode, reg_cir->opcode_modifier, reg_cir->event,
        	reg_cir->input_modifier, reg_cir->status, reg_cir->token, data, reg_cir, len);

        switch (opmod) {
        case SX_MB_OUTPUT: /* An output mailbox is returned, struct query_fw_output_mbox_layout */
        	if ((rc = acquire_go_bit(client, data, (void *)reg_cir, len, offset, in_out_buf,
                                         _cmd_start, _cmd_finish_mbox)) < 0)
			return rc;
        	break;
        case SX_IMM_OUTPUT_1: /* The output parameter is an immediate value, struct output_param_opmode1 */
        	if ((rc = acquire_go_bit(client, data, (void *)reg_cir, len, offset, in_out_buf,
                                         _cmd_start, query_fw_cmd_finish_imm)) < 0)
			return rc;
        	break;
        case SX_IMM_OUTPUT_2: /* The output parameter is an immediate values, struct output_param_opmode2 */
		rc = -ENOTSUPP;
        	break;
        }

        return (rc > 0) ? 0 : -1;
}

static int query_boardinfo_cmd(struct device *dev, int len, int offset, u8 *in_out_buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct switchx_data *data = i2c_get_clientdata(client);
	struct reg_cir_layout *reg_cir = &data->reg_cir;
	int rc;

        memset(reg_cir, 0, sizeof(struct reg_cir_layout));
        reg_cir->opcode = SX_QUERY_BOARDINFO_OP;  /* QUERY_BOARDINFO command opcode 0x06 */
        reg_cir->event = CIR_EVENT_NO_REPORT;     /* Do not report event (software will poll the GO bit) */

        if ((rc = acquire_go_bit(client, data, (void *)reg_cir, len, offset, in_out_buf,
                                 _cmd_start, _cmd_finish_mbox)) != 0)
		return rc;

        return rc;
}

static int config_profile_cmd(struct device *dev, enum config_profile_opcode_mod opmod,
				int len, int offset, u8 *in_out_buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct switchx_data *data = i2c_get_clientdata(client);
	struct reg_cir_layout *reg_cir = &data->reg_cir;
	int rc;

        memset(reg_cir, 0, sizeof(struct reg_cir_layout)); 
        reg_cir->opcode = SX_CONFIG_PROFILE_OP;     /* CONFIG_PROFILE command opcode TBD */
        reg_cir->opcode_modifier = opmod;  
        reg_cir->event = CIR_EVENT_NO_REPORT;       /* Do not report event (software will poll the GO bit) */

        if (opmod == SX_CONF_PROFILE_SET)
                rc = acquire_go_bit(client, data, (void *)reg_cir, len, offset, in_out_buf,
                                    _cmd_start, NULL);
        else
                rc = acquire_go_bit(client, data, (void *)reg_cir, len,  offset, in_out_buf,
                                    _cmd_start, _cmd_finish_mbox);

        return rc;
}

static int access_reg_cmd_finish_get(struct i2c_client *client, 
				     struct switchx_data *data,
				     struct reg_cir_layout *reg_cir,
				     int len, int offset, u8 *in_out_buf)
{
	struct access_reg_layout *acc_reg = (struct access_reg_layout *)in_out_buf; // ????
	struct i2c_msg msgs[2];
	u8 msgbuf[SWITCHX_ADDR32];
	int rc;

	switch (reg_cir->status) {
		case SX_STAT_OK:
			break;
		case SX_STAT_INTERNAL_ERR:
			printk(KERN_INFO "%s: Internal bus error occurred while processing command, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case SX_STAT_BAD_SYS_STATE:
			printk(KERN_INFO "%s: Switch is not initialized, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case SX_STAT_BAD_OP:
			printk(KERN_INFO "%s: Operation not supported, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	default:
			printk(KERN_INFO "%s: Unexpected status is reported, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
	}

	/* ACCESS_REG commands uses mailboxes. In order to use mailboxes through the i2c, special area is
	   reserved on the i2c address space that can be used for input and output mailboxes. Such mailboxes
	   are called Local Mailboxes. When using a local mailbox, software should specify 0 as the 
	   Input/Output parameters. The location of the Local Mailbox addresses on the i2c space can be retrieved
	   through the QUERY_FW command*/
	memset(msgs, 0, 2*sizeof(struct i2c_msg));
#if !defined(__i386__) && !defined(__x86_64__)
	rc = translate_offset_addr(client->addr, data->mb_offset_out, len, SWITCHX_ADDR32,
				msgbuf, msgs, in_out_buf, I2C_M_RD);
	rc = i2c_transfer(client->adapter, msgs, 2);
#else
	rc = translate_offset_addr(client->addr, data->mb_offset_out, len, SWITCHX_ADDR32,
				msgbuf, msgs, in_out_buf, I2C_M_RD);
#endif
	memset(data->cache, 0, SX_MAX_REG_SIZE);
	if (rc < 0)
		return -EIO;
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s read input at 0x%08x (len=%d) status=%x\n", 
		__FUNCTION__, data->mb_offset_out, len, acc_reg->oper_tlv.status);

	switch (acc_reg->oper_tlv.status) {
        	case OPER_TLV_STATUS_OK:
			break;
        	case OPER_TLV_STATUS_BUSY:
			printk(KERN_INFO "%s: Device is busy, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case OPER_TLV_STATUS_VER_NSUPP:
			printk(KERN_INFO "%s: Version not supported, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case OPER_TLV_STATUS_UNKNWN:
			printk(KERN_INFO "%s: Unknown TLV, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case OPER_TLV_STATUS_REG_NSUPP:
			printk(KERN_INFO "%s: Register not supported, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case OPER_TLV_STATUS_CLASS_NSUPP:
			printk(KERN_INFO "%s: Class not supported, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case OPER_TLV_STATUS_METHOD_NSUPP:
			printk(KERN_INFO "%s: Method not supported, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case OPER_TLV_STATUS_BAD_PARAM:
			printk(KERN_INFO "%s: Bad parameter, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case OPER_TLV_STATUS_RSRC_NAVAIL:
			printk(KERN_INFO "%s: Resource not available, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	case OPER_TLV_STATUS_MSG_ACK:
			printk(KERN_INFO "%s: Message Receipt Acknowledgement, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
        	default:
			printk(KERN_INFO "%s: Unexpected TLV status is reported, client %s addr 0x%02x\n",
				__FUNCTION__, client->name, client->addr);
			return -EIO;
	}

	return 0;
}

static int access_reg_cmd(struct device *dev, u32 reg_id, int reg_len, enum oper_tlv_method method,
			  int len, int offset, u8 *in_out_buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct switchx_data *data = i2c_get_clientdata(client);
        struct operation_tlv oper_tlv;
        struct reg_access_tlv reg_accs_tlv;
        struct end_tlv e_tlv;
	u8 tlv_buf[sizeof(struct operation_tlv) + offsetof(struct reg_access_tlv, switchx_reg) +
		   SX_MAX_REG_SIZE + sizeof(struct end_tlv)];
	struct reg_cir_layout *reg_cir = &data->reg_cir;
	int off = 0;
	int offreg;
	int rc = 0;
 
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
               "%s input buf (len=%d) reg_len=%d reg_id=%x method=%d\n", 
               __FUNCTION__, len, reg_len, reg_id, method);

	offreg = offsetof(struct reg_access_tlv, switchx_reg);
	memset(tlv_buf, 0, sizeof(struct operation_tlv) + offreg +
		   SX_MAX_REG_SIZE + sizeof(struct end_tlv));
	memset(&oper_tlv, 0, sizeof(struct operation_tlv));
	memset(&reg_accs_tlv, 0, offreg);
	memset(&e_tlv, 0, sizeof(struct end_tlv));
	memset(reg_cir, 0, sizeof(struct reg_cir_layout));
	oper_tlv.type = SX_OPERATION_TLV;
	oper_tlv.len = sizeof(struct operation_tlv) / 4; // size in dwords
	oper_tlv.dr = DIRECT_ROUTE_NOT_SET;
	oper_tlv.status = OPER_TLV_STATUS_OK;
	oper_tlv.register_id = reg_id;
	oper_tlv.res_req = OPER_TLV_REQ;
	oper_tlv.method = method;
	oper_tlv.class = SX_EMAD_REG_ACCESS_CLASS;
	oper_tlv.tid = data->transact_id++;
	reg_accs_tlv.type = SX_REG_ACCESS_TLV;
        reg_accs_tlv.len = (offreg + reg_len) / 4; // size in dwords
        e_tlv.type = SX_END_OF_TLV;
        e_tlv.len = sizeof(struct end_tlv) / 4; // size in dwords

#if defined(__i386__) || defined(__x86_64__)
	//convert_oper_tlv((u32 *)&oper_tlv);
	//convert_reg_accs_tlv((u32 *)&reg_accs_tlv);
	//convert_end_tlv((u32 *)&e_tlv);
#endif
        /* Copy data to buffer, which should be passed to input mail box */
        memcpy(&tlv_buf[0], &oper_tlv, sizeof(struct operation_tlv)); /* Put Oper TLV                     */
        off = sizeof(struct operation_tlv);
        memcpy(&tlv_buf[off], &reg_accs_tlv, offreg);                 /* Put Reg TLV constant part        */
        off += offreg;
        if (method == OPER_TLV_METHOD_WRITE)                          /* Put Reg TLV variable part        */
                memcpy(&tlv_buf[off + offset], in_out_buf, len);      /*  write - copy user buffer        */
        else
                memcpy(&tlv_buf[off + offset], data->cache, reg_len); /*  read - fill index field if any  */
                // add flag: if called from sysfs - use data->cache, in other case in_out_buf
        off += reg_len;                                               
        memcpy(&tlv_buf[off], &e_tlv, sizeof(struct end_tlv));        /* Put End TLV                      */

        reg_cir->opcode = SX_ACCESS_REG_OP;                           /* ACCESS_REG command opcode 0x40   */
        reg_cir->event = CIR_EVENT_NO_REPORT;
 
        switch (method) {
        case OPER_TLV_METHOD_QUERY:
                if ((rc = acquire_go_bit(client, data, (void *)reg_cir, off + sizeof(struct end_tlv),
                                         offset, tlv_buf, _cmd_start, access_reg_cmd_finish_get)) != 0)
                        return rc;
                memcpy(in_out_buf, &tlv_buf[sizeof(struct operation_tlv) + offreg + offset], len);
                break;
        case OPER_TLV_METHOD_WRITE:
                if ((rc = acquire_go_bit(client, data, (void *)reg_cir, off + sizeof(struct end_tlv),
                                         offset, tlv_buf, _cmd_start, NULL)) != 0)
                        return rc;
                break;
        case OPER_TLV_METHOD_SEND:
        case OPER_TLV_METHOD_EVENT:
                rc = -ENOTSUPP;
                break;
        }

        return rc;
}

static ssize_t show_pspa_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_pspa_layout *p_pspa_reg = (struct reg_pspa_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;
 
	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_pspa_layout))
		return 0;
	if (off + count > sizeof(struct reg_pspa_layout))
		count = sizeof(struct reg_pspa_layout) - off;

	rc = access_reg_cmd(&client->dev, PSPA_REG_ID, sizeof(struct reg_pspa_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s (rc=%d) 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, rc, p_pspa_reg->swid, p_pspa_reg->local_port,
		p_pspa_reg->sub_port);
	return count;
}

static ssize_t set_pspa_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_pspa_layout *p_pspa_reg = (struct reg_pspa_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;
 
	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_pspa_layout))
		return -1;
	if (off + count > sizeof(struct reg_pspa_layout))
		count = sizeof(struct reg_pspa_layout) - off;

	rc = access_reg_cmd(&client->dev, PSPA_REG_ID, sizeof(struct reg_pspa_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s (rc=%d) 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, rc, p_pspa_reg->swid, p_pspa_reg->local_port,
		p_pspa_reg->sub_port);
	return count;
}

static ssize_t show_qsptc_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_qsptc_layout *p_qsptc_reg = (struct reg_qsptc_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_qsptc_layout))
		return 0;
	if (off + count > sizeof(struct reg_qsptc_layout))
		count = sizeof(struct reg_qsptc_layout) - off;

	rc = access_reg_cmd(&client->dev, QSPTC_REG_ID, sizeof(struct reg_qsptc_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, p_qsptc_reg->local_iport, p_qsptc_reg->local_eport,
		p_qsptc_reg->tclass, p_qsptc_reg->tclass);
	return count;
}

static ssize_t set_qsptc_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_qsptc_layout *p_qsptc_reg = (struct reg_qsptc_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_qsptc_layout))
		return -1;
	if (off + count > sizeof(struct reg_qsptc_layout))
		count = sizeof(struct reg_qsptc_layout) - off;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, p_qsptc_reg->local_iport, p_qsptc_reg->local_eport,
		p_qsptc_reg->tclass, p_qsptc_reg->tclass);
	rc = access_reg_cmd(&client->dev, QSPTC_REG_ID, sizeof(struct reg_qsptc_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_qstct_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_qstct_layout *p_qstct_reg = (struct reg_qstct_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_qstct_layout))
		return 0;
	if (off + count > sizeof(struct reg_qstct_layout))
		count = sizeof(struct reg_qstct_layout) - off;

	rc = access_reg_cmd(&client->dev, QSTCT_REG_ID, sizeof(struct reg_qstct_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, p_qstct_reg->swid, p_qstct_reg->prio,
		p_qstct_reg->utclass, p_qstct_reg->mtclass);
	return count;
}

static ssize_t set_qstct_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_qstct_layout *p_qstct_reg = (struct reg_qstct_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_qstct_layout))
		return -1;
	if (off + count > sizeof(struct reg_qstct_layout))
		count = sizeof(struct reg_qstct_layout) - off;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, p_qstct_reg->swid, p_qstct_reg->prio,
		p_qstct_reg->utclass, p_qstct_reg->mtclass);
	rc = access_reg_cmd(&client->dev, QSTCT_REG_ID, sizeof(struct reg_qstct_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_pmtu_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_pmtu_layout *p_pmtu_reg = (struct reg_pmtu_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_pmtu_layout))
		return 0;
	if (off + count > sizeof(struct reg_pmtu_layout))
		count = sizeof(struct reg_pmtu_layout) - off;

	rc = access_reg_cmd(&client->dev, PMTU_REG_ID, sizeof(struct reg_pmtu_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x %04x %04x %04x\n",
		__FUNCTION__, p_pmtu_reg->local_port, p_pmtu_reg->max_mtu,
		p_pmtu_reg->admin_mtu, p_pmtu_reg->oper_mtu);
	return count;
}

static ssize_t set_pmtu_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_pmtu_layout *p_pmtu_reg = (struct reg_pmtu_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_pmtu_layout))
		return -1;
	if (off + count > sizeof(struct reg_pmtu_layout))
		count = sizeof(struct reg_pmtu_layout) - off;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x %04x %04x %04x\n",
		__FUNCTION__, p_pmtu_reg->local_port, p_pmtu_reg->max_mtu,
		p_pmtu_reg->admin_mtu, p_pmtu_reg->oper_mtu);
	rc = access_reg_cmd(&client->dev, PMTU_REG_ID, sizeof(struct reg_pmtu_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_ptys_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_ptys_layout *p_ptys_reg = (struct reg_ptys_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_ptys_layout))
		return 0;
	if (off + count > sizeof(struct reg_ptys_layout))
		count = sizeof(struct reg_ptys_layout) - off;

	rc = access_reg_cmd(&client->dev, PTYS_REG_ID, sizeof(struct reg_ptys_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x %08x %08x %08x %08x %08x %08x %08x %08x %08x\n", __FUNCTION__, 
		p_ptys_reg->local_port, p_ptys_reg->eth_proto_capability, p_ptys_reg->fc_proto_capability,
		p_ptys_reg->ib_proto_capability, p_ptys_reg->eth_proto_admin, p_ptys_reg->fc_proto_admin,
		p_ptys_reg->ib_proto_admin, p_ptys_reg->eth_proto_oper, p_ptys_reg->fc_proto_oper,
		p_ptys_reg->ib_proto_oper);
	return count;
}

static ssize_t set_ptys_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_ptys_layout *p_ptys_reg = (struct reg_ptys_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_ptys_layout))
		return -1;
	if (off + count > sizeof(struct reg_ptys_layout))
		count = sizeof(struct reg_ptys_layout) - off;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x %08x %08x %08x %08x %08x %08x %08x %08x %08x\n", __FUNCTION__, 
		p_ptys_reg->local_port, p_ptys_reg->eth_proto_capability, p_ptys_reg->fc_proto_capability,
		p_ptys_reg->ib_proto_capability, p_ptys_reg->eth_proto_admin, p_ptys_reg->fc_proto_admin,
		p_ptys_reg->ib_proto_admin, p_ptys_reg->eth_proto_oper, p_ptys_reg->fc_proto_oper,
		p_ptys_reg->ib_proto_oper);
	rc = access_reg_cmd(&client->dev, PTYS_REG_ID, sizeof(struct reg_ptys_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_pmlp_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_pmlp_layout *p_pmlp_reg = (struct reg_pmlp_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_pmlp_layout))
		return 0;
	if (off + count > sizeof(struct reg_pmlp_layout))
		count = sizeof(struct reg_pmlp_layout) - off;

	rc = access_reg_cmd(&client->dev, PMLP_REG_ID, sizeof(struct reg_pmlp_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x %04x 0x%02x %04x 0x%02x %04x 0x%02x %04x 0x%02x\n", __FUNCTION__,
		p_pmlp_reg->local_port, p_pmlp_reg->width, p_pmlp_reg->lane_0, p_pmlp_reg->module_0,  
		p_pmlp_reg->lane_1, p_pmlp_reg->module_1, p_pmlp_reg->lane_2, p_pmlp_reg->module_2, 
		p_pmlp_reg->lane_3, p_pmlp_reg->module_3);
	return count;
}

static ssize_t set_pmlp_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_pmlp_layout *p_pmlp_reg = (struct reg_pmlp_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_pmlp_layout))
		return -1;
	if (off + count > sizeof(struct reg_pmlp_layout))
		count = sizeof(struct reg_pmlp_layout) - off;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x %04x 0x%02x %04x 0x%02x %04x 0x%02x %04x 0x%02x\n", __FUNCTION__,
		p_pmlp_reg->local_port, p_pmlp_reg->width, p_pmlp_reg->lane_0, p_pmlp_reg->module_0,  
		p_pmlp_reg->lane_1, p_pmlp_reg->module_1, p_pmlp_reg->lane_2, p_pmlp_reg->module_2, 
		p_pmlp_reg->lane_3, p_pmlp_reg->module_3);
	rc = access_reg_cmd(&client->dev, PMLP_REG_ID, sizeof(struct reg_pmlp_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_plib_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_plib_layout *p_plib_reg = (struct reg_plib_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_plib_layout))
		return 0;
	if (off + count > sizeof(struct reg_plib_layout))
		count = sizeof(struct reg_plib_layout) - off;

	rc = access_reg_cmd(&client->dev, PLIB_REG_ID, sizeof(struct reg_plib_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, p_plib_reg->map, p_plib_reg->local_port, p_plib_reg->ib_port);
	return count;
}

static ssize_t set_plib_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_plib_layout *p_plib_reg = (struct reg_plib_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_plib_layout))
		return -1;
	if (off + count > sizeof(struct reg_plib_layout))
		count = sizeof(struct reg_plib_layout) - off;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, p_plib_reg->map, p_plib_reg->local_port, p_plib_reg->ib_port);
	rc = access_reg_cmd(&client->dev, PLIB_REG_ID, sizeof(struct reg_plib_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_paos_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_paos_layout *p_paos_reg = (struct reg_paos_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_paos_layout))
		return 0;
	if (off + count > sizeof(struct reg_paos_layout))
		count = sizeof(struct reg_paos_layout) - off;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x 0x%02x 0x%08x\n",
		__FUNCTION__, p_paos_reg->local_port, p_paos_reg->admin_status,
		p_paos_reg->oper_status, p_paos_reg->event_gen);
	rc = access_reg_cmd(&client->dev, PAOS_REG_ID, sizeof(struct reg_paos_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_paos_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_paos_layout *p_paos_reg = (struct reg_paos_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_paos_layout))
		return -1;
	if (off + count > sizeof(struct reg_paos_layout))
		count = sizeof(struct reg_paos_layout) - off;

	rc = access_reg_cmd(&client->dev, PAOS_REG_ID, sizeof(struct reg_paos_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x 0x%02x %08x\n",
		__FUNCTION__, p_paos_reg->local_port, p_paos_reg->admin_status,
		p_paos_reg->oper_status, p_paos_reg->event_gen);
	return count;
}

static ssize_t show_spzr_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_spzr_layout *p_spzr_reg = (struct reg_spzr_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_spzr_layout))
		return 0;
	if (off + count > sizeof(struct reg_spzr_layout))
		count = sizeof(struct reg_spzr_layout) - off;

	rc = access_reg_cmd(&client->dev, SPZR_REG_ID, sizeof(struct reg_spzr_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n", __FUNCTION__, 
		p_spzr_reg->swid, p_spzr_reg->ndm, p_spzr_reg->enh_sw_p0_mask, p_spzr_reg->cap_mask, 
		p_spzr_reg->v_key, p_spzr_reg->mpkey, p_spzr_reg->sys_guid, p_spzr_reg->node_guid, 
		p_spzr_reg->guid0, p_spzr_reg->enh_sw_p0);
	return count;
}

static ssize_t set_spzr_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_spzr_layout *p_spzr_reg = (struct reg_spzr_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_spzr_layout))
		return -1;
	if (off + count > sizeof(struct reg_spzr_layout))
		count = sizeof(struct reg_spzr_layout) - off;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n", __FUNCTION__, 
		p_spzr_reg->swid, p_spzr_reg->ndm, p_spzr_reg->enh_sw_p0_mask, p_spzr_reg->cap_mask, 
		p_spzr_reg->v_key, p_spzr_reg->mpkey, p_spzr_reg->sys_guid, p_spzr_reg->node_guid, 
		p_spzr_reg->guid0, p_spzr_reg->enh_sw_p0);
	rc = access_reg_cmd(&client->dev, SPZR_REG_ID, sizeof(struct reg_spzr_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_hcap_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_hcap_layout))
		return 0;
	if (off + count > sizeof(struct reg_hcap_layout))
		count = sizeof(struct reg_hcap_layout) - off;

	rc = access_reg_cmd(&client->dev, HCAP_REG_ID, sizeof(struct reg_hcap_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_hcap_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_hcap_layout))
		return -1;
	if (off + count > sizeof(struct reg_hcap_layout))
		count = sizeof(struct reg_hcap_layout) - off;

	rc = access_reg_cmd(&client->dev, HCAP_REG_ID, sizeof(struct reg_hcap_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_htgt_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_htgt_layout))
		return 0;
	if (off + count > sizeof(struct reg_htgt_layout))
		count = sizeof(struct reg_htgt_layout) - off;

	rc = access_reg_cmd(&client->dev, HTGT_REG_ID, sizeof(struct reg_htgt_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_htgt_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_htgt_layout))
		return -1;
	if (off + count > sizeof(struct reg_htgt_layout))
		count = sizeof(struct reg_htgt_layout) - off;

	rc = access_reg_cmd(&client->dev, HTGT_REG_ID, sizeof(struct reg_htgt_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_hpkt_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_hpkt_layout))
		return 0;
	if (off + count > sizeof(struct reg_hpkt_layout))
		count = sizeof(struct reg_hpkt_layout) - off;

	rc = access_reg_cmd(&client->dev, HPKT_REG_ID, sizeof(struct reg_hpkt_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_hpkt_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_hpkt_layout))
		return -1;
	if (off + count > sizeof(struct reg_hpkt_layout))
		count = sizeof(struct reg_hpkt_layout) - off;

	rc = access_reg_cmd(&client->dev, HPKT_REG_ID, sizeof(struct reg_hpkt_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_hdrt_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_hdrt_layout))
		return 0;
	if (off + count > sizeof(struct reg_hdrt_layout))
		count = sizeof(struct reg_hdrt_layout) - off;

	rc = access_reg_cmd(&client->dev, HDRT_REG_ID, sizeof(struct reg_hdrt_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_hdrt_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_hdrt_layout))
		return -1;
	if (off + count > sizeof(struct reg_hdrt_layout))
		count = sizeof(struct reg_hdrt_layout) - off;

	rc = access_reg_cmd(&client->dev, HDRT_REG_ID, sizeof(struct reg_hdrt_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_mfcr_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_mfcr_layout))
		return 0;
	if (off + count > sizeof(struct reg_mfcr_layout))
		count = sizeof(struct reg_mfcr_layout) - off;

	rc = access_reg_cmd(&client->dev, MFCR_REG_ID, sizeof(struct reg_mfcr_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_mfcr_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_mfcr_layout))
		return -1;
	if (off + count > sizeof(struct reg_mfcr_layout))
		count = sizeof(struct reg_mfcr_layout) - off;

	rc = access_reg_cmd(&client->dev, MFCR_REG_ID, sizeof(struct reg_mfcr_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_mfsc_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_mfsc_layout))
		return 0;
	if (off + count > sizeof(struct reg_mfsc_layout))
		count = sizeof(struct reg_mfsc_layout) - off;

	rc = access_reg_cmd(&client->dev, MFSC_REG_ID, sizeof(struct reg_mfsc_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_mfsc_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_mfsc_layout))
		return -1;
	if (off + count > sizeof(struct reg_mfsc_layout))
		count = sizeof(struct reg_mfsc_layout) - off;

	rc = access_reg_cmd(&client->dev, MFSC_REG_ID, sizeof(struct reg_mfsc_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_mfsm_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_mfsm_layout))
		return 0;
	if (off + count > sizeof(struct reg_mfsm_layout))
		count = sizeof(struct reg_mfsm_layout) - off;

	rc = access_reg_cmd(&client->dev, MFSM_REG_ID, sizeof(struct reg_mfsm_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_mfsm_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_mfsm_layout))
		return -1;
	if (off + count > sizeof(struct reg_mfsm_layout))
		count = sizeof(struct reg_mfsm_layout) - off;

	rc = access_reg_cmd(&client->dev, MFSM_REG_ID, sizeof(struct reg_mfsm_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_mfsl_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_mfsl_layout))
		return 0;
	if (off + count > sizeof(struct reg_mfsl_layout))
		count = sizeof(struct reg_mfsl_layout) - off;

	rc = access_reg_cmd(&client->dev, MFSL_REG_ID, sizeof(struct reg_mfsl_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_mfsl_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_mfsl_layout))
		return -1;
	if (off + count > sizeof(struct reg_mfsl_layout))
		count = sizeof(struct reg_mfsl_layout) - off;

	rc = access_reg_cmd(&client->dev, MFSL_REG_ID, sizeof(struct reg_mfsl_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_fore_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_fore_layout))
		return 0;
	if (off + count > sizeof(struct reg_fore_layout))
		count = sizeof(struct reg_fore_layout) - off;

	rc = access_reg_cmd(&client->dev, FORE_REG_ID, sizeof(struct reg_fore_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_fore_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_fore_layout))
		return -1;
	if (off + count > sizeof(struct reg_fore_layout))
		count = sizeof(struct reg_fore_layout) - off;

	rc = access_reg_cmd(&client->dev, FORE_REG_ID, sizeof(struct reg_fore_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_mcia_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct reg_mcia_layout))
		return 0;
	if (off + count > sizeof(struct reg_mcia_layout))
		count = sizeof(struct reg_mcia_layout) - off;

	rc = access_reg_cmd(&client->dev, MCIA_REG_ID, sizeof(struct reg_mcia_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_mcia_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_mcia_layout))
		return -1;
	if (off + count > sizeof(struct reg_mcia_layout))
		count = sizeof(struct reg_mcia_layout) - off;

	rc = access_reg_cmd(&client->dev, MCIA_REG_ID, sizeof(struct reg_mcia_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_mtmp_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_mtmp_layout))
		return -1;
	if (off + count > sizeof(struct reg_mtmp_layout))
		count = sizeof(struct reg_mtmp_layout) - off;

	rc = access_reg_cmd(&client->dev, MTMP_REG_ID, sizeof(struct reg_mtmp_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_mtmp_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_mtmp_layout))
		return -1;
	if (off + count > sizeof(struct reg_mtmp_layout))
		count = sizeof(struct reg_mtmp_layout) - off;

	rc = access_reg_cmd(&client->dev, MTMP_REG_ID, sizeof(struct reg_mtmp_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_mpgr_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_mpgr_layout))
		return -1;
	if (off + count > sizeof(struct reg_mpgr_layout))
		count = sizeof(struct reg_mpgr_layout) - off;

	rc = access_reg_cmd(&client->dev, MPGR_REG_ID, sizeof(struct reg_mpgr_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
	return count;
}

static ssize_t set_mpgr_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_mpgr_layout))
		return -1;
	if (off + count > sizeof(struct reg_mpgr_layout))
		count = sizeof(struct reg_mpgr_layout) - off;

	rc = access_reg_cmd(&client->dev, MPGR_REG_ID, sizeof(struct reg_mpgr_layout),
				OPER_TLV_METHOD_WRITE, count, off, buf);
	return count;
}

static ssize_t show_mgir_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
#ifdef SXDBGMODE
	struct reg_mgir_layout *p_mgir_reg = (struct reg_mgir_layout *)buf;
#endif
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_mgir_layout))
		return -1;
	if (off + count > sizeof(struct reg_mgir_layout))
		count = sizeof(struct reg_mgir_layout) - off;

	rc = access_reg_cmd(&client->dev, MGIR_REG_ID, sizeof(struct reg_mgir_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);
#if !defined(__i386__) && !defined(__x86_64__)
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s sw_major=0x%02x sw_minor=%04x sw_sub_minor=%04x dvfs=%04x rev=%04x dev_id=%04x uptime=%04x\n"
		"major=%04x minor=%04x sub_minor=%04x build_id=%04x month=%04x day=%04x year=%04x hour=%04x\n"
		"ini_file_version=%08x extended_major=%08x extended_minor=%08x extended_sub_minor=%08x\n",
		__FUNCTION__, p_mgir_reg->sw_info.major,
		p_mgir_reg->sw_info.minor, p_mgir_reg->sw_info.sub_minor,
		p_mgir_reg->hw_info.dvfs, p_mgir_reg->hw_info.device_hw_revesion,
		p_mgir_reg->hw_info.device_id, p_mgir_reg->hw_info.uptime,
		p_mgir_reg->fw_info.major, p_mgir_reg->fw_info.minor, p_mgir_reg->fw_info.sub_minor, 
		p_mgir_reg->fw_info.build_id, p_mgir_reg->fw_info.month, p_mgir_reg->fw_info.day,
		p_mgir_reg->fw_info.year, p_mgir_reg->fw_info.hour, p_mgir_reg->fw_info.ini_file_version,
		p_mgir_reg->fw_info.extended_major, p_mgir_reg->fw_info.extended_minor,
		p_mgir_reg->fw_info.extended_sub_minor);
#else
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s (%p) 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x " \
		"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
		__FUNCTION__, p_mgir_reg, *buf, *(buf+1), *(buf+2), *(buf+3), *(buf+4), *(buf+5),
		*(buf+6), *(buf+7), *(buf+8), *(buf+9), *(buf+10), *(buf+11), *(buf+12), *(buf+13),
		*(buf+14), *(buf+15));

#endif
	return count;
}

static ssize_t set_cache(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	struct switchx_data *data = i2c_get_clientdata(client);

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off >= SX_MAX_REG_SIZE)
		return -1;
	if (off + count > SX_MAX_REG_SIZE)
		count = SX_MAX_REG_SIZE - off;

	memcpy(data->cache, buf, count);

	return count;
}

static ssize_t switchx_sw_reset(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);

	return sx_i2c_software_reset(&client->dev);
}
 
static ssize_t query_fw_read(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "query_fw_read(p=%p, off=%lli, c=%zi)\n",
		buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct query_fw_output_mbox_layout))
		return 0;
	if (off + count > sizeof(struct query_fw_output_mbox_layout))
		count = sizeof(struct query_fw_output_mbox_layout) - off;

	/* Read from mailbox */
	rc = query_fw_cmd(&client->dev, SX_MB_OUTPUT, SX_BAR0_OFF_NONE,
			  CIR_EVENT_REPORT, count, off, buf);
 
	return count;
}

static ssize_t boardinfo_read(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "boardinfo_read(p=%p, off=%lli, c=%zi)\n",
		buf, off, count);

	if (off % 4)
		return 0;
	if (off >= sizeof(struct query_boardinfo_layout))
		return 0;
	if (off + count > sizeof(struct query_boardinfo_layout))
		count = sizeof(struct query_boardinfo_layout) - off;

	rc = query_boardinfo_cmd(&client->dev, count, off, buf);

	return count;
}

static ssize_t config_profile_read(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				   char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return 0;
	if (off >= sizeof(struct config_profile_mbox_layout))
		return 0;
	if (off + count > sizeof(struct config_profile_mbox_layout))
		count = sizeof(struct config_profile_mbox_layout) - off;
 
	rc = config_profile_cmd(&client->dev, SX_CONF_PROFILE_GET, count, off, buf);

	return count;
}

static ssize_t config_profile_write(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				    char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct config_profile_mbox_layout))
		return -1;
	if (off + count > sizeof(struct config_profile_mbox_layout))
		count = sizeof(struct config_profile_mbox_layout) - off;

	rc = config_profile_cmd(&client->dev, SX_CONF_PROFILE_SET, count, off, buf);

	return count;
}

static ssize_t show_msci_reg(_FL_  struct kobject *kobj, struct bin_attribute *attr,
				  char *buf, loff_t off, size_t count)
{
	struct i2c_client *client = kobj_to_i2c_client(kobj);
	int rc;

	dev_dbg(&client->dev, "%s(p=%p, off=%lli, c=%zi)\n",
		__FUNCTION__, buf, off, count);
	if (off % 4)
		return -1;
	if (off >= sizeof(struct reg_msci_layout))
		return -1;
	if (off + count > sizeof(struct reg_msci_layout))
		count = sizeof(struct reg_msci_layout) - off;

	rc = access_reg_cmd(&client->dev, MSCI_REG_ID, sizeof(struct reg_msci_layout),
				OPER_TLV_METHOD_QUERY, count, off, buf);

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s cpld_ver 0x%x\n", __FUNCTION__, *(buf+4));

	return count;
}

static int __i2c_match_addr(struct device *dev, void *addrp)
{
	struct i2c_client *client = i2c_verify_client(dev);
	int addr = *(int *)addrp;

        mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
                "%s client=%p addr=%x\n", __FUNCTION__, client, addr);

	if (client && client->addr == addr)
		return 1;
	return 0;
}

static int get_dev_by_bus_id_and_addr(int bus_id, int addr, struct device *dev)
{
	struct device *adap_dev;
        struct device *client_dev;
        struct i2c_adapter *adap;

        mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
                "%s bus=%d addr=%x\n", __FUNCTION__, bus_id, addr);
        adap = i2c_get_adapter(bus_id);
	if (adap == NULL) {
                mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
			"%s adapter not found bus=%d addr=%x\n", __FUNCTION__, bus_id, addr);
		return -EINVAL;
        }
	adap_dev = &adap->dev;
	client_dev = device_find_child(adap_dev, &addr, __i2c_match_addr);
	if(!client_dev) {
                mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
			"%s client not found bus=%d addr=%x\n", __FUNCTION__, bus_id, addr);
		return -EINVAL;
        }
	dev = client_dev;

	return 0;
}

struct i2c_client *sx_get_client_by_i2c_bus_and_addr(int bus_id, int addr)
{
	struct switchx_data *swd;

	list_for_each_entry(swd, &switchx_cfg.switchx_list, list) {
		if ((bus_id == swd->bus_id) && (addr == swd->i2c_dev_addr))
			return swd->client;
	}
	return NULL;
}

struct switchx_probe_failed * find_dev_probe_context(int bus_id)
{
	struct switchx_probe_failed *swpf, *next;

	list_for_each_entry_safe(swpf, next, &switchx_cfg.probe_failed_list, list) {
		if (bus_id == swpf->bus_id)
			return swpf; /* device on this bus failed probing */
	}
	return NULL;
}

static int sx_i2c_probing_test(struct i2c_client *client, struct switchx_data *data)
{
	struct output_param_opmode1 output_param;
	struct switchx_probe_failed *p_probe_failed = NULL;
	int rc;

	/* Find probing context */
	if (client->adapter) {
		if ((p_probe_failed = find_dev_probe_context(client->adapter->nr)) == NULL) {
			p_probe_failed->probe_cntr++;
			return -EINVAL;
		}
	}

	memset(&output_param, 0, sizeof(struct output_param_opmode1));
	/* Take mutex in case it's not held yet */
	if (!data->route_enforced)
		mutex_lock(&__i2c_mlnx_lock);
	if((rc = query_fw_cmd(&client->dev, SX_IMM_OUTPUT_1, SX_BAR0_OFF_NONE, 
				CIR_EVENT_NO_REPORT, sizeof(struct output_param_opmode1),
				0, (u8 *)&output_param)) < 0) {
		if (!data->route_enforced)
			mutex_unlock(&__i2c_mlnx_lock);
		/* Mark probing failure - such device should be isolated */
		p_probe_failed->probe_status = PROBE_STAT_FAILED;
		return -EIO;
	}
	if (!data->route_enforced)
		mutex_unlock(&__i2c_mlnx_lock);
	p_probe_failed->probe_status = PROBE_STAT_OK;
	return 0;
}

int sx_access_reg_cmd(int i2c_dev_id, u32 reg_id, int reg_len, enum oper_tlv_method method,
			int len, int offset, u8 *in_out_buf)
{
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	struct device *dev = NULL;

        mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
               "%s i2c_dev_id=%x reg_id=%x bus_id=%d addr=%x\n",
               __FUNCTION__, i2c_dev_id, reg_id, bus_id, addr);

	if (get_dev_by_bus_id_and_addr(bus_id, addr, dev))
		return -EINVAL;

        mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
               "%s bus_id=%d addr=%x\n", __FUNCTION__, bus_id, addr);

	return access_reg_cmd(dev, reg_id, reg_len, method, len, offset, in_out_buf);
}
EXPORT_SYMBOL(sx_access_reg_cmd);

int sx_config_profile_cmd(int i2c_dev_id, enum config_profile_opcode_mod opmod,
			  int len, int offset, u8 *in_out_buf)
{
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	struct device *dev = NULL;

	if (!(get_dev_by_bus_id_and_addr(bus_id, addr, dev)))
		return -EINVAL;

	return config_profile_cmd(dev, opmod, len, offset, in_out_buf);
}
EXPORT_SYMBOL(sx_config_profile_cmd);

int sx_query_boardinfo_cmd(int i2c_dev_id, int len, int offset, u8 *in_out_buf)
{
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	struct device *dev = NULL;

	if (!(get_dev_by_bus_id_and_addr(bus_id, addr, dev)))
		return -EINVAL;

	return query_boardinfo_cmd(dev, len, offset, in_out_buf);
}
EXPORT_SYMBOL(sx_query_boardinfo_cmd);

int sx_query_fw_cmd(int i2c_dev_id, enum query_fw_opcode_mod opmod, 
			enum query_fw_input_mod inmod, u8 event, int len, int offset, u8 *in_out_buf)
{
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	struct device *dev = NULL;

	if (!(get_dev_by_bus_id_and_addr(bus_id, addr, dev)))
		return -EINVAL;
	return query_fw_cmd(dev, opmod, inmod, event, len, offset, in_out_buf);
}
EXPORT_SYMBOL(sx_query_fw_cmd);

int sx_software_reset(int i2c_dev_id)
{
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	struct device *dev = NULL;

	if (!(get_dev_by_bus_id_and_addr(bus_id, addr, dev)))
		return -EINVAL;
	return sx_i2c_software_reset(dev);
}
EXPORT_SYMBOL(sx_software_reset);

int sx_i2c_get_probe_status(struct i2c_client *client, u8 *probe_status, int *probe_cntr, u8 *go_bit_status,
				u8 *hcr_acess_status)
{
	struct switchx_probe_failed *p_probe_failed = NULL;
	struct switchx_data *data = NULL;

	if((!client) || (!client->adapter))
		return -EINVAL;

	/* Check whether this device failed last probing - if yes remove return failed status */
	if (!(p_probe_failed = find_dev_probe_context(client->adapter->nr)))
		return -EINVAL;
	*probe_status = p_probe_failed->probe_status;
	*probe_cntr = p_probe_failed->probe_cntr;
	if ((data = i2c_get_clientdata(client))) {
		*go_bit_status = data->go_bit_status;
		*hcr_acess_status = data->hcr_access_status;
	}
	return 0;
}
EXPORT_SYMBOL(sx_i2c_get_probe_status);

int sx_i2c_get_fw_rev(int i2c_dev_id, u64* fw_rev)
{
	struct i2c_client *client;
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	struct query_fw_output_mbox_layout query_fw;

	client = sx_get_client_by_i2c_bus_and_addr(bus_id, addr);
	if(!client)
		return -EUNATCH;

	if (!query_fw_cmd(&client->dev, SX_MB_OUTPUT, SX_BAR0_OFF_NONE, \
					  CIR_EVENT_REPORT, sizeof(struct query_fw_output_mbox_layout), 0, (u8*)&query_fw))
		return -EIO;

	*fw_rev = (u64)query_fw.fw_rev_major << 32 | (u64)query_fw.fw_rev_minor << 16 | query_fw.fw_rev_subminor;

	return 0;
}

int sx_i2c_set_go_bit_stuck(int i2c_dev_id)
{
	struct switchx_data *data = NULL;
	struct i2c_client *client;
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);

	client = sx_get_client_by_i2c_bus_and_addr(bus_id, addr);
	if(!client)
		return -EUNATCH;
	data = i2c_get_clientdata(client);
	if(!data)
		return -EINVAL;
	if(data->signature != SX_I2C_DEV_SIGNATURE)
		return -EINVAL;
	data->go_bit_status = GO_BIT_STAT_STUCK;
	return 0;
}

int sx_i2c_get_local_mbox(int i2c_dev_id, u32 *mb_size_in, u32 *mb_offset_in, u32 *mb_size_out, u32 *mb_offset_out)
{
	struct switchx_data *data;
	struct i2c_client *client;
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);

        if (!mb_size_in || !mb_offset_in || !mb_size_out || !mb_offset_out)
		return -EINVAL;
	client = sx_get_client_by_i2c_bus_and_addr(bus_id, addr);
	if(!client)
		return -EUNATCH;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	if(client->adapter->blocked == I2C_ADAP_BLK) {
		printk(KERN_INFO "%s (blocked) i2c_dev_id=%x bus=%d addr=%x client=%p\n",
			__FUNCTION__, i2c_dev_id, bus_id, addr, client);
		return -EUNATCH;
	}
#endif

	data = i2c_get_clientdata(client);
	if(!data)
		return -EINVAL;

	*mb_size_in = data->mb_size_in;
	*mb_offset_in = data->mb_offset_in;
	*mb_size_out = data->mb_size_out;
	*mb_offset_out = data->mb_offset_out;

	return 0;
}

int sx_i2c_enforce_route(int i2c_dev_id)
{
	struct switchx_data *data;
	struct i2c_client *client;
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	int rc = 0;
#ifdef SX_CACHE_OPTIMIZED
	int count = SX_ROUTE_DEPTH - 1;
#endif

	/* No need of enfocrce/release for bus of TOR */
	if (bus_id == 2)
		return 0;

	client = sx_get_client_by_i2c_bus_and_addr(bus_id, addr);
	if(!client)
		return -EUNATCH;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	if(client->adapter->blocked == I2C_ADAP_BLK) {
		printk(KERN_INFO "%s (blocked) i2c_dev_id=%x bus=%d addr=%x client=%p\n",
			__FUNCTION__, i2c_dev_id, bus_id, addr, client);
		return -EUNATCH;
	}
#endif

	if(asic_bad_health(bus_id))
		return -ECANCELED;

	data = i2c_get_clientdata(client);
	if(!data)
		return -EINVAL;

	mutex_lock(&__i2c_mlnx_lock);

	/* No need optimization for enfocrce/release for bus of TOR */
	if (bus_id == 2) {
		data->route_enforced = 1;
		return 0;
	}
	if (switchx_cfg.last_access_dev_id && (i2c_dev_id != switchx_cfg.last_access_dev_id))
		printk(KERN_INFO "%s mismatch detected i2c_dev_id=%d, last_access_dev_id=%d\n",
			__FUNCTION__, i2c_dev_id, switchx_cfg.last_access_dev_id);
	switchx_cfg.last_access_dev_id = i2c_dev_id;

#if defined SX_CACHE_OPTIMIZED
	for (count = SX_ROUTE_DEPTH - 1; count > 0; count--) {
		struct i2c_mux_priv *priv = data->mux_priv[count];
		data->deselect[count - 1] = priv->deselect;
		priv->deselect = NULL;
 	}
#elif defined SX_CACHE_OPTIMIZED_2
	mlnx_i2c_save_nailed_up_route(client->adapter);
#endif
	data->route_enforced = 1;

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s i2c_dev_id=%x bus=%d addr=%x client=%p\n", __FUNCTION__, i2c_dev_id, bus_id, addr, client);

	return rc;
}

int sx_i2c_release_route(int i2c_dev_id)
{
	struct switchx_data *data = NULL;
	struct i2c_client *client;
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	int rc = 0;
#ifdef SX_CACHE_OPTIMIZED
	int count = 0;
#endif

	/* No need of enfocrce/release for bus of TOR */
	if (bus_id == 2) {
		__last_accessed_bus_id = LAST_I2C_ACCESS_FAILED;
		return 0;
	}

	if (switchx_cfg.last_access_dev_id && (i2c_dev_id != switchx_cfg.last_access_dev_id))
		printk(KERN_INFO "%s mismatch detected i2c_dev_id=%d, last_access_dev_id=%d\n",
			__FUNCTION__, i2c_dev_id, switchx_cfg.last_access_dev_id);
	switchx_cfg.last_access_dev_id = 0;
	client = sx_get_client_by_i2c_bus_and_addr(bus_id, addr);
	if(!client) {
		rc = -EUNATCH;
		goto cleanup;
	}
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	if(client->adapter->blocked == I2C_ADAP_BLK) {
		printk(KERN_INFO "%s (blocked) i2c_dev_id=%x bus=%d addr=%x client=%p\n",
			__FUNCTION__, i2c_dev_id, bus_id, addr, client);
		rc = -EUNATCH;
		goto cleanup;
	}
#endif
	data = i2c_get_clientdata(client);
	if(!data) {
		rc = -EINVAL;
		goto cleanup;
	}

	/* No need optimization for enfocrce/release for bus of TOR */
	if (bus_id == 2) {
		goto cleanup;
	}
#ifdef SX_CACHE_OPTIMIZED
	for (count = 1; count < SX_ROUTE_DEPTH; count++) {
		struct i2c_mux_priv *priv = data->mux_priv[count];
		priv->deselect = data->deselect[count - 1];
                if(priv->deselect)
                	/*rc =*/ data->mux_priv[count]->deselect(priv->parent, priv->mux_dev, priv->chan_id);
	}
#endif

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s i2c_dev_id=%x bus=%d addr=%x client=%p\n", __FUNCTION__, i2c_dev_id, bus_id, addr, client);

cleanup:
	mutex_unlock(&__i2c_mlnx_lock);
	if (data)
		data->route_enforced = 0;
	if (rc)
		__last_accessed_bus_id = LAST_I2C_ACCESS_FAILED;
	else
		__last_accessed_bus_id = bus_id;
	return rc;
}

int sx_i2c_write(int i2c_dev_id, int offset, int len, u8 *in_out_buf)
{
	struct switchx_data *data;
	struct i2c_client *client;
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	struct i2c_msg msgs[1];
	u8 msgbuf[SWITCHX_ADDR32];
	u8 tbuf[MAX_I2C_BUFF_SIZE + 64];
	unsigned long orig_jiffies; 
	int try = 0;
	int rc = 0;

#if defined(__i386__) || defined(__x86_64__)
        int num = len/LPCI2C_BLOCK_MAX;
        int off = offset;
        int chunk_size = (len > LPCI2C_BLOCK_MAX) ? LPCI2C_BLOCK_MAX : len;
	u8 *pbuf = tbuf;
        int i;

        if (len%LPCI2C_BLOCK_MAX)
                num++;
#endif

	client = sx_get_client_by_i2c_bus_and_addr(bus_id, addr);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s i2c_dev_id=%x bus=%d addr=%x client=%p\n", __FUNCTION__, i2c_dev_id, bus_id, addr, client);
	if(!client)
		return -EUNATCH;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	if(client->adapter->blocked == I2C_ADAP_BLK) {	
		printk(KERN_INFO "%s (blocked) i2c_dev_id=%x bus=%d addr=%x client=%p\n",
			__FUNCTION__, i2c_dev_id, bus_id, addr, client);
		return -EUNATCH;
	}
#endif
	data = i2c_get_clientdata(client);
	if(!data)
		return -EINVAL;

	memset(msgs, 0, sizeof(struct i2c_msg));
	memset(tbuf, 0, len + SWITCHX_ADDR32);
	memcpy(tbuf + SWITCHX_ADDR32, in_out_buf, len);

#if defined(__i386__) || defined(__x86_64__)
        for (i = 0; i < num; i++) {
		chunk_size = (len > LPCI2C_BLOCK_MAX) ? LPCI2C_BLOCK_MAX : len;
		rc = translate_offset_addr(client->addr, off, chunk_size, SWITCHX_ADDR32,
						msgbuf, msgs, pbuf, 0);
#else

	translate_offset_addr(client->addr, offset, len, SWITCHX_ADDR32,
				msgbuf, msgs, tbuf, 0);
#endif

	orig_jiffies = jiffies;
	for (try = 0; try <= 0; try++) {
		rc = i2c_transfer(client->adapter, msgs, 1);
		if (rc > 0)
			break;
	}
	data->retry_cntr += try;

#if defined(__i386__) || defined(__x86_64__)
		pbuf += chunk_size;
		off += chunk_size;
		len -= chunk_size;
		memset(msgs, 0, 2*sizeof(struct i2c_msg));
        }
#endif


	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s rc=%d try=%d\n", __FUNCTION__, rc, try);

	if (rc < 0) {
		/* Transaction problem - test device with probing */
		sx_i2c_probing_test(client, data);
	}

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s i2c_dev_id=%x bus=%d addr=%x client=%p\n", __FUNCTION__, i2c_dev_id, bus_id, addr, client);


	return (rc > 0) ? 0 : rc;			   		 
}

int sx_i2c_read(int i2c_dev_id, int offset, int len, u8 *in_out_buf)
{
	struct switchx_data *data;
	struct i2c_client *client;
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	struct i2c_msg msgs[2];
	u8 msgbuf[SWITCHX_ADDR32];
	unsigned long orig_jiffies; 
	int try = 0;
	int rc = 0;

#if defined(__i386__) || defined(__x86_64__)
        int num = len/LPCI2C_BLOCK_MAX;
        int off = offset;
        int chunk_size = (len > LPCI2C_BLOCK_MAX) ? LPCI2C_BLOCK_MAX : len;
	u8 *pbuf = in_out_buf;
        int i;

        if (len%LPCI2C_BLOCK_MAX)
                num++;
#endif

	client = sx_get_client_by_i2c_bus_and_addr(bus_id, addr);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s i2c_dev_id=%x bus=%d addr=%x client=%p\n", __FUNCTION__, i2c_dev_id, bus_id, addr, client);
	if(!client)
		return -EUNATCH;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	if(client->adapter->blocked == I2C_ADAP_BLK) {	
		printk(KERN_INFO "%s (blocked) i2c_dev_id=%x bus=%d addr=%x client=%p\n",
			__FUNCTION__, i2c_dev_id, bus_id, addr, client);
		return -EUNATCH;
	}
#endif
	data = i2c_get_clientdata(client);
	if(!data)
		return -EINVAL;

	memset(msgs, 0, 2*sizeof(struct i2c_msg));

#if defined(__i386__) || defined(__x86_64__)
        for (i = 0; i < num; i++) {
		chunk_size = (len > LPCI2C_BLOCK_MAX) ? LPCI2C_BLOCK_MAX : len;
		rc = translate_offset_addr(client->addr, off, chunk_size, SWITCHX_ADDR32,
						msgbuf, msgs, pbuf, I2C_M_RD);
#else
	translate_offset_addr(client->addr, offset, len, SWITCHX_ADDR32,
				msgbuf, msgs, in_out_buf, I2C_M_RD);
#endif

	orig_jiffies = jiffies;
	for (try = 0; try <= 0; try++) {
		rc = i2c_transfer(client->adapter, msgs, 2);
		if (rc > 0)
			break;
	}
	data->retry_cntr += try;

#if defined(__i386__) || defined(__x86_64__)
		pbuf += chunk_size;
		off += chunk_size;
		len -= chunk_size;
		memset(msgs, 0, 2*sizeof(struct i2c_msg));
        }
#endif

	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s rc=%d try=%d\n", __FUNCTION__, rc, try);
 
	if (rc < 0) {
		/* Transaction problem - test device with probing */
		sx_i2c_probing_test(client, data);
	}

	return (rc > 0) ? 0 : rc;
}

int sx_i2c_write_dword(int i2c_dev_id, int offset, u32 val)
{
	struct switchx_data *data;
	struct i2c_client *client;
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	struct i2c_msg msgs[1];
	u8 msgbuf[SWITCHX_ADDR32];
	u8 tbuf[SWITCHX_ADDR32 + 4];
	unsigned long orig_jiffies; 
	int try;
	int rc = 0;

	client = sx_get_client_by_i2c_bus_and_addr(bus_id, addr);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX, "%s i2c_dev_id=%x bus=%d addr=%x client=%p\n",
		    __FUNCTION__, i2c_dev_id, bus_id, addr, client);
	if(!client)
		return -EUNATCH;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	if(client->adapter->blocked == I2C_ADAP_BLK) {
		printk(KERN_INFO "%s (blocked) i2c_dev_id=%x bus=%d addr=%x client=%p\n",
			__FUNCTION__, i2c_dev_id, bus_id, addr, client);
		return -EUNATCH;
	}
#endif
	data = i2c_get_clientdata(client);
	if(!data)
		return -EINVAL;

	memset(msgs, 0, sizeof(struct i2c_msg));
	memset(tbuf, 0, SWITCHX_ADDR32 + 4);
	memcpy(tbuf + SWITCHX_ADDR32, (u8 *)&val, 4);
 
	translate_offset_addr(client->addr, offset, 4, SWITCHX_ADDR32,
				msgbuf, msgs, tbuf, 0);

	orig_jiffies = jiffies;
	for (try = 0; try <= 0; try++) {
		rc = i2c_transfer(client->adapter, msgs, 1);
		if (rc > 0)
			break;
	}
	data->retry_cntr += try;
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s rc=%d try=%d\n", __FUNCTION__, rc, try);

	if (rc < 0) {
		/* Transaction problem - test device with probing */
		sx_i2c_probing_test(client, data);
	}
	return (rc > 0) ? 0 : rc;				   		 
}
 
int sx_i2c_read_dword(int i2c_dev_id, int offset, u32 *val)
{
	struct switchx_data *data;
	struct i2c_client *client;
	int bus_id = BUS_ID_FROM_ID(i2c_dev_id);
	u8 addr = ADDR_FROM_DEV_ID(i2c_dev_id);
	struct i2c_msg msgs[2];
	u8 msgbuf[SWITCHX_ADDR32];
	unsigned long orig_jiffies; 
	int try;
	int rc = 0;

	client = sx_get_client_by_i2c_bus_and_addr(bus_id, addr);
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX, "%s i2c_dev_id=%x bus=%d addr=%x client=%p\n",
		  __FUNCTION__, i2c_dev_id, bus_id, addr, client);
	if(!client)
		return -EUNATCH;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	if(client->adapter->blocked == I2C_ADAP_BLK) {
		printk(KERN_INFO "%s (blocked) i2c_dev_id=%x bus=%d addr=%x client=%p\n",
			__FUNCTION__, i2c_dev_id, bus_id, addr, client);
		return -EUNATCH;
	}
#endif
	data = i2c_get_clientdata(client);
	if(!data)
		return -EINVAL;

	memset(msgs, 0, 2*sizeof(struct i2c_msg));
	translate_offset_addr(client->addr, offset, 4, SWITCHX_ADDR32,
				msgbuf, msgs, (u8 *)val, I2C_M_RD);

	orig_jiffies = jiffies;
	for (try = 0; try <= 0; try++) {
		rc = i2c_transfer(client->adapter, msgs, 2);
		if (rc > 0) {
			data->hcr_access_status = HCR_ACCESS_OK;
			break;
		}
		/* This API is used for access to HCR register. In case this 
                   transaction failed - status is to be provided, since such
                   failure in certain scenario can cause i2c bus stuck */
		data->hcr_access_status = HCR_ACCESS_NOT_OK;
	}
	data->retry_cntr += try;
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s rc=%d try=%d\n", __FUNCTION__, rc, try);

	if (rc < 0) {
		/* Transaction problem - test device with probing */
		sx_i2c_probing_test(client, data);
	}

	return (rc > 0) ? 0 : rc;
}

static struct bin_attribute query_fm_attr = {
	.attr = {
		.name = "queryfw",
		.mode = S_IRUGO,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct query_fw_output_mbox_layout),
	.read = query_fw_read,
};

static struct bin_attribute boardinfo_attr = {
	.attr = {
		.name = "boardinfo",
		.mode = S_IRUGO,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct query_boardinfo_layout),
	.read = boardinfo_read,
};

static struct bin_attribute configprofile_attr = {
	.attr = {
		.name = "configprofile",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct config_profile_mbox_layout),
	.read = config_profile_read,
	.write = config_profile_write,
};

static struct bin_attribute pspa_reg_attr = {
	.attr = {
		.name = "pspa_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_pspa_layout),
	.read = show_pspa_reg,
	.write = set_pspa_reg,
};

static struct bin_attribute qsptc_reg_attr = {
	.attr = {
		.name = "qsptc_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_qsptc_layout),
	.read = show_qsptc_reg,
	.write = set_qsptc_reg,
};

static struct bin_attribute qstct_reg_attr = {
	.attr = {
		.name = "qstct_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_qstct_layout),
	.read = show_qstct_reg,
	.write = set_qstct_reg,
};

static struct bin_attribute pmtu_reg_attr = {
	.attr = {
		.name = "pmtu_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_pmtu_layout),
	.read = show_pmtu_reg,
	.write = set_pmtu_reg,
};

static struct bin_attribute ptys_reg_attr = {
	.attr = {
		.name = "ptys_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_ptys_layout),
	.read = show_ptys_reg,
	.write = set_ptys_reg,
};

static struct bin_attribute pmlp_reg_attr = {
	.attr = {
		.name = "pmlp_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_pmlp_layout),
	.read = show_pmlp_reg,
	.write = set_pmlp_reg,
};

static struct bin_attribute plib_reg_attr = {
	.attr = {
		.name = "plib_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_plib_layout),
	.read = show_plib_reg,
	.write = set_plib_reg,
};

static struct bin_attribute paos_reg_attr = {
	.attr = {
		.name = "paos_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_paos_layout),
	.read = show_paos_reg,
	.write = set_paos_reg,
};

static struct bin_attribute spzr_reg_attr = {
	.attr = {
		.name = "spzr_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_spzr_layout),
	.read = show_spzr_reg,
	.write = set_spzr_reg,
};

static struct bin_attribute sw_reset_attr = {
	.attr = {
		.name = "sw_reset",
		.mode = S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_sw_reset_layout),
	.write = switchx_sw_reset,
};

static struct bin_attribute hcap_reg_attr = {
	.attr = {
		.name = "hcap_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_hcap_layout),
	.read = show_hcap_reg,
	.write = set_hcap_reg,
};

static struct bin_attribute htgt_reg_attr = {
	.attr = {
		.name = "htgt_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_htgt_layout),
	.read = show_htgt_reg,
	.write = set_htgt_reg,
};

static struct bin_attribute hpkt_reg_attr = {
	.attr = {
		.name = "hpkt_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_hpkt_layout),
	.read = show_hpkt_reg,
	.write = set_hpkt_reg,
};

static struct bin_attribute hdrt_reg_attr = {
	.attr = {
		.name = "hdrt_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_hdrt_layout),
	.read = show_hdrt_reg,
	.write = set_hdrt_reg,
};

static struct bin_attribute mfcr_reg_attr = {
	.attr = {
		.name = "mfcr_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_mfcr_layout),
	.read = show_mfcr_reg,
	.write = set_mfcr_reg,
};

static struct bin_attribute mfsc_reg_attr = {
	.attr = {
		.name = "mfsc_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_mfsc_layout),
	.read = show_mfsc_reg,
	.write = set_mfsc_reg,
};

static struct bin_attribute mfsm_reg_attr = {
	.attr = {
		.name = "mfsm_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_mfsm_layout),
	.read = show_mfsm_reg,
	.write = set_mfsm_reg,
};

static struct bin_attribute mfsl_reg_attr = {
	.attr = {
		.name = "mfsl_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_mfsl_layout),
	.read = show_mfsl_reg,
	.write = set_mfsl_reg,
};

static struct bin_attribute fore_reg_attr = {
	.attr = {
		.name = "fore_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_fore_layout),
	.read = show_fore_reg,
	.write = set_fore_reg,
};

static struct bin_attribute mcia_reg_attr = {
	.attr = {
		.name = "mcia_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_mcia_layout),
	.read = show_mcia_reg,
	.write = set_mcia_reg,
};

static struct bin_attribute mtmp_reg_attr = {
	.attr = {
		.name = "mtmp_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_mtmp_layout),
	.read = show_mtmp_reg,
	.write = set_mtmp_reg,
};

static struct bin_attribute mpgr_reg_attr = {
	.attr = {
		.name = "mpgr_reg",
		.mode = S_IRUGO|S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_mpgr_layout),
	.read = show_mpgr_reg,
	.write = set_mpgr_reg,
};

static struct bin_attribute mgir_reg_attr = {
	.attr = {
		.name = "mgir_reg",
		.mode = S_IRUGO,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_mgir_layout),
	.read = show_mgir_reg,
};

static struct bin_attribute cache_attr = {
	.attr = {
		.name = "cache",
		.mode = S_IWUSR,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = SX_MAX_REG_SIZE,
	.write = set_cache,
};

static struct bin_attribute msci_reg_attr = {
	.attr = {
		.name = "msci_reg",
		.mode = S_IRUGO,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		.owner = THIS_MODULE,
#endif
	},
	.size = sizeof(struct reg_msci_layout),
	.read = show_msci_reg,
};

/* Return 0 if detection is successful, -ENODEV otherwise */
static int switchx_detect(struct i2c_client *client,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
			 int kind,
#endif
			 struct i2c_board_info *info)
{
	struct i2c_adapter *adapter = client->adapter;

	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA 
				     | I2C_FUNC_SMBUS_WORD_DATA 
				     | I2C_FUNC_SMBUS_WRITE_BYTE))
		return -ENODEV;

	strlcpy(info->type, "switchx", I2C_NAME_SIZE);

	return 0;
}

static int switchx_probe(struct i2c_client *client,
			const struct i2c_device_id *id)
{
	struct switchx_data *data = NULL;
	struct output_param_opmode1 output_param;
	int fail_point = 0;
	int count = SX_ROUTE_DEPTH - 1;
        struct i2c_adapter *parent = NULL;
	struct i2c_mux_priv *priv = NULL;
	struct i2c_adapter *adap = NULL;
	struct switchx_probe_failed *p_probe_failed = NULL;
	struct query_fw_output_mbox_layout query_fw;
	int rc = -1;

	/* Check whether this device failed last probing - if yes remove from failed list */
	if (client->adapter) {
		if ((p_probe_failed = find_dev_probe_context(client->adapter->nr))) {
			p_probe_failed->probe_cntr++;
		}
		else {
			p_probe_failed = kmem_cache_alloc(switchx_probe_failed_cache, GFP_ATOMIC);
			if (!p_probe_failed) {
				rc = -ENOMEM;
				goto exit;
			}
			p_probe_failed->bus_id = client->adapter->nr;
			p_probe_failed->probe_cntr = 1;
			p_probe_failed->probe_status = PROBE_STAT_OK;
			list_add(&p_probe_failed->list, &switchx_cfg.probe_failed_list);
		}
	}
 
	data = kzalloc(sizeof(struct switchx_data), GFP_KERNEL);
	if (!data) {
		rc = -ENOMEM;
		goto exit;
	}
	memset(data, 0, sizeof(struct switchx_data));
	data->signature = SX_I2C_DEV_SIGNATURE;

	i2c_set_clientdata(client, data);
	mutex_init(&data->cmd_lock);

	if (sysfs_create_bin_file(&client->dev.kobj, &query_fm_attr))
		goto exit_free;
	if (sysfs_create_bin_file(&client->dev.kobj, &boardinfo_attr))
		goto exit_sysfs_attr_1;
	if (sysfs_create_bin_file(&client->dev.kobj, &configprofile_attr))
		goto exit_sysfs_attr_2;
	if (sysfs_create_bin_file(&client->dev.kobj, &pspa_reg_attr))
		goto exit_sysfs_attr_3;
	if (sysfs_create_bin_file(&client->dev.kobj, &qsptc_reg_attr))
		goto exit_sysfs_attr_4;
	if (sysfs_create_bin_file(&client->dev.kobj, &qstct_reg_attr))
		goto exit_sysfs_attr_5;
	if (sysfs_create_bin_file(&client->dev.kobj, &pmtu_reg_attr))
		goto exit_sysfs_attr_6;
	if (sysfs_create_bin_file(&client->dev.kobj, &ptys_reg_attr))
		goto exit_sysfs_attr_7;
	if (sysfs_create_bin_file(&client->dev.kobj, &pmlp_reg_attr))
		goto exit_sysfs_attr_8;
	if (sysfs_create_bin_file(&client->dev.kobj, &plib_reg_attr))
		goto exit_sysfs_attr_9;
	if (sysfs_create_bin_file(&client->dev.kobj, &paos_reg_attr))
		goto exit_sysfs_attr_10;
	if (sysfs_create_bin_file(&client->dev.kobj, &spzr_reg_attr))
		goto exit_sysfs_attr_11;
	if (sysfs_create_bin_file(&client->dev.kobj, &hcap_reg_attr))
		goto exit_sysfs_attr_12;
	if (sysfs_create_bin_file(&client->dev.kobj, &htgt_reg_attr))
		goto exit_sysfs_attr_13;
	if (sysfs_create_bin_file(&client->dev.kobj, &hpkt_reg_attr))
		goto exit_sysfs_attr_14;
	if (sysfs_create_bin_file(&client->dev.kobj, &hdrt_reg_attr))
		goto exit_sysfs_attr_15;
	if (sysfs_create_bin_file(&client->dev.kobj, &mfcr_reg_attr))
		goto exit_sysfs_attr_16;
	if (sysfs_create_bin_file(&client->dev.kobj, &mfsc_reg_attr))
		goto exit_sysfs_attr_17;
	if (sysfs_create_bin_file(&client->dev.kobj, &mfsm_reg_attr))
		goto exit_sysfs_attr_18;
	if (sysfs_create_bin_file(&client->dev.kobj, &mfsl_reg_attr))
		goto exit_sysfs_attr_19;
	if (sysfs_create_bin_file(&client->dev.kobj, &fore_reg_attr))
		goto exit_sysfs_attr_20;
	if (sysfs_create_bin_file(&client->dev.kobj, &mcia_reg_attr))
		goto exit_sysfs_attr_21;
	if (sysfs_create_bin_file(&client->dev.kobj, &mtmp_reg_attr))
		goto exit_sysfs_attr_22;
	if (sysfs_create_bin_file(&client->dev.kobj, &sw_reset_attr))
		goto exit_sysfs_attr_23;
	if (sysfs_create_bin_file(&client->dev.kobj, &cache_attr))
		goto exit_sysfs_attr_24;
	if (sysfs_create_bin_file(&client->dev.kobj, &mpgr_reg_attr))
		goto exit_sysfs_attr_25;
	if (sysfs_create_bin_file(&client->dev.kobj, &mgir_reg_attr))
		goto exit_sysfs_attr_26;
    if (sysfs_create_bin_file(&client->dev.kobj, &msci_reg_attr))
		goto exit_sysfs_attr_27;

	data->retries = 10;    // use 10 retries only during probing, then callibrate it accrding
			       // to number of retries of query_fm at it first probing call
	data->timeout = 500;
	data->transact_id = 0;
	data->bus_id = client->adapter->nr;
	data->i2c_dev_addr = client->addr;
	data->i2c_dev = &client->dev;
	data->client = client;
	data->route_enforced = 0;
	adap = client->adapter;
	/* Set path for access via virtual i2c */
	while(adap) {
		priv = adap->algo_data;
		if (!priv)
			break;
		parent = priv->parent;
		if (!parent)
			break;
		mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
			"%s (%d) selected count=%d bus=%d (%p) parent bus=%d (%p)\n",
			__FUNCTION__, __LINE__, count, adap->nr, adap, priv->parent->nr, priv->parent);
		data->parent[count] = adap;
		data->mux_priv[count] = priv;
		data->parent[count - 1] = parent;
		count--;
		adap = parent;
	}

	/* ACCESS_REG and CONFIG_PROFILE commands use mailboxes for input and for output parameters.
	   QUERY_FW and QUERY_BOARDINFO commands uses mailboxes for output parameters. 
	   In order to use mailboxes through the i2c, special area is reserved on the i2c address space that
	   can be used for input and output mailboxes. 
	   Such mailboxes are called Local Mailboxes. When using a local mailbox, software should specify 0 as the 
	   Input/Output parameters. The location of the Local Mailbox addresses on the i2c space can be retrieved
	   through the QUERY_FW command.
	   For this purpose QUERY_FW is to be issued with opcode modifier equal SX_IMM_OUTPUT_1 (0x01).
	   For such command the output parameter is an immediate value, struct output_param_opmode1.
	   Invoke QUERY_FW command for swicthx probing and for getting local mailboxes addresses.
	   Read from immedate output parameters.
	*/
	memset(&output_param, 0, sizeof(struct output_param_opmode1));
#if defined SX_CACHE_OPTIMIZED_2
	mlnx_i2c_save_nailed_up_route(client->adapter);
#endif
	mutex_lock(&__i2c_mlnx_lock);
	if((rc = query_fw_cmd(&client->dev, SX_IMM_OUTPUT_1, SX_BAR0_OFF_NONE, 
				CIR_EVENT_NO_REPORT, sizeof(struct output_param_opmode1),
				0, (u8 *)&output_param)) < 0) {
#if defined SX_CACHE_OPTIMIZED_2
		mlnx_i2c_break_nailed_up_route(client->adapter);
#endif
		mutex_unlock(&__i2c_mlnx_lock);
		p_probe_failed->probe_status = PROBE_STAT_FAILED;
		goto exit_query_fw_cmd;
	}
	mutex_unlock(&__i2c_mlnx_lock);
	p_probe_failed->probe_status = PROBE_STAT_OK;
	list_add(&data->list, &switchx_cfg.switchx_list);

#if defined(__i386__) || defined(__x86_64__)
	convert_mbox((u32 *)&output_param);
#endif
	mlnx_log((MLNX_LOG_SXI2C_MIN + 2), MLNX_LOG_SXI2C_MAX,
		"%s(%d) 0x%08x 0x%08x 0x%08x 0x%08x\n", 
		__FUNCTION__, __LINE__,
		output_param.out_param_h.local_mb_size, output_param.out_param_h.local_mb_bar0_offset,
		output_param.out_param_l.local_mb_size, output_param.out_param_l.local_mb_bar0_offset);
 
	data->mb_size_in = output_param.out_param_h.local_mb_size;
	data->mb_offset_in = output_param.out_param_h.local_mb_bar0_offset;
	data->mb_size_out = output_param.out_param_l.local_mb_size;
	data->mb_offset_out = output_param.out_param_l.local_mb_bar0_offset;
	data->retries = data->retry_cntr + 3;
	data->go_bit_status = GO_BIT_STAT_OK;
	data->hcr_access_status = HCR_ACCESS_OK;
	memset(data->cache, 0, SX_MAX_REG_SIZE);
	memset(&query_fw, 0, sizeof(struct query_fw_output_mbox_layout));
	mutex_lock(&__i2c_mlnx_lock);
	query_fw_cmd(&client->dev, SX_MB_OUTPUT, SX_BAR0_OFF_NONE, CIR_EVENT_REPORT,
			sizeof(struct query_fw_output_mbox_layout), 0, (u8*)&query_fw);
#if defined SX_CACHE_OPTIMIZED_2
	mlnx_i2c_break_nailed_up_route(client->adapter);
#endif

	mutex_unlock(&__i2c_mlnx_lock);
#if defined(__i386__) || defined(__x86_64__)
	convert_query_fw_mbox((u32 *)&query_fw);
#endif
	printk(KERN_INFO "%s (rtry=%d) probe on bus %d (cnt=%d) fw_rev: major=%d minor=%d submin=%d in mb size=%x off=0x%08x out mb size=%x off=0x%08x\n",
		__FUNCTION__, data->retry_cntr, p_probe_failed->bus_id, p_probe_failed->probe_cntr,
		query_fw.fw_rev_major, query_fw.fw_rev_minor, query_fw.fw_rev_subminor,
		data->mb_size_in, data->mb_offset_in, data->mb_size_out, data->mb_offset_out);
	return rc;

      exit_query_fw_cmd:
	fail_point = 30;
	sysfs_remove_bin_file(&client->dev.kobj, &msci_reg_attr);
      exit_sysfs_attr_27:
	fail_point = (fail_point == 0) ? 30 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &mgir_reg_attr);
      exit_sysfs_attr_26:
	fail_point = (fail_point == 0) ? 29 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &mpgr_reg_attr);
      exit_sysfs_attr_25:
	fail_point = (fail_point == 0) ? 28 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &cache_attr);
      exit_sysfs_attr_24:
	fail_point = (fail_point == 0) ? 27 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &sw_reset_attr);
      exit_sysfs_attr_23:
	fail_point = (fail_point == 0) ? 26 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &mtmp_reg_attr);
      exit_sysfs_attr_22:
	fail_point = (fail_point == 0) ? 25 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &mcia_reg_attr); 
      exit_sysfs_attr_21:
	fail_point = (fail_point == 0) ? 24 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &fore_reg_attr);  
      exit_sysfs_attr_20:
	fail_point = (fail_point == 0) ? 23 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &mfsl_reg_attr); 
      exit_sysfs_attr_19:
	fail_point = (fail_point == 0) ? 22 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &mfsm_reg_attr); 
      exit_sysfs_attr_18:
	fail_point = (fail_point == 0) ? 21 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &mfsc_reg_attr);  
      exit_sysfs_attr_17:
	fail_point = (fail_point == 0) ? 20 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &mfcr_reg_attr); 
      exit_sysfs_attr_16:
	fail_point = (fail_point == 0) ? 19 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &hdrt_reg_attr);
      exit_sysfs_attr_15:
	fail_point = (fail_point == 0) ? 18 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &hpkt_reg_attr);  
      exit_sysfs_attr_14:
	fail_point = (fail_point == 0) ? 17 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &htgt_reg_attr); 
      exit_sysfs_attr_13:
	fail_point = (fail_point == 0) ? 16 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &hcap_reg_attr);
      exit_sysfs_attr_12:
	fail_point = (fail_point == 0) ? 15 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &spzr_reg_attr);
      exit_sysfs_attr_11:
	fail_point = (fail_point == 0) ? 14 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &paos_reg_attr);
      exit_sysfs_attr_10:
	fail_point = (fail_point == 0) ? 13 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &plib_reg_attr);
      exit_sysfs_attr_9:
	fail_point = (fail_point == 0) ? 12 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &pmlp_reg_attr);
      exit_sysfs_attr_8:
	fail_point = (fail_point == 0) ? 11 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &ptys_reg_attr);
      exit_sysfs_attr_7:
	fail_point = (fail_point == 0) ? 10 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &pmtu_reg_attr);
      exit_sysfs_attr_6:
	fail_point = (fail_point == 0) ? 9 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &qstct_reg_attr);
      exit_sysfs_attr_5:
	fail_point = (fail_point == 0) ? 8 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &qsptc_reg_attr);
      exit_sysfs_attr_4:
	fail_point = (fail_point == 0) ? 7 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &pspa_reg_attr);
      exit_sysfs_attr_3:
	fail_point = (fail_point == 0) ? 6 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &configprofile_attr);
      exit_sysfs_attr_2:
	sysfs_remove_bin_file(&client->dev.kobj, &boardinfo_attr);
      exit_sysfs_attr_1:
	fail_point = (fail_point == 0) ? 5 : fail_point;
	sysfs_remove_bin_file(&client->dev.kobj, &query_fm_attr);
      exit_free:
	fail_point = (fail_point == 0) ? 2 : fail_point;
	i2c_set_clientdata(client, NULL);
	kfree(data);
      exit:
	fail_point = (fail_point == 0) ? 1 : fail_point;
	printk(KERN_INFO "%s: failed fail_point=%d\n", __FUNCTION__, fail_point);
     
	return rc;
}

static int switchx_remove(struct i2c_client *client)
{
	struct switchx_data *data = i2c_get_clientdata(client);
	sysfs_remove_bin_file(&client->dev.kobj, &msci_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &mgir_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &mpgr_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &cache_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &sw_reset_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &spzr_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &paos_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &plib_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &pmlp_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &ptys_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &pmtu_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &qstct_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &qsptc_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &pspa_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &hcap_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &htgt_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &hpkt_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &hdrt_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &mfcr_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &mfsc_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &mfsm_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &mfsl_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &fore_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &mcia_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &mtmp_reg_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &configprofile_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &boardinfo_attr);
	sysfs_remove_bin_file(&client->dev.kobj, &query_fm_attr);
	if (!list_empty(&switchx_cfg.switchx_list))
		list_del_rcu(&data->list);
	if(data)
		kfree(data);

	return 0;
}

static const struct i2c_device_id switchx_id[] = {
	{ "switchx", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, switchx_id);

/* This is the driver that will be inserted */
static struct i2c_driver switchx_driver = {
	.class		= I2C_CLASS_HWMON,
	.driver = {
		.name	= "switchx",
	},
	.probe		= switchx_probe,
	.remove		= switchx_remove,
	.id_table	= switchx_id,
	.detect		= switchx_detect,
};

static void *sx_symbol = NULL;
int *mlnx_debug_level_symbol = NULL;
static int __init switchx_init(void)
{
	int count;

	switchx_probe_failed_cache = kmem_cache_create("switchx_probe_failed_cache",
					 sizeof(struct switchx_probe_failed),
					 0,
					 SLAB_HWCACHE_ALIGN, NULL);
	if (!switchx_probe_failed_cache)
		return -ENOMEM;

	INIT_LIST_HEAD(&switchx_cfg.switchx_list);
	INIT_LIST_HEAD(&switchx_cfg.probe_failed_list);
	if (cir_reg_num == 0)
		switchx_cfg.reg_cir_offset = SX_BAR0_OFFSET_PRIM;
	else
		switchx_cfg.reg_cir_offset = SX_BAR0_OFFSET_SECOND;
	switchx_cfg.reg_sw_reset_offset = SX_SW_RESET_OFFSET;
	if (i2c_add_driver(&switchx_driver) != 0)
		return -1;
	switchx_cfg.last_access_dev_id = 0;
	for (count = 0; count < SX_ROUTE_DEPTH - 1; count++)
		__p_last_bus_cache_path[count] = &cache_template[count];

	printk(KERN_INFO "%s: reg_cir_offset=0x%08x reg_sw_reset_offset=0x%08x\n", 
		__FUNCTION__, switchx_cfg.reg_cir_offset, switchx_cfg.reg_sw_reset_offset);

	sx_symbol = __symbol_get("sx_i2c_get_probe_status_hook");
	if (sx_symbol == NULL) goto sx_i2c_get_probe_status_hook_failed;
	sx_symbol = __symbol_get("i2c_write_hook");
	if (sx_symbol == NULL) goto i2c_write_hook_failed;
	sx_symbol = __symbol_get("i2c_read_hook");
	if (sx_symbol == NULL) goto i2c_read_hook_failed;
	sx_symbol = __symbol_get("i2c_write_dword_hook");
	if (sx_symbol == NULL) goto i2c_write_dword_hook_failed;
	sx_symbol = __symbol_get("i2c_read_dword_hook");
	if (sx_symbol == NULL) goto i2c_read_dword_hook_failed;
	sx_symbol = __symbol_get("i2c_enforce_hook");
	if (sx_symbol == NULL) goto i2c_enforce_hook_failed;
	sx_symbol = __symbol_get("i2c_release_hook");
	if (sx_symbol == NULL) goto i2c_release_hook_failed;
	sx_symbol = __symbol_get("i2c_get_local_mbox_hook");
	if (sx_symbol == NULL) goto i2c_get_local_mbox_hook_failed;
	sx_symbol = __symbol_get("i2c_get_fw_rev_hook");
	if (sx_symbol == NULL) goto i2c_get_fw_rev_hook_failed;
	sx_symbol = __symbol_get("i2c_set_go_bit_stuck_hook");
	if (sx_symbol == NULL) goto i2c_set_go_bit_stuck_hook_failed;

	rcu_assign_pointer(sx_i2c_get_probe_status_hook, sx_i2c_get_probe_status);
	rcu_assign_pointer(i2c_write_hook, sx_i2c_write);
	rcu_assign_pointer(i2c_read_hook, sx_i2c_read);
	rcu_assign_pointer(i2c_write_dword_hook, sx_i2c_write_dword);
	rcu_assign_pointer(i2c_read_dword_hook, sx_i2c_read_dword);
	rcu_assign_pointer(i2c_enforce_hook, sx_i2c_enforce_route);
	rcu_assign_pointer(i2c_release_hook, sx_i2c_release_route);
	rcu_assign_pointer(i2c_get_local_mbox_hook, sx_i2c_get_local_mbox);
	rcu_assign_pointer(i2c_get_fw_rev_hook, sx_i2c_get_fw_rev);
	rcu_assign_pointer(i2c_set_go_bit_stuck_hook, sx_i2c_set_go_bit_stuck);

	mlnx_debug_level_symbol = __symbol_get("mlnx_debug_level");

	return 0;

i2c_set_go_bit_stuck_hook_failed: 
	__symbol_put("i2c_get_fw_rev_hook");
i2c_get_fw_rev_hook_failed:
	__symbol_put("i2c_get_local_mbox_hook");
i2c_get_local_mbox_hook_failed:
	__symbol_put("i2c_release_hook");
i2c_release_hook_failed:
	__symbol_put("i2c_enforce_hook");
i2c_enforce_hook_failed:
	__symbol_put("i2c_read_dword_hook");
i2c_read_dword_hook_failed:
	__symbol_put("i2c_write_dword_hook");
i2c_write_dword_hook_failed:
	__symbol_put("i2c_read_hook");
i2c_read_hook_failed:
	__symbol_put("i2c_write_hook");
i2c_write_hook_failed:
	__symbol_put("sx_i2c_get_probe_status_hook");
sx_i2c_get_probe_status_hook_failed:

	return 0;
}

static void __exit switchx_exit(void)
{
#if defined(__i386__) || defined(__x86_64__)
    struct switchx_probe_failed *swpf, *next;
#endif

	if (sx_symbol != NULL) {
		__symbol_put("sx_i2c_get_probe_status_hook");
		__symbol_put("i2c_write_hook");
		__symbol_put("i2c_read_hook");
		__symbol_put("i2c_write_dword_hook");
		__symbol_put("i2c_read_dword_hook");
		__symbol_put("i2c_enforce_hook");
		__symbol_put("i2c_release_hook");
		__symbol_put("i2c_get_local_mbox_hook");
		__symbol_put("i2c_get_fw_rev_hook");
		__symbol_put("i2c_set_go_bit_stuck_hook");
		rcu_assign_pointer(sx_i2c_get_probe_status_hook, NULL);
		rcu_assign_pointer(i2c_write_hook, NULL);
		rcu_assign_pointer(i2c_read_hook, NULL);
		rcu_assign_pointer(i2c_write_dword_hook, NULL);
		rcu_assign_pointer(i2c_read_dword_hook, NULL);
		rcu_assign_pointer(i2c_enforce_hook, NULL);
		rcu_assign_pointer(i2c_release_hook, NULL);
		rcu_assign_pointer(i2c_get_local_mbox_hook, NULL);
		rcu_assign_pointer(i2c_get_fw_rev_hook, NULL);
		rcu_assign_pointer(i2c_set_go_bit_stuck_hook, NULL);
	}

	i2c_del_driver(&switchx_driver);

	if (switchx_probe_failed_cache) {
#if defined(__i386__) || defined(__x86_64__)
                list_for_each_entry_safe(swpf, next, &switchx_cfg.probe_failed_list, list) {
                        list_del(&swpf->list);
                        kmem_cache_free(switchx_probe_failed_cache, swpf);
                }
#endif
		kmem_cache_destroy(switchx_probe_failed_cache);
    }

	if (mlnx_debug_level_symbol)
		__symbol_put("mlnx_debug_level");
}


MODULE_AUTHOR("Vadim Pasternak <vadimp@mellanox.com>");
MODULE_DESCRIPTION("SWITCHX I2C/SMbus driver");
MODULE_LICENSE("GPL v2");

module_init(switchx_init);
module_exit(switchx_exit);
