/**
 *
 * Copyright (C) 2011-2014, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <linux/mod_devicetable.h>
#include <asm/io.h>
#include <linux/rculist.h>
#include <linux/jhash.h>
#include <linux/i2c.h>
#include <linux/i2c/pca954x.h>
#include <linux/i2c/pca953x.h>
#include <linux/gpio.h>
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
#include <linux/gpio-i2cmux.h>
#endif
#include <linux/platform_device.h>
#include "cpld_handler.h"
#include "mlnx_i2c_if.h"
#include "mlnx_log.h"
#include "mlnx_if.h"
#include "mlnx_if_int.h"
#if defined(__x86_64__)
#include "cpld_mux.h"
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
#define I2C_ADAP_BLK     0
#define I2C_ADAP_UNBLK   1
#endif

#define PHYSICAL_BUS_NUM 1
#define SWITCH_MAX_DEPTH 3

#define PCA954x_HASH_BITS 8
#define PCA954x_HASH_SIZE (1 << PCA954x_HASH_BITS)
#define I2C_DEV_HASH_BITS 8
#define I2C_DEV_HASH_SIZE (1 << I2C_DEV_HASH_BITS)
#define I2C_BUS_HASH_BITS 8
#define I2C_BUS_HASH_SIZE (1 << I2C_BUS_HASH_BITS)
#define CPLD_HASH_BITS 4
#define CPLD_HASH_SIZE (1 << CPLD_HASH_BITS)
#define TRANSACT_TO       500

static int mellaggra_dbg = 0;
#define DBG(args...) {\
	if (mellaggra_dbg) {\
                printk(KERN_INFO args);\
	}\
}

DEFINE_MUTEX(__i2c_mlnx_lock);
EXPORT_SYMBOL_GPL(__i2c_mlnx_lock);
DEFINE_MUTEX(__lpc_mlnx_lock);
EXPORT_SYMBOL_GPL(__lpc_mlnx_lock);

//#define SX_CACHE_OPTIMIZED_2 1
int __last_accessed_bus_id = LAST_I2C_ACCESS_FAILED;
EXPORT_SYMBOL_GPL(__last_accessed_bus_id);
struct i2c_mux_priv_cache *__p_last_bus_cache_path[SX_ROUTE_DEPTH - 1];
EXPORT_SYMBOL_GPL(__p_last_bus_cache_path);

/* External hooks */
int (*sx_i2c_get_probe_status_hook)(struct i2c_client *client, u8 *probe_status, int *probe_cntr,
				    u8 *go_bit_status, u8 *hcr_acess_status) = NULL;
int (*cpld_handler_action_subscribe_hook)(struct cpld_handler_action_entry *action_entry) = NULL;
int (*cpld_handler_action_unsubscribe_hook)(struct cpld_handler_action_entry *action_entry) = NULL;

EXPORT_SYMBOL(sx_i2c_get_probe_status_hook);
EXPORT_SYMBOL(cpld_handler_action_subscribe_hook);
EXPORT_SYMBOL(cpld_handler_action_unsubscribe_hook);

int (*mlnx_cpld_led_config_symbol_hook)(struct mlnx_cpld_leds_config_ioctl_params *p_cpld_led_config) = NULL;
int (*mlnx_asic_fan_config_symbol_hook)(struct mlnx_asic_fan_config_ioctl_params *p_asic_fan_config) = NULL;
int (*mlnx_asic_temp_sensor_config_symbol_hook)(struct mlnx_asic_temp_sensor_config_ioctl_params *p_asic_temp_sensor) = NULL;

EXPORT_SYMBOL(mlnx_cpld_led_config_symbol_hook);
EXPORT_SYMBOL(mlnx_asic_fan_config_symbol_hook);
EXPORT_SYMBOL(mlnx_asic_temp_sensor_config_symbol_hook);

static struct pca954x_platform_mode mlnx_pca9548_switch_mode[] = {  {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0} };
static struct pca954x_platform_data mlnx_pca9548_switch_data = {
    .modes     = mlnx_pca9548_switch_mode,
    .num_modes = ARRAY_SIZE(mlnx_pca9548_switch_mode),
};
static struct i2c_board_info i2c_bus_devices[MAX_DEV_INREQ_NUM];

#if defined(__x86_64__)
static struct cpld_mux_platform_mode cpld_mux_tor_modes[] = {  {0, 1}, {0, 1}, {0, 1}, {0, 1}, {0, 1}, {0, 1}, {0, 1}, {0, 1} };
#endif

struct pca954x_table {
	struct hlist_node hlist;
	struct list_head list;
	struct list_head attached_switch_list;
	u32 switch_id;
	u32 parent_bus_num;
	struct i2c_adapter *adap;
	struct i2c_client *client;
	int adap_id_array[LEGS_NUM];
	char switch_name[SWITCH_NAME_SIZE];
	u8 attached_dev_cntr;
	u8 attached_switch_cntr;
	u8 status;
	int sel_reg_addr;
};

struct i2c_devices_table {
	struct hlist_node hlist;
	u32 parent_bus_num;
	struct i2c_adapter *adap;
	struct i2c_client *client;
	u32 user_handle;
	char i2c_dev_type[I2C_NAME_SIZE];
};

struct i2c_bus_table {
	struct hlist_node hlist;
	struct list_head attached_gpio_bus_list;
	u32 bus_id;
	u32 switch_id;
};

#ifdef GPIO_BUS
struct i2c_gpio_table {
	struct hlist_node hlist;
	struct list_head list;
	struct platform_device *plat_dev;
	struct gpio_i2cmux_platform_data mux_data;
	char mux_name[I2C_NAME_SIZE];
	unsigned mux_values[MAX_I2C_GPIO_BUSES];
	unsigned mux_gpios[MAX_GPIO_NUM];
	int gpio_num;
};
#endif /* GPIO_BUS */

struct irq_action_entry {
	u_int8_t rw_func_id:2;    /* read-write function Id - up to 4 functions,
			            0x0 - read-write for device attached to I2C_BUS_TYPE;
			            0x1 - read-write for device attached to I2C_BUS_TYPE;
			            0x2 0x3 - no function is assigned */
	u_int8_t reserved:1;
	u_int8_t flags:5;
};

struct irq_set_entry {
	__u8 status_range[2];        /* Status range -  from address to address */
	__u8 event_range[2];         /* Event range -  from address to address */
	__u8 mask_range[2];          /* Mask range -  from address to address */
	__u32 mask;                  /* Interrupt mask */
	struct irq_config_entry cfg; /* Interrupt maske, event, status configuration */
	struct irq_action_entry act; /* Entry to action table */
	union hw_dev status_hwdev;   /* Access parameters */
	union hw_dev event_hwdev;    /* Access parameters */
	union hw_dev mask_hwdev;     /* Access parameters */
};

struct irq_get_entry {
	__u32 status[2];             /* Status */
	__u32 event[2];              /* Event */
	struct irq_action_entry act; /* Entry to action table */
	union {                      /* Access parameters */
		struct mlnx_i2c_dev_params i2c_params;
		struct mlnx_lpc_dev_params lpc_params;
	} hwdev;
};

struct mlnx_cpld_interrupt_set_params_action {
	union {
		struct cpld_mgmt_set_action {
			struct irq_set_entry s_entry[MAX_MGMT_IRQS]; /* Interrupt management  */
		} mgmt_set_action;
		struct cpld_module_set_action {
			struct irq_set_entry s_entry[MAX_MOD_IRQS];  /* Interrupt modules'  */
		} module_set_action;
		struct cpld_tor_set_action {
			struct irq_set_entry s_entry[MAX_TOR_IRQS];  /* Interrupt Top Of the Rack'  */
		} tor_set_action;
	} u;
};

struct cpld_device_table {
	struct hlist_node hlist;
	u32 device_id;
	u16 device_type;                                               /* device attached to I2C_BUS_TYPE or to LPC_BUS_TYPE */
	u8 card_type;                                                  /* card type: management, module, TOR */
	irq_signal_e wake_on_signal;                                   /* For aggregation - signal which is not ignored */
	int usr_handle;                                                /* upper user hanlde */
	u8 status;                                                     /* current status of device */
	u8 hw_acc_tab_num;                                             /* real number of interrupt capable devices */
	u8 hw_term_acc_tab_num;                                        /* real number of interrupt capable termination devices */
	struct dev_access_block hw_acc_tab[DEV_INT_CAP_MAX];           /* table of interrupt capable devices */
	struct dev_access_block hw_term_acc_tab[DEV_TERM_INT_CAP_MAX]; /* table of interrupt capable termination devices */
	struct bus_reset_block bus_rst;                                /* bus physical reset block */
	struct cpld_handler_action_entry action_entry;                 /* irq handler pre-action and post-action */
	struct mod_digest digest_entry;                                /* module digst entry */
	enum irq_level_e int_level;                                    /* interrupt's hierarchical level */
        u8 irq_entry_num;                                              /* number of entries in irq table */
	struct mlnx_cpld_interrupt_set_params_action *p_irqs;          /* mask irq mapping */
	struct mlnx_cpld_interrupt_get_params *p_get_irqs;             /* status and event irq mapping */
	struct mlnx_cpld_interrupt_get_params *p_auxilary_cache_blk;   /* status and event irq mapping in shadow */
	struct mlnx_cpld_interrupt_get_params *p_master_cache_blk;     /* changed status and event irq mapping */
};

static struct kmem_cache *pca954x_table_cache __read_mostly;
static struct kmem_cache *i2c_devices_table_cache __read_mostly;
static struct kmem_cache *i2c_bus_table_cache __read_mostly;
#ifdef GPIO_BUS
static struct kmem_cache *i2c_gpio_table_cache __read_mostly;
#endif /* GPIO_BUS */
static struct kmem_cache *cpld_device_table_cache __read_mostly;

struct i2c_cntrl_ctx {
	struct platform_device* plat_dev;
	struct hlist_head pca954x_table_hash[PCA954x_HASH_SIZE];
	struct hlist_head i2c_devices_table_hash[I2C_DEV_HASH_SIZE];
	struct hlist_head i2c_bus_table_hash[I2C_BUS_HASH_SIZE];
#ifdef GPIO_BUS
	struct hlist_head i2c_gpio_table_hash[I2C_BUS_HASH_SIZE];
#endif /* GPIO_BUS */
} i2c_ctx;

struct cpld_cntrl_ctx {
	struct platform_device* plat_dev;
	struct hlist_head cpld_device_table_hash[CPLD_HASH_SIZE];
	rw_bus_access_func rw_func[MAX_BUS_TYPES];
	rw_bus_reset_func rst_func[MAX_BUS_TYPES];
	__u8 wake_on_signal_id_tab[MAX_MODULE_NUM]; /* Aggregated signals passed to underlying levels' actions (_none_sig - nothing) */
	__u8 wake_on_signal_cntr;                   /* Counter of the above signals */
} cpld_ctx;

static inline int switch_hash_func(int *switch_id)
{
	u32 *key = (u32 *)switch_id;
	return jhash(key, 4, 0) & (PCA954x_HASH_SIZE - 1);
}

static inline int device_hash_func(int *bus_id)
{
	u32 *key = (u32 *)bus_id;
	return jhash(key, 4, 0) & (I2C_BUS_HASH_SIZE - 1);
}

static inline int bus_hash_func(int *bus_id)
{
	u32 *key = (u32 *)bus_id;
	return jhash(key, 4, 0) & (I2C_DEV_HASH_SIZE - 1);
}

static inline int cpld_device_hash_func(int *device_id)
{
	u32 *key = (u32 *)device_id;
	return jhash(key, 4, 0) & (CPLD_HASH_SIZE - 1);
}

static inline struct pca954x_table *find_pca954x_switch(u32 switch_id)
{
	struct hlist_node *h;

	hlist_for_each(h, &i2c_ctx.pca954x_table_hash[switch_hash_func(&switch_id)]) {
		struct pca954x_table *t;
		t = hlist_entry(h, struct pca954x_table, hlist);
		if (t->switch_id == switch_id)
			return t;
	}
	return NULL;
}

static inline struct pca954x_table *find_pca954x_switch_by_name(const char *name)
{
	int cntr;

	for (cntr = 0; cntr < PCA954x_HASH_SIZE; cntr++) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		struct hlist_node *h = NULL, *n = NULL;
#else
		struct hlist_node *n = NULL;
#endif
		static struct pca954x_table *t = NULL;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		hlist_for_each_entry_safe(t, h, n, &i2c_ctx.pca954x_table_hash[cntr], hlist) {
#else
		hlist_for_each_entry_safe(t, n, &i2c_ctx.pca954x_table_hash[cntr], hlist) {
#endif
			if (!strncmp(t->switch_name, name, SWITCH_NAME_SIZE))
				return t;
		}
	}
	return NULL;
}

static inline struct pca954x_table *find_pca954x_main_switch(void)
{
	int cntr;

	for (cntr = 0; cntr < PCA954x_HASH_SIZE; cntr++) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		struct hlist_node *h = NULL, *n = NULL;
#else
		struct hlist_node *n = NULL;
#endif
		static struct pca954x_table *t = NULL;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		hlist_for_each_entry_safe(t, h, n, &i2c_ctx.pca954x_table_hash[cntr], hlist) {
#else
		hlist_for_each_entry_safe(t, n, &i2c_ctx.pca954x_table_hash[cntr], hlist) {
#endif
			if(t->parent_bus_num == PHYSICAL_BUS_NUM)
				return t;
		}
	}
	return NULL;
}

static inline struct i2c_devices_table *find_i2c_device(u32 bus_id, u8 *is_switch, u8 *addr)
{
	struct hlist_node *h;

	*is_switch = 0;

	hlist_for_each(h, &i2c_ctx.i2c_devices_table_hash[device_hash_func(&bus_id)]) {
		struct i2c_devices_table *t;
		t = hlist_entry(h, struct i2c_devices_table, hlist);
		if (t->parent_bus_num == bus_id) {
			if (!strcmp(t->client->name, "pca9548"))
				*is_switch = 1;
			*addr = t->client->addr;
			return t;
		}
	}
	return NULL;
}

static inline struct i2c_devices_table *find_i2c_device_by_addr(u32 bus_id, u8 addr, u8 *is_switch)
{
	struct hlist_node *h;

	*is_switch = 0;

	hlist_for_each(h, &i2c_ctx.i2c_devices_table_hash[device_hash_func(&bus_id)]) {
		struct i2c_devices_table *t;
		t = hlist_entry(h, struct i2c_devices_table, hlist);
		if ((t->parent_bus_num == bus_id) && (t->client->addr == addr)) {
			if (!strcmp(t->client->name, "pca9548"))
				*is_switch = 1;
			return t;
		}
	}
	return NULL;
}

static inline struct i2c_devices_table *get_i2c_device(u32 bus_id, u8 addr)
{
	struct hlist_node *h;

	hlist_for_each(h, &i2c_ctx.i2c_devices_table_hash[device_hash_func(&bus_id)]) {
		struct i2c_devices_table *t;
		t = hlist_entry(h, struct i2c_devices_table, hlist);
		if ((t->parent_bus_num == bus_id) && (t->client->addr == addr)) {
			return t;
		}
	}
	return NULL;
}

static inline struct i2c_bus_table *find_i2c_bus_entry(u32 bus_id)
{
	struct hlist_node *h;

	hlist_for_each(h, &i2c_ctx.i2c_bus_table_hash[bus_hash_func(&bus_id)]) {
		struct i2c_bus_table *b;
		b = hlist_entry(h, struct i2c_bus_table, hlist);
		if (b->bus_id == bus_id)
			return b;
	}
	return NULL;
}

#ifdef GPIO_BUS
static inline struct i2c_gpio_table *find_i2c_gpio_entry(u32 bus_id)
{
	struct hlist_node *h;

	hlist_for_each(h, &i2c_ctx.i2c_gpio_table_hash[bus_hash_func(&bus_id)]) {
		struct i2c_gpio_table *g;
		g = hlist_entry(h, struct i2c_gpio_table, hlist);
		if (g->mux_data.parent == bus_id)
			return g;
	}
	return NULL;
}
#endif /* GPIO_BUS */

static inline struct cpld_device_table *find_cpld_device_entry(u32 device_id)
{
	struct hlist_node *h;

	hlist_for_each(h, &cpld_ctx.cpld_device_table_hash[cpld_device_hash_func(&device_id)]) {
		struct cpld_device_table *c;
		c = hlist_entry(h, struct cpld_device_table, hlist);
		if (c->device_id == device_id)
			return c;
	}
	return NULL;
}


static inline int __i2c_match_addr(struct device *dev, void *addrp)
{
	struct i2c_client *client = i2c_verify_client(dev);
	int addr = *(int *)addrp;
        DBG("%s client=%p addr=%x\n", __FUNCTION__, client, addr);
	if (client && client->addr == addr)
		return 1;
	return 0;
}

static inline struct i2c_client *__get_client_by_adapter(struct i2c_adapter *adap, int addr)
{
        struct device *client_dev;
	client_dev = device_find_child(&adap->dev, &addr, __i2c_match_addr);
	if(!client_dev) {
                printk(KERN_INFO "%s client not found bus=%d addr=%x\n", __FUNCTION__, adap->nr, addr);
		return NULL;
        }
	return to_i2c_client(client_dev);
}

static int lpc_rw(u16 base, u16 size, u8 offset,
                        int len, u8 rw_flag, u8 *p_data)
{
	struct mlnx_lpc_kernel_params params;

        DBG("%s size=%d offset=%d rw_flag=%d base=%x len=%x\n",
                __FUNCTION__, size, offset, rw_flag,
                base, len);

	memset(&params, 0, sizeof(struct mlnx_lpc_kernel_params));
	params.base = base;           /* base address of LPC bus */
	params.offset= offset;        /* offset within LPC device from which to start access */
	params.read_write = rw_flag;  /* write - 0, read 1 */
	params.datalen = len;         /* Length of data */
	params.data = p_data;         /* Pointer to data */

	mlnx_rw_lpc_kernel(&params);

	return 0;
}

int i2c_bus_access_func(union hw_dev *hw_data, __u8 from_range,
			__u8 to_range, __u8 rw_flag, __u8 *p_data, __u8 lock_flag)
{
        union i2c_smbus_data data;
	int size;
        int len;
        int rc = 0;

        mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MIN,
                "%s from_range=%x to_range=%x rw_flag=%d bus=%d addr=%x addr_space=%d adap=%p\n",
                __FUNCTION__, from_range, to_range, rw_flag,
                hw_data->i2c_params.bus_id, hw_data->i2c_params.addr,
                hw_data->i2c_params.addr_space, hw_data->i2c_params.adap);

        len = to_range - from_range + 1;

        if (len == 1) {       // size is byte
                size = I2C_SMBUS_BYTE_DATA;
                if (rw_flag == I2C_SMBUS_WRITE)
                        data.byte = *p_data;
        }
        else if (len == 2) {  // size is word
                size = I2C_SMBUS_WORD_DATA;
                if (rw_flag == I2C_SMBUS_WRITE)
                        memcpy(&data.word, p_data, 2);
        }
        else {                // size is block
                size = I2C_SMBUS_I2C_BLOCK_DATA;
                data.block[0] = len;
                if (rw_flag == I2C_SMBUS_WRITE)
                        memcpy(&data.block[1], p_data, data.block[0]);
        }

        mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MIN,
                "%s offset=%x %s addr=%x size=%d data=%x\n",
                __FUNCTION__, from_range, (rw_flag == I2C_SMBUS_WRITE)?"write":"read",
                hw_data->i2c_params.addr, size, *p_data);

        if (lock_flag == _B_LOCK)
                mutex_lock(&__i2c_mlnx_lock);
        rc = i2c_smbus_xfer(hw_data->i2c_params.adap, hw_data->i2c_params.addr, 0,
                            rw_flag, from_range, size, &data);
        if (lock_flag == _B_LOCK)
                mutex_unlock(&__i2c_mlnx_lock);

        if (rc < 0) {
                printk(KERN_INFO "%s(%d) access failed bus=%d addr=0x%x off=0x%x size=%d on %s lock=%d rc=%d\n",
                                  __FUNCTION__, __LINE__, hw_data->i2c_params.bus_id,
                                  hw_data->i2c_params.addr, from_range, size,
                                  (rw_flag == I2C_SMBUS_WRITE)?"write":"read",
                                  lock_flag, rc);
                return rc;
        }

        mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
                "%s rc=%d\n", __FUNCTION__, rc);

        if (rw_flag == I2C_SMBUS_READ) {
                if (len == 1)
                        *p_data = data.byte;
                else if (len == 2)
                        memcpy(p_data, &data.word, 2);
                else
                        memcpy(p_data, &data.block[1], data.block[0]);
        }
#if 0
        hex_dump(p_data, len, MLNX_LOG_MELLAGGRA_MIN, MLNX_LOG_MELLAGGRA_MIN);
#endif
        return rc;
}
EXPORT_SYMBOL(i2c_bus_access_func);

int lpc_bus_access_func(union hw_dev *hw_data, __u8 from_range,
			__u8 to_range, __u8 rw_flag, __u8 *p_data, __u8 lock_flag)
{
        int rc;
        int len = to_range - from_range + 1;
        mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MIN,
                "%s from_range=%d to_range=%d rw_flag=%d base=%x size=%x len=%d\n",
                __FUNCTION__, from_range, to_range, rw_flag,
                hw_data->lpc_params.base, hw_data->lpc_params.size, len);
        if (lock_flag == _B_LOCK)
                mutex_lock(&__lpc_mlnx_lock);
        rc = lpc_rw(hw_data->lpc_params.base, hw_data->lpc_params.size,
                    from_range, len, rw_flag, p_data);
        if (lock_flag == _B_LOCK)
                mutex_unlock(&__lpc_mlnx_lock);
#if 0
        hex_dump(p_data, len, MLNX_LOG_MELLAGGRA_MIN, MLNX_LOG_MELLAGGRA_MIN);
#endif
	return 0;
}
EXPORT_SYMBOL(lpc_bus_access_func);

int i2c_bus_reset_func(struct bus_reset_block *bus_rst, __u8 lock_flag)
{
	__u8 reset_mask;
	int rc;

	reset_mask = ~bus_rst->toggle_off_mask;
	rc = i2c_bus_access_func(&bus_rst->hwdev, bus_rst->offset, bus_rst->offset,
				 0, &reset_mask, lock_flag);
	if (rc < 0)
		return rc;
	msleep(100);

	reset_mask = ~bus_rst->toggle_on_mask;
	rc = i2c_bus_access_func(&bus_rst->hwdev, bus_rst->offset, bus_rst->offset,
				 0, &reset_mask, lock_flag);

	return rc;
}
EXPORT_SYMBOL(i2c_bus_reset_func);

int lpc_bus_reset_func(struct bus_reset_block *bus_rst, __u8 lock_flag)
{
	return 0;
}
EXPORT_SYMBOL(lpc_bus_reset_func);

void _lock_bus(__u8 lock_unlock_flag)
{
	if (lock_unlock_flag == LPC_BUS_TYPE)
		mutex_lock(&__lpc_mlnx_lock);
	if (lock_unlock_flag == I2C_BUS_TYPE)
		mutex_lock(&__i2c_mlnx_lock);
}
EXPORT_SYMBOL(_lock_bus);

void _unlock_bus(__u8 lock_unlock_flag)
{
	if (lock_unlock_flag == I2C_BUS_TYPE)
		mutex_unlock(&__i2c_mlnx_lock);
	if (lock_unlock_flag == LPC_BUS_TYPE)
		mutex_unlock(&__lpc_mlnx_lock);
}
EXPORT_SYMBOL(_unlock_bus);

int mlnx_add_i2c_virtual_buses(struct mlnx_i2c_virt_bus_ioctl_params *params)
{
	int adap_cntr;
	int switch_cntr;
	struct i2c_adapter *adap;
	struct i2c_client *client;
	struct i2c_switch_params *sw_params;
	int parent_bus_num = params->parent_bus_num;
	int switch_id;
	int parent_switch_id = params->upper_switch_id;
	struct hlist_head *head;
	struct hlist_head *head_bus;
	struct pca954x_table *p_pca954x_tbl[MAX_SW_INREQ_NUM];
	struct pca954x_table *p_pca954x_parent = NULL;
	struct i2c_bus_table *p_bus_tbl[MAX_SW_INREQ_NUM][LEGS_NUM];
#if defined(__x86_64__)
	struct cpld_mux_platform_data* cpld_mux_data = NULL;
	int i, cpld_mux_req = 0;
#endif
        params->status = 0;
        if (params->element_num > MAX_SW_INREQ_NUM)
	        return -EFAULT;
        if (params->element_num == 1 && (params->flag == 0)) {
                if ((p_pca954x_tbl[0] = find_pca954x_switch(params->sw[0].switch_id))) {
			if (p_pca954x_tbl[0]->status == I2C_ADAP_UNBLK) {
                        	/* Swicth is exist.
                        	   Attempt to unblock non-blocked switch or reopen opened buses */
				if (p_pca954x_tbl[0]->adap && p_pca954x_tbl[0]->client) {
			        	printk(KERN_INFO "%s: client at bus %d addr %x is already exist\n",
				        	__FUNCTION__, p_pca954x_tbl[0]->client->adapter->nr,
						p_pca954x_tbl[0]->client->addr);
			        	params->status = 1;
                        	}
				else {
		        		return -EFAULT;
				}
		        	return 0;
			}
			p_pca954x_tbl[0]->status = I2C_ADAP_UNBLK;
                        /* swicth is exist */
                        if (parent_switch_id == 0) {
                                printk(KERN_INFO "%s: client (%p on adap %p) at bus %d (%d) addr %x is held\n",
                                        __FUNCTION__, p_pca954x_tbl[0]->client, p_pca954x_tbl[0]->client->adapter,
                                        p_pca954x_tbl[0]->client->adapter->nr,
                                        p_pca954x_tbl[0]->adap->nr, p_pca954x_tbl[0]->client->addr);
                                return 0;
                        }
			if (p_pca954x_tbl[0]->adap) {
		        	mutex_lock(&__i2c_mlnx_lock);
				p_pca954x_tbl[0]->adap->blocked = I2C_ADAP_UNBLK;
			        if (p_pca954x_tbl[0]->client) {
						if (!(p_pca954x_tbl[0]->sel_reg_addr)) {
							pca954x_block_unblock(p_pca954x_tbl[0]->client, I2C_ADAP_UNBLK);
							pca954x_reset(p_pca954x_tbl[0]->client);
						}
#if defined(__x86_64__)
						else {
							cpld_mux_block_unblock(p_pca954x_tbl[0]->client, I2C_ADAP_UNBLK);
							cpld_mux_reset(p_pca954x_tbl[0]->client);
						}
#endif
			        }
		        	mutex_unlock(&__i2c_mlnx_lock);
			        printk(KERN_INFO "%s: client (%p on adap %p) at bus %d addr %x is unblocked\n",
				        __FUNCTION__, p_pca954x_tbl[0]->client, p_pca954x_tbl[0]->client->adapter,
				        p_pca954x_tbl[0]->client->adapter->nr, p_pca954x_tbl[0]->client->addr);
                        }

		        return 0;
                }
        }

        memset(p_pca954x_tbl, 0, MAX_SW_INREQ_NUM * sizeof(struct pca954x_table *));
        memset(p_bus_tbl, 0, MAX_SW_INREQ_NUM * LEGS_NUM * sizeof(struct i2c_bus_table *));
        memset(&i2c_bus_devices, 0, sizeof(struct i2c_board_info) * MAX_SW_INREQ_NUM);
        if (parent_switch_id)
                p_pca954x_parent = find_pca954x_switch(parent_switch_id);

        for (switch_cntr = 0; switch_cntr < params->element_num; switch_cntr++) {

		sw_params = &params->sw[switch_cntr];
		switch_id = sw_params->switch_id;
		head =  &i2c_ctx.pca954x_table_hash[switch_hash_func(&switch_id)];
		p_pca954x_tbl[switch_cntr] = kmem_cache_alloc(pca954x_table_cache, GFP_ATOMIC);
		memset(p_pca954x_tbl[switch_cntr], 0, sizeof(struct pca954x_table));
		INIT_HLIST_NODE(&p_pca954x_tbl[switch_cntr]->hlist);
		INIT_LIST_HEAD(&p_pca954x_tbl[switch_cntr]->attached_switch_list);
		INIT_LIST_HEAD(&p_pca954x_tbl[switch_cntr]->list);
		hlist_add_head_rcu(&p_pca954x_tbl[switch_cntr]->hlist, head);
		//mlnx_pca9548_switch_data.num_modes = sw_params->adap_num;
		for (adap_cntr = 0; adap_cntr < sw_params->adap_num; adap_cntr++) {
                	mlnx_pca9548_switch_mode[adap_cntr].adap_id = sw_params->adap_id_array[adap_cntr];
                	mlnx_pca9548_switch_mode[adap_cntr].deselect_on_exit = sw_params->deselect_on_exit;
                	p_pca954x_tbl[switch_cntr]->adap_id_array[adap_cntr] = sw_params->adap_id_array[adap_cntr];
			head_bus = &i2c_ctx.i2c_bus_table_hash[bus_hash_func(&sw_params->adap_id_array[adap_cntr])];
			p_bus_tbl[switch_cntr][adap_cntr] = kmem_cache_alloc(i2c_bus_table_cache, GFP_ATOMIC);
			INIT_HLIST_NODE(&p_bus_tbl[switch_cntr][adap_cntr]->hlist);
			INIT_LIST_HEAD(&p_bus_tbl[switch_cntr][adap_cntr]->attached_gpio_bus_list);
			hlist_add_head_rcu(&p_bus_tbl[switch_cntr][adap_cntr]->hlist, head_bus);
			p_bus_tbl[switch_cntr][adap_cntr]->switch_id = sw_params->switch_id;
			p_bus_tbl[switch_cntr][adap_cntr]->bus_id = sw_params->adap_id_array[adap_cntr];
        	}
#if defined(__x86_64__)
		if (!(sw_params->sel_reg_addr)) {
			// pca954x
			mlnx_pca9548_switch_data.num_modes = sw_params->adap_num;
			i2c_bus_devices[switch_cntr].platform_data = &mlnx_pca9548_switch_data;
		}
		else {
			/* Cpld_mux platforma data is individual for each cpld_mux.
			   It's allocated here and freed by cpld_mux driver. */
			i2c_bus_devices[switch_cntr].platform_data = NULL;
			cpld_mux_data = kzalloc(sizeof(struct cpld_mux_platform_data), GFP_KERNEL);
			if (!cpld_mux_data) {
				for (i = 0; i < switch_cntr; i++) {
					cpld_mux_data = i2c_bus_devices[i].platform_data;
					if (cpld_mux_data)
						kfree(cpld_mux_data);
				}
				return -ENOMEM;
			}
			cpld_mux_data->modes = cpld_mux_tor_modes;
			cpld_mux_data->num_modes = sw_params->adap_num;
			cpld_mux_data->id = sw_params->switch_id;
			cpld_mux_data->sel_reg_addr = sw_params->sel_reg_addr;
			cpld_mux_data->first_channel = sw_params->adap_id_array[0];
			cpld_mux_data->addr = sw_params->i2c_dev_addr;
			i2c_bus_devices[switch_cntr].platform_data = cpld_mux_data;
			i2c_bus_devices[switch_cntr].flags = I2C_CLIENT_TEN;
			cpld_mux_req = 1;
		}
#else
		mlnx_pca9548_switch_data.num_modes = sw_params->adap_num;
		i2c_bus_devices[switch_cntr].platform_data = &mlnx_pca9548_switch_data;
#endif
		memcpy(i2c_bus_devices[switch_cntr].type, sw_params->i2c_dev_type, I2C_NAME_SIZE);
		i2c_bus_devices[switch_cntr].addr = 0xFF & ((unsigned short)sw_params->i2c_dev_addr);
		adap = i2c_get_adapter(parent_bus_num);
		if (adap == NULL) {
			printk(KERN_INFO "mlnx_add_i2c_virtual_buses: failed to get adapter for bus %d\n",
			        parent_bus_num);
#if defined(__x86_64__)
			if (cpld_mux_data)
				kfree(cpld_mux_data);
#endif
			return -EFAULT;
		}

		adap->blocked = I2C_ADAP_UNBLK;
		client = i2c_new_device(adap, (struct i2c_board_info const*)&i2c_bus_devices[switch_cntr]);
		if (client == NULL) {
			printk(KERN_INFO "mlnx_add_i2c_virtual_buses: failed to create client for bus %d\n",
			        parent_bus_num);
			goto err_exit;
		}
		p_pca954x_tbl[switch_cntr]->switch_id = sw_params->switch_id;
		p_pca954x_tbl[switch_cntr]->parent_bus_num = parent_bus_num;
		p_pca954x_tbl[switch_cntr]->adap = adap;
		p_pca954x_tbl[switch_cntr]->status = I2C_ADAP_UNBLK;
		p_pca954x_tbl[switch_cntr]->client = client;
		p_pca954x_tbl[switch_cntr]->sel_reg_addr = sw_params->sel_reg_addr;
		memcpy(p_pca954x_tbl[switch_cntr]->switch_name, sw_params->switch_name, SWITCH_NAME_SIZE);
		if (p_pca954x_parent) {
			p_pca954x_parent->attached_switch_cntr++;
			list_add_rcu(&p_pca954x_tbl[switch_cntr]->list, &p_pca954x_parent->attached_switch_list);
		}
	} /* for (switch_cntr = 0; switch_cntr < params->element_num; swicth_cntr++) */
	return 0;
err_exit:
        /* In case of error clear all elements */
        for (switch_cntr = 0; switch_cntr < params->element_num; switch_cntr++) {
#if defined(__x86_64__)
			if (cpld_mux_req) {
				cpld_mux_data = i2c_bus_devices[switch_cntr].platform_data;
				if (cpld_mux_data)
					kfree(cpld_mux_data);
			}
#endif
		    if (p_pca954x_tbl[switch_cntr]) {
		        if (p_pca954x_tbl[switch_cntr]->client) {
			        if (p_pca954x_tbl[switch_cntr]->adap)
					mutex_lock(&__i2c_mlnx_lock); //i2c_lock_adapter(p_pca954x_tbl[switch_cntr]->adap);
			        i2c_unregister_device(p_pca954x_tbl[switch_cntr]->client);
			        if (p_pca954x_tbl[switch_cntr]->adap)
					mutex_unlock(&__i2c_mlnx_lock); //i2c_unlock_adapter(p_pca954x_tbl[switch_cntr]->adap);
		        }
		        if (p_pca954x_tbl[switch_cntr]->adap)
			        i2c_put_adapter(p_pca954x_tbl[switch_cntr]->adap);
	                p_pca954x_tbl[switch_cntr]->adap = NULL;
	                p_pca954x_tbl[switch_cntr]->client = NULL;
	                p_pca954x_tbl[switch_cntr]->switch_id = PCA954x_SWITCH_NONE;
                        for (adap_cntr = 0; adap_cntr < sw_params->adap_num; adap_cntr++) {
                                if (p_bus_tbl[switch_cntr][adap_cntr]) {
#ifdef GPIO_BUS
			                struct i2c_gpio_table *p, *n;
			                list_for_each_entry_safe(p, n,
			                        &p_bus_tbl[switch_cntr][adap_cntr]->attached_gpio_bus_list, list) {
				                struct mlnx_gpio_i2c_del_ioctl_params gpio_params;

				                memset(&gpio_params, 0, sizeof(struct mlnx_gpio_i2c_del_ioctl_params));
				                gpio_params.element_num = 1;
				                gpio_params.parent_bus_num[0] = p->mux_data.parent;
				                mlnx_del_gpio_i2c_buses(&gpio_params);
			                }
#endif /*GPIO_BUS*/
                                        hlist_del_rcu(&p_bus_tbl[switch_cntr][adap_cntr]->hlist);
		                        kmem_cache_free(i2c_bus_table_cache, p_bus_tbl[switch_cntr][adap_cntr]);
                                }
                        }
		        if (p_pca954x_parent) {
			        p_pca954x_parent->attached_switch_cntr--;
			        list_del_rcu(&p_pca954x_tbl[switch_cntr]->list);
                        }
		        hlist_del_rcu(&p_pca954x_tbl[switch_cntr]->hlist);
		        kmem_cache_free(pca954x_table_cache, p_pca954x_tbl[switch_cntr]);
             }
	}

	return -EFAULT;
}

int mlnx_delete_i2c_virtual_buses(struct mlnx_i2c_switch_ioctl_params *params)
{
	int switch_cntr;
	int adap_cntr;
	struct i2c_adapter *adap;
	struct i2c_client *client;
	int parent_switch_id = params->upper_switch_id;
	struct pca954x_table *p_pca954x_tbl[MAX_SW_INREQ_NUM];
	struct pca954x_table *p_pca954x_parent = NULL;
	struct i2c_bus_table *p_bus;

        if (params->element_num > MAX_SW_INREQ_NUM)
	        return -EFAULT;
        if ((params->element_num == 1) && (params->flag == 0)) {
                /* flag != 0 - don't remove, just block and remove attached devices */
		p_pca954x_tbl[0] = find_pca954x_switch(params->sw[0]);
		if(!p_pca954x_tbl[0])
	                return -EFAULT;
		if (((p_pca954x_tbl[0]->attached_dev_cntr == 0) || (parent_switch_id != 0)) &&
                    (p_pca954x_tbl[0]->adap_id_array[0] < (BASE_VIRT_BUS + LEGS_NUM))) {
		        p_pca954x_tbl[0]->status = I2C_ADAP_BLK;
		        if (parent_switch_id == 0)
		                return 0;
                        if (p_pca954x_tbl[0]->attached_dev_cntr > 0 ) {
                                /* !!! Cleanup: switch has attached devices - delete it */
                	        struct mlnx_i2c_switch_ioctl_params dev_params;

                                memset(&dev_params, 0, sizeof(struct mlnx_i2c_switch_ioctl_params));
                                dev_params.element_num = 1;
                                dev_params.upper_switch_id = parent_switch_id;
                                dev_params.flag = 0;
                                dev_params.sw[0] = params->sw[0];

	        		mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
	        		        "%s(%d) clean device with attached=%d upper_switch_id=%d switch_id=%d\n",
	        		        __FUNCTION__, __LINE__, p_pca954x_tbl[0]->attached_dev_cntr,
	        		        dev_params.upper_switch_id, dev_params.sw[0]);

                                mlnx_delete_devices_from_vbus(&dev_params);
                        }
                        return mlnx_block_i2c_virtual_buses(params);
                }
        }

        memset(p_pca954x_tbl, 0, MAX_SW_INREQ_NUM*sizeof(struct pca954x_table *));
        if (parent_switch_id)
                p_pca954x_parent = find_pca954x_switch(parent_switch_id);
        for (switch_cntr = 0; switch_cntr < params->element_num; switch_cntr++) {
		p_pca954x_tbl[switch_cntr] = find_pca954x_switch(params->sw[switch_cntr]);
		if(!p_pca954x_tbl[switch_cntr])
	                return -EFAULT;
		adap = p_pca954x_tbl[switch_cntr]->adap;
		client = p_pca954x_tbl[switch_cntr]->client;
                if (p_pca954x_tbl[switch_cntr]->attached_dev_cntr > 0 ) {
                	/* !!! Cleanup: switch has attache devices - delete it now */
                	struct mlnx_i2c_switch_ioctl_params dev_params;

                        printk(KERN_INFO "%s: switch 0x%08x bus 0x%08x has %d attached dev\n",
                                __FUNCTION__, params->sw[switch_cntr],
                                GET_BUS_ID_BY_SWITCH_ID(params->sw[switch_cntr]),
                                p_pca954x_tbl[switch_cntr]->attached_dev_cntr);
                        memset(&dev_params, 0, sizeof(struct mlnx_i2c_switch_ioctl_params));
                        dev_params.element_num = 1;
                        dev_params.upper_switch_id = parent_switch_id;
                        dev_params.sw[0] = params->sw[switch_cntr];

	        	mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
	        	        "%s(%d) multi clean device switch_cntr=%d with attached=%d upper_switch_id=%d switch_id=%x\n",
	        	        __FUNCTION__, __LINE__, switch_cntr, p_pca954x_tbl[switch_cntr]->attached_dev_cntr,
	        	        dev_params.upper_switch_id, dev_params.sw[0]);

                        mlnx_delete_devices_from_vbus(&dev_params);
                }
		if (p_pca954x_tbl[switch_cntr]->attached_switch_cntr > 0 ) {
			struct pca954x_table *p, *n;
			printk(KERN_INFO "%s: switch 0x%08x has %d switches attached\n",
				__FUNCTION__, params->sw[switch_cntr],
				p_pca954x_tbl[switch_cntr]->attached_switch_cntr);

			list_for_each_entry_safe(p, n, &p_pca954x_tbl[switch_cntr]->attached_switch_list, list) {
				struct mlnx_i2c_switch_ioctl_params sw_params;

				memset(&sw_params, 0, sizeof(struct mlnx_i2c_switch_ioctl_params));
				sw_params.element_num = 1;
				sw_params.upper_switch_id = p_pca954x_tbl[switch_cntr]->switch_id;
				sw_params.sw[0] = p->switch_id;

	        	        mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
	        	                "%s(%d) clean bus switch_cntr=%d with attached=%d upper_switch_id=%d switch_id=%d\n",
	        	                __FUNCTION__, __LINE__, switch_cntr, p_pca954x_tbl[switch_cntr]->attached_dev_cntr,
	        	                sw_params.upper_switch_id, sw_params.sw[0]);

				mlnx_delete_i2c_virtual_buses(&sw_params);
			}
			list_del_rcu(&p_pca954x_tbl[switch_cntr]->attached_switch_list);
		}

		if (client) {
			printk(KERN_INFO "%s: client at bus %d addr %x is unregistered\n",
				__FUNCTION__, client->adapter->nr, client->addr);
			if (adap)
		        	mutex_lock(&__i2c_mlnx_lock); //i2c_lock_adapter(adap);
			i2c_unregister_device(client);
			if (adap)
				mutex_unlock(&__i2c_mlnx_lock); //i2c_unlock_adapter(adap);
		}
		if (adap)
			i2c_put_adapter(adap);
		p_pca954x_tbl[switch_cntr]->adap = NULL;
		p_pca954x_tbl[switch_cntr]->client = NULL;
                for (adap_cntr = 0; adap_cntr < LEGS_NUM; adap_cntr++) {
			p_bus = find_i2c_bus_entry(p_pca954x_tbl[switch_cntr]->adap_id_array[adap_cntr]);
			if (p_bus) {
			        hlist_del_rcu(&p_bus->hlist);
			        kmem_cache_free(i2c_bus_table_cache, p_bus);
			}
		}
		if (p_pca954x_parent) {
			p_pca954x_parent->attached_switch_cntr--;
			list_del_rcu(&p_pca954x_tbl[switch_cntr]->list);
		}
		hlist_del_rcu(&p_pca954x_tbl[switch_cntr]->hlist);
		p_pca954x_tbl[switch_cntr]->switch_id = PCA954x_SWITCH_NONE;
		kmem_cache_free(pca954x_table_cache, p_pca954x_tbl[switch_cntr]);
        }

	return 0;
}

int mlnx_block_i2c_virtual_buses(struct mlnx_i2c_switch_ioctl_params *params)
{
	int switch_cntr;
	struct pca954x_table *p_pca954x_tbl = NULL;

        for (switch_cntr = 0; switch_cntr < params->element_num; switch_cntr++) {
	        p_pca954x_tbl = find_pca954x_switch(params->sw[switch_cntr]);
	        if (p_pca954x_tbl) {
		        mutex_lock(&__i2c_mlnx_lock);
		        if (p_pca954x_tbl->adap)
			        p_pca954x_tbl->adap->blocked = I2C_ADAP_BLK;
		        mutex_unlock(&__i2c_mlnx_lock);
			printk(KERN_INFO "%s: client (%p on adap %p) at bus %d addr %x is blocked\n",
				__FUNCTION__, p_pca954x_tbl->client, p_pca954x_tbl->client->adapter,
				p_pca954x_tbl->client->adapter->nr, p_pca954x_tbl->client->addr);
                }
        }
	return 0;
}

int mlnx_block_unblock_i2c_virtual_bus_kernel(unsigned short virt_bus_num, unsigned char state)
{
    struct i2c_adapter *adap = NULL;
    int blocked;

    if (state)
        blocked = I2C_ADAP_BLK;
    else
        blocked = I2C_ADAP_UNBLK;

    adap = i2c_get_adapter(virt_bus_num);
    if(!adap) {
        printk(KERN_INFO "%s(%d): failed get adapter of bus %d\n", __FUNCTION__, __LINE__, virt_bus_num);
        return -ENXIO;
    }

    mutex_lock(&__i2c_mlnx_lock);
    adap->blocked = blocked;
    mutex_unlock(&__i2c_mlnx_lock);

    i2c_put_adapter(adap);

    printk(KERN_INFO "%s: bus %d is %s\n", __FUNCTION__,
           virt_bus_num, (state ? "blocked": "unblocked"));

    return 0;
}
EXPORT_SYMBOL(mlnx_block_unblock_i2c_virtual_bus_kernel);

int mlnx_unblock_i2c_virtual_buses(struct mlnx_i2c_switch_ioctl_params *params)
{
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	int switch_cntr;
	struct pca954x_table *p_pca954x_tbl = NULL;

        for (switch_cntr = 0; switch_cntr < params->element_num; switch_cntr++) {
	        p_pca954x_tbl = find_pca954x_switch(params->sw[switch_cntr]);
	        if (p_pca954x_tbl) {
		        mutex_lock(&__i2c_mlnx_lock);
			if (p_pca954x_tbl->adap)
				p_pca954x_tbl->adap->blocked = I2C_ADAP_UNBLK;
			if (p_pca954x_tbl->client) {
				if (!(p_pca954x_tbl->sel_reg_addr)) {
					pca954x_block_unblock(p_pca954x_tbl->client, I2C_ADAP_UNBLK);
					pca954x_reset(p_pca954x_tbl->client);
				}
#if defined(__x86_64__)
				else {
					cpld_mux_block_unblock(p_pca954x_tbl->client, I2C_ADAP_UNBLK);
					cpld_mux_reset(p_pca954x_tbl->client);
				}
#endif
			}
			else {
		        mutex_unlock(&__i2c_mlnx_lock);
				printk(KERN_INFO "%s: client is NULL\n", __FUNCTION__);
				return -1;
			}
		    mutex_unlock(&__i2c_mlnx_lock);
			printk(KERN_INFO "%s: client at bus %d addr %x is unblocked\n",
				__FUNCTION__, p_pca954x_tbl->client->adapter->nr, p_pca954x_tbl->client->addr);
		}
	}
#endif
	return 0;
}

int mlnx_reset_i2c_virtual_buses_and_devices(struct mlnx_i2c_switch_ioctl_params *params)
{
	struct pca954x_table *p_pca954x_tbl = NULL;
	struct i2c_devices_table *p_i2c_dev_tbl = NULL;
	int switch_cntr;
	int adap_cntr;
	int rc = 0;

        for (switch_cntr = 0; switch_cntr < params->element_num; switch_cntr++) {
	        p_pca954x_tbl = find_pca954x_switch(params->sw[switch_cntr]);
	        if (p_pca954x_tbl) {
	                unsigned int bus_id;

		        mutex_lock(&__i2c_mlnx_lock);
			if (p_pca954x_tbl->adap)
				p_pca954x_tbl->adap->blocked = I2C_ADAP_UNBLK;
			if (p_pca954x_tbl->client && (params->flag & RESET_FLAG_RESET_SWITCH)) {
				if (!(p_pca954x_tbl->sel_reg_addr)) {
					pca954x_block_unblock(p_pca954x_tbl->client, I2C_ADAP_UNBLK);
					pca954x_reset(p_pca954x_tbl->client);
				}
#if defined(__x86_64__)
				else {
					cpld_mux_block_unblock(p_pca954x_tbl->client, I2C_ADAP_UNBLK);
					cpld_mux_reset(p_pca954x_tbl->client);
				}
#endif
			}
		    mutex_unlock(&__i2c_mlnx_lock);
			if (p_pca954x_tbl->client && (params->flag & RESET_FLAG_RESET_SWITCH))
				printk(KERN_INFO "%s: client at bus %d addr %x has been reset\n",
					__FUNCTION__, p_pca954x_tbl->client->adapter->nr, p_pca954x_tbl->client->addr);

                        for (adap_cntr = 0; adap_cntr < LEGS_NUM; adap_cntr++) {
				u8 is_switch = 0;
				u8 addr = 0;

	                        bus_id = p_pca954x_tbl->adap_id_array[adap_cntr];
	                        if (bus_id == 0)
	                                continue;

	                        if ( (p_i2c_dev_tbl = find_i2c_device(bus_id, &is_switch, &addr)) ) {
                	                if (is_switch)
                                                continue;
                	                if (p_i2c_dev_tbl->client) {
                			        /* !!! Reset: switch has attached devices - delete and add again */
                			        struct mlnx_i2c_one_device_ioctl_params dev_params;

                        		        memset(&dev_params, 0, sizeof(struct mlnx_i2c_one_device_ioctl_params));
                        		        dev_params.parent_bus_num = p_i2c_dev_tbl->parent_bus_num;
                        		        dev_params.dev.user_handle = p_i2c_dev_tbl->user_handle;
                        		        dev_params.dev.i2c_dev_addr = p_i2c_dev_tbl->client->addr;
                        		        memcpy(&dev_params.dev.i2c_dev_type, &p_i2c_dev_tbl->i2c_dev_type,
                        		        	I2C_NAME_SIZE);
                        		        if (params->flag & RESET_FLAG_REMOVE_DEV) {
                        		                rc = mlnx_delete_one_device_from_vbus(&dev_params);
                        		                printk(KERN_INFO "%s: delete device at switch 0x%08x bus 0x%08x\n",
                                		                __FUNCTION__, params->sw[switch_cntr],
                                		                GET_BUS_ID_BY_SWITCH_ID(params->sw[switch_cntr]));
                                                }
                        		        if (params->flag & RESET_FLAG_ADD_DEV) {
                        		                rc = (rc == 0) ? mlnx_add_one_device_to_vbus(&dev_params):rc;
                        		                printk(KERN_INFO "%s: add device at switch 0x%08x bus 0x%08x\n",
                                		                __FUNCTION__, params->sw[switch_cntr],
                                		                GET_BUS_ID_BY_SWITCH_ID(params->sw[switch_cntr]));
                                                }
                                        }
                	        }
			}
		}
	}
	return rc;
}

int mlnx_add_devices_to_vbus(struct mlnx_i2c_devices_ioctl_params *params)
{
	int parent_bus_num = params->parent_bus_num;
	u8 element_num = params->element_num;
	int dev_cntr;
	int parent_switch_id = 0;
	int rc = 0;
	struct hlist_head *head;
	struct i2c_devices_table *p_i2c_dev_tbl[MAX_DEV_INREQ_NUM];
	struct pca954x_table *p_pca954x_parent = NULL;
	struct i2c_bus_table *p_bus = NULL;
	u8 go_bit_status = 0;
	u8 probe_status;
	u8 hcr_acess_status;
	int probe_cntr;

        if (element_num > MAX_DEV_INREQ_NUM)
	        return -EFAULT;
        if (parent_bus_num < 0)
	        return -EFAULT;

        params->status = 0;
        if (params->element_num == 1) {
                if ( (p_i2c_dev_tbl[0] = get_i2c_device(parent_bus_num, params->dev[0].i2c_dev_addr)) != NULL ) {
                        printk(KERN_INFO "%s device at bus %d addr %x already exist\n", __FUNCTION__,
                                parent_bus_num, params->dev[0].i2c_dev_addr);
                        params->status = 1;
                        return 0;
                }
        }

        memset(&i2c_bus_devices, 0, sizeof(struct i2c_board_info)*MAX_DEV_INREQ_NUM);
        p_bus = find_i2c_bus_entry(parent_bus_num);
        if (p_bus)
                parent_switch_id = p_bus->switch_id;
        if (parent_switch_id)
                p_pca954x_parent = find_pca954x_switch(parent_switch_id);

	head =  &i2c_ctx.i2c_devices_table_hash[device_hash_func(&parent_bus_num)];
	for (dev_cntr = 0; dev_cntr < element_num; dev_cntr++) {
		p_i2c_dev_tbl[dev_cntr] = kmem_cache_alloc(i2c_devices_table_cache, GFP_ATOMIC);
		memset(p_i2c_dev_tbl[dev_cntr], 0, sizeof(struct i2c_devices_table));
		hlist_add_head_rcu(&p_i2c_dev_tbl[dev_cntr]->hlist, head);
		i2c_bus_devices[dev_cntr].addr = params->dev[dev_cntr].i2c_dev_addr;
		memcpy(&i2c_bus_devices[dev_cntr].type, &params->dev[dev_cntr].i2c_dev_type, I2C_NAME_SIZE);
		p_i2c_dev_tbl[dev_cntr]->adap = i2c_get_adapter(parent_bus_num);
		if (p_i2c_dev_tbl[dev_cntr]->adap == NULL) {
			printk(KERN_INFO "mlnx_add_devices_to_vbus: failed to get adapter for bus %d\n",
				parent_bus_num);
			goto err_exit_adap;
		}
		p_i2c_dev_tbl[dev_cntr]->client = i2c_new_device(p_i2c_dev_tbl[dev_cntr]->adap,
						(struct i2c_board_info const*)&i2c_bus_devices[dev_cntr]);
		if (p_i2c_dev_tbl[dev_cntr]->client == NULL) {
			printk(KERN_INFO "mlnx_add_devices_to_vbus: failed to create client for bus %d\n",
				parent_bus_num);
			goto err_exit_client;
		}
		p_i2c_dev_tbl[dev_cntr]->user_handle = params->dev[dev_cntr].user_handle;
		memcpy(p_i2c_dev_tbl[dev_cntr]->i2c_dev_type, &params->dev[dev_cntr].i2c_dev_type, I2C_NAME_SIZE);
		/* Special case for SX chip */
		if (sx_i2c_get_probe_status_hook) {
		if (sx_i2c_get_probe_status_hook && (params->dev[dev_cntr].i2c_dev_addr == 0x48) &&
	   	    (!strncmp(&params->dev[dev_cntr].i2c_dev_type[0], "switchx", I2C_NAME_SIZE)))
			if (sx_i2c_get_probe_status_hook(p_i2c_dev_tbl[dev_cntr]->client,
							&probe_status, &probe_cntr,
							&go_bit_status, &hcr_acess_status)
							|| (probe_status != 0)) {
				printk(KERN_INFO "mlnx_add_one_device_to_vbus: failed on getting asic status for client on bus %d %x %s\n",
					parent_bus_num, params->dev[dev_cntr].i2c_dev_addr, &params->dev[dev_cntr].i2c_dev_type[0]);
				goto err_exit_client;
			}
		}
		p_i2c_dev_tbl[dev_cntr]->parent_bus_num = parent_bus_num;
		if (p_pca954x_parent)
			p_pca954x_parent->attached_dev_cntr++;
	}

	return rc;
err_exit_client:
	i2c_put_adapter(p_i2c_dev_tbl[dev_cntr]->adap);
err_exit_adap:
	/* Clear only failed device - don't delete all devices already added,
	   but don't continue to add*/
	hlist_del_rcu(&p_i2c_dev_tbl[dev_cntr]->hlist);
 	kmem_cache_free(i2c_devices_table_cache, p_i2c_dev_tbl[dev_cntr]);

	return -EFAULT;
}

int mlnx_delete_devices_from_vbus(struct mlnx_i2c_switch_ioctl_params *params)
{
	int switch_cntr;
	unsigned int adap_cntr;
	unsigned int bus_id;
	u8 is_switch = 0;
	u8 addr = 0;
	struct i2c_devices_table *p_i2c_dev_tbl = NULL;
	struct pca954x_table *p_pca954x_tbl[MAX_SW_INREQ_NUM];

        if (params->element_num > MAX_SW_INREQ_NUM)
	        return -EFAULT;

        memset(p_pca954x_tbl, 0, MAX_SW_INREQ_NUM*sizeof(struct pca954x_table *));
        for (switch_cntr = 0; switch_cntr < params->element_num; switch_cntr++) {
                p_pca954x_tbl[switch_cntr] = find_pca954x_switch(params->sw[switch_cntr]);
                if (!p_pca954x_tbl[switch_cntr])
                        return -EFAULT;
                for (adap_cntr = 0; adap_cntr < LEGS_NUM; adap_cntr++) {
	                bus_id = p_pca954x_tbl[switch_cntr]->adap_id_array[adap_cntr];
	                if (bus_id == 0)
	                        continue;
	                while ( (p_i2c_dev_tbl = find_i2c_device(bus_id, &is_switch, &addr)) ) {
				struct i2c_bus_table *p_bus = NULL;
                	        if (is_switch) {
                	                /* !!! Cleanup: switch has been not deleted - delete it know */
                	                struct mlnx_i2c_switch_ioctl_params sw_params;

                	                memset(&sw_params, 0, sizeof(struct mlnx_i2c_switch_ioctl_params));
                	                sw_params.element_num = 1;
                                        p_bus = find_i2c_bus_entry(bus_id);
                                        if (!p_bus)
                                                continue;
                                        sw_params.upper_switch_id = p_bus->switch_id;
	                	        sw_params.sw[0] = SET_SWITCH_ID(0, bus_id, addr);

        			        mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
	        		        	"%s(%d) delete mux upper_switch_id=%d switch_id=%d bus_id=%d\n",
	        		        	__FUNCTION__, __LINE__,
	        		        	sw_params.upper_switch_id, sw_params.sw[0], bus_id);

                	                mlnx_delete_i2c_virtual_buses(&sw_params);
                	        }
                	        if (p_i2c_dev_tbl->client) {

        			        mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
	        		        	"%s(%d) unregister client  at addr=%x on bus_id=%d\n",
	        		        	__FUNCTION__, __LINE__,
	        		        	p_i2c_dev_tbl->client->addr, p_i2c_dev_tbl->client->adapter->nr);

			                if (p_i2c_dev_tbl->adap)
				                mutex_lock(&__i2c_mlnx_lock); //i2c_lock_adapter(p_i2c_dev_tbl->adap);
				        i2c_unregister_device(p_i2c_dev_tbl->client);
			                if (p_i2c_dev_tbl->adap)
				                mutex_unlock(&__i2c_mlnx_lock); //i2c_unlock_adapter(p_i2c_dev_tbl->adap);
				        p_i2c_dev_tbl->client = NULL;
                	                p_pca954x_tbl[switch_cntr]->attached_dev_cntr--;
                	        }
                	        if (p_i2c_dev_tbl->adap) {
				        i2c_put_adapter(p_i2c_dev_tbl->adap);
				        p_i2c_dev_tbl->adap = NULL;
                	        }
			        hlist_del_rcu(&p_i2c_dev_tbl->hlist);
			        kmem_cache_free(i2c_devices_table_cache, p_i2c_dev_tbl);
			}
			//p_pca954x_tbl[switch_cntr]->adap_id_array[adap_cntr] = 0;
                }
	}
	return 0;
}

int mlnx_add_one_device_to_vbus(struct mlnx_i2c_one_device_ioctl_params *params)
{
	int parent_bus_num = params->parent_bus_num;
	int parent_switch_id = 0;
	int rc = 0;
	struct hlist_head *head;
	struct i2c_devices_table *p_i2c_dev_tbl;
	struct pca954x_table *p_pca954x_parent = NULL;
	struct i2c_bus_table *p_bus = NULL;
	u8 go_bit_status = 0;
	u8 probe_status;
	u8 hcr_acess_status;
	int probe_cntr;

        if (parent_bus_num < 0)
	        return -EFAULT;

        params->status = 0;
        if ( (p_i2c_dev_tbl = get_i2c_device(parent_bus_num, params->dev.i2c_dev_addr)) != NULL ) {
                printk(KERN_INFO "%s device at bus %d addr %x already exist\n", __FUNCTION__,
                         parent_bus_num, params->dev.i2c_dev_addr);
                params->status = 1;
                return 0;
        }

        memset(&i2c_bus_devices, 0, sizeof(struct i2c_board_info));
        p_bus = find_i2c_bus_entry(parent_bus_num);
        if (p_bus)
                parent_switch_id = p_bus->switch_id;
        if (parent_switch_id)
                p_pca954x_parent = find_pca954x_switch(parent_switch_id);

	head =  &i2c_ctx.i2c_devices_table_hash[device_hash_func(&parent_bus_num)];
	p_i2c_dev_tbl = kmem_cache_alloc(i2c_devices_table_cache, GFP_ATOMIC);
	memset(p_i2c_dev_tbl, 0, sizeof(struct i2c_devices_table));
	hlist_add_head_rcu(&p_i2c_dev_tbl->hlist, head);
	i2c_bus_devices[0].addr = params->dev.i2c_dev_addr;
	memcpy(&i2c_bus_devices[0].type, &params->dev.i2c_dev_type, I2C_NAME_SIZE);
	p_i2c_dev_tbl->adap = i2c_get_adapter(parent_bus_num);
	if (p_i2c_dev_tbl->adap == NULL) {
		printk(KERN_INFO "mlnx_add_devices_to_vbus: failed to get adapter for bus %d\n",
			parent_bus_num);
		goto err_exit_adap;
	}
	p_i2c_dev_tbl->client = i2c_new_device(p_i2c_dev_tbl->adap,
					(struct i2c_board_info const*)&i2c_bus_devices);
	if (p_i2c_dev_tbl->client == NULL) {
		printk(KERN_INFO "mlnx_add_devices_to_vbus: failed to create client for bus %d\n",
			parent_bus_num);
		goto err_exit_client;
	}
	p_i2c_dev_tbl->user_handle = params->dev.user_handle;
	memcpy(p_i2c_dev_tbl->i2c_dev_type, &params->dev.i2c_dev_type, I2C_NAME_SIZE);
	/* Special case for SX chip */
	if (sx_i2c_get_probe_status_hook && (params->dev.i2c_dev_addr == 0x48) &&
	   (!strncmp(params->dev.i2c_dev_type, "switchx", I2C_NAME_SIZE))) {
		if (sx_i2c_get_probe_status_hook(p_i2c_dev_tbl->client, &probe_status,
						&probe_cntr, &go_bit_status,
						&hcr_acess_status) || (probe_status != 0)) {
			printk(KERN_INFO "mlnx_add_one_device_to_vbus: failed on getting asic status for client on bus %d %x %s\n",
				parent_bus_num, params->dev.i2c_dev_addr, params->dev.i2c_dev_type);
			goto err_exit_client;
		}
	}
	p_i2c_dev_tbl->parent_bus_num = parent_bus_num;
	if (p_pca954x_parent)
		p_pca954x_parent->attached_dev_cntr++;

	return rc;
err_exit_client:
	i2c_put_adapter(p_i2c_dev_tbl->adap);
err_exit_adap:
	/* Clear only failed device - don't delete all devices already added,
	   but don't continue to add*/
	hlist_del_rcu(&p_i2c_dev_tbl->hlist);
 	kmem_cache_free(i2c_devices_table_cache, p_i2c_dev_tbl);

	return -EFAULT;
}

int mlnx_delete_one_device_from_vbus(struct mlnx_i2c_one_device_ioctl_params *params)
{
	int parent_bus_num = params->parent_bus_num;
	u8 addr = params->dev.i2c_dev_addr;
	u8 is_switch = 0;
	struct pca954x_table *p_pca954x_tbl = NULL;
	struct i2c_bus_table *p_bus;
	struct i2c_devices_table *p_i2c_dev_tbl;

        if ( (p_i2c_dev_tbl = find_i2c_device_by_addr(parent_bus_num, addr, &is_switch)) == NULL ) {
                printk(KERN_INFO "%s device already removed at bus %d addr %x\n", __FUNCTION__,
                        parent_bus_num, addr);
                return -1;
        }
        if ( (p_bus = find_i2c_bus_entry(parent_bus_num)) == NULL ) {
                printk(KERN_INFO "%s invalid request bus %d addr %x\n", __FUNCTION__,
                        parent_bus_num, addr);
                //return -1; // don't exit to support tools
        }
	if (p_bus) {
		p_pca954x_tbl = find_pca954x_switch(p_bus->switch_id);
        	if (is_switch) {
                	/* !!! Cleanup: switch is reqeusted to be deleted */
                	struct mlnx_i2c_switch_ioctl_params sw_params;

                	memset(&sw_params, 0, sizeof(struct mlnx_i2c_switch_ioctl_params));
                	sw_params.element_num = 1;
                	sw_params.upper_switch_id = p_bus->switch_id;
                	sw_params.sw[0] = SET_SWITCH_ID(0, parent_bus_num, addr);
                	mlnx_delete_i2c_virtual_buses(&sw_params);
        	}
        }
        if (p_i2c_dev_tbl->client) {
		if (p_i2c_dev_tbl->adap)
			mutex_lock(&__i2c_mlnx_lock); //i2c_lock_adapter(p_i2c_dev_tbl->adap);
		i2c_unregister_device(p_i2c_dev_tbl->client);
		if (p_i2c_dev_tbl->adap)
			mutex_unlock(&__i2c_mlnx_lock); //i2c_unlock_adapter(p_i2c_dev_tbl->adap);
		p_i2c_dev_tbl->client = NULL;
		if(p_pca954x_tbl)
                        p_pca954x_tbl->attached_dev_cntr--;
        }
        if (p_i2c_dev_tbl->adap) {
		i2c_put_adapter(p_i2c_dev_tbl->adap);
		p_i2c_dev_tbl->adap = NULL;
        }
	hlist_del_rcu(&p_i2c_dev_tbl->hlist);
	kmem_cache_free(i2c_devices_table_cache, p_i2c_dev_tbl);

	return 0;
}

int mlnx_add_gpio_i2c_buses(struct mlnx_gpio_i2c_ioctl_params *params)
{
#ifdef GPIO_BUS
        struct i2c_gpio_params *mux_params;
        struct i2c_gpio_table *p_gpio_tbl[MAX_I2C_GPIO_INREQ_NUM];
        struct i2c_bus_table *p_bus;
	struct hlist_head *head;
        int mux_cntr;
        int mux_cntre;
        int cntr = 0;
        int cntre;
        int rc;

        if (params->element_num > MAX_I2C_GPIO_INREQ_NUM)
	        return -EFAULT;

        for (mux_cntr = 0; mux_cntr < params->element_num; mux_cntr++) {
		mux_params = &params->exp[mux_cntr];
		p_bus = find_i2c_bus_entry(mux_params->parent_bus_num);
		if (!p_bus)
                        goto err_exit;
		head =  &i2c_ctx.i2c_gpio_table_hash[bus_hash_func(&mux_params->parent_bus_num)];
		p_gpio_tbl[mux_cntr] = kmem_cache_alloc(i2c_gpio_table_cache, GFP_ATOMIC);
		memset(p_gpio_tbl[mux_cntr], 0, sizeof(struct i2c_gpio_table));
		INIT_HLIST_NODE(&p_gpio_tbl[mux_cntr]->hlist);
		p_gpio_tbl[mux_cntr]->mux_data.parent = mux_params->parent_bus_num;
		p_gpio_tbl[mux_cntr]->mux_data.base_nr = mux_params->base_bus_num;
                p_gpio_tbl[mux_cntr]->mux_data.n_values = mux_params->buses_num;
                p_gpio_tbl[mux_cntr]->mux_data.n_gpios = mux_params->gpios_num;
		p_gpio_tbl[mux_cntr]->mux_data.values = p_gpio_tbl[mux_cntr]->mux_values;
		p_gpio_tbl[mux_cntr]->mux_data.gpios = p_gpio_tbl[mux_cntr]->mux_gpios;
                for (cntr = 0; cntr <  mux_params->buses_num; cntr++)
                        p_gpio_tbl[mux_cntr]->mux_values[cntr] = mux_params->buses[cntr];
                for (cntr = 0; cntr <  mux_params->gpios_num; cntr++)
                        p_gpio_tbl[mux_cntr]->mux_gpios[cntr] = mux_params->gpios[cntr];
                cntr = 0;
                p_gpio_tbl[mux_cntr]->mux_data.idle = mux_params->idle;
		memcpy(p_gpio_tbl[mux_cntr]->mux_name, mux_params->i2c_dev_type, I2C_NAME_SIZE);
		hlist_add_head_rcu(&p_gpio_tbl[mux_cntr]->hlist, head);
		list_add_rcu(&p_gpio_tbl[mux_cntr]->list, &p_bus->attached_gpio_bus_list);
		p_gpio_tbl[mux_cntr]->plat_dev = platform_device_alloc(p_gpio_tbl[mux_cntr]->mux_name,
									mux_params->chip_id);
		if (!p_gpio_tbl[mux_cntr]->plat_dev) {
			rc = -ENOMEM;
			printk(KERN_INFO "%s: Device allocation failed\n", __FUNCTION__);
			goto err_exit;
		}
		rc = platform_device_add_data(p_gpio_tbl[mux_cntr]->plat_dev,
				              &p_gpio_tbl[mux_cntr]->mux_data,
				              sizeof(p_gpio_tbl[mux_cntr]->mux_data));
		if (rc) {
			printk(KERN_INFO "%s: Platform data allocation failed (%d)\n", __FUNCTION__, rc);
			platform_device_put(p_gpio_tbl[mux_cntr]->plat_dev);
			goto err_exit;
		}
		rc = platform_device_add(p_gpio_tbl[mux_cntr]->plat_dev);
		if (rc) {
			printk(KERN_INFO "%s Device addition failed (%d)\n",  __FUNCTION__, rc);
			platform_device_put(p_gpio_tbl[mux_cntr]->plat_dev);
                        goto err_exit;
		}

		/* Export gpio - this will result in the pseudo-files for pin n
                   showing up under /sys/class/gpio/gpio<n> as:
                   /sys/class/gpio/gpio<n>/direction
                   /sys/class/gpio/gpio<n>/value
                   For this pin should be set direction to either 'in' or 'out'.
                   And then pin could be set and get through GPIO set/get interface.
		*/
                for (cntr = 0; cntr < mux_params->gpios_num; cntr++) {
                        unsigned gpio = mux_params->gpios[cntr];
                        p_gpio_tbl[mux_cntr]->mux_gpios[cntr] = gpio;
			rc = gpio_export(gpio, true);
			if (rc < 0) {
				gpio_free(gpio);
                        	goto err_exit;
			}
                        if (mux_params->gpios_dir[cntr] == GPIO_HIGH)
                                rc = gpio_direction_output(gpio, 1);
                        else if ( (mux_params->gpios_dir[cntr] == GPIO_OUT) ||
                                  (mux_params->gpios_dir[cntr] == GPIO_LOW) )
                                rc = gpio_direction_output(gpio, 0);
                        else
                                rc = gpio_direction_input(gpio);
                }
                p_gpio_tbl[mux_cntr]->gpio_num = mux_params->gpios_num;
        }
        return 0;

err_exit:
        /* In case of error clear all elements */
        for (cntre = 0; cntre < cntr; cntre++) {
                 unsigned gpio = mux_params->gpios[cntr];
                 gpio_unexport(gpio);
                 gpio_free(gpio);
        }
        for (mux_cntre = 0; mux_cntre < mux_cntr; mux_cntre++) {
                platform_device_del(p_gpio_tbl[mux_cntre]->plat_dev);
		platform_device_put(p_gpio_tbl[mux_cntre]->plat_dev);
                hlist_del_rcu(&p_gpio_tbl[mux_cntre]->hlist);
		list_del_rcu(&p_gpio_tbl[mux_cntre]->list);
                kmem_cache_free(i2c_gpio_table_cache, p_gpio_tbl[mux_cntre]);
	}

	return -EFAULT;
#endif /* GPIO_BUS */
        return 0;
}

int mlnx_del_gpio_i2c_buses(struct mlnx_gpio_i2c_del_ioctl_params *params)
{
#ifdef GPIO_BUS
        struct i2c_gpio_table *p_gpio_tbl;
        int mux_cntr;
        int cntr;

        if (params->element_num > MAX_I2C_GPIO_INREQ_NUM)
	        return -EFAULT;

        for (mux_cntr = 0; mux_cntr < params->element_num; mux_cntr++) {
                p_gpio_tbl = find_i2c_gpio_entry(params->parent_bus_num[mux_cntr]);
                if (!p_gpio_tbl) {
			printk(KERN_INFO "%s: failed to find gpio bus %d - bad input\n",
			        __FUNCTION__, params->parent_bus_num[mux_cntr]);
			return -EFAULT;
                }
                for (cntr = 0; cntr < p_gpio_tbl->gpio_num; cntr++) {
                        unsigned gpio = p_gpio_tbl->mux_gpios[cntr];
                        gpio_unexport(gpio);
                        gpio_free(gpio);
                }
                platform_device_del(p_gpio_tbl->plat_dev);
		platform_device_put(p_gpio_tbl->plat_dev);
                hlist_del_rcu(&p_gpio_tbl->hlist);
		list_del_rcu(&p_gpio_tbl->list);
                kmem_cache_free(i2c_gpio_table_cache, p_gpio_tbl);
	}
        return 0;
#endif /* GPIO_BUS */
        return 0;
}

int mlnx_rw_i2c_device(struct mlnx_rdwr_i2c_ioctl_params *params)
{
	struct i2c_adapter *adap;
	int rc = 0;
	int ind = 0;
	struct i2c_msg msg[I2C_MAX_RW_MSGS];
	unsigned char msgbuf[I2C_MAX_RW_MSGS][I2C_BLOCK_MAX];
	struct mlnx_rdwr_i2c_ioctl_params rdwr_arg;

	if (copy_from_user(&rdwr_arg, (struct mlnx_rdwr_i2c_ioctl_params __user *)params, sizeof(rdwr_arg)))
		return -EFAULT;

	/* Put an arbitrary limit on the number of messages that can
	   be sent at once */
	if (rdwr_arg.nmsgs > I2C_MAX_RW_MSGS)
		return -EINVAL;

	for (ind = 0; ind < rdwr_arg.nmsgs; ind++) {
		memcpy(&msg[ind], &rdwr_arg.msgs[ind], sizeof(struct i2c_msg));
		msg[ind].buf = msgbuf[ind];
		if (copy_from_user(msg[ind].buf, rdwr_arg.msgs[ind].buf, rdwr_arg.msgs[ind].len)) {
			rc = -EFAULT;
			break;
		}
	}

	mutex_lock(&__i2c_mlnx_lock);
	adap = i2c_get_adapter(rdwr_arg.bus);
	if (!adap) {
		mutex_unlock(&__i2c_mlnx_lock);
		rc = -EINVAL;
		goto exit;
	}

#ifdef SX_CACHE_OPTIMIZED_2
	mlnx_i2c_save_nailed_up_route(adap);
#endif

	mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
		"bus=%d addr=%x len=%d read_write=%x nmsgs=%d last bus=%d\n",
		params->bus, params->msgs[0].addr, params->msgs[0].len,
		params->msgs[0].flags, params->nmsgs, __last_accessed_bus_id);

	rc = i2c_transfer(adap, msg, rdwr_arg.nmsgs);
	mutex_unlock(&__i2c_mlnx_lock);

	for (ind = 0; ind < rdwr_arg.nmsgs; ind++) {
		if (rc >= 0 && (rdwr_arg.msgs[ind].flags & I2C_M_RD)) {
			if (copy_to_user(rdwr_arg.msgs[ind].buf, msg[ind].buf, rdwr_arg.msgs[ind].len)) {
				rc = -EFAULT;
				break;
			}
		}
	}
	if (copy_to_user(&((struct mlnx_rdwr_i2c_ioctl_params __user *)params)->rc, &rc, sizeof(rc))) {
		if (adap)
	    	i2c_put_adapter(adap);
	    return -EFAULT;
	}
exit:
    if (adap)
        i2c_put_adapter(adap);
	return rc;
}
EXPORT_SYMBOL(mlnx_rw_i2c_device);

int mlnx_rw_i2c_kernel(struct mlnx_rdwr_i2c_ioctl_params *params)
{
	struct i2c_adapter *adap;
	int rc = 0;

	if (params->nmsgs > I2C_MAX_RW_MSGS)
		return -EINVAL;

	mutex_lock(&__i2c_mlnx_lock);
	adap = i2c_get_adapter(params->bus);
	if (!adap) {
		mutex_unlock(&__i2c_mlnx_lock);
		rc = -EINVAL;
		goto exit;
	}
#ifdef SX_CACHE_OPTIMIZED_2
	mlnx_i2c_save_nailed_up_route(adap);
#endif

	mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
		"bus=%d addr=%x len=%d read_write=%x nmsgs=%d last bus=%d\n",
		params->bus, params->msgs[0].addr, params->msgs[0].len,
		params->msgs[0].flags, params->nmsgs, __last_accessed_bus_id);

	rc = i2c_transfer(adap, params->msgs, params->nmsgs);
	mutex_unlock(&__i2c_mlnx_lock);

	params->rc = rc;
exit:
    if (rc > 0)
        rc = 0;

    if (adap)
        i2c_put_adapter(adap);

	return rc;
}
EXPORT_SYMBOL(mlnx_rw_i2c_kernel);

int mlnx_rw_smbus_device(struct mlnx_rdwr_smbus_ioctl_params *params)
{
	struct i2c_adapter *adap = NULL;
	struct mlnx_rdwr_smbus_ioctl_params rdwr_arg;
	union i2c_smbus_data temp;
	int datasize;
	int rc = 0;

	if (copy_from_user(&rdwr_arg, (struct mlnx_rdwr_smbus_ioctl_params __user *)params, sizeof(rdwr_arg)))
		return -EFAULT;

	adap = i2c_get_adapter(rdwr_arg.bus);
	if (!adap)
		return -EINVAL;

	/* Put an arbitrary limit on the number of messages that can
	   be sent at once */
        /* ... */
	if ((rdwr_arg.size != I2C_SMBUS_BYTE) &&
	    (rdwr_arg.size != I2C_SMBUS_QUICK) &&
	    (rdwr_arg.size != I2C_SMBUS_BYTE_DATA) &&
	    (rdwr_arg.size != I2C_SMBUS_WORD_DATA) &&
	    (rdwr_arg.size != I2C_SMBUS_PROC_CALL) &&
	    (rdwr_arg.size != I2C_SMBUS_BLOCK_DATA) &&
	    (rdwr_arg.size != I2C_SMBUS_I2C_BLOCK_BROKEN) &&
	    (rdwr_arg.size != I2C_SMBUS_I2C_BLOCK_DATA) &&
	    (rdwr_arg.size != I2C_SMBUS_BLOCK_PROC_CALL)) {
		printk(KERN_INFO "bus %d:size out of range (%x) in ioctl I2C_SMBUS.\n",
			rdwr_arg.bus, rdwr_arg.size);
		i2c_put_adapter(adap);
		return -EINVAL;
	}
	/* Note that I2C_SMBUS_READ and I2C_SMBUS_WRITE are 0 and 1,
	   so the check is valid if size==I2C_SMBUS_QUICK too. */
	if ((rdwr_arg.read_write != I2C_SMBUS_READ) &&
	    (rdwr_arg.read_write != I2C_SMBUS_WRITE)) {
		printk(KERN_INFO "bus %d: read_write out of range (%x) in ioctl I2C_SMBUS.\n",
			rdwr_arg.bus, rdwr_arg.read_write);
		i2c_put_adapter(adap);
		return -EINVAL;
	}
	/* Note that command values are always valid! */
	mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
		"bus %d (%d): addr %x flags %x read_write %x command %x size %x block[0] %x %x %x %x %x %x %x %x %x %x %x %x %x\n",
		rdwr_arg.bus, __LINE__, rdwr_arg.addr, rdwr_arg.flags, rdwr_arg.read_write,
		rdwr_arg.command, rdwr_arg.size, rdwr_arg.data.block[0],
		temp.block[1], temp.block[2], temp.block[3], temp.block[4],
		temp.block[5], temp.block[6], temp.block[7], temp.block[8],
		temp.block[9], temp.block[10], temp.block[11], temp.block[12]);
	if ((rdwr_arg.size == I2C_SMBUS_QUICK) ||
	    ((rdwr_arg.size == I2C_SMBUS_BYTE) &&
	    (rdwr_arg.read_write == I2C_SMBUS_WRITE))) { //??????
		/* These are special: we do not use data */
		mutex_lock(&__i2c_mlnx_lock);
#ifdef SX_CACHE_OPTIMIZED_2
		mlnx_i2c_save_nailed_up_route(adap);
#endif
		mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
			"bus %d (%d): addr %x flags %x read_write %x command %x size %x block[0] %x\n",
			rdwr_arg.bus, __LINE__, rdwr_arg.addr, rdwr_arg.flags, rdwr_arg.read_write,
			rdwr_arg.command, rdwr_arg.size, rdwr_arg.data.block[0]);
		if (rdwr_arg.read_write == I2C_SMBUS_WRITE) {
			rc = i2c_smbus_xfer(adap, rdwr_arg.addr, rdwr_arg.flags,
		   				rdwr_arg.read_write, rdwr_arg.command,
		   				rdwr_arg.size, &temp);
		}
		else {
			rc = i2c_smbus_xfer(adap, rdwr_arg.addr, rdwr_arg.flags,
		   				rdwr_arg.read_write, 0,
		   				rdwr_arg.size, &temp);
			rdwr_arg.data.byte = temp.byte;
		}
		mutex_unlock(&__i2c_mlnx_lock);

		i2c_put_adapter(adap);
		return rc;
	}

	memset(&temp, 0, sizeof(union i2c_smbus_data));
	if ((rdwr_arg.size == I2C_SMBUS_BYTE_DATA) || (rdwr_arg.size == I2C_SMBUS_BYTE)) {
		datasize = sizeof(rdwr_arg.data.byte);
		temp.byte = rdwr_arg.data.byte;
	}
	else if ((rdwr_arg.size == I2C_SMBUS_WORD_DATA) || (rdwr_arg.size == I2C_SMBUS_PROC_CALL)) {
		datasize = sizeof(rdwr_arg.data.word);
		if (rdwr_arg.addr_width == 2)
			memcpy(&temp.block[0], &rdwr_arg.data.block[0], datasize + 1);
		temp.word = rdwr_arg.data.word;
	}
	else { /* size == smbus block, i2c block, or block proc. call */
		datasize = rdwr_arg.data.block[0];
		memcpy(&temp.block[0], &rdwr_arg.data.block[0], datasize + 1);
	}
	memcpy(&temp, &rdwr_arg.data.block[0], datasize);
	if ((rdwr_arg.size == I2C_SMBUS_PROC_CALL) ||
	    (rdwr_arg.size == I2C_SMBUS_BLOCK_PROC_CALL) ||
	    (rdwr_arg.size == I2C_SMBUS_I2C_BLOCK_DATA) ||
	    (rdwr_arg.size == I2C_SMBUS_BLOCK_DATA) ||
	    (rdwr_arg.read_write == I2C_SMBUS_WRITE)) {
		mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
			"bus %d (%d): addr %x flags %x read_write %x command %x size %x block[0] %x %x %x %x %x %x %x %x %x %x %x %x %x\n",
			rdwr_arg.bus, __LINE__, rdwr_arg.addr, rdwr_arg.flags, rdwr_arg.read_write,
			rdwr_arg.command, rdwr_arg.size, rdwr_arg.data.block[0],
			temp.block[1], temp.block[2], temp.block[3], temp.block[4],
			temp.block[5], temp.block[6], temp.block[7], temp.block[8],
			temp.block[9], temp.block[10], temp.block[11], temp.block[12]);

	}
	if (rdwr_arg.size == I2C_SMBUS_I2C_BLOCK_BROKEN) {
		/* Convert old I2C block commands to the new
		   convention. This preserves binary compatibility. */
		rdwr_arg.size = I2C_SMBUS_I2C_BLOCK_DATA;
		if (rdwr_arg.read_write == I2C_SMBUS_READ)
			temp.block[0] = I2C_SMBUS_BLOCK_MAX;
	}
	mutex_lock(&__i2c_mlnx_lock);
	adap = i2c_get_adapter(rdwr_arg.bus);
#ifdef SX_CACHE_OPTIMIZED_2
	mlnx_i2c_save_nailed_up_route(adap);
#endif
	rc = i2c_smbus_xfer(adap, rdwr_arg.addr, rdwr_arg.flags,
	      rdwr_arg.read_write, rdwr_arg.command, rdwr_arg.size, &temp);
	mutex_unlock(&__i2c_mlnx_lock);
	mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
		"bus %d (%d): rc=%d addr %x flags %x read_write %x command %x size %x temp->block[0] %x %x %x %x %x %x %x %x %x %x %x %x %x\n",
		rdwr_arg.bus, __LINE__, rc, rdwr_arg.addr, rdwr_arg.flags, rdwr_arg.read_write,
		rdwr_arg.command, rdwr_arg.size,
		temp.block[0], temp.block[1], temp.block[2], temp.block[3],
		temp.block[4], temp.block[5], temp.block[6], temp.block[7], temp.block[8],
		temp.block[9], temp.block[10], temp.block[11], temp.block[12]);
	if (!rc && ((rdwr_arg.size == I2C_SMBUS_PROC_CALL) ||
		     (rdwr_arg.size == I2C_SMBUS_BLOCK_PROC_CALL) ||
		     (rdwr_arg.read_write == I2C_SMBUS_READ))) {
		memcpy(&rdwr_arg.data, &temp, datasize);
		mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
			"bus %d (%d): addr %x datasize=%d temp->block[0] %x %x %x %x %x %x %x %x\n",
			rdwr_arg.bus, __LINE__, rdwr_arg.addr, datasize, temp.block[0], temp.block[1],
			temp.block[2], temp.block[3], temp.block[4],
			temp.block[5], temp.block[6], temp.block[7]);
	}
        rdwr_arg.rc = rc;
	if (copy_to_user((struct mlnx_rdwr_smbus_ioctl_params __user *)params, &rdwr_arg, sizeof(rdwr_arg)))
		rc = -EFAULT;
	if (adap)
	    i2c_put_adapter(adap);
	return rc;
}
EXPORT_SYMBOL(mlnx_rw_smbus_device);

#define SMBUS_XFER(adap, rw_smbus_buf, read_write, command, size)        \
	try = 0;                                                         \
	rc = 0;                                                          \
	DBG("%s(%d) try=%d rc=%d\n", __FUNCTION__, __LINE__, try, rc);   \
	orig_jiffies = jiffies;                                          \
	for (try = 0; try <= 3; try++) {                                 \
		rc = i2c_smbus_xfer(adap, addr, flags, read_write,       \
		   			command, size, &rw_smbus_buf);   \
		if (rc >= 0) {		                                 \
			break;                                           \
		}                                                        \
		if (time_after(jiffies, orig_jiffies + TRANSACT_TO)) {   \
			rc = -EIO;                                       \
			break;                                           \
		}                                                        \
		cond_resched();                                          \
	}                                                                \
	DBG("%s(%d) try=%d rc=%d\n", __FUNCTION__, __LINE__, try, rc);   \
	if (rc < 0)                                                      \
		return rc;

static inline int smbus_xfer_chunk(struct i2c_adapter *adap, u16 addr, u32 offset, u32 datalen,
				    u16 flags, u8 rw_oper, int addr_width, u8 *in_out_buf)
{
	//struct i2c_client *client;
	union i2c_smbus_data rw_smbus_buf;
	char read_write;
	u8 command = (u8)offset;
	int size;
	u8 *p_buf = in_out_buf;
	u32 len = datalen;
	unsigned long orig_jiffies;
	u8 smbus_block_max = I2C_SMBUS_BLOCK_MAX;
	int chunk_size = I2C_SMBUS_BLOCK_MAX;
	int try = 0;
        int rc = 0;

        DBG("%s(%d) bus=%d addr=0x%x rw_oper=%d offset=0x%x datalen=%d len=%d addr_width=%d\n",
	        __FUNCTION__, __LINE__, adap->nr, addr, rw_oper, offset, datalen, len, addr_width);
	memset(&rw_smbus_buf, 0, sizeof(union i2c_smbus_data));
	switch (addr_width) {
		case 4:
			/* Write first buffer with addres and data */
			size = I2C_SMBUS_I2C_BLOCK_DATA;
			command = (offset >> 24) & 0x00ff;
			/* Stole three data bytes for address pointing */
			rw_smbus_buf.block[1] = (offset >> 16) & 0x00ff;
			rw_smbus_buf.block[2] = (offset >> 8) & 0x00ff;
			rw_smbus_buf.block[3] = offset & 0x00ff;
			read_write = I2C_SMBUS_WRITE;
			rw_smbus_buf.block[0] = 3;
			if (rw_oper == I2C_SMBUS_WRITE) {
				chunk_size = (len > smbus_block_max - 3) ?
						(smbus_block_max - 3) : len;
				rw_smbus_buf.block[0] += chunk_size;
				SMBUS_XFER(adap, rw_smbus_buf, read_write, command, size);
				//rc = i2c_smbus_write_i2c_block_data(client, command, rw_smbus_buf.block[0],
				//					&rw_smbus_buf.block[1]);
				memcpy(p_buf, &rw_smbus_buf.block[4], chunk_size);
				p_buf += chunk_size;
				datalen -= chunk_size;
			}
			else {
                                /* Pointing internal address for read */
				SMBUS_XFER(adap, rw_smbus_buf, read_write, command, size);
				//rc = i2c_smbus_write_i2c_block_data(client, command, rw_smbus_buf.block[0],
				//					&rw_smbus_buf.block[1]);
			}
			break;
		case 2:
			if (rw_oper == I2C_SMBUS_WRITE) {
				switch (datalen) {
					/* Stole one data byte for address pointing */
					case 1:
						rw_smbus_buf.byte = offset & 0x00ff;
						size = I2C_SMBUS_WORD_DATA;
						break;
					case 2:
						size = I2C_SMBUS_I2C_BLOCK_DATA;
						rw_smbus_buf.block[1] = offset & 0x00ff;
						rw_smbus_buf.block[0] = 3;
						chunk_size = 2;
						memcpy(&rw_smbus_buf.block[2], p_buf, chunk_size);
						break;
					default:
						size = I2C_SMBUS_I2C_BLOCK_DATA;
						rw_smbus_buf.block[1] = offset & 0x00ff;
						chunk_size = (len > smbus_block_max - 1) ?
								(smbus_block_max - 1) : len;
						rw_smbus_buf.block[0] = 1 + chunk_size;
						memcpy(&rw_smbus_buf.block[2], p_buf, chunk_size);
						break;
				}
				command = (offset >> 8) & 0x00ff;
				DBG("%s(%d) bus=%d addr=0x%x offset=0x%x datalen=%d len=%d chunk_size=%d command=%x data: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
		   	    		__FUNCTION__, __LINE__, adap->nr, addr, offset, datalen,
                                	len, chunk_size, command, rw_smbus_buf.block[0], rw_smbus_buf.block[1],
                                	rw_smbus_buf.block[2], rw_smbus_buf.block[3],
                                	rw_smbus_buf.block[4], rw_smbus_buf.block[5],
                                	rw_smbus_buf.block[6], rw_smbus_buf.block[7]);
				SMBUS_XFER(adap, rw_smbus_buf, I2C_SMBUS_WRITE, command, size);
				//rc = i2c_smbus_write_byte_data(client, command, rw_smbus_buf.byte);
				p_buf += chunk_size;
				datalen -= chunk_size;
			}
			else {
				size = I2C_SMBUS_BYTE_DATA;
				command = (offset >> 8) & 0x00ff;
				/* Stole one data byte for address pointing */
				rw_smbus_buf.byte = offset & 0x00ff;
				read_write = I2C_SMBUS_WRITE;
				DBG("%s(%d) point addr bus=%d addr=0x%x offset=0x%x datalen=%d len=%d rc=%d command=%x byte 0x%x word 0x%x block 0x%x 0x%x\n",
		   	    		__FUNCTION__, __LINE__, adap->nr, addr, offset, datalen,
                                	len, rc, command, rw_smbus_buf.byte, rw_smbus_buf.word,
                                	rw_smbus_buf.block[0], rw_smbus_buf.block[1]);
				SMBUS_XFER(adap, rw_smbus_buf, read_write, command, size);
				//rc = i2c_smbus_write_byte_data(client, command, rw_smbus_buf.byte);
			}
			break;
		case 1:
			break;
	        case 0: /* Offset not specified */
			size = I2C_SMBUS_QUICK;
			command = 0;
			SMBUS_XFER(adap, rw_smbus_buf, rw_oper, command, size);
			return rc;
		default:
			return -EFAULT;
	}
	switch (addr_width) {
		case 4:
			break;
		case 2:
			/* For device with address width 2 (EEPROM), read byte by byte only */
			if (rw_oper == I2C_SMBUS_READ) {
				memset(&rw_smbus_buf, 0, sizeof(union i2c_smbus_data));
				read_write = rw_oper;
				size = I2C_SMBUS_BYTE;
				len = datalen;
				while (len > 0) {
					//command = (offset >> 8) & 0x00ff;
					//rw_smbus_buf.byte = offset & 0x00ff;
					DBG("%s(%d) bus=%d addr=0x%x offset=0x%x datalen=%d len=%d command=%x byte=%x\n",
		    				__FUNCTION__, __LINE__, adap->nr, addr,
		    				offset, datalen, len, command, rw_smbus_buf.byte);
					//SMBUS_XFER(adap, rw_smbus_buf, read_write, command, size);
					//memcpy(p_buf, &rw_smbus_buf.block[0], 1);
					if ((rc = i2c_smbus_xfer(adap, addr, 0, I2C_SMBUS_READ, 0,
								I2C_SMBUS_BYTE, &rw_smbus_buf)) < 0)
						return rc;
					*p_buf = rw_smbus_buf.byte;
					len -= 1;
					offset += 1;
					p_buf += 1;
					datalen = len;
					DBG("%s(%d) bus=%d addr=0x%x offset=0x%x datalen=%d len=%d rc=%d command=%x byte 0x%x word 0x%x block 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
		   	    			__FUNCTION__, __LINE__, adap->nr, addr, offset, datalen,
                                		len, rc, command, rw_smbus_buf.byte,
                                		rw_smbus_buf.word, rw_smbus_buf.block[0],
		   	    			rw_smbus_buf.block[1], rw_smbus_buf.block[2],
		   	    			rw_smbus_buf.block[3], rw_smbus_buf.block[4],
                                		rw_smbus_buf.block[5], rw_smbus_buf.block[6]);
				} /*while*/
				return rc;
			} /*if (rw_oper == I2C_SMBUS_READ)*/
			break;
		case 1:
			break;
		case 0:
			break;
		default:
			break;
	}
	switch (datalen) {
		case 0:
			return rc;
		case 1:
		case 2:
			read_write = rw_oper;
			command = offset;
			if (datalen == 1)
				size = I2C_SMBUS_BYTE_DATA;
			else
				size = I2C_SMBUS_WORD_DATA;
			if (rw_oper == I2C_SMBUS_WRITE) {
				memcpy(&rw_smbus_buf.block[0], p_buf, datalen);
				//rc = i2c_smbus_write_byte_data(client, command, &rw_smbus_buf.byte);
				//rc = i2c_smbus_write_word_data(client, command, &rw_smbus_buf.word);
			}
			DBG("%s(%d) bus=%d addr=0x%x offset=0x%x datalen=%d *p_buf=%x len=%d rc=%d command=%x read_write=%x byte 0x%x word 0x%x block 0x%x 0x%x\n",
		   	    	__FUNCTION__, __LINE__, adap->nr, addr, offset, datalen, *p_buf,
                                len, rc, command, read_write, rw_smbus_buf.byte, rw_smbus_buf.word,
                                rw_smbus_buf.block[0], rw_smbus_buf.block[1]);
			SMBUS_XFER(adap, rw_smbus_buf, read_write, command, size);
			DBG("%s(%d) bus=%d addr=0x%x offset=0x%x datalen=%d len=%d rc=%d command=%x read_write=%x byte 0x%x word 0x%x block 0x%x 0x%x\n",
		   	    	__FUNCTION__, __LINE__, adap->nr, addr, offset, datalen,
                                len, rc, command, read_write, rw_smbus_buf.byte, rw_smbus_buf.word,
                                rw_smbus_buf.block[0], rw_smbus_buf.block[1]);
			if (rw_oper == I2C_SMBUS_READ) {
				//rc = p_buf = i2c_smbus_read_byte_data(client, command);
				//rc = p_buf = i2c_smbus_read_word_data(client, command);
				memcpy(p_buf, &rw_smbus_buf.block[0], datalen);
			}
			break;
		default:
			len = datalen;
			while (len > 0) {
				read_write = rw_oper;
				command = offset;
				if (len > smbus_block_max)
					datalen = smbus_block_max;
				else
					datalen = len;
				if (datalen == 2) {
					size = I2C_SMBUS_WORD_DATA;
				} else if (datalen == 1) {
					size = I2C_SMBUS_BYTE_DATA;
				} else {
					rw_smbus_buf.block[0] = smbus_block_max;
					size = I2C_SMBUS_I2C_BLOCK_DATA;
				}
				DBG("%s(%d) bus=%d addr=0x%x offset=0x%x datalen=%d len=%d\n",
		    			__FUNCTION__, __LINE__, adap->nr, addr,
		    			offset, datalen, len);
				//rc = i2c_smbus_write_byte_data(client, command, &rw_smbus_buf.byte);
				//rc = i2c_smbus_write_word_data(client, command, &rw_smbus_buf.word);
				//rc = i2c_smbus_write_i2c_block_data(client, command, rw_smbus_buf.block[0],
				//					&rw_smbus_buf.block[1]);
				if (rw_oper == I2C_SMBUS_WRITE) {
					if (len <= 2) {
						memcpy(p_buf, &rw_smbus_buf.block[0], datalen);
					} else {
						memcpy(p_buf, &rw_smbus_buf.block[1], datalen);
					}
				}
				SMBUS_XFER(adap, rw_smbus_buf, read_write, command, size);
				//rc = p_buf = i2c_smbus_read_byte_data(client, command);
				//rc = p_buf = i2c_smbus_read_word_data(client, command);
				//rc = i2c_smbus_read_i2c_block_data(client, command, datalen, p_buf);
				if (rw_oper == I2C_SMBUS_READ) {
					if (len <= 2) {
						memcpy(p_buf, &rw_smbus_buf.block[0], datalen);
					} else {
						memcpy(p_buf, &rw_smbus_buf.block[1], datalen);
					}
				}
				len -= datalen;
				offset += datalen;
				p_buf += datalen;
				datalen = len;
				DBG("%s(%d) bus=%d addr=0x%x offset=0x%x datalen=%d len=%d rc=%d command=%x byte 0x%x word 0x%x block 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
		   	    		__FUNCTION__, __LINE__, adap->nr, addr, offset, datalen,
                                	len, rc, command, rw_smbus_buf.byte,
                                	rw_smbus_buf.word, rw_smbus_buf.block[0],
		   	    		rw_smbus_buf.block[1], rw_smbus_buf.block[2],
		   	    		rw_smbus_buf.block[3], rw_smbus_buf.block[4],
                                	rw_smbus_buf.block[5], rw_smbus_buf.block[6]);
			} /*while*/
		break; /*default*/
	} /* switch(datalen)*/
	return rc;
}

int mlnx_rw_smbus(struct mlnx_rdwr_smbus_ioctl_params_2 *params)
{
	struct i2c_adapter *adap = NULL;
	u8 read_write;
	int rc;

        DBG("%s(%d) bus=%d addr=0x%x rw_oper=%d offset=0x%x datalen=%d\n", __FUNCTION__, __LINE__,
                params->bus, params->addr, params->read_write, params->offset, params->datalen);
        adap = i2c_get_adapter(params->bus);
        if (!adap)
                return -EFAULT;
        read_write = (params->read_write != 0) ? I2C_SMBUS_READ : I2C_SMBUS_WRITE;
        DBG("%s(%d) bus=%d addr=0x%x rw_oper=%d offset=0x%x datalen=%d\n", __FUNCTION__, __LINE__,
                adap->nr, params->addr, read_write, params->offset, params->datalen);

        rc = smbus_xfer_chunk(adap, params->addr, params->offset, params->datalen, 0,
				read_write, params->addr_width, params->data);
        if (adap)
            i2c_put_adapter(adap);
        return rc;
}
EXPORT_SYMBOL(mlnx_rw_smbus);

static DEFINE_MUTEX(lpc_mutex);
union io_data {
	struct {
		u16 v1;
		u8  v2;
	} __attribute__ ((packed)) val;
	u8 buf[3];
};

int mlnx_rw_lpc_device(struct mlnx_lpc_ioctl_params *params)
{
	int i, rc = 0;
	u32 addr;
	u8 rem = 0, widx = 0;
	struct mlnx_lpc_ioctl_params lpc_data;
	u8 data_buf[LPCI2C_DATA_REG_SZ];

	if (copy_from_user(&lpc_data, (struct mlnx_lpc_ioctl_params __user *)params, sizeof(lpc_data)))
		return -EFAULT;

	addr = lpc_data.base + lpc_data.offset;
	if (lpc_data.read_write == 0) {
	    DBG("LPC write: base: 0x%x, offs: 0x%x, data 0x%x\n", lpc_data.base, lpc_data.offset, *(u8*)lpc_data.data);
	    switch (lpc_data.datalen) {
		case 1:
			outb(*((u8*)lpc_data.data), addr);
			break;
		case 2:
			outw(*((u16*)lpc_data.data), addr);
			break;
		case 3:
			outw(*((u16*)lpc_data.data), addr);
			outb(*((u8*)lpc_data.data + 2), addr + 2);
			break;
		case 4:
			outl(*((u32*)lpc_data.data), addr);
			break;
		default:
			rem = lpc_data.datalen % 4;
			widx = lpc_data.datalen / 4;
			for (i = 0; i < widx; i++)
				outl(*((u32*)lpc_data.data + i), addr + i*4);
			for (i = 0; i < rem; i++)
				outb(*((u8*)lpc_data.data + widx*4 + i), addr + widx*4 + i);
			break;
		}
	}
	else {
	    DBG("LPC read: base: 0x%x, offs: 0x%x\n", lpc_data.base, lpc_data.offset);
	    switch (lpc_data.datalen) {
		case 1:
			data_buf[0] = inb(addr);
			break;
		case 2:
			*(u16*)data_buf = inw(addr);
		break;
		case 3:
			{
				union io_data* data = (union io_data*)data_buf;
				data->val.v1 = inw(addr);
				data->val.v2 = inb(addr + 2);
			}
			break;
		case 4:
			*(u32*)data_buf = inl(addr);
			break;
		default:
			rem = lpc_data.datalen % 4;
			widx = lpc_data.datalen / 4;
            for (i = 0; i < widx; i++)
                *((u32*)data_buf + i) = inl(addr + i*4);
            for (i = 0; i < rem; i++)
                *(data_buf + widx*4 + i) = inb(addr + widx*4 + i);
			break;
		}
		if (copy_to_user((void*)(((struct mlnx_lpc_ioctl_params*)params)->data), &data_buf, lpc_data.datalen))
			return -EFAULT;
	}
    return rc;
}

void mlnx_rw_lpc_kernel(struct mlnx_lpc_kernel_params *params)
{
	u32 i, addr;
	u8 rem = 0, widx = 0;

	addr = params->base + params->offset;
	if (params->read_write == 0) {
		switch (params->datalen) {
		case 4:
			outl(*((u32*)params->data), addr);
			break;
		case 3:
			outw(*((u16*)params->data), addr);
			outb(*((u8*)params->data + 2), addr + 2);
			break;
		case 2:
			outw(*((u16*)params->data), addr);
			break;
		case 1:
			outb(*((u8*)params->data), addr);
			break;
		default:
			rem = params->datalen % 4;
			widx = params->datalen / 4;
            for (i = 0; i < widx; i++)
				outl(*((u32*)params->data + i), addr + i*4);
			for (i = 0; i < rem; i++)
				outb(*((u8*)params->data + widx*4 + i), addr + widx*4 + i);
			break;
		}
	}
	else {
		switch (params->datalen) {
		case 4:
			*((u32*)params->data) = inl(addr);
			break;
		case 3:
			*((u16*)params->data) = inw(addr);
			*((u8*)(params->data + 2)) = inb(addr+2);
			break;
		case 2:
			*((u16*)params->data) = inw(addr);
			break;
		case 1:
			*((u8*)params->data) = inb(addr);
			break;
		default:
			rem = params->datalen % 4;
			widx = params->datalen / 4;
            for (i = 0; i < widx; i++)
                *((u32*)params->data + i) = inl(addr + i*4);
            for (i = 0; i < rem; i++)
                *((u8*)params->data + widx*4 + i) = inb(addr + widx*4 + i);
			break;
		}
	}
}
EXPORT_SYMBOL(mlnx_rw_lpc_kernel);

int mlnx_test_vbus_exist(struct mlnx_i2c_virt_bus_scan_ioctl_params *params)
{
	struct i2c_bus_table *p_bus = NULL;
	struct pca954x_table *p_pca954x_parent = NULL;

        if (params->parent_bus_num <= 0)
	        return -EFAULT;

        params->status = 0;
        p_bus = find_i2c_bus_entry(params->parent_bus_num);
        if (!p_bus)
	        return -EFAULT;
        if (p_bus->switch_id) {
                p_pca954x_parent = find_pca954x_switch(p_bus->switch_id);
                if (p_pca954x_parent)
                        params->status = p_pca954x_parent->status;
        }
        return 0;
}

int mlnx_test_device_exist(struct mlnx_i2c_one_device_ioctl_params *params)
{
	struct i2c_devices_table *p_i2c_dev_tbl;
	u8 go_bit_status = 0;
	u8 probe_status = 0;
	u8 hcr_acess_status = 0;
	int probe_cntr = 0;

        if (params->parent_bus_num <= 0)
	        return -EFAULT;
        params->status = DEV_NOT_EXIST;
        if ( (p_i2c_dev_tbl = get_i2c_device(params->parent_bus_num, params->dev.i2c_dev_addr)) != NULL ) {
                if (sx_i2c_get_probe_status_hook && p_i2c_dev_tbl->client)
                        sx_i2c_get_probe_status_hook(p_i2c_dev_tbl->client, &probe_status,
                                                        &probe_cntr, &go_bit_status, &hcr_acess_status);
                params->status = DEV_EXIST_OK;
                if (probe_status == PROBE_STAT_FAILED)
                        params->status |= DEV_FAILED_PROBE;
                if (go_bit_status == GO_BIT_STAT_STUCK)
                        params->status |= DEV_BAD_GO_BIT;
                if (hcr_acess_status == HCR_ACCESS_NOT_OK)
                        params->status |= DEV_FAILED_HCR_ACCESS;
                return 0;
        }
        return -EFAULT;
}

static inline int get_irq_entry_num(struct cpld_device_table *p_cpld_device)
{
        int entry_num;
        switch (p_cpld_device->card_type) {
                case CARD_TYPE_MGMT:
                        entry_num = MAX_MGMT_IRQS;
                        break;
                case CARD_TYPE_MODULE:
                        entry_num = MAX_MOD_IRQS;
                        break;
                case CARD_TYPE_TOR:
                        entry_num = MAX_TOR_IRQS;
                        break;
                case CARD_TYPE_PORT:
                case CARD_TYPE_NONE:
                default:
                        printk(KERN_INFO "%s(%d) bad card type %d\n",
                                __FUNCTION__, __LINE__, p_cpld_device->card_type);
                                return -EINVAL;
        }
        return entry_num;
}

static inline int cpld_match_dev_index(struct cpld_device_table *p_cpld_device, int dev_index)
{
        int i;
        for (i = 0; (i < DEV_INT_CAP_MAX) &&
                (p_cpld_device->hw_acc_tab[i].dev_index != DEV_INDEX_NONE); i++) {

                	mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
	                            "%s(%d) i=%d dev_ind %d %d\n",
                                    __FUNCTION__, __LINE__, i, dev_index,
                                    p_cpld_device->hw_acc_tab[i].dev_index);

                if (p_cpld_device->hw_acc_tab[i].dev_index == dev_index) {
                        return i;
                }
        }
        return -1;
}

static int cpld_mask_and_read_irqs(struct cpld_device_table *p_cpld_device,
                                        struct int_digest_blk_t *p_digest_blk,
                                        struct mod_digest *p_digest_entry,
                                        int module_ind,
                                        int level)
{
        struct cpld_mgmt_set_action *p_set_act = &p_cpld_device->p_irqs->u.mgmt_set_action;
        struct cpld_mgmt_get *p_get = &p_cpld_device->p_get_irqs->u.mgmt_get;
        struct cpld_mgmt_get *p_auxilary_cache_blk = &p_cpld_device->p_auxilary_cache_blk->u.mgmt_get;
        struct cpld_mgmt_get *p_master_cache_blk = &p_cpld_device->p_master_cache_blk->u.mgmt_get;
        int entry_num = 0;
        int i;
        int rc = 0;
        u8 no_event_reg = 0;
        u8 no_status_reg = 0;
        u8 status_has_changed = 0;
        u8 retry_level_1_access_cntr;

        if ((entry_num = get_irq_entry_num(p_cpld_device)) < 0)
                return entry_num;

        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	          "%s(%d) entry_num=%d tab size=%zd level=%d wake_on_signal=%x\n",
                  __FUNCTION__, __LINE__, entry_num, sizeof(struct cpld_device_table),
                 level, p_cpld_device->wake_on_signal);

        p_cpld_device->action_entry.aggregation = 0;
        p_cpld_device->action_entry.recovered = 0;
        p_cpld_device->action_entry.termination = 0;
        for (i = 0; i < entry_num; i++) {
	        if (p_set_act->s_entry[i].cfg.level != level)
                        continue;

	        if (p_set_act->s_entry[i].mask != ~CPLD_MASK_NOP8) {
	                /* Mask interrupts */
	                if ((p_set_act->s_entry[i].cfg.e_mask.range[0] != CPLD_MASK_NOP8) &&
	                    (p_set_act->s_entry[i].cfg.e_mask.range[1] != CPLD_MASK_NOP8) &&
	                    (p_set_act->s_entry[i].cfg.profile & _E_AVAIL)) {
		                if (cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id]) {
			                u8 set_mask = 0;

			                if ((p_set_act->s_entry[i].cfg.profile & _E_INV) &&
			                    (!(p_set_act->s_entry[i].cfg.profile & _E_INV_M)))
			                        set_mask = 0xff;

                                        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                                        "%s(%d) i=%d set_mask=%x mask addr=%x mask=%x\n",
                                                __FUNCTION__, __LINE__, i, set_mask,
                                                p_set_act->s_entry[i].cfg.e_mask.range[0],
                                                p_set_act->s_entry[i].cfg.mask);

			                retry_level_1_access_cntr = 3;

retry_level_1_mask_access:
			                rc = cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id](
						                &p_set_act->s_entry[i].mask_hwdev,
						                p_set_act->s_entry[i].cfg.e_mask.range[0],
						                p_set_act->s_entry[i].cfg.e_mask.range[1],
						                0, &set_mask, _B_NOLOCK);
		                        if (rc < 0) {
			                        if ((level == INT_LEVEL_1) && retry_level_1_access_cntr) {
			                                /* Try to reset bus only at the first level access */
			                                if (cpld_ctx.rst_func[p_set_act->s_entry[i].act.rw_func_id])
			                                        rc = cpld_ctx.rst_func[p_set_act->s_entry[i].act.rw_func_id](
						                                        &p_cpld_device->bus_rst, _B_NOLOCK);
			                                if (rc < 0) {
			                                        printk(KERN_INFO "%s(%d) Can not reset auxilary bus\n", __func__, __LINE__);
			                                        return rc;
			                                }
			                                retry_level_1_access_cntr--;
			                                goto retry_level_1_mask_access;
			                        }
			                        return rc;
		                        }
		                        if (retry_level_1_access_cntr < 3)
		                                p_cpld_device->action_entry.recovered = 1;
		                }
	                }
	        }
        }

        for (i = 0; i < entry_num; i++) {
	        if (p_set_act->s_entry[i].cfg.level != level)
                        continue;
                if (p_set_act->s_entry[i].mask == ~CPLD_MASK_NOP8)
                        continue;
                p_get->g_entry[i].status = 0;
                p_get->g_entry[i].event = 0;

	        /* Read statuses */
	        /* Note: for non CPLD based system (status equivalent to event
	                 and reading status also acknowledges interrupt */
	        if ((p_set_act->s_entry[i].cfg.e_status.range[0] != CPLD_MASK_NOP8) &&
	            (p_set_act->s_entry[i].cfg.e_status.range[1] != CPLD_MASK_NOP8) &&
	            (p_set_act->s_entry[i].cfg.profile & _E_AVAIL)) {
		        if (cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id]) {
	                        retry_level_1_access_cntr = 3;
retry_level_1_status_access:
	                        rc = cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id](
						        &p_set_act->s_entry[i].status_hwdev,
						        p_set_act->s_entry[i].cfg.e_status.range[0],
						        p_set_act->s_entry[i].cfg.e_status.range[1],
						        1, (u8 *)&p_get->g_entry[i].status, _B_NOLOCK);

		        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		                "%s(%d) i=%d status=0x%02x mask=0x%02x profile=0x%02x\n",
		                __FUNCTION__, __LINE__, i,  p_get->g_entry[i].status, p_set_act->s_entry[i].cfg.mask,
		                p_set_act->s_entry[i].cfg.profile);

		                if (rc < 0) {
			                if ((level == INT_LEVEL_1) && retry_level_1_access_cntr) {
			                        /* Try to reset bus only at the first level access */

			                       mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
			                       	        "%s(%d) dev=%p reset i2c: adap=%p bus=%d addr=%x addr_space=%d %x %x %x %x\n",
			                       	        __FUNCTION__, __LINE__, p_cpld_device,
			                       	        p_cpld_device->bus_rst.hwdev.i2c_params.adap,
			                       	        p_cpld_device->bus_rst.hwdev.i2c_params.bus_id,
			                       	        p_cpld_device->bus_rst.hwdev.i2c_params.addr,
			                       	        p_cpld_device->bus_rst.hwdev.i2c_params.addr_space,
			                       	        p_cpld_device->bus_rst.offset, p_cpld_device->bus_rst.offset_wp,
			                       	        p_cpld_device->bus_rst.toggle_off_mask, p_cpld_device->bus_rst.toggle_on_mask);

			                       if (cpld_ctx.rst_func[p_set_act->s_entry[i].act.rw_func_id])
			                               rc = cpld_ctx.rst_func[p_set_act->s_entry[i].act.rw_func_id](
						                               &p_cpld_device->bus_rst, _B_NOLOCK);
			                        if (rc < 0) {
			                                printk(KERN_INFO "%s(%d) Can not reset auxilary bus\n", __func__, __LINE__);
			                                return rc;
			                        }
			                        retry_level_1_access_cntr --;
			                        goto retry_level_1_status_access;
			                }
			                return rc;
		                }
		                if (retry_level_1_access_cntr < 3)
		                        p_cpld_device->action_entry.recovered = 1;
		        }

		        p_get->g_entry[i].status &= p_set_act->s_entry[i].cfg.mask;

		        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		                "%s(%d) i=%d status=0x%02x mask=0x%02x profile=0x%02x\n",
		                __FUNCTION__, __LINE__, i,  p_get->g_entry[i].status, p_set_act->s_entry[i].cfg.mask,
		                p_set_act->s_entry[i].cfg.profile);

		        no_status_reg = 0;
                }
                else {
                        /* There are some special registers (watchdog) that do not have
                           status register.
                        */
                        no_status_reg  = 1;
	        }
	        /* Read events */
	        if ((p_set_act->s_entry[i].cfg.e_event.range[0] != CPLD_MASK_NOP8) &&
	            (p_set_act->s_entry[i].cfg.e_event.range[1] != CPLD_MASK_NOP8) &&
	            (p_set_act->s_entry[i].cfg.profile & _E_AVAIL)) {
		        if (!(p_set_act->s_entry[i].cfg.profile & _E_INV)) {
		                if (cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id]) {
			                rc = cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id](
						                &p_set_act->s_entry[i].event_hwdev,
						                p_set_act->s_entry[i].cfg.e_event.range[0],
						                p_set_act->s_entry[i].cfg.e_event.range[1],
						                1, (u8 *)&p_get->g_entry[i].event, _B_NOLOCK);
		                        if (rc < 0)
			                        return rc;
		                }
		                p_get->g_entry[i].event &= p_set_act->s_entry[i].cfg.mask;
		        }
		        else {
		                /* Set event as status */
		                p_get->g_entry[i].event = p_get->g_entry[i].status;
		        }

		        no_event_reg = 0;
                }
                else {
                        /* There are some special aggregation registers that do not have
                           event register.
                        */
                        no_event_reg  = 1;
                }

                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                "%s(%d) i=%d no_ev_reg=%d no_st_reg=%d s=%x e=%x old aux s=%x e=%x master s=%x e=%x\n",
                        __FUNCTION__, __LINE__, i, no_event_reg, no_status_reg,
                        p_get->g_entry[i].status, p_get->g_entry[i].event,
                        p_auxilary_cache_blk->g_entry[i].status,
                        p_auxilary_cache_blk->g_entry[i].event,
                        p_master_cache_blk->g_entry[i].status,
                        p_master_cache_blk->g_entry[i].event);

                /* Compare values against shadow - set CPLD_STATUS_CHANGED and
                   CPLD_EVENT_OCCURRED for bits which have been changed.
                   At the beginning:

                   Get changed filed list and store new values in digest block
                   Fill data from digest block into caller's container.
                   After verifiyng:
                   p_master_cache_blk   - keeps deltas between previos interrupt map and current
                                          interrupt map
                   p_auxilary_cache_blk - keeps new map, in case something has been chaged,
                                          or stays unchanged, in other case
                   if p_master_cache_blk  checksum is not zero - compose digest block from the
                                          signals that has been inserted
                */
                /* Get vector for changed statuses */
	        status_has_changed = 0;
	        if (p_auxilary_cache_blk->g_entry[i].status != p_get->g_entry[i].status) {
                        p_cpld_device->action_entry.status_checksum++;

                        if (p_set_act->s_entry[i].cfg.profile & _E_AGGR) {
                                /* This is aggregated signal - it should be passed to next
                                   lower level for wake_on_signal handling, so only relevant
                                   handler will react. */
                                /* Handle aggregation registers that do not have event register. */
                                p_cpld_device->action_entry.aggregation++;

                                if ((level > INT_LEVEL_1) /*&& (p_cpld_device->usr_handle != MAIN_MODULE_HANDLE)*/) {
                                        u8 wake_on_signal_mask;
	                                int j, k;

                                        wake_on_signal_mask =
                                                (p_auxilary_cache_blk->g_entry[i].status ^ p_get->g_entry[i].status);

                                        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
                                                "%s(%d) wake_on_signal_mask=0x%08x\n",
                                                __FUNCTION__, __LINE__, wake_on_signal_mask);
                                        if (wake_on_signal_mask) {
                                                for (j = 0, k = 1; j <= 7; j++) {
                                                        if (wake_on_signal_mask & k) {
                                                                cpld_ctx.wake_on_signal_id_tab[cpld_ctx.wake_on_signal_cntr++] =
                                                                        p_set_act->s_entry[i].cfg.sig[j].signal_id;

                                                                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                                                                "%s(%d) trigger signal: i=%d j=%d k=0x%02x mod_ind=%d wake_on=%x cntr=%d\n",
                                                                        __FUNCTION__, __LINE__, i, j, k, module_ind,
                                                                        p_set_act->s_entry[i].cfg.sig[j].signal_id,
                                                                        cpld_ctx.wake_on_signal_cntr);
                                                        }
                                                        k = k << 1;
                                                }
                                        }
                                }
                        }
                        else {
                                if (!no_status_reg) {
                                        /* Update status cache block only for non-aggregated entries.
                                           For aggregated entries it should be done automatically
                                           when lower level related non-aggregated interrupt is acknowledged.
                                        */
                                        status_has_changed = 1;
                                        /* This is specail case for non CPLD devices */
                                        if (!(p_set_act->s_entry[i].cfg.profile & _E_AGGR) &&  /* not aggregation */
                                             (p_set_act->s_entry[i].cfg.profile & _E_INV)) {   /* inversed */
                                                p_auxilary_cache_blk->g_entry[i].event = p_auxilary_cache_blk->g_entry[i].status;
                                                p_get->g_entry[i].event = p_get->g_entry[i].status;
                                        }
                                        p_master_cache_blk->g_entry[i].status =
                                                (p_auxilary_cache_blk->g_entry[i].status ^ p_get->g_entry[i].status);
                                        p_auxilary_cache_blk->g_entry[i].status = p_get->g_entry[i].status;
                                }
                        }
                        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                        "%s(%d) i=%d no_stat_reg=%d stat_changed=%d no_ev_reg=%d aggr=%d stat=%x aux stat=%x ev=%x aux ev=%x\n",
                                __FUNCTION__, __LINE__, i, no_status_reg, status_has_changed,
                                no_event_reg, p_cpld_device->action_entry.aggregation,
                                p_master_cache_blk->g_entry[i].status, p_auxilary_cache_blk->g_entry[i].status,
                                p_auxilary_cache_blk->g_entry[i].event, p_get->g_entry[i].event);
                }
                /* Get vector for changed events. */
                p_master_cache_blk->g_entry[i].event = 0;
                if (p_auxilary_cache_blk->g_entry[i].event != p_get->g_entry[i].event) {
                        /* Events have been arrived if newly read event bit/bits are set 1. */
                        p_cpld_device->action_entry.event_checksum++;
                        p_master_cache_blk->g_entry[i].event =
                                (p_auxilary_cache_blk->g_entry[i].event ^ p_get->g_entry[i].event);
                        p_auxilary_cache_blk->g_entry[i].event = p_get->g_entry[i].event;
                }

                /* Insert entry for passing to user space over netlink.
                   Only entry which has non empty mask, changed event bit and
                   changed status bit (in case status bit is exist should be
                   inserted.
                */
                if ((p_set_act->s_entry[i].mask != ~CPLD_MASK_NOP8) &&
                    ( (p_master_cache_blk->g_entry[i].event && (!no_status_reg)) ||
                      (!(p_set_act->s_entry[i].cfg.profile & _E_AGGR))) &&
                    status_has_changed) {
	        	p_digest_entry->int_digest[i].int_vector = p_master_cache_blk->g_entry[i].event;
	        	if ((p_set_act->s_entry[i].cfg.profile & _E_INV) &&
	        	    (p_set_act->s_entry[i].cfg.profile & _E_INV_S))
	        		p_digest_entry->int_digest[i].int_status = p_get->g_entry[i].status;
	        	else if ((p_set_act->s_entry[i].cfg.profile & _E_INV) ||
	        	         (p_set_act->s_entry[i].cfg.profile & _E_INV_S))
	        		p_digest_entry->int_digest[i].int_status = ~p_get->g_entry[i].status;
	        	else
	        		p_digest_entry->int_digest[i].int_status = p_get->g_entry[i].status;
                	p_digest_entry->change_cntr++; // event detected for entry i
                	p_digest_blk->digest_table[module_ind].int_digest[i].int_vector =
                	                	                	p_master_cache_blk->g_entry[i].event;
	        	p_digest_blk->digest_table[module_ind].int_digest[i].int_status =
	        		p_digest_entry->int_digest[i].int_status;

                	p_digest_blk->digest_table[module_ind].change_cntr++;
                	p_digest_blk->digest_table[module_ind].usr_handle = p_cpld_device->usr_handle;

                	if (p_set_act->s_entry[i].cfg.profile & _E_TERM)
                		p_cpld_device->action_entry.termination = 1;

                	mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                	"%s(%d) i=%d term=%d ch_cntr=%d mod_ind=%d int_s=%x int_v=%x old s=%x e=%x master s=%x e=%x h=%x\n",
                        	__FUNCTION__, __LINE__, i, p_cpld_device->action_entry.termination,
                        	p_digest_entry->change_cntr, module_ind,
                        	p_digest_entry->int_digest[i].int_status,
                        	p_digest_entry->int_digest[i].int_vector,
                        	p_auxilary_cache_blk->g_entry[i].status,
                        	p_auxilary_cache_blk->g_entry[i].event,
                        	p_master_cache_blk->g_entry[i].status,
                        	p_master_cache_blk->g_entry[i].event,
                        	p_digest_blk->digest_table[module_ind].usr_handle);
                }
        }

        return rc;
}

static int cpld_ack_and_unmask_irqs(struct cpld_device_table *p_cpld_device, int level)
{
        struct cpld_mgmt_set_action *p_set_act = &p_cpld_device->p_irqs->u.mgmt_set_action;
        struct cpld_mgmt_get *p_get = &p_cpld_device->p_get_irqs->u.mgmt_get;
        struct cpld_mgmt_get *p_auxilary_cache_blk = &p_cpld_device->p_auxilary_cache_blk->u.mgmt_get;
        struct cpld_mgmt_get *p_master_cache_blk = &p_cpld_device->p_master_cache_blk->u.mgmt_get;
        int entry_num = 0;
        int i;
        int rc = 0;

        if ((entry_num = get_irq_entry_num(p_cpld_device)) < 0)
                return entry_num;

        mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
	        "%s(%d) entry_num=%d tab size=%zd level=%d\n",
                __FUNCTION__, __LINE__, entry_num, sizeof(struct cpld_device_table),
                level);

        for (i = 0; i < entry_num; i++) {
	        if (p_set_act->s_entry[i].cfg.level != level)
                        continue;

	        /* acknowledge events */
	        if ((p_set_act->s_entry[i].cfg.e_event.range[0] != CPLD_MASK_NOP8) &&
                    (p_set_act->s_entry[i].cfg.e_event.range[1] != CPLD_MASK_NOP8) &&
                    (!(p_set_act->s_entry[i].cfg.profile & _E_INV) &&
                    (p_set_act->s_entry[i].cfg.profile & _E_AVAIL))) {
                        if (cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id]) {
	                        struct irq_sig_entry *p_sig;
	                        u8 ack_event = 0;
	                        u8 rw_oper = 0;
	                        u8 rw_size = 0;
	                        u8 rw_data[4];
	                        int j, k, l, m;

	                        /* If there are any termination devices - acknowledge them */
	                        for (j = 1, k = 1; j <= 8; j++) {
	                                p_sig = &p_set_act->s_entry[i].cfg.sig[j - 1];
	                                if ((p_sig->dev_signal.dev_index =! DEV_INDEX_NONE) &&
	                                    (p_get->g_entry[i].event & k)) {
	                                        for (l = 0; l < p_cpld_device->hw_term_acc_tab_num; l++) {
	                                                if (p_sig->dev_signal.dev_index ==
	                                                    p_cpld_device->hw_term_acc_tab[l].dev_index) {
	                                                        rw_oper = RW_T_OPER(p_sig->dev_signal.rw_oper);
	                                                        rw_size = RW_T_SIZE(p_sig->dev_signal.rw_oper);
	                                                        if (!rw_oper) {
	                        	                                /* Write for clear */
	                                                                for (m = 0; m < rw_size; m++)
	                                                                        rw_data[m] = p_sig->dev_signal.clear_mask;
	                                                        }
	                                                        rc = i2c_bus_access_func(
	                                                                (union hw_dev*)&p_cpld_device->hw_term_acc_tab[l].hwdev,
	                                                                p_sig->dev_signal.clear_addr,
	                                                                p_sig->dev_signal.clear_addr + rw_size - 1,
	                                                                rw_oper, rw_data, _B_NOLOCK);
	                                                        if (rc < 0)
	                                                                return rc;

                                                                 mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
	                                                                 "%s(%d) %d %d %d %d addr=%x size=%x oper=%x " \
	                                                                 "data %x %x %x %x adap=%p bus=%d dev addr=%x\n",
                                                                         __FUNCTION__, __LINE__, i, j, k, l,
                                                                         p_sig->dev_signal.clear_addr, rw_size, rw_oper,
                                                                         rw_data[0], rw_data[1], rw_data[2], rw_data[3],
                                                                         p_cpld_device->hw_term_acc_tab[l].hwdev.i2c_params.adap,
                                                                         p_cpld_device->hw_term_acc_tab[l].hwdev.i2c_params.bus_id,
                                                                         p_cpld_device->hw_term_acc_tab[l].hwdev.i2c_params.addr);
	                                                }
	                                        }
	                                }
	                                k = k << 1;
	                        }

	                        rc = cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id](
						        &p_set_act->s_entry[i].event_hwdev,
						        p_set_act->s_entry[i].cfg.e_event.range[0],
						        p_set_act->s_entry[i].cfg.e_event.range[1],
						        0, &ack_event, _B_NOLOCK);
	                        if (rc < 0)
                                        return rc;

	                        p_auxilary_cache_blk->g_entry[i].event = ack_event;

                                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                                "%s(%d) i=%d event addr=%x s=%x e=%x aux e=%x master e=%x\n",
                                        __FUNCTION__, __LINE__, i,
                                        p_set_act->s_entry[i].cfg.e_event.range[0],
                                        p_get->g_entry[i].status,
                                        p_get->g_entry[i].event,
                                        p_auxilary_cache_blk->g_entry[i].event,
                                        p_master_cache_blk->g_entry[i].event);
                        }
	        }
        }

        for (i = 0; i < entry_num; i++) {
	        if (p_set_act->s_entry[i].cfg.level != level)
                        continue;
	        /* unmask interrupts */
	        if ((p_set_act->s_entry[i].cfg.e_mask.range[0] != CPLD_MASK_NOP8) &&
	            (p_set_act->s_entry[i].cfg.e_mask.range[1] != CPLD_MASK_NOP8) &&
	            (p_set_act->s_entry[i].cfg.profile & _E_AVAIL)) {
	                if (cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id]) {
			        u8 set_mask;

	                        if ((p_set_act->s_entry[i].cfg.profile & _E_INV) &&
	        		     (!(p_set_act->s_entry[i].cfg.profile & _E_INV_M)))
                                        set_mask = ~p_set_act->s_entry[i].cfg.mask;
	                        else
                                        set_mask = p_set_act->s_entry[i].cfg.mask;

                                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                                "%s(%d) i=%d set_mask=%x mask addr=%x mask=%x\n",
                                         __FUNCTION__, __LINE__, i, set_mask,
                                         p_set_act->s_entry[i].cfg.e_mask.range[0],
                                         p_set_act->s_entry[i].cfg.mask);

	                        rc = cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id](
						                &p_set_act->s_entry[i].mask_hwdev,
						                p_set_act->s_entry[i].cfg.e_mask.range[0],
						                p_set_act->s_entry[i].cfg.e_mask.range[1],
						                0, &set_mask, _B_NOLOCK);
	                        if (rc < 0)
	                                return rc;

                                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                                "%s(%d) i=%d set_mask=%x mask addr=%x mask=%x\n",
                                        __FUNCTION__, __LINE__, i, set_mask,
                                        p_set_act->s_entry[i].cfg.e_mask.range[0],
                                        p_set_act->s_entry[i].cfg.mask);
		                }
	                }
        }

        return rc;
}

static __u8 inline is_device_wake_on_signal(struct cpld_device_table *p_cpld_device)
{
        int i;

        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
                "%s(%d) wake_on_signal_cntr=%d wake_on_signal=%x\n",
                __FUNCTION__, __LINE__, cpld_ctx.wake_on_signal_cntr,
                p_cpld_device->wake_on_signal);
        if (cpld_ctx.wake_on_signal_cntr == 0)
                return 0;
        for (i = 0; i < cpld_ctx.wake_on_signal_cntr; i++) {
                if (p_cpld_device->wake_on_signal == cpld_ctx.wake_on_signal_id_tab[i])
                        return 1;
        }
        return 0;
}

/*
  cpld_i2c_pre_action:
    Acts on i2c device.
    Perfroms "mask interrupts" -> "read interrupts" sequence and updates
    detected interrupts within master cache block.
*/
int cpld_i2c_pre_action(void *subscriber_data,
			struct int_digest_blk_t *p_digest_blk,
			struct mod_digest *p_digest_entry,
			int module_ind,
			int level)
{
	int rc = 0;
	struct cpld_device_table *p_cpld_device = (struct cpld_device_table *)subscriber_data;
        struct cpld_mgmt_get *p_master_cache_blk = &p_cpld_device->p_master_cache_blk->u.mgmt_get;
        struct cpld_mgmt_set_action *p_set_act;

	if (!p_cpld_device) {
		printk(KERN_INFO "%s: bad subscriber data\n", __FUNCTION__);
		return -EFAULT;
	}

        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	        "%s(%d) cpld_device=%p digest_entry=%p digest_blk=%p usr_handle=(%x) wake_on_signal=%x level=%d\n",
	        __FUNCTION__, __LINE__,
	        p_cpld_device, p_digest_entry, p_digest_blk, p_cpld_device->usr_handle,
	        p_cpld_device->wake_on_signal, level);

        if (level == INT_LEVEL_1) {
                /* Cleanup at level 1 wake_on_signal table */
                int i;

                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                "%s(%d) lock lock buses on start\n",
	                __FUNCTION__, __LINE__);
	        _lock_bus(/*_B_ALL*/ p_cpld_device->device_type);
                cpld_ctx.wake_on_signal_cntr = 0;
                for (i = 0; i < MAX_MODULE_NUM; i++)
                        cpld_ctx.wake_on_signal_id_tab[i] = _none_sig;
                memset(p_digest_entry, 0, sizeof(struct mod_digest));  // Not correct
                memset(p_master_cache_blk, 0, sizeof(struct cpld_mgmt_get));
        }
        else {
                /* Skip not relevant signals */
                if ((p_cpld_device->usr_handle != MAIN_MODULE_HANDLE) &&
                    (!is_device_wake_on_signal(p_cpld_device)))
                        return rc;
        }

        rc = cpld_mask_and_read_irqs(p_cpld_device, p_digest_blk, p_digest_entry,
					module_ind, level);
        if (rc < 0) {
	        _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);
                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                "%s(%d) unlock lock buses on failure rc=%d\n",
	                __FUNCTION__, __LINE__, rc);
	        return rc;
        }
	p_set_act = &p_cpld_device->p_irqs->u.mgmt_set_action;
	p_digest_entry->usr_handle = p_cpld_device->usr_handle;
	p_master_cache_blk = &p_cpld_device->p_master_cache_blk->u.mgmt_get;

        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	        "%s(%d) p_cpld_device=%p p_digest_entry=%p usr_handle=(%x) (%x) cntr=%d all_change_cntr=%d\n",
	        __FUNCTION__, __LINE__,
            p_cpld_device, p_digest_entry, p_digest_entry->usr_handle, p_cpld_device->usr_handle,
            p_digest_entry->change_cntr, p_digest_blk->all_change_cntr);

	return rc;
}

/*
  cpld_i2c_post_action:
    Acts on i2c device.
    Perfroms "acknowledge interrupts" -> "unmask interrupts" sequence and updates
    detected interrupts within auxiliary cache blok.
*/
int cpld_i2c_post_action(void *subscriber_data,
			 struct int_digest_blk_t *p_digest_blk,
			 struct mod_digest *p_digest_entry,
                         int module_ind,
			 int level)
{
	int rc = 0;
	struct cpld_device_table *p_cpld_device = (struct cpld_device_table *)subscriber_data;

	if (!p_cpld_device) {
	        _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);
		printk(KERN_INFO "%s: bad subscriber data\n", __FUNCTION__);
		return -EFAULT;
	}

        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) cpld_device=%p digest_entry=%p digest_blk=%p usr_handle=(%x) cntr=%d wake_on_signal=%x change_cntr=%d " \
		"all_change_cntr=%d usr_handle=%x\n",
		__FUNCTION__, __LINE__,
		p_cpld_device, p_digest_entry, p_digest_blk, p_cpld_device->usr_handle, p_digest_entry->change_cntr,
		p_cpld_device->wake_on_signal, p_digest_entry->change_cntr, p_digest_blk->all_change_cntr,
		p_digest_entry->usr_handle);

        /* Skip not relevant signals */
        if ((p_cpld_device->usr_handle != MAIN_MODULE_HANDLE) &&
            (!is_device_wake_on_signal(p_cpld_device)))
                        return rc;

        rc = cpld_ack_and_unmask_irqs(p_cpld_device, level);
        if (rc < 0) {
	        _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);
                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                "%s(%d) unlock buses on failure rc=%d\n",
	                __FUNCTION__, __LINE__, rc);
	        return rc;
        }

        if (level == INT_LEVEL_1) {
	        _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);
                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                "%s(%d) unlock buses on complete rc=%d\n",
	                __FUNCTION__, __LINE__, rc);
        }

	return rc;
}

/*
  cpld_lpc_pre_action:
    Acts on lpc device.
    Perfroms "mask interrupts" -> "read interrupts" sequence and updates
    detected interrupts within master cache blok.
*/
int cpld_lpc_pre_action(void *subscriber_data,
			struct int_digest_blk_t *p_digest_blk,
			struct mod_digest *p_digest_entry,
			int module_ind,
			int level)
{
	int rc = 0;
	struct cpld_device_table *p_cpld_device = (struct cpld_device_table *)subscriber_data;
        struct cpld_mgmt_get *p_master_cache_blk = &p_cpld_device->p_master_cache_blk->u.mgmt_get;
        struct cpld_mgmt_set_action *p_set_act;

	if (!p_cpld_device) {
		printk(KERN_INFO "%s: bad subscriber data\n", __FUNCTION__);
		return -EFAULT;
	}

        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	        "%s(%d) p_cpld_device=%p p_digest_entry=%p usr_handle=(%x) wake_on_signal=%x\n",
	        __FUNCTION__, __LINE__,
	        p_cpld_device, p_digest_entry, p_cpld_device->usr_handle,
	        p_cpld_device->wake_on_signal);

        if (level == INT_LEVEL_1) {
                /* Cleanup at level 1 wake_on_signal table */
                int i;

                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                "%s(%d) lock buses on start\n",
	                __FUNCTION__, __LINE__);
	        _lock_bus(/*_B_ALL*/ p_cpld_device->device_type);
                cpld_ctx.wake_on_signal_cntr = 0;
                for (i = 0; i < MAX_MODULE_NUM; i++)
                        cpld_ctx.wake_on_signal_id_tab[i] = _none_sig;
                memset(p_digest_entry, 0, sizeof(struct mod_digest));  // Not correct
                memset(p_master_cache_blk, 0, sizeof(struct cpld_mgmt_get));
        }
        else {
                /* Skip not relevant signals */
                if ((p_cpld_device->usr_handle != MAIN_MODULE_HANDLE) &&
                    (!is_device_wake_on_signal(p_cpld_device)))
                        return rc;
        }

        rc = cpld_mask_and_read_irqs(p_cpld_device, p_digest_blk, p_digest_entry,
					module_ind, level);
        if (rc < 0) {
	        _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);
                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                "%s(%d) unlock lock buses on failure rc=%d\n",
	                __FUNCTION__, __LINE__, rc);
	        return rc;
        }
	p_set_act = &p_cpld_device->p_irqs->u.mgmt_set_action;
	p_digest_entry->usr_handle = p_cpld_device->usr_handle;
	p_master_cache_blk = &p_cpld_device->p_master_cache_blk->u.mgmt_get;

        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	        "%s(%d) p_digest_entry=%p usr_handle=(%x)\n",
	        __FUNCTION__, __LINE__, p_digest_entry, p_digest_entry->usr_handle);

	return rc;
}

/*
  cpld_lpc_post_action:
    Acts on lpc device.
    Perfroms "acknowledge interrupts" -> "unmask interrupts" sequence and updates
    detected interrupts within auxiliary cache blok.
*/
int cpld_lpc_post_action(void *subscriber_data,
			 struct int_digest_blk_t *p_digest_blk,
			 struct mod_digest *p_digest_entry,
                         int module_ind,
			 int level)
{
	int rc = 0;
	struct cpld_device_table *p_cpld_device = (struct cpld_device_table *)subscriber_data;

	if (!p_cpld_device) {
	        _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);
                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                "%s(%d) unlock buses on failure rc=%d\n",
	                __FUNCTION__, __LINE__, rc);
		printk(KERN_INFO "%s: bad subscriber data\n", __FUNCTION__);
		return -EFAULT;
	}

        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
		"%s(%d) p_cpld_device=%p p_digest_entry=%p usr_handle=(%x) cntr=%d wake_on_signal=%x all_change_cntr=%d usr_handle=%x\n",
		__FUNCTION__, __LINE__,
		p_cpld_device, p_digest_entry, p_cpld_device->usr_handle, p_digest_entry->change_cntr,
		p_cpld_device->wake_on_signal, p_digest_blk->all_change_cntr, p_digest_entry->usr_handle);

        if ((p_cpld_device->usr_handle != MAIN_MODULE_HANDLE) &&
            (!is_device_wake_on_signal(p_cpld_device)))
                return rc;

        rc = cpld_ack_and_unmask_irqs(p_cpld_device, level);
        if (rc < 0) {
	        _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);
                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                "%s(%d) unlock buses on failure rc=%d\n",
	                __FUNCTION__, __LINE__, rc);
	        return rc;
        }

        if (level == INT_LEVEL_1) {
	        _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);
                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                "%s(%d) unlock buses on complete rc=%d\n",
	                __FUNCTION__, __LINE__, rc);
        }

	return rc;
}

/* Routine performs interrupts' masking accumulation.
   The possible situation when interrupt related to
   the same module are located on different devices
   (CPLD or GPIO expenders).
   In such case all interrupts are to be accumulated and
   set with the proper access rules within module related
   context.
   At this points the necessary interrupts are masked
   and kept not effective.
*/
static int cpld_accumulate_irqs(struct cpld_device_table *p_cpld_device,
                                struct mlnx_cpld_subscribe_ioctl_params *params)
{
        struct mlnx_cpld_interrupt_set_params *irqs = &params->irqs;
        struct cpld_mgmt_set *p_set = &irqs->u.mgmt_set;
        struct cpld_mgmt_set_action *p_set_act = &p_cpld_device->p_irqs->u.mgmt_set_action;
        int entry_num = 0;
        int dev_status_entry = -1;
        int dev_event_entry = -1;
        int dev_mask_entry = -1;
        int i;
        int rc = 0;

        mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
	        "%s(%d) dev type=%d card type=%d dev_id=%d usr_handle=(%x)\n",
	        __FUNCTION__, __LINE__,
	        p_cpld_device->device_type, p_cpld_device->card_type,
	        p_cpld_device->device_id, p_cpld_device->usr_handle);

	if (params->device_type == I2C_BUS_TYPE) {
		p_cpld_device->action_entry.pre_action = cpld_i2c_pre_action;   /* interrupt handler pre-action */
		p_cpld_device->action_entry.post_action = cpld_i2c_post_action; /* interrupt handler post-action */

                for (i = 0; (i < DEV_INT_CAP_MAX) &&
                	(params->hw_acc_tab[i].dev_index != DEV_INDEX_NONE); i++) {
	        	struct i2c_adapter *adap = NULL;

                        mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
	                        "%s(%d) i2c entry=%d dev_index=%d bus_id=%d\n",
                                __FUNCTION__, __LINE__, i, params->hw_acc_tab[i].dev_index,
                                params->hw_acc_tab[i].hwdev.i2c_params.bus_id);
                	adap = i2c_get_adapter(params->hw_acc_tab[i].hwdev.i2c_params.bus_id);
                	if (!adap) {
		        	printk(KERN_INFO "%s: bad adaptor for device 0x%08x dev_ind=%d bus_id=%d\n",
		        		__FUNCTION__, params->dev_id,
		        		params->hw_acc_tab[i].dev_index,
		        		params->hw_acc_tab[i].hwdev.i2c_params.bus_id);
                        	return -EFAULT;
                	}

                        memcpy(&p_cpld_device->hw_acc_tab[i], &params->hw_acc_tab[i], sizeof(struct dev_access_block));
			p_cpld_device->hw_acc_tab[i].hwdev.i2c_params.adap = adap;
                        mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
	                        "%s(%d) dev=%p entry=%d dev_index=%d i2c: adap=%p bus=%d addr=%x addr_space=%d\n",
	                        __FUNCTION__, __LINE__, p_cpld_device, i,
	                        p_cpld_device->hw_acc_tab[i].dev_index,
	                        p_cpld_device->hw_acc_tab[i].hwdev.i2c_params.adap,
	                        p_cpld_device->hw_acc_tab[i].hwdev.i2c_params.bus_id,
	                        p_cpld_device->hw_acc_tab[i].hwdev.i2c_params.addr,
	                        p_cpld_device->hw_acc_tab[i].hwdev.i2c_params.addr_space);
                }
	        p_cpld_device->hw_acc_tab_num = i;
	}
	else if (params->device_type == LPC_BUS_TYPE) {
		p_cpld_device->action_entry.pre_action = cpld_lpc_pre_action;
		p_cpld_device->action_entry.post_action = cpld_lpc_post_action;

                for (i = 0; (i < DEV_INT_CAP_MAX) &&
                	(params->hw_acc_tab[i].dev_index != DEV_INDEX_NONE); i++) {
                        memcpy(&p_cpld_device->hw_acc_tab[i], &params->hw_acc_tab[i], sizeof(struct dev_access_block));
                        mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
                                "%s(%d) dev=%p entry=%d dev_index=%d i2c: lpc: size=%d base=%x\n",
                                __FUNCTION__, __LINE__, p_cpld_device, i,
                                p_cpld_device->hw_acc_tab[i].dev_index,
                                p_cpld_device->hw_acc_tab[i].hwdev.lpc_params.size,
                                p_cpld_device->hw_acc_tab[i].hwdev.lpc_params.base);
                }
	        p_cpld_device->hw_acc_tab_num = i;
	}

	/* Add termination devices if any (could be attached only to I2C bus */
        for (i = 0; (i < DEV_TERM_INT_CAP_MAX) &&
	     (params->hw_term_acc_tab[i].dev_index != DEV_INDEX_NONE); i++) {
		struct i2c_adapter *adap = NULL;

		mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
			"%s(%d) i2c entry=%d dev_index=%d bus_id=%d\n",
			__FUNCTION__, __LINE__, i, params->hw_acc_tab[i].dev_index,
			params->hw_term_acc_tab[i].hwdev.i2c_params.bus_id);
		adap = i2c_get_adapter(params->hw_term_acc_tab[i].hwdev.i2c_params.bus_id);
		if (!adap) {
			printk(KERN_INFO "%s: bad adaptor for termination device at bus %d addr 0x%08x\n",
			        __FUNCTION__, params->hw_term_acc_tab[i].hwdev.i2c_params.bus_id,
			        params->hw_term_acc_tab[i].hwdev.i2c_params.addr);
			return -EFAULT;
		}

		memcpy(&p_cpld_device->hw_term_acc_tab[i], &params->hw_term_acc_tab[i], sizeof(struct dev_access_block));
		p_cpld_device->hw_term_acc_tab[i].hwdev.i2c_params.adap = adap;
		mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
			"%s(%d) dev=%p entry=%d termination dev_index=%d i2c: adap=%p bus=%d addr=%x addr_space=%d\n",
			__FUNCTION__, __LINE__, p_cpld_device, i,
			p_cpld_device->hw_term_acc_tab[i].dev_index,
			p_cpld_device->hw_term_acc_tab[i].hwdev.i2c_params.adap,
			p_cpld_device->hw_term_acc_tab[i].hwdev.i2c_params.bus_id,
			p_cpld_device->hw_term_acc_tab[i].hwdev.i2c_params.addr,
			p_cpld_device->hw_term_acc_tab[i].hwdev.i2c_params.addr_space);
	}
	p_cpld_device->hw_term_acc_tab_num = i;

	/* Add reset block */
	memcpy(&p_cpld_device->bus_rst, &params->bus_rst, sizeof(struct bus_reset_block));
	if (params->device_type == I2C_BUS_TYPE) {

		p_cpld_device->bus_rst.hwdev.i2c_params.adap = i2c_get_adapter(params->bus_rst.hwdev.i2c_params.bus_id);
		if (!p_cpld_device->bus_rst.hwdev.i2c_params.adap) {
			printk(KERN_INFO "%s: bad adaptor for termination device at bus %d addr 0x%08x\n",
			        __FUNCTION__, params->bus_rst.hwdev.i2c_params.bus_id,
			        params->bus_rst.hwdev.i2c_params.addr);
			return -EFAULT;
		}
		mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
			"%s(%d) dev=%p reset i2c: adap=%p bus=%d addr=%x addr_space=%d\n",
			__FUNCTION__, __LINE__, p_cpld_device,
			p_cpld_device->bus_rst.hwdev.i2c_params.adap,
			p_cpld_device->bus_rst.hwdev.i2c_params.bus_id,
			p_cpld_device->bus_rst.hwdev.i2c_params.addr,
			p_cpld_device->bus_rst.hwdev.i2c_params.addr_space);
	}

        /* Initial values for action byte field is NOP8, for double word fields
           is NOP32.
           For currunt value for action range just copy value from
           relevant parameter. Range for particular interrupt can't
           be defined on two different devices.
           For mask just OR current value with input parameter value.
        */

        if ((entry_num = get_irq_entry_num(p_cpld_device)) < 0)
            return entry_num;

        for (i = 0; i < entry_num; i++) {
                mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
                        "%s(%d) entry=%d profile=0x%04x mask 0x%08x status=0x%08x..0x%08x (%d)\n" \
                        "event=0x%08x..0x%08x (%d) mask=0x%08x..0x%08x (%d)\n" \
                        "signal_id dev_signal.dev_index: (0x%02x 0x%02x) (0x%02x 0x%02x)\n" \
                        "(0x%02x 0x%02x) (0x%02x 0x%02x) (0x%02x 0x%02x) (0x%02x 0x%02x)\n" \
                        "(0x%02x 0x%02x) (0x%02x 0x%02x)\n",
                        __FUNCTION__, __LINE__, i, p_set->s_entry[i].profile,
                        p_set->s_entry[i].mask,
                        p_set->s_entry[i].e_status.range[0], p_set->s_entry[i].e_status.range[1],
                        p_set->s_entry[i].e_status.dev_index,
                        p_set->s_entry[i].e_event.range[0], p_set->s_entry[i].e_event.range[1],
                        p_set->s_entry[i].e_event.dev_index,
                        p_set->s_entry[i].e_mask.range[0], p_set->s_entry[i].e_mask.range[1],
                        p_set->s_entry[i].e_mask.dev_index,
                        p_set->s_entry[i].sig[0].signal_id, p_set->s_entry[i].sig[0].dev_signal.dev_index,
                        p_set->s_entry[i].sig[1].signal_id, p_set->s_entry[i].sig[1].dev_signal.dev_index,
                        p_set->s_entry[i].sig[2].signal_id, p_set->s_entry[i].sig[2].dev_signal.dev_index,
                        p_set->s_entry[i].sig[3].signal_id, p_set->s_entry[i].sig[3].dev_signal.dev_index,
                        p_set->s_entry[i].sig[4].signal_id, p_set->s_entry[i].sig[4].dev_signal.dev_index,
                        p_set->s_entry[i].sig[5].signal_id, p_set->s_entry[i].sig[5].dev_signal.dev_index,
                        p_set->s_entry[i].sig[6].signal_id, p_set->s_entry[i].sig[6].dev_signal.dev_index,
                        p_set->s_entry[i].sig[7].signal_id, p_set->s_entry[i].sig[7].dev_signal.dev_index);

		memcpy(&p_set_act->s_entry[i].cfg.e_status.range, &p_set->s_entry[i].e_status.range,
		        sizeof(p_set->s_entry[i].e_mask.range));
		memcpy(&p_set_act->s_entry[i].cfg.e_event.range, &p_set->s_entry[i].e_event.range,
		        sizeof(p_set->s_entry[i].e_event.range));
		memcpy(&p_set_act->s_entry[i].cfg.e_mask.range, &p_set->s_entry[i].e_mask.range,
		        sizeof(p_set->s_entry[i].e_mask.range));
		memcpy(&p_set_act->s_entry[i].cfg.sig, &p_set->s_entry[i].sig,
		        sizeof(struct irq_sig_entry)*8);
		p_set_act->s_entry[i].mask = p_set->s_entry[i].mask;
		p_set_act->s_entry[i].cfg.mask = p_set->s_entry[i].mask;
		p_set_act->s_entry[i].cfg.level = p_set->s_entry[i].level;
		p_set_act->s_entry[i].cfg.profile = p_set->s_entry[i].profile;
		p_set_act->s_entry[i].act.rw_func_id = p_cpld_device->device_type;
		p_set_act->s_entry[i].cfg.e_status.dev_index = p_set->s_entry[i].e_status.dev_index;
		p_set_act->s_entry[i].cfg.e_event.dev_index = p_set->s_entry[i].e_event.dev_index;
		p_set_act->s_entry[i].cfg.e_mask.dev_index = p_set->s_entry[i].e_mask.dev_index;
                if(p_set->s_entry[i].e_status.dev_index != CPLD_NO_DEV) {
		        dev_status_entry = cpld_match_dev_index(p_cpld_device, p_set->s_entry[i].e_status.dev_index);
		        if (dev_status_entry < 0) {
		                printk(KERN_INFO "%s: device status entry not found in table\n", __FUNCTION__);
                        return -EFAULT;
                        }
		        memcpy(&p_set_act->s_entry[i].status_hwdev, &p_cpld_device->hw_acc_tab[dev_status_entry].hwdev,
		                sizeof(union hw_dev));
                }
                if(p_set->s_entry[i].e_event.dev_index != CPLD_NO_DEV) {
		        dev_event_entry = cpld_match_dev_index(p_cpld_device, p_set->s_entry[i].e_event.dev_index);
		        if (dev_event_entry < 0) {
		                printk(KERN_INFO "%s: device event entry not found in table\n", __FUNCTION__);
                        return -EFAULT;
                        }
		        memcpy(&p_set_act->s_entry[i].event_hwdev, &p_cpld_device->hw_acc_tab[dev_event_entry].hwdev,
		                sizeof(union hw_dev));
                }
                if(p_set->s_entry[i].e_mask.dev_index != CPLD_NO_DEV) {
		        dev_mask_entry = cpld_match_dev_index(p_cpld_device, p_set->s_entry[i].e_mask.dev_index);
		        if (dev_mask_entry < 0) {
		                printk(KERN_INFO "%s: device mask entry not found in table\n", __FUNCTION__);
                        return -EFAULT;
                        }
		        memcpy(&p_set_act->s_entry[i].mask_hwdev, &p_cpld_device->hw_acc_tab[dev_mask_entry].hwdev,
		                sizeof(union hw_dev));
                }
        }

        return rc;
}

/* Perfroms interrupt unmasking on physical device:
   all masks that have been stored within module
   context are written out to physical devices.
*/
static inline int cpld_set_irqs(struct cpld_device_table *p_cpld_device)
{
        struct cpld_mgmt_set_action *p_set_act = &p_cpld_device->p_irqs->u.mgmt_set_action;
	struct cpld_mgmt_get *p_master_cache_blk = &p_cpld_device->p_master_cache_blk->u.mgmt_get; // consider overlay
        struct cpld_mgmt_get *p_auxilary_cache_blk = &p_cpld_device->p_auxilary_cache_blk->u.mgmt_get; // consider overlay
        int entry_num = 0;
        int i;
        int rc = 0;

        if ((entry_num = get_irq_entry_num(p_cpld_device)) < 0)
                return entry_num;

        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
                "%s(%d) entry_num=%d p_cpld_device=%p wake_on_signal=%x\n",
                __FUNCTION__, __LINE__, entry_num, p_cpld_device,
                p_cpld_device->wake_on_signal);

        for (i = 0; i < entry_num; i++) {
	        /* clean status value for master_cache_blk and auxilary_cache_blk */
	        p_master_cache_blk->g_entry[i].status = 0;
	        p_auxilary_cache_blk->g_entry[i].status = 0;
	        /* set initial event value for master_cache_blk and auxilary_cache_blk */
	        p_master_cache_blk->g_entry[i].event = 0;
	        p_auxilary_cache_blk->g_entry[i].event = 0;
	        if (cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id]) {
		        mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
	                        "%s(%d) rw_func_id=%d\n",
                        	__FUNCTION__, __LINE__,
                        	p_set_act->s_entry[i].act.rw_func_id);

		        /* Read statuses and store them in master_cache_blk */
                        if ((p_set_act->s_entry[i].cfg.e_status.range[0] != CPLD_MASK_NOP8) &&
                            (p_set_act->s_entry[i].cfg.e_status.range[1] != CPLD_MASK_NOP8) &&
	                    (p_set_act->s_entry[i].cfg.profile & _E_AVAIL)) {
		        	u8 read_status = 0;
		        	rc = cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id](
					        	&p_set_act->s_entry[i].status_hwdev,
					        	p_set_act->s_entry[i].cfg.e_status.range[0],
					        	p_set_act->s_entry[i].cfg.e_status.range[1],
					        	1, &read_status, _B_NOLOCK);
	                        if (rc < 0)
		                        return rc;
	                        /* set initial status value for master_cache_blk and auxilary_cache_blk */
	                        p_master_cache_blk->g_entry[i].status = read_status & p_set_act->s_entry[i].cfg.mask;
	                        p_auxilary_cache_blk->g_entry[i].status = read_status & p_set_act->s_entry[i].cfg.mask;

                                mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                                "%s(%d) i=%d status (addr %x) = %x %x profile 0x%04x\n",
                                        __FUNCTION__, __LINE__, i, p_set_act->s_entry[i].cfg.e_status.range[0],
                                        p_master_cache_blk->g_entry[i].status,
                                        read_status, p_set_act->s_entry[i].cfg.profile);
                        }
		        /* Clear events */
                        if (!(p_set_act->s_entry[i].cfg.profile & _E_INV)) {
                                if ((p_set_act->s_entry[i].cfg.e_event.range[0] != CPLD_MASK_NOP8) &&
                                    (p_set_act->s_entry[i].cfg.e_event.range[1] != CPLD_MASK_NOP8) &&
                                    (p_set_act->s_entry[i].cfg.profile & _E_AVAIL)) {
		        	        u8 clear_event = 0;
		        	        rc = cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id](
					        	        &p_set_act->s_entry[i].event_hwdev,
					        	        p_set_act->s_entry[i].cfg.e_event.range[0],
					        	        p_set_act->s_entry[i].cfg.e_event.range[1],
					        	        0, &clear_event, _B_NOLOCK);
	                                if (rc < 0)
		                                return rc;

                                        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                                        "%s(%d) i=%d clear event (addr %x)\n",
                                                __FUNCTION__, __LINE__, i, p_set_act->s_entry[i].cfg.e_event.range[0]);
                                }
                        }
                }
        }

        for (i = 0; i < entry_num; i++) {
	        if (cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id]) {
		        u8 set_mask = 0;
		        /* Unmask */
		        if ((p_set_act->s_entry[i].cfg.profile & _E_INV) &&
		            (!(p_set_act->s_entry[i].cfg.profile & _E_INV_M)))
		                set_mask = ~p_set_act->s_entry[i].cfg.mask;
		        else
		                set_mask = p_set_act->s_entry[i].cfg.mask;

                        if ((p_set_act->s_entry[i].cfg.e_mask.range[0] != CPLD_MASK_NOP8) &&
                            (p_set_act->s_entry[i].cfg.e_mask.range[1] != CPLD_MASK_NOP8) &&
                            (p_set_act->s_entry[i].cfg.profile & _E_AVAIL)) {
		        	rc = cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id](
					        	&p_set_act->s_entry[i].mask_hwdev,
					        	p_set_act->s_entry[i].cfg.e_mask.range[0],
					        	p_set_act->s_entry[i].cfg.e_mask.range[1],
					        	0, &set_mask, _B_NOLOCK);
	                        if (rc < 0)
		                        return rc;

	                        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                                 "%s(%d) i=%d addr=%x ind=%d base=%x mask=%x set mask=%x prof=%x level=%x\n" \
	                                 "sig: (%x %x) (%x %x) (%x %x) (%x %x) (%x %x) (%x %x) (%x %x) (%x %x)\n",
                                         __FUNCTION__, __LINE__, i, p_set_act->s_entry[i].cfg.e_mask.range[0],
                                         p_set_act->s_entry[i].cfg.e_mask.dev_index,
                                         p_set_act->s_entry[i].mask_hwdev.lpc_params.base,
                                         p_set_act->s_entry[i].cfg.mask, set_mask,
                                         p_set_act->s_entry[i].cfg.profile,
                                         p_set_act->s_entry[i].cfg.level,
                                         p_set_act->s_entry[i].cfg.sig[0].signal_id,
                                         p_set_act->s_entry[i].cfg.sig[0].dev_signal.dev_index,
                                         p_set_act->s_entry[i].cfg.sig[1].signal_id,
                                         p_set_act->s_entry[i].cfg.sig[1].dev_signal.dev_index,
                                         p_set_act->s_entry[i].cfg.sig[2].signal_id,
                                         p_set_act->s_entry[i].cfg.sig[2].dev_signal.dev_index,
                                         p_set_act->s_entry[i].cfg.sig[3].signal_id,
                                         p_set_act->s_entry[i].cfg.sig[3].dev_signal.dev_index,
                                         p_set_act->s_entry[i].cfg.sig[4].signal_id,
                                         p_set_act->s_entry[i].cfg.sig[4].dev_signal.dev_index,
                                         p_set_act->s_entry[i].cfg.sig[5].signal_id,
                                         p_set_act->s_entry[i].cfg.sig[5].dev_signal.dev_index,
                                         p_set_act->s_entry[i].cfg.sig[6].signal_id,
                                         p_set_act->s_entry[i].cfg.sig[6].dev_signal.dev_index,
                                         p_set_act->s_entry[i].cfg.sig[7].signal_id,
                                         p_set_act->s_entry[i].cfg.sig[7].dev_signal.dev_index);
                        }
                }
        }

        return rc;
}

static inline int cpld_irqs_mask(struct cpld_device_table *p_cpld_device)
{
        struct cpld_mgmt_set_action *p_set_act = &p_cpld_device->p_irqs->u.mgmt_set_action;
        int entry_num = 0;
        u8 retry_access_cntr = 1;
        int i;
        int rc = 0;

        if ((entry_num = get_irq_entry_num(p_cpld_device)) < 0)
                return entry_num;

        _lock_bus(/*_B_ALL*/ p_cpld_device->device_type);

        for (i = 0; i < entry_num; i++) {
	        if (p_set_act->s_entry[i].mask != ~CPLD_MASK_NOP8) {
                        /* Mask interrupts */
	                if ((p_set_act->s_entry[i].cfg.e_mask.range[0] != CPLD_MASK_NOP8) &&
	                     (p_set_act->s_entry[i].cfg.e_mask.range[1] != CPLD_MASK_NOP8) &&
	                     (p_set_act->s_entry[i].cfg.profile & _E_AVAIL)) {
	        	        if (cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id]) {
	        	                u8 set_mask = 0;

	        		        if ((p_set_act->s_entry[i].cfg.profile & _E_INV) &&
	        		            (!(p_set_act->s_entry[i].cfg.profile & _E_INV_M)))
	        		                set_mask = 0xff;

                                        mlnx_log((MLNX_LOG_IRQ_MIN + 2), MLNX_LOG_IRQ_MAX,
	                                        "%s(%d) i=%d set_mask=%x mask addr=%x mask=%x\n",
                                                __FUNCTION__, __LINE__, i, set_mask,
                                                p_set_act->s_entry[i].cfg.e_mask.range[0],
                                                p_set_act->s_entry[i].cfg.mask);

retry_access:
	        		        rc = cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id](
			        			        &p_set_act->s_entry[i].mask_hwdev,
				        		        p_set_act->s_entry[i].cfg.e_mask.range[0],
				        		        p_set_act->s_entry[i].cfg.e_mask.range[1],
				        		        0, &set_mask, _B_NOLOCK);
		                        if ((rc < 0) && retry_access_cntr) {
		                         	/* Try to reset bus */
        			         	if (cpld_ctx.rst_func[p_set_act->s_entry[i].act.rw_func_id]) {
        			                         rc = cpld_ctx.rst_func[p_set_act->s_entry[i].act.rw_func_id](
        						                                        &p_cpld_device->bus_rst, _B_NOLOCK);
        			                         if (rc < 0) {
        			                                 _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);
        			                                 return rc;
        			                         }
        			                         retry_access_cntr = 0;
        			                         goto retry_access;

        			                         _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);

        			                         return rc;
        		                        }
        		                }

        	                } /*if (cpld_ctx.rw_func[p_set_act->s_entry[i].act.rw_func_id])*/
        	        } /*if ((p_set_act->s_entry[i].cfg.e_mask.range[0] !=*/
                } /*(p_set_act->s_entry[i].mask != ~CPLD_MASK_NOP8)*/
        } /*for (i = 0; i < entry_num; i++) {*/

        _unlock_bus(/*_B_ALL*/ p_cpld_device->device_type);

        return rc;
}

/*
  Interrupt handling schema operates with the following entities:
  - Master cache block:
     this block contains initial interrupt snapshot at the moment work
     queue entity is started.
     When work queue entity is start running. Each work queue entity
     starts interrupt collection from particular device by "mask interrupt"->
     "read interrupts" sequence and updates detected interrupts on this block.
     When new interrupt mapping is submitted to interrupted handler,
     the related region of master cache block is to be cleared -
     in this way all the interrupts raised before submitting are not taken
     into consideration.
  - Auxiliary cache block:
     this block also contains initial interrupt snapshot at the moment
     work queue entity is started. Each work queue entity handles interrupt
     generating devices by hierarchical nested traversing.
     It completes particular device by "acknowledge interrupt" ->
     "unmask interrupts" sequence, acknowledging only
     the interrupts collected in master block. Then it reads interrupts
     again and updates detected interrupts on this block. In such way
     it collects unhandled interrupts, which can be raised during
     traversing. These are deferred interrupts, which should be handled
     by next work queue entity. At the end of completion auxiliary cache
     block is copied to master cache block. In case auxiliary cache block
     checksum is different from master block cache some, work queue is
     rescheduled for deferred interrupts handling.
  - Digest block:
     this block contains interrupt digest, collected by work queue entity run.
     It contains digest per each system module (management, line, spine, fan, PS).
     This block is provided to user space as a result of interrupt handling.
     It requires implementation of shared-exclusive lock to support multiple
     readers/single-writer functionality. On this matter read-copy-update (RCU)
     wait-free for readers algorithm Linux algorithm will be applied.
     It will allow concurrent access to digest block for read operations
     and exclusive access for write operations.
*/
int mlnx_cpld_irq_subscribe(struct mlnx_cpld_subscribe_ioctl_params *params)
{
	struct cpld_device_table *p_cpld_device;
	struct hlist_head *head;
	int rc = 0;

	p_cpld_device = kmem_cache_alloc(cpld_device_table_cache, GFP_ATOMIC);
	if (!p_cpld_device) {
		printk(KERN_INFO "%s: failed to allocate cpld device 0x%08x from cache\n",
			__FUNCTION__, params->dev_id);
		return -EFAULT;
	}

	  mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
	  	  "%s(%d) dev_id=%d int_level=%d (dev_index, bus_id) (%d,%d)(%d,%d)(%d,%d)(%d,%d)\n",
	  	  __FUNCTION__, __LINE__, params->dev_id, params->int_level,
	  	  params->hw_acc_tab[0].dev_index, params->hw_acc_tab[0].hwdev.i2c_params.bus_id,
	  	  params->hw_acc_tab[1].dev_index, params->hw_acc_tab[1].hwdev.i2c_params.bus_id,
	  	  params->hw_acc_tab[2].dev_index, params->hw_acc_tab[2].hwdev.i2c_params.bus_id,
		  params->hw_acc_tab[3].dev_index, params->hw_acc_tab[3].hwdev.i2c_params.bus_id);

	  mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
		  "%s(%d) termination devices (dev_index, bus_id) (%d,%d)(%d,%d)(%d,%d)(%d,%d)\n",
	  	  __FUNCTION__, __LINE__,
	  	  params->hw_term_acc_tab[0].dev_index, params->hw_term_acc_tab[0].hwdev.i2c_params.bus_id,
	  	  params->hw_term_acc_tab[1].dev_index, params->hw_term_acc_tab[1].hwdev.i2c_params.bus_id,
	  	  params->hw_term_acc_tab[2].dev_index, params->hw_term_acc_tab[2].hwdev.i2c_params.bus_id,
		  params->hw_term_acc_tab[3].dev_index, params->hw_term_acc_tab[3].hwdev.i2c_params.bus_id);

	memset(&p_cpld_device->action_entry, 0, sizeof(struct cpld_handler_action_entry));
	memset(&p_cpld_device->digest_entry, 0, sizeof(struct mod_digest));
	p_cpld_device->action_entry.digest_entry = &p_cpld_device->digest_entry;
	INIT_HLIST_NODE(&p_cpld_device->hlist);
	p_cpld_device->device_id = params->dev_id;
	p_cpld_device->device_type = params->device_type;
	p_cpld_device->card_type = params->card_type;
	p_cpld_device->wake_on_signal = params->wake_on_signal;
	head =  &cpld_ctx.cpld_device_table_hash[cpld_device_hash_func(&params->dev_id)];
	hlist_add_head_rcu(&p_cpld_device->hlist, head);
	INIT_LIST_HEAD(&p_cpld_device->action_entry.action_list_1);
	INIT_LIST_HEAD(&p_cpld_device->action_entry.action_list_2);
	INIT_LIST_HEAD(&p_cpld_device->action_entry.action_list_3);
	INIT_LIST_HEAD(&p_cpld_device->action_entry.action_list_ha);
	p_cpld_device->action_entry.subscriber_data = (void *)p_cpld_device;         /* User data */
	p_cpld_device->action_entry.index = params->int_id;                          /* Device interrupt line id */
	p_cpld_device->action_entry.int_level = (enum irq_level_e)params->int_level; /* Device interrupt level */
	p_cpld_device->irq_entry_num = params->irq_entry_num;                        /* Number of entries in irq table */
	p_cpld_device->action_entry.usr_handle = params->dev_id;                     /* Upper user hanlde (module handle) */
	p_cpld_device->usr_handle = params->dev_id;                                  /* Upper user hanlde (module handle) */
	p_cpld_device->int_level = (enum irq_level_e)params->int_level;              /* Interrupt's hierarchical level */ // ??

	/* Allocate interrupt entries */
	p_cpld_device->p_irqs = kmalloc(sizeof(struct mlnx_cpld_interrupt_set_params_action), GFP_KERNEL);
	if (!p_cpld_device->p_irqs)
		goto cpld_device_alloc_failed;
	p_cpld_device->p_get_irqs = kmalloc(sizeof(struct mlnx_cpld_interrupt_get_params), GFP_KERNEL);
	if (!p_cpld_device->p_get_irqs) {
		kfree(p_cpld_device->p_irqs);
		goto cpld_device_alloc_failed;
	}
	p_cpld_device->p_auxilary_cache_blk = kmalloc(sizeof(struct mlnx_cpld_interrupt_get_params), GFP_KERNEL);
	if (!p_cpld_device->p_auxilary_cache_blk) {
		kfree(p_cpld_device->p_irqs);
		kfree(p_cpld_device->p_get_irqs);
		goto cpld_device_alloc_failed;
	}
	p_cpld_device->p_master_cache_blk = kmalloc(sizeof(struct mlnx_cpld_interrupt_get_params), GFP_KERNEL);
	if (!p_cpld_device->p_master_cache_blk) {
		kfree(p_cpld_device->p_irqs);
		kfree(p_cpld_device->p_get_irqs);
		kfree(p_cpld_device->p_auxilary_cache_blk);
		goto cpld_device_alloc_failed;
	}

	  mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
		"%s(%d) type=%d index=%d usr_handle %d %d int_level=%d %d\n",
		__FUNCTION__, __LINE__,
		params->device_type, p_cpld_device->action_entry.index,
		p_cpld_device->action_entry.usr_handle, p_cpld_device->usr_handle,
		p_cpld_device->int_level, p_cpld_device->action_entry.int_level);

	rc = cpld_accumulate_irqs(p_cpld_device, params);
	mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
		"%s(%d) rc=%d hook=%p\n", __FUNCTION__, __LINE__, rc,
		cpld_handler_action_subscribe_hook);
	if (cpld_handler_action_subscribe_hook) {
		rc = (rc >= 0) ? cpld_handler_action_subscribe_hook(&p_cpld_device->action_entry):rc;
		mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
			"%s(%d) rc=%d\n", __FUNCTION__, __LINE__, rc);
	}

cpld_device_alloc_failed:
	if (rc < 0)
		kmem_cache_free(cpld_device_table_cache, p_cpld_device);
	return rc;
}

int mlnx_cpld_irq_unsubscribe(struct mlnx_cpld_subscribe_ioctl_params *params)
{
	struct cpld_device_table *p_cpld_device;
	int rc = 0;

	mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
		"%s(%d)\n", __FUNCTION__, __LINE__);
	p_cpld_device = find_cpld_device_entry(params->dev_id);
	if (!p_cpld_device) {
		printk(KERN_INFO "%s: failed to find cpld device 0x%08x from cache\n",
			__FUNCTION__, params->dev_id);
		return -EFAULT;
	}

	/* De-allocate interrupt entries */
	if (!p_cpld_device->p_irqs)
		kfree(p_cpld_device->p_irqs);
	if (!p_cpld_device->p_get_irqs)
		kfree(p_cpld_device->p_get_irqs);
	if (!p_cpld_device->p_auxilary_cache_blk)
		kfree(p_cpld_device->p_auxilary_cache_blk);
	if (!p_cpld_device->p_master_cache_blk)
		kfree(p_cpld_device->p_master_cache_blk);

	hlist_del_rcu(&p_cpld_device->hlist);
	if (cpld_handler_action_unsubscribe_hook) {
		rc = cpld_handler_action_unsubscribe_hook(&p_cpld_device->action_entry);
	}
	kmem_cache_free(cpld_device_table_cache, p_cpld_device);
	return rc;
}

int mlnx_cpld_irq_set(struct mlnx_cpld_set_ioctl_params *params)
{
	struct cpld_device_table *p_cpld_device;
	int rc = 0;
	p_cpld_device = find_cpld_device_entry(params->dev_id);

        mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
		"%s(%d) params->dev_id=%d dev=%p\n",
		__FUNCTION__, __LINE__, params->dev_id, p_cpld_device);

	if (!p_cpld_device) {
                printk(KERN_INFO "%s bad device id 0x%08x\n", __FUNCTION__, params->dev_id);
                return -EFAULT;
	}
        mlnx_log((MLNX_LOG_MEL_IRQ_MIN + 2), MLNX_LOG_MEL_IRQ_MAX,
		"%s(%d) dev type=%d card type=%d\n",
		__FUNCTION__, __LINE__,
		p_cpld_device->device_type,
		p_cpld_device->card_type);
	if ((params->device_type == I2C_BUS_TYPE) ||
	    (params->device_type == LPC_BUS_TYPE)) {
	        _lock_bus(params->device_type);
	        rc = cpld_set_irqs(p_cpld_device);
	        _unlock_bus(params->device_type);
	}

        return rc;
}

int mlnx_cpld_irq_get_data(struct mlnx_cpld_get_data_ioctl_params *params)
{
	struct cpld_device_table *p_cpld_device;
        struct cpld_mgmt_set_action *p_set_act;
	struct cpld_mgmt_get *p_master_cache_blk;
	struct cpld_mgmt_get *p_get_params;
	int entry_num = 0;
	int i;
	int rc = 0;

	p_cpld_device = find_cpld_device_entry(params->dev_id);
	if (!p_cpld_device) {
                printk(KERN_INFO "%s bad device id 0x%08x\n", __FUNCTION__, params->dev_id);
                return -EFAULT;
	}

        if ((entry_num = get_irq_entry_num(p_cpld_device)) < 0)
                return entry_num;

        p_set_act = &p_cpld_device->p_irqs->u.mgmt_set_action;
	p_master_cache_blk = &p_cpld_device->p_master_cache_blk->u.mgmt_get;
	p_get_params  = &params->get_irqs.u.mgmt_get;

        for (i = 0; i < entry_num; i++) {
                if (p_set_act->s_entry[i].mask == ~CPLD_MASK_NOP8)
                        continue;
                p_get_params->g_entry[i].status = p_master_cache_blk->g_entry[i].status;
                p_get_params->g_entry[i].event = p_master_cache_blk->g_entry[i].event;
        }

	return rc;
}

void inline translate_offset_addr(u32 addr, u32 offset, u32 len, u8 width, char *offset_buf,
				  struct i2c_msg *msg, u8 *in_out_buf, u8 flag)
{
	char *msg0;
	struct i2c_msg *msg_i2c = msg;

	if (flag == 1) {
		msg0 = offset_buf;
		switch (width) {
			case 4:
				*msg0++ = offset >> 24;
				*msg0++ = offset >> 16;
				*msg0++ = offset >> 8;
				*msg0 = offset;
				break;
			case 2:
				offset &= 0xffff;
				*msg0++ = offset >> 8;
				*msg0 = offset;
				break;
			case 1:
				offset &= 0xff;
				*msg0 = offset;
				break;
		}
		msg_i2c->addr = addr;
		msg_i2c->buf = offset_buf;
		msg_i2c->len = width;
		msg_i2c++;
		msg_i2c->addr = addr;
		msg_i2c->flags = 1;
		msg_i2c->buf = in_out_buf;
		msg_i2c->len = len;
	}
	else {
		msg0 = in_out_buf;
		switch (width) {
			case 4:
				*msg0++ = offset >> 24;
				*msg0++ = offset >> 16;
				*msg0++ = offset >> 8;
				*msg0 = offset;
				break;
			case 2:
				offset &= 0xffff;
				*msg0++ = offset >> 8;
				*msg0 = offset;
				break;
			case 1:
				offset &= 0xff;
				*msg0 = offset;
				break;
		}
		msg_i2c->addr = addr;
		msg_i2c->flags = 0;
		msg_i2c->len = width + len;
		msg_i2c->buf = in_out_buf;
	}
}

int mlnx_vpd_rw_limited(struct mlnx_rdwr_vpd_i2c_ioctl_params *params)
{
	struct i2c_adapter *adap;
	struct i2c_adapter *adap_prot = NULL;
	int rc = 0;
	int rc1 = 1;
	struct i2c_msg msgs[2];
	struct i2c_msg msg_prot;
	unsigned char msgbuf_prot[2];
	struct mlnx_rdwr_vpd_i2c_ioctl_params vpd_rdwr_arg;
	int limit = 0;
	int border = 0;
	u32 offset;
	unsigned long timeout = 0;
	unsigned long rchunk_del;
	unsigned long wchunk_del;
	unsigned long post_del;
	int nmsgs;
	int datalen;
	int len;
	u8 offbuf[4];
	u8 tempbuf[I2C_CHUNK_MAX + 4];
	u8 databuf[2*I2C_CHUNK_MAX];
	u8 *dataptr;

	if (copy_from_user(&vpd_rdwr_arg, (void __user *)params, sizeof(struct mlnx_rdwr_vpd_i2c_ioctl_params))) {
		rc = -EFAULT;
		goto exit;
	}

	memset(&msg_prot, 0, sizeof(struct i2c_msg));
	memset(databuf, 0, 2*I2C_CHUNK_MAX);
	if  (vpd_rdwr_arg.flags == 0)
		memcpy(databuf, vpd_rdwr_arg.databuf, vpd_rdwr_arg.datalen);

	limit = vpd_rdwr_arg.limit;
	border = vpd_rdwr_arg.border;
	rchunk_del = 1000 * vpd_rdwr_arg.read_chunk_delay; // Input in msec - covert to usec
	wchunk_del = 1000 * vpd_rdwr_arg.write_chunk_delay; // Input in msec - covert to usec
	post_del = 1000 * vpd_rdwr_arg.post_delay; // Input in msec - covert to usec

	mutex_lock(&__i2c_mlnx_lock);

	if (limit > 0)
		timeout = jiffies + msecs_to_jiffies(limit);

	if (vpd_rdwr_arg.bus_prot >= 0) {
		adap_prot = i2c_get_adapter(vpd_rdwr_arg.bus_prot);
		if (!adap_prot) {
			mutex_unlock(&__i2c_mlnx_lock);
			rc = -EINVAL;
			goto exit;
		}
		msgbuf_prot[0] = vpd_rdwr_arg.reg_prot & 0xff;
		msgbuf_prot[1] = vpd_rdwr_arg.get_prot;
		msg_prot.addr = vpd_rdwr_arg.addr_prot;
		msg_prot.flags = 0;
		msg_prot.len = 2;
		msg_prot.buf = msgbuf_prot;

		/* Provide delay before taking protected bus */
		if (post_del > 0)
			udelay(post_del);

		rc = i2c_transfer(adap_prot, &msg_prot, 1);
		if (rc <= 0) {
			printk(KERN_ERR "%s(%d) rc=%d bus=%d addr=%x reg=%x get=%x\n",
				__FUNCTION__, __LINE__, rc,
				vpd_rdwr_arg.bus_prot, vpd_rdwr_arg.addr_prot,
				vpd_rdwr_arg.reg_prot, vpd_rdwr_arg.get_prot);
			goto failure;
		}
	}

	adap = i2c_get_adapter(vpd_rdwr_arg.bus);
	if (!adap) {
		rc = -EINVAL;
		goto failure;
	}


	offset = vpd_rdwr_arg.offset;
	datalen = vpd_rdwr_arg.datalen;
	dataptr = databuf;
	len = datalen;
	while (len > 0) {
		memset(msgs, 0, 2 * sizeof(struct i2c_msg));
		if (border > 0) {
			if (((offset % border) + len) > border) {
				if (len <= border)
					datalen = border - (offset % border);
				else
					datalen = (border - (offset % border)) % border;
			}
		}
		if (vpd_rdwr_arg.flags == 1) {
			nmsgs = 2; // Read
			translate_offset_addr(vpd_rdwr_arg.addr, offset, datalen,
						vpd_rdwr_arg.addr_width, offbuf, msgs, dataptr, 1);
		}
		else {
			nmsgs = 1; // Write
			memset(tempbuf, 0, (datalen + vpd_rdwr_arg.addr_width));
			memcpy(tempbuf + vpd_rdwr_arg.addr_width, dataptr, datalen);
			translate_offset_addr(vpd_rdwr_arg.addr, offset, datalen,
						vpd_rdwr_arg.addr_width, offbuf, msgs, tempbuf, 0);
		}

		rc = i2c_transfer(adap, msgs, nmsgs);
		if (rc <= 0) {
			printk(KERN_ERR "%s(%d) rc=%d nmsgs=%d bus_pr=%d addr_pr=%x reg_prot=%x get=%x put=%x bus=%d addr=%x width=%d off=%x flag=%d datalen=%d len=%d lim=%d border=%d post_delay=%d addr0=%x, flags0=%d addr1=%x, flags1=%d msg0=%p msg1=%p\n",
                                __FUNCTION__, __LINE__, rc, nmsgs, vpd_rdwr_arg.bus_prot, vpd_rdwr_arg.addr_prot,
				                vpd_rdwr_arg.reg_prot, vpd_rdwr_arg.get_prot,
                                vpd_rdwr_arg.put_prot, vpd_rdwr_arg.bus, vpd_rdwr_arg.addr, vpd_rdwr_arg.addr_width,
				                vpd_rdwr_arg.offset, vpd_rdwr_arg.flags, vpd_rdwr_arg.datalen, len, vpd_rdwr_arg.limit,
				                vpd_rdwr_arg.border, vpd_rdwr_arg.post_delay, msgs[0].addr, msgs[0].flags,
				                msgs[1].addr, msgs[1].flags, &msgs[0], &msgs[1]);
			goto failure;
		}

		if ((limit > 0) && (!time_before(jiffies, timeout))) {
			rc = -ETIME;
			goto failure;
		}

		len -= datalen;
		offset += datalen;
		dataptr += datalen;
		datalen = len;

		/* Proceed if more data to be sent and delay between chunks is required */
		if (len > 0) {
			if ((vpd_rdwr_arg.flags == 1) && (rchunk_del > 0)) {
				udelay(rchunk_del);
				if ((limit > 0) && (!time_before(jiffies, timeout))) {
					rc = -ETIME;
					goto failure;
				}
			}
			else if ((vpd_rdwr_arg.flags == 0) && (wchunk_del > 0)) {
				udelay(wchunk_del);
				if ((limit > 0) && (!time_before(jiffies, timeout))) {
					rc = -ETIME;
					goto failure;
				}
			}
		} /*if (len > 0)*/
	} /*while (len > 0)*/

failure:
	if ((vpd_rdwr_arg.bus_prot >= 0) && adap_prot) {
		msgbuf_prot[0] = vpd_rdwr_arg.reg_prot & 0xff;
		msgbuf_prot[1] = vpd_rdwr_arg.put_prot;
		msg_prot.addr = vpd_rdwr_arg.addr_prot;
		msg_prot.flags = 0;
		msg_prot.len = 2;
		msg_prot.buf = msgbuf_prot;

		/* Provide delay before release protected bus */
		if (post_del > 0)
			udelay(post_del);

		rc1 = i2c_transfer(adap_prot, &msg_prot, 1);
		if (rc1 <= 0) {
			printk(KERN_ERR "%s(%d) rc=%d fbus=%d addr=%x reg=%x put=%x\n",
				__FUNCTION__, __LINE__, rc1,
				vpd_rdwr_arg.bus_prot, vpd_rdwr_arg.addr_prot,
				vpd_rdwr_arg.reg_prot, vpd_rdwr_arg.put_prot);
		}
	}

	mutex_unlock(&__i2c_mlnx_lock);


	if ((rc >= 0) && (vpd_rdwr_arg.flags == 1))
		memcpy(vpd_rdwr_arg.databuf, databuf, vpd_rdwr_arg.datalen);
        vpd_rdwr_arg.rc = rc;

        if (copy_to_user((void __user *)params, &vpd_rdwr_arg, sizeof(struct mlnx_rdwr_vpd_i2c_ioctl_params)))
               return -EFAULT;

exit:
	/* rc - return code from data transaction or bus get, rc1 - return code from bus put */
	if (rc1 <= 0)
		return rc1;
	return rc;
}
EXPORT_SYMBOL(mlnx_vpd_rw_limited);

static int __i2c_check_addr(struct device *dev, void *addrp)
{
	struct i2c_client	*client = i2c_verify_client(dev);
	int			addr = *(int *)addrp;

	if (client && client->addr == addr)
		return -EBUSY;
	return 0;
}

static int i2c_check_addr(struct i2c_adapter *adapter, int addr)
{
	return device_for_each_child(&adapter->dev, &addr, __i2c_check_addr);
}

int mlnx_scan_switch_devices(struct mlnx_i2c_virt_bus_scan_ioctl_params *params)
{
	struct i2c_adapter *adap;
	u32 parent_bus_num;
	int addr;
	int ind = 0x04;
	int ind1 = 0;

        parent_bus_num = params->parent_bus_num;
        adap = i2c_get_adapter(parent_bus_num);
        if (!adap)
                return -EFAULT;
        memset(&params->presence, 0, (sizeof(struct mlnx_i2c_virt_bus_scan_ioctl_params) -
                offsetof(struct mlnx_i2c_virt_bus_scan_ioctl_params, presence)));

        for (addr = 0x04; addr <= 0xff; addr++, ind++) {
		/* Check whether address is already in use */
		if (i2c_check_addr(adap, addr) < 0) {
			params->presence[ind1] |= ((ind%32) << 1);
		}
		else {
			/* Make sure there is something at this address */
			if (i2c_smbus_xfer(adap, addr, 0, 0, 0, I2C_SMBUS_QUICK, NULL) >= 0)
				params->presence[ind1] |= ((ind%32) << 1);
		}
		if (ind%32 == 0)
			ind1++;
	}
	i2c_put_adapter(adap);

	return 0;
}
EXPORT_SYMBOL(mlnx_scan_switch_devices);

#define LEVEL_DEPTH      4
static int switch_level = 0;
static struct _switch_bus {
	u32 swicth_addr;
	u32 bus_num;
} swicth_bus[LEVEL_DEPTH];

static int _mlnx_is_device_switch(int addr)
{
	if((addr >= 0x70) && (addr <= 0x77))
		return 1;
	return 0;
}

static int _mlnx_is_device_bus_master_selector(int addr)
{
	if(addr == 0x79)
		return 1;
	return 0;
}

static int mlnx_scan_bus(u32 bus_num, int first_addr, int last_addr, int skip_addr,
				char *name, char *buf, ssize_t *p_cnt)
{
	struct i2c_adapter *adap = NULL;
	struct pca954x_table *p_switch = NULL;
	u8 bus_master_selector = 0;
	int bus_master_selector_addr = 0;
	int addr;
	int adap_cntr;
	int rc;

        adap = i2c_get_adapter(bus_num);
        if (!adap)
                return -EFAULT;

        switch_level++;
        for (addr = last_addr; addr >= first_addr; addr--) {
		/* Check whether address is already in use */
		switch (switch_level) {
		        case 1:
		                break;
		        case 2:
		                if ((addr >= 0x76 && addr <= 0x77) || (addr == skip_addr))
		                        continue;
		                break;
		        case 3:
		                if ((addr >= 0x72 && addr <= 0x77) || (addr == skip_addr))
		                        continue;
		                break;
		        case 4:
		                if ((addr >= 0x70 && addr <= 0x77) || (addr == skip_addr))
		                        continue;
		                break;
		        default:
		                return 0;
		}
		/* Make sure there is something at this address */
		rc = i2c_smbus_xfer(adap, addr, 0, 0, 0, I2C_SMBUS_QUICK, NULL);
		if (rc >= 0) {
			switch(switch_level) {
				case 2:
					printk(KERN_INFO
					//(*p_cnt) += sprintf(buf + (*p_cnt),
						"%*c bus=%4d addr=0x%02x %s route [%1d 0x%02x]\n",
						5*switch_level, '<', bus_num, addr, (name == NULL)?"NONE":name,
						swicth_bus[1].bus_num, swicth_bus[1].swicth_addr);
					break;
				case 3:
					printk(KERN_INFO
					//(*p_cnt) += sprintf(buf + (*p_cnt),
						"%*c bus=%4d addr=0x%02x %s route [%1d 0x%02x %1d 0x%02x]\n",
						5*switch_level, '<', bus_num, addr, (name == NULL)?"NONE":name,
						swicth_bus[1].bus_num, swicth_bus[1].swicth_addr,
						swicth_bus[2].bus_num, swicth_bus[2].swicth_addr);
					break;
				case 4:
					if (_mlnx_is_device_bus_master_selector(addr)) {
						bus_master_selector_addr = addr;
						bus_master_selector = 1;
						break;
					}
					if (bus_master_selector)
						printk(KERN_INFO
						//(*p_cnt) += sprintf(buf + (*p_cnt),
							"%*c bus=%4d addr=0x%02x %s route [%1d 0x%02x %1d 0x%02x %2d 0x%02x via bus master selector pca9541 at 0x%02x]\n",
							5*switch_level, '<', bus_num, addr, (name == NULL)?"NONE":name,
							swicth_bus[1].bus_num, swicth_bus[1].swicth_addr,
							swicth_bus[2].bus_num, swicth_bus[2].swicth_addr,
							swicth_bus[3].bus_num, swicth_bus[3].swicth_addr,
							bus_master_selector_addr);
					else
						printk(KERN_INFO
						//(*p_cnt) += sprintf(buf + (*p_cnt),
							"%*c bus=%4d addr=0x%02x %s route [%1d 0x%02x %1d 0x%02x %2d 0x%02x]\n",
							5*switch_level, '<', bus_num, addr, (name == NULL)?"NONE":name,
							swicth_bus[1].bus_num, swicth_bus[1].swicth_addr,
							swicth_bus[2].bus_num, swicth_bus[2].swicth_addr,
							swicth_bus[3].bus_num, swicth_bus[3].swicth_addr);
					break;
				default:
					break;
			}
        		if (_mlnx_is_device_switch(addr)) {
                		p_switch = find_pca954x_switch(SET_SWITCH_ID(0, bus_num, addr));
				if (p_switch) {
	        			swicth_bus[switch_level].swicth_addr = addr;
	        			swicth_bus[switch_level].bus_num = bus_num;
                			for (adap_cntr = 0; adap_cntr < LEGS_NUM; adap_cntr++)
						mlnx_scan_bus(p_switch->adap_id_array[adap_cntr], first_addr,
								last_addr, addr, p_switch->switch_name, buf, p_cnt);
				}
			}
		}
	}
	i2c_put_adapter(adap);
	swicth_bus[switch_level].swicth_addr = 0;
	swicth_bus[switch_level].bus_num = 0;
        switch_level--;
        return 0;
}

ssize_t virt_i2c_tree_show(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
	ssize_t cnt = 0;
	mutex_lock(&__i2c_mlnx_lock);
	swicth_bus[switch_level].swicth_addr = 0;
	swicth_bus[switch_level].bus_num = PHYSICAL_BUS_NUM;
	mlnx_scan_bus(PHYSICAL_BUS_NUM, 0x20, 0x79, 0x0, NULL, buf, &cnt);
	mutex_unlock(&__i2c_mlnx_lock);
	return cnt+1;
}

ssize_t show_module_static_info(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
	ssize_t cnt = 0;
	int index;
	struct pca954x_table *p_switch = NULL;

	p_switch = find_pca954x_switch_by_name(attr->attr.name);
	if (p_switch) {
		struct i2c_devices_table *p_i2c_dev_tbl = NULL;
		u8 is_switch = 0;
		u8 addr = 0;
		u32 bus_num;
		cnt += sprintf(buf + cnt, "Module %s parent bus %4d [switch id 0x%06x status %s] with own buses:\n",
				p_switch->switch_name, p_switch->parent_bus_num,
				p_switch->switch_id, (p_switch->status == I2C_ADAP_BLK)?"BLK":"UNBLK");
		for (index = 0; index < LEGS_NUM; index++) {
			bus_num = p_switch->adap_id_array[index];
			cnt += sprintf(buf + cnt, "%*c %4d", 10, '<', bus_num);
	        	if ( (p_i2c_dev_tbl = find_i2c_device(bus_num, &is_switch, &addr)) ) {
                		if (is_switch)
					cnt += sprintf(buf + cnt, " has child switch at 0x%02x adaptor status %s", addr,
                        				(p_i2c_dev_tbl->adap->blocked == I2C_ADAP_BLK)?"BLK":"UNBLK");
                	        else
					cnt += sprintf(buf + cnt,
                        				" has attached device %s at 0x%02x dev id 0x%x [0x%06x] adaptor is %s",
                        				p_i2c_dev_tbl->i2c_dev_type, addr, p_i2c_dev_tbl->user_handle,
                        				(((bus_num & 0x0000ffff) << 8) | (addr & 0x7f)),
                        				(p_i2c_dev_tbl->adap->blocked == I2C_ADAP_BLK)?"BLK":"UNBLK");
			}
			cnt += sprintf(buf + cnt, "%*c \n", 10, '>');
		}
	}
	else
		cnt += sprintf(buf + cnt, "Module %s not present\n", attr->attr.name);
	return cnt+1;
}

ssize_t show_mgmt_module_static_info(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
	ssize_t cnt = 0;
	struct pca954x_table *p_switch = NULL;
	struct pca954x_table *p = NULL;
	struct pca954x_table *n = NULL;
	u32 bus_num;
	int index;

	p_switch = find_pca954x_main_switch();
        if (!p_switch) {
                cnt += sprintf(buf + cnt, "Module %s not present\n", attr->attr.name);
                return cnt+1;
        }
	cnt += sprintf(buf + cnt, "Management module: parent bus %4d [switch id 0x%06x status %s] with own buses:\n",
		p_switch->parent_bus_num, p_switch->switch_id, (p_switch->status == I2C_ADAP_BLK)?"BLK":"UNBLK");
	for (index = 0; index < LEGS_NUM; index++) {
		bus_num = p_switch->adap_id_array[index];
		cnt += sprintf(buf + cnt, "%*c %4d %*c\n", 2, '<', bus_num,  2, '>');
	}
	list_for_each_entry_safe_reverse(p, n, &p_switch->attached_switch_list, list) {
		cnt += sprintf(buf + cnt, "parent bus %4d [switch id 0x%06x status %s] with own buses:\n",
			p->parent_bus_num, p->switch_id, (p->status == I2C_ADAP_BLK)?"BLK":"UNBLK");
		for (index = 0; index < LEGS_NUM; index++) {
			bus_num = p->adap_id_array[index];
			cnt += sprintf(buf + cnt, "%*c %4d %*c\n", 4, '<', bus_num,  4, '>');
		}
	}
	return cnt+1;
}

static struct kobj_attribute virt_i2c_tree_attr =
	__ATTR(virt_i2c_tree, S_IRUGO, virt_i2c_tree_show, NULL);

static struct kobj_attribute fru_module_attr[] = {
	__ATTR(S01, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S02, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S03, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S04, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S05, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S06, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S07, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S08, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S09, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S10, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S11, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S12, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S13, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S14, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S15, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S16, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S17, S_IRUGO, show_module_static_info, NULL),
	__ATTR(S18, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L01, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L02, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L03, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L04, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L05, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L06, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L07, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L08, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L09, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L10, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L11, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L12, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L13, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L14, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L15, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L16, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L17, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L18, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L19, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L20, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L21, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L22, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L23, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L24, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L25, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L26, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L27, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L28, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L29, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L30, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L31, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L32, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L33, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L34, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L35, S_IRUGO, show_module_static_info, NULL),
	__ATTR(L36, S_IRUGO, show_module_static_info, NULL),
	__ATTR(MGMT, S_IRUGO, show_mgmt_module_static_info, NULL),
};

static int (*cpld_handler_action_subscribe_symbol)(struct cpld_handler_action_entry *action_entry) = NULL;
static int (*cpld_handler_action_unsubscribe_symbol)(struct cpld_handler_action_entry *action_entry) = NULL;
int (*mlnx_watchdog_config_symbol)(struct mlnx_watchdog_config_ioctl_params *wd_config_params,
				   rw_bus_access_func rw_func) = NULL;
int (*mlnx_watchdog_enable_disable_symbol)(struct mlnx_watchdog_enable_disable_ioctl_params *wd_ed_params) = NULL;
int (*mlnx_watchdog_timer_clear_symbol)(wd_obj_e wd_id) = NULL;
int (*mlnx_watchdog_counter_clear_symbol)(wd_obj_e wd_id) = NULL;
int *mlnx_debug_level_symbol = NULL;

int mlnx_i2c_init(void)
{
	int rc;
	int index;
#if 0
	int index1;
#endif

	pca954x_table_cache = kmem_cache_create("pca954x_table_cache",
					 sizeof(struct pca954x_table),
					 0,
					 SLAB_HWCACHE_ALIGN, NULL);
	if (!pca954x_table_cache)
		return -ENOMEM;
	i2c_devices_table_cache = kmem_cache_create("i2c_devices_table_cache",
					 sizeof(struct i2c_devices_table),
					 0,
					 SLAB_HWCACHE_ALIGN, NULL);
	if (!i2c_devices_table_cache) {
		kmem_cache_destroy(pca954x_table_cache);
		return -ENOMEM;
	}
	i2c_bus_table_cache = kmem_cache_create("i2c_bus_table_cache",
					 sizeof(struct i2c_bus_table),
					 0,
					 SLAB_HWCACHE_ALIGN, NULL);
	if (!i2c_bus_table_cache) {
		kmem_cache_destroy(i2c_devices_table_cache);
		kmem_cache_destroy(pca954x_table_cache);
		return -ENOMEM;
	}
#ifdef GPIO_BUS
	i2c_gpio_table_cache = kmem_cache_create("i2c_gpio_table_cache",
					 sizeof(struct i2c_gpio_table),
					 0,
					 SLAB_HWCACHE_ALIGN, NULL);
	if (!i2c_gpio_table_cache) {
		kmem_cache_destroy(i2c_bus_table_cache);
		kmem_cache_destroy(i2c_devices_table_cache);
		kmem_cache_destroy(pca954x_table_cache);
		return -ENOMEM;
	}
#endif /* GPIO_BUS */
	cpld_device_table_cache = kmem_cache_create("cpld_device_table_cache",
					 sizeof(struct cpld_device_table),
					 0,
					 SLAB_HWCACHE_ALIGN, NULL);
	if (!cpld_device_table_cache) {
#ifdef GPIO_BUS
		kmem_cache_destroy(i2c_gpio_table_cache);
#endif
		kmem_cache_destroy(i2c_bus_table_cache);
		kmem_cache_destroy(i2c_devices_table_cache);
		kmem_cache_destroy(pca954x_table_cache);
		return -ENOMEM;
	}
	memset(&i2c_ctx, 0, sizeof(struct i2c_cntrl_ctx));
	memset(&cpld_ctx, 0, sizeof(struct cpld_cntrl_ctx));
	cpld_ctx.rw_func[I2C_BUS_TYPE] = i2c_bus_access_func;
	cpld_ctx.rw_func[LPC_BUS_TYPE] = lpc_bus_access_func;
	cpld_ctx.rst_func[I2C_BUS_TYPE] = i2c_bus_reset_func;
	cpld_ctx.rst_func[LPC_BUS_TYPE] = lpc_bus_reset_func;

	i2c_ctx.plat_dev = platform_device_alloc("mlnx_i2c", -1);
	if (!i2c_ctx.plat_dev) {
		rc = -ENOMEM;
		printk(KERN_INFO "%s failed to allocate virtual i2c tree platform device\n",  __FUNCTION__);
		goto fail_mlnx_i2c_0;
	}
	rc = platform_device_add(i2c_ctx.plat_dev);
	if (rc) {
		printk(KERN_INFO "%s failed to add cpld platform device\n", __FUNCTION__);
		goto fail_mlnx_i2c_1;
	}
	rc = sysfs_create_file(&(i2c_ctx.plat_dev->dev.kobj), &virt_i2c_tree_attr.attr);
	if (rc) {
		printk(KERN_INFO "%s create sysfs file failed, error %d\n", __FUNCTION__, rc);
		goto fail_mlnx_i2c_2;
	}
	for (index = 0; index < ARRAY_SIZE(fru_module_attr); index++) {
		rc = sysfs_create_file(&(i2c_ctx.plat_dev->dev.kobj), &fru_module_attr[index].attr);
		if (rc) {
			printk(KERN_INFO "%s create sysfs file failed, error %d\n", __FUNCTION__, rc);
			goto fail_mlnx_i2c_modules;
		}
	}
	rcu_assign_pointer(sx_i2c_get_probe_status_hook, NULL);
	cpld_handler_action_subscribe_symbol = __symbol_get("cpld_handler_action_subscribe");
	cpld_handler_action_unsubscribe_symbol = __symbol_get("cpld_handler_action_unsubscribe");
	mlnx_log((MLNX_LOG_MELLAGGRA_MIN + 2), MLNX_LOG_MELLAGGRA_MAX,
		"%s(%d) rc=%d subscribe=%p unsubscribe=%p\n", __FUNCTION__, __LINE__, rc,
		cpld_handler_action_subscribe_symbol, cpld_handler_action_unsubscribe_symbol);
	if(cpld_handler_action_subscribe_symbol)
		rcu_assign_pointer(cpld_handler_action_subscribe_hook, cpld_handler_action_subscribe_symbol);
	if(cpld_handler_action_unsubscribe_symbol)
		rcu_assign_pointer(cpld_handler_action_unsubscribe_hook, cpld_handler_action_unsubscribe_symbol);

	mlnx_watchdog_config_symbol = __symbol_get("mlnx_watchdog_config");
	mlnx_watchdog_enable_disable_symbol = __symbol_get("mlnx_watchdog_enable_disable");
	mlnx_watchdog_timer_clear_symbol = __symbol_get("mlnx_watchdog_timer_clear");
	mlnx_watchdog_counter_clear_symbol = __symbol_get("mlnx_watchdog_counter_clear");
	mlnx_debug_level_symbol = __symbol_get("mlnx_debug_level");


	return 0;

fail_mlnx_i2c_modules:
#if 0
	for (index1 = index - 1; index1 >= 0; index1--)
		sysfs_remove_file(&(i2c_ctx.plat_dev->dev.kobj), &fru_module_attr[index1].attr);
	sysfs_remove_file(&(i2c_ctx.plat_dev->dev.kobj), &virt_i2c_tree_attr.attr);
#endif
fail_mlnx_i2c_2:
	platform_device_del(i2c_ctx.plat_dev);
fail_mlnx_i2c_1:
	platform_device_put(i2c_ctx.plat_dev);
fail_mlnx_i2c_0:
	return 0;
}
EXPORT_SYMBOL(mlnx_i2c_init);

void mlnx_i2c_exit(void)
{
	int cntr;
	int cntr1;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
	struct hlist_node *h = NULL, *n = NULL;
#else
	struct hlist_node *n = NULL;
#endif
#ifdef GPIO_BUS
	struct i2c_gpio_table *gt;
#endif
	struct pca954x_table *st;
	struct i2c_devices_table *dt;
#if defined(__i386__) || defined(__x86_64__)
	struct cpld_device_table *ct;
	struct i2c_bus_table *bt;
#endif

#if defined(__i386__) || defined(__x86_64__)
	for (cntr = 0; cntr < CPLD_HASH_SIZE; cntr++) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		hlist_for_each_entry_safe(ct, h, n, &cpld_ctx.cpld_device_table_hash[cntr], hlist) {
#else
		hlist_for_each_entry_safe(ct, n, &cpld_ctx.cpld_device_table_hash[cntr], hlist) {
#endif
			hlist_del_rcu(&ct->hlist);
			kmem_cache_free(cpld_device_table_cache, ct);
		}
	}
#endif

#ifdef GPIO_BUS
	for (cntr1 = 0; cntr1 < I2C_BUS_HASH_SIZE; cntr1++) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		hlist_for_each_entry_safe(gt, h, n, &i2c_ctx.i2c_gpio_table_hash[cntr1], hlist) {
#else
		hlist_for_each_entry_safe(gt, n, &i2c_ctx.i2c_gpio_table_hash[cntr1], hlist) {
#endif
			if (gt->plat_dev) {
                		for (cntr = 0; cntr < gt->gpio_num; cntr++) {
                        		unsigned gpio = gt->mux_gpios[cntr];
                        		gpio_unexport(gpio);
                        		gpio_free(gpio);
                		}
                		platform_device_del(gt->plat_dev);
				platform_device_put(gt->plat_dev);
			}
		}
	}
#endif
	for (cntr = 0; cntr < I2C_DEV_HASH_SIZE; cntr++) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		hlist_for_each_entry_safe(dt, h, n, &i2c_ctx.i2c_devices_table_hash[cntr], hlist) {
#else
		hlist_for_each_entry_safe(dt, n, &i2c_ctx.i2c_devices_table_hash[cntr], hlist) {
#endif
			if (dt->client) {
#if 1
				if (dt->adap)
					mutex_lock(&__i2c_mlnx_lock); //i2c_lock_adapter(dt->adap);
				i2c_unregister_device(dt->client);
				if (dt->adap)
					mutex_unlock(&__i2c_mlnx_lock); //i2c_unlock_adapter(dt->adap);
#endif
			}
			if (dt->adap)
				i2c_put_adapter(dt->adap);
			dt->adap = NULL;
			dt->client = NULL;
			dt->parent_bus_num = I2C_BUS_NONE;
#if defined(__i386__) || defined(__x86_64__)
			hlist_del_rcu(&dt->hlist);
			kmem_cache_free(i2c_devices_table_cache, dt);
#endif
		}
	}
	for (cntr1 = 0; cntr1 <= SWITCH_MAX_DEPTH; cntr1++) {
		for (cntr = 0; cntr < PCA954x_HASH_SIZE; cntr++) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
			hlist_for_each_entry_safe(st, h, n, &i2c_ctx.pca954x_table_hash[cntr], hlist) {
#else
			hlist_for_each_entry_safe(st, n, &i2c_ctx.pca954x_table_hash[cntr], hlist) {
#endif
				if (st->client) {
					if(st->attached_switch_cntr == 0) {
						struct pca954x_table *p_parent = NULL;
						struct i2c_bus_table *p_bus = NULL;
#if 1
						if (st->adap)
							mutex_lock(&__i2c_mlnx_lock);
						i2c_unregister_device(st->client);
						if (st->adap)
							mutex_unlock(&__i2c_mlnx_lock);
#endif
        					if (st->parent_bus_num > 0) {
        						p_bus = find_i2c_bus_entry(st->parent_bus_num);
        						if (p_bus) {
        							if (p_bus->switch_id) {
                							p_parent = find_pca954x_switch(p_bus->switch_id);
                							if (p_parent)
                								p_parent->attached_switch_cntr--;
        							}
        						}
        					}
						if (st->adap)
							i2c_put_adapter(st->adap);
						st->adap = NULL;
						st->client = NULL;
						st->switch_id = PCA954x_SWITCH_NONE;
						st->attached_dev_cntr = 0;
						st->attached_switch_cntr = 0;
#if defined(__i386__) || defined(__x86_64__)
						hlist_del_rcu(&st->hlist);
						kmem_cache_free(pca954x_table_cache, st);
#endif
					}
				}
			}
		}
	}
#if defined(__i386__) || defined(__x86_64__)
	for (cntr = 0; cntr <= I2C_BUS_HASH_SIZE; cntr++) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,32)
		hlist_for_each_entry_safe(bt, h, n, &i2c_ctx.i2c_bus_table_hash[cntr], hlist) {
#else
		hlist_for_each_entry_safe(bt, n, &i2c_ctx.i2c_bus_table_hash[cntr], hlist) {
#endif
			hlist_del_rcu(&bt->hlist);
			kmem_cache_free(i2c_bus_table_cache, bt);
		}
	}
#endif

	if (cpld_device_table_cache)
		kmem_cache_destroy(cpld_device_table_cache);
#ifdef GPIO_BUS
	if (i2c_gpio_table_cache)
		kmem_cache_destroy(i2c_gpio_table_cache);
#endif
	if (i2c_bus_table_cache)
		kmem_cache_destroy(i2c_bus_table_cache);
	if (i2c_devices_table_cache)
		kmem_cache_destroy(i2c_devices_table_cache);
	if (pca954x_table_cache)
		kmem_cache_destroy(pca954x_table_cache);

// MS #if !defined(__i386__) && !defined(__x86_64__)
    /* MS Check. Done in platform_device_del
    for (cntr = 0; cntr < ARRAY_SIZE(fru_module_attr); cntr++)
		sysfs_remove_file(&(i2c_ctx.plat_dev->dev.kobj), &fru_module_attr[cntr].attr);
	sysfs_remove_file(&(i2c_ctx.plat_dev->dev.kobj), &virt_i2c_tree_attr.attr); */
	platform_device_del(i2c_ctx.plat_dev);
	platform_device_put(i2c_ctx.plat_dev);
// MS #endif
	rcu_assign_pointer(sx_i2c_get_probe_status_hook, NULL);
	rcu_assign_pointer(cpld_handler_action_subscribe_hook, NULL);
	rcu_assign_pointer(cpld_handler_action_unsubscribe_hook, NULL);
	if (cpld_handler_action_subscribe_symbol != NULL)
		__symbol_put("cpld_handler_action_subscribe");
	if (cpld_handler_action_unsubscribe_symbol != NULL)
		__symbol_put("cpld_handler_action_unsubscribe");
	if (mlnx_watchdog_config_symbol != NULL)
		__symbol_put("mlnx_watchdog_config");
	if (mlnx_watchdog_enable_disable_symbol != NULL)
		__symbol_put("mlnx_watchdog_enable_disable");
	if (mlnx_watchdog_timer_clear_symbol != NULL)
		__symbol_put("mlnx_watchdog_timer_clear");
	if (mlnx_watchdog_counter_clear_symbol != NULL)
		__symbol_put("mlnx_watchdog_counter_clear");
	if (mlnx_debug_level_symbol)
		__symbol_put("mlnx_debug_level");

	printk(KERN_INFO "%s module mellaggra_mod has been unloaded\n", __FUNCTION__);
}
EXPORT_SYMBOL(mlnx_i2c_exit);

MODULE_AUTHOR("Vadim Pasternak <vadimp@mellanox.com>");
MODULE_DESCRIPTION("MELLAGGRA: MEllanox Low Level AGGRegated Api");
MODULE_LICENSE("GPL v2");


